/*
 * CreateFilterBank1D.c
 *
 *  Function creates a 1D multiband Gabor filter bank in a wavelet 
 *  tesselation.  
 * 
 *  A pointer to the new filter bank is returned through the function name.  
 *
 *  The 2D filter design equations are found on pp. 111 - 112 of Book VI.
 *  REVISION:  8/14/94: pp. 82 - 83 book VIII.
 *
 *  The 1D filterbank is just a section through 1 RAY of the 2D bank.
 *
 *  The 1D design equations are on pp. 96 - 100 of book VIII.
 *
 *  NOTE that the InitialRadialFreq parameter to this function is specified in
 *       units of CYCLES PER PIXEL.
 *
 *  jph 9/19/94
 *
 */

#ifndef CREATEFILTERBANK1D_C
#define CREATEFILTERBANK1D_C
#include "teager2D.h"
#include "Esa1D.h"
#include <stdio.h>

filterbank *CreateFilterBank1D(r0,OBW,Eta,kappa)

  double      r0;                  /* radial center freq. of the first filter */
  float       OBW;                 /* Eta-peak octave bandwidth of filters */
  float       Eta;                 /* OBW is defined in terms of Eta-peak. */
  float       kappa;               /* postfilter sigma scaling factor */

{

  filterbank *G;                   /* pointer to the new filter bank */
  float       R;                   /* common ratio of radii of adjcnt filtrs */
  int         i;                   /* loop counter */

 /*
  * Allocate the filter bank on the heap
  */
  if ((G = (filterbank *)malloc(sizeof(filterbank))) == NULL) {
    printf("\n\nCreateFilterBank: free store exhausted.\n");
    exit(-1);
  }

 /*
  * Assiged filter bank parameters
  */
  G->r0    = r0;
  G->OBW   = OBW;
  G->Eta   = Eta;
  G->kappa = kappa;

 /*
  * Computed filter bank parameters
  */
  G->R = (float)pow((double)2.0,(double)OBW);
  G->NumFiltersPerRay  = (int)floor(pow((double)-OBW,(double)-1.0) * 
                   ((log((double)G->r0) / log((double)2.0)) + (double)1.0)) + 1;
  G->gamma = pow(pow((double)2.0,(double)G->OBW) - (double)1.0, (double)2.0) /
             pow(pow((double)2.0,(double)G->OBW) + (double)1.0, (double)2.0);
  G->Theta0 = (double)0.0;
  G->Lambda = (double)0.0;
  G->NumRays = 1;
  G->TotNumFilters = G->NumFiltersPerRay;

 /*
  * Create the filters
  */
  if ((G->filt = (FilterPtr *)malloc(G->TotNumFilters * sizeof(FilterPtr))) 
      == NULL) {
    printf("\n\nCreateFilterBank: free store exhausted.\n");
    exit(-1);
  }
  for (i=0; i<G->TotNumFilters; i++) {
    G->filt[i] = CreateGaborFilter1D(G,i);
  }

  return(G);
}
#endif 
