/*
 * HardyConvolveGabor1D.c
 *
 *  Function convolves hardy input sequence x with a specified 1D gabor filter 
 *  by multiplication in the frequency domain.  The space domain and spectral
 *  representations of Y are computed.
 *
 *  On invocation, it is assumed that the X and Y sequences have the same size 
 *  and *  have been fully allocated, and that the X sequence has been fully 
 *  initialized.  It is also assumed that the size, square, rows, and cols 
 *  fields of the Y image have been initialized.
 *
 *  jph 9/24/94
 *
 *-----------------------------------
 *
 *  Major Revision: 12/19/2000:
 *
 *    Make agreement with discrete Hilbert transform as in Dissertation/Cizek.
 *    See book XII, pp. 94-98.
 *
 *    The +/- N/2 coefs were being zero'd.  Now they are not.
 */

#ifndef HARDYCONVOLVEGABOR1D_C
#define HARDYCONVOLVEGABOR1D_C
#include <stdio.h>
#include "teager2D.h"
#include "Esa1D.h"

void HardyConvolveGabor1D(Y,X,G)

  pFloatImage   Y;  /* output signal */
  pFloatImage   X;  /* input  signal */
  FilterPtr     G;  /* pointer to the Gabor filter */

{

  double      u;       /* freq in cycles per sample */ 
  double      fincr;   /* incr in freq per sample of the sequence */
  int         i;       /* loop counter */
  int         so2;     /* size over 2 */

 /*
  * Useful values
  */
  fincr = (double)1.0 / (double)X->size;
  so2   = X->size >> 1;

 /*
  * Output spectrum for index zero frequency, u = -0.5 Hz per sample
  */
  i=0;
  u = (double)(i-so2) * fincr; 
  fCMult(&Y->FFTReal[i],    &Y->FFTImag[i], 
          X->FFTReal[i],     X->FFTImag[i],
          G_EvalReal1D(G,u), G_EvalImag1D(G,u));

 /*
  * Freq index 1 to N/2-1 (negative frequencies except -0.5)
  */
  for (i=1; i < so2; i++) {       
      Y->FFTReal[i] = Y->FFTImag[i] = (float)0.0;
  }

 /*
  * Freq index N/2 to N-1 (DC and positive frequencies)
  */
  for ( ; i < X->size; i++) {  
    u = (double)(i-so2) * fincr; 
    fCMult(&Y->FFTReal[i],    &Y->FFTImag[i], 
            X->FFTReal[i],     X->FFTImag[i],
            G_EvalReal1D(G,u), G_EvalImag1D(G,u));
  }

 /*
  * Compute the image domain response by inverse fft
  */
  fft1d(Y->Real,Y->Imag,Y->FFTReal,Y->FFTImag,Y->size,-1);

  return;
} 
#endif
