/*
 * HardyFiltLF1D.c
 *
 *  Function convolves hardy input sequence x with a specified 1D gabor filter 
 *  by multiplication in the frequency domain.  The space domain and spectral
 *  representations of Y are computed.
 *
 *  This routine compensates for edge effects in very low frequency filters
 *  by making sure that the signal is at least 3 NSIG times the filter space
 *  constant in length.  See book XI, p. 22.  If necessary, the signal is
 *  extended by padding and reflection.
 *
 *  The possibly extended signal is filtered using the routine
 *  HardyConvolveGabor1D.
 *
 *  jph 2/16/96
 */

#ifndef HARDYFILTLF1D_C
#define HARDYFILTLF1D_C
#define NSIG 0
#include <stdio.h>
#include "teager2D.h"
#include "Esa1D.h"

void HardyFiltLF1D(Y,X,G,ZP,Reflect)

  pFloatImage   Y;            /* output image */
  pFloatImage   X;            /* input image */
  FilterPtr     G;            /* pointer to the Gabor filter */
  int           ZP;           /* 1 is zero padding is in use */
  int           Reflect;      /* 1 if padded areas filled by reflection */

{
  int           OrigSize;     /* original size of signal */
  int           NewSize;      /* size of extended signal */
  int           CopySize;     /* size of block transfers in data copy */
  int           k;            /* cursor and counter */
  int           j;            /* cursor and counter */
  int           P;            /* 2^P >=  NSIG sigma, book XI, p. 22 */
  int           M;            /* 2^M >= 3*NSIG sigma, book XI, p. 22 */
  int           Q;            /* Q = 2^M - 3 * 2^P, book XI, p. 22 */
  double        Log2;         /* log base e of 2 */
  double        TwoToP;       /* 2^P */
  float        *srcR;         /* source pointer for Real data copy */
  float        *dstR;         /* destination pointer for Real data copy */
  float        *srcI;         /* source pointer for Imag data copy */
  float        *dstI;         /* destination pointer for Imag data copy */
  pFloatImage   XX;           /* extended input image */
  pFloatImage   YY;           /* extended output image */
  
 /*
  * If signal long enough, no extension required
  */
  if (((double)X->size > (double)3.0 * (double)NSIG * G->sigma) 
                                            || (!ZP) || (!Reflect)) {
    HardyConvolveGabor1D(Y,X,G);
    return;
  }

 /*
  * Signal too short.  Must extend it.
  */
  OrigSize = X->size >> 1;
  Log2 = log((double)2.0);
  P = (int)ceil(log((double)NSIG * G->sigma / (double)OrigSize) / Log2);
  TwoToP = pow((double)2.0,(double)P);
  M = (int)ceil(log((double)3.0 * TwoToP * OrigSize)  / Log2);
  NewSize = pow((double)2.0,(double)M);
  Q = NewSize - (double)3.0 * TwoToP * OrigSize; 

 /*
  * Allocate the extended input and output signals
  */
  XX = (pFloatImage)malloc(sizeof(FloatImage));
  YY = (pFloatImage)malloc(sizeof(FloatImage));
  AllocFloatSequence(XX,NewSize);
  AllocFloatSequence(YY,NewSize);

 /*
  * Copy signal data into extended input signal
  */
  CopySize = X->size;
  for (k=0,srcR=X->Real,srcI=X->Imag,dstR=XX->Real,dstI=XX->Imag;
       k+CopySize < NewSize; dstR+=CopySize,dstI+=CopySize,k+=CopySize) {
    memcpy(dstR,srcR,CopySize*sizeof(float));
    memcpy(dstI,srcI,CopySize*sizeof(float));
  }
  for(j=0; k < NewSize; j++,k++) {
    XX->Real[k] = X->Real[j];
    XX->Imag[k] = X->Imag[j];
  }

 /*
  * Compute the spectrum of the extended input signal
  */
  fft1d(XX->Real,XX->Imag,XX->FFTReal,XX->FFTImag,XX->size,1);

 /*
  * Convolve
  */
  HardyConvolveGabor1D(YY,XX,G);

 /*
  * Copy the result from YY to Y
  */
  for (k=0,j=(int)TwoToP * OrigSize; k < X->size; k++,j++) {
    Y->Real[k] = YY->Real[j];
    Y->Imag[k] = YY->Imag[j];
  }

 /*
  * Deallocate XX and YY
  */
  XX = KillFloat1D(XX);
  YY = KillFloat1D(YY);

  return;
}
#endif
