/*
 * PtEsa1DNoFilt.c
 *
 *  Function evaluates the 1D Unfiltered ESA at one point in a 1D signal.
 *
 *  - This version uses the COMBINED ALGORITHM:  see book VII p. 49
 *
 *  - Edge conditions: see book XI, p. 18.
 *
 *  - The amplitude estimate is returned in "a" and the frequency estimate 
 *    is returned in "u".  u_valid is one if the corresponding 
 *    frequency estimate is valid, and zero otherwise (e.g. if there is a 
 *    zero denominator in the esa).  
 *
 *  - Edge conditions are handled by reflection.
 *   
 *  jph 2/2/96
 */

#ifndef PTESA1DNOFILT_C
#define PTESA1DNOFILT_C
#include <stdio.h>
#include "teager2D.h"
#include "Esa1D.h"

#define FLOAT_TOO_SMALL ((float)0.00001)

void PtEsa1DNoFilt(t,i,u,a,u_valid,a_valid)

  pFloatImage   t;                  /* response image */
  int           i;                  /* domain pt at which to evaluate esa */
  float        *u;                  /* estimated inst freq */
  float        *a;                  /* estimated amplitude */
  BYTE         *u_valid;            /* 1 if u is valid */
  BYTE         *a_valid;            /* 1 if a is valid */
{

  int           j,k,m;              /* cursors for indexing the image */
  float         ahat;               /* magnitude of response image */
  float         MagDenom;           /* magnitude of esa denominator */
  double        one_on_MagDenom;    /* one over magnitude of esa denominator */
  double        ArgDenom;           /* argument of esa denominator */
  float         uMagNumer;          /* magnitude of esa Numerator for u calc */
  double        uArgNumer;          /* argument of esa Numerator for u calc */
  double        uCos;               /* computed cosine of u */
  double        uSin;               /* computed sine of u */
  double        sgn_uSin;           /* sign of Sine of u */
  float         A,B;                /* numerator terms from u calc */

 /*
  * Handle Edge Conditions
  */
  if (i == t->size-1) {
    j = t->size - 1;
    m = t->size - 2;
    k = t->size - 3;
  }
  else {
    if (i == 0) {
      j = 2;
      m = 1;             /* NOTE: j > k */
      k = 0;
    }
    else {
      j = i + 1;
      m = i;
      k = i - 1;
    }
  }

 /*
  * Evaluate the esa
  */
  ahat = FNorm(t->Real[i],t->Imag[i]);
  MagDenom = FNorm(t->Real[m],t->Imag[m]); 
  if ((ahat < FLOAT_TOO_SMALL) || (MagDenom < FLOAT_TOO_SMALL)) { 
    *a = (float)0.0;
    *u = (float)0.0;      /* Re & Im were 0 to within float prcsn */
    *a_valid = 0;
    *u_valid = 0;
    return;
  }
  one_on_MagDenom = (double)1.0 / MagDenom;
  ArgDenom = atan2((double)t->Imag[m],(double)t->Real[m]);

 /*
  * Find uCos:  This block computes the value by Cos Alg.
  */
  A = (float)0.5 * (t->Real[j] + t->Real[k]); 
  B = (float)0.5 * (t->Imag[j] + t->Imag[k]); 
  uMagNumer = FNorm(A,B); 
  if (uMagNumer < FLOAT_TOO_SMALL) { /* A & B both 0 to within precision */
    *u = (float) 0.0;
    *u_valid = 0; 
  } 
  else { 
    uArgNumer = atan2((double)B, (double)A); 
    uCos = (double)uMagNumer * one_on_MagDenom * cos(uArgNumer - ArgDenom); 
    if (uCos < (double)-1.0) { 
      uCos = (double) -1.0; 
    } 
    else { 
      if (uCos > (double)1.0) { 
        uCos = (double)1.0; 
      } 
    } 
    *u = (float) 0.5 * acos(uCos) / (double)M_PI; 
    *u_valid = 1; 
  } 

 /*
  * Find uSin:  This block corrects the SIGN using Sine Alg.
  */
  A = (float)0.5 * (t->Real[k] - t->Real[j]);
  B = (float)0.5 * (t->Imag[j] - t->Imag[k]);
  uMagNumer = FNorm(A,B);
  if (uMagNumer < FLOAT_TOO_SMALL) {     /* Sine alg numerator zero */
    if (!u_valid) {
      *u = (float) 0.0;
      *u_valid = 0; 
    }
  } 
  else {
    uArgNumer = atan2((double)A, (double)B);
    uSin = (double)uMagNumer * one_on_MagDenom * cos(uArgNumer - ArgDenom);
    if (uSin < (double)-1.0) {
      uSin = (double) -1.0;
    }
    else {
      if (uSin > (double)1.0) {
        uSin = (double)1.0;
      }
    }
    if (!u_valid) {
      *u = (float) 0.5 * asin(uSin) / (double)M_PI;
      *u_valid = 1;
    } else {
      if (asin(uSin) < (double)0.0) sgn_uSin = (double) -1.0;
      else                          sgn_uSin = (double)  1.0;
      *u *= sgn_uSin;
    }
  }

 /*
  * Find uSin: This block computes the value by Sine Alg.
  */
/*  A = (float)0.5 * (t->Real[k] - t->Real[j]);*/
/*  B = (float)0.5 * (t->Imag[j] - t->Imag[k]);*/
/*  uMagNumer = FNorm(A,B);*/
/*  if (uMagNumer < FLOAT_TOO_SMALL) {*/     /* Sine alg numerator zero */
/*    *u = (float) 0.0;*/
/*    *u_valid = 0; */
/*  } */
/*  else { */
/*    uArgNumer = atan2((double)A, (double)B);*/
/*    uSin = (double)uMagNumer * one_on_MagDenom * cos(uArgNumer - ArgDenom);*/
/*    if (uSin < (double)-1.0) {*/
/*      uSin = (double) -1.0;*/
/*    }*/
/*    else {*/
/*      if (uSin > (double)1.0) {*/
/*        uSin = (double)1.0;*/
/*      }*/
/*    }*/
/*    *u = (float) 0.5 * asin(uSin) / (double)M_PI; */
/*    *u_valid = 1; */
/*  } */

 /*
  * Estimate a
  */
  if (*u_valid) {
      *a = ahat;
      *a_valid = 1;
  }   /* if u_valid */
  else {              /* u and/or v not valid, can't  */
    *a = (float)0.0;  /* estimate a here using this   */
    *a_valid = 0;     /* response image.              */
  }  /* else */

  return;
} 
#endif
