/*
 * PtEsa1D_clip.c
 *
 *  Function evaluates the 1D ICIP ESA at one point in a 1D signal.
 *
 *  IN THIS VERSION, if amp > 1, it is clipped and "invalid" is returned.
 *
 *  - This version uses the COMBINED ALGORITHM:  see book VII p. 49
 *
 *  - The magnitude of the response image is divided by the max 
 *    response of the filter. This is compared to the filter selection 
 *    criterion value.  If less, zero is returned and no further action is 
 *    taken (values of all the other parameters are preserved).
 *  - Otherwise, the esa is computed and one is returned.  The amplitude 
 *    estimate is returned in "a" and the frequency estimates are returned 
 *    in "u" and "v".  u_valid and v_valid are one if the corresponding 
 *    frequency estimate is valid, and zero otherwise (e.g. if there is a 
 *    zero denominator in the esa).  The filter selection criterion is updated 
 *    to reflect the new value.
 *  - Edge conditions are handled by reflection.
 *   
 *  jph 9/24/94
 */

#ifndef PTESA1D_CLIP_C
#define PTESA1D_CLIP_C
#include <stdio.h>
#include "teager2D.h"
#include "Esa1D.h"

#define FLOAT_TOO_SMALL ((float)0.00001)

int PtEsa1D_clip(t,Asm,col,G,FilterSelCriterion,u,a,u_valid,a_valid)

  pFloatImage   t;                  /* response image */
  pFloatImage   Asm;                /* Postfiltered Amp Response */
  int           col;                /* domain pt at which to evaluate esa */
  FilterPtr     G;                  /* pointer to the Gabor filter */
  float        *FilterSelCriterion; /* amp / max filt resp. must exceed this */
  float        *u;                  /* estimated inst freq */
  float        *a;                  /* estimated amplitude */
  BYTE         *u_valid;            /* 1 if u is valid */
  BYTE         *a_valid;            /* 1 if a is valid */
{

  int           i,j,k;              /* cursors for indexing the image */
  float         ahat;               /* magnitude of response image */
  float         criterion;          /* filter selection criterion value */
  float         MagDenom;           /* magnitude of esa denominator */
  double        one_on_MagDenom;    /* one over magnitude of esa denominator */
  double        ArgDenom;           /* argument of esa denominator */
  float         uMagNumer;          /* magnitude of esa Numerator for u calc */
  double        uArgNumer;          /* argument of esa Numerator for u calc */
  double        uCos;               /* computed cosine of u */
  double        uSin;               /* computed sine of u */
  double        sgn_uSin;           /* sign of Sine of u */
  float         A,B;                /* numerator terms from u calc */
  float         E,F;                /* numerator terms from v calc */
  float         FilterMag;          /* magnitude of filter at estimated u,v */
  BYTE          A_valid;            /* temp placeholder for a_valid value */
  BYTE          U_valid;            /* temp placeholder for u_valid value */
  float         Amp;                /* temp placeholder for a */
  float         U;                  /* temp placeholder for u */
  float         Criterion;          /* temp placeholder for FilterSelCriterion*/

 /*
  * Evaluate the filter selection criterion
  */
  A_valid = U_valid = Amp = U = (float)0.0;
  i = col;
  ahat = MagDenom = fNorm(t->Real[i],t->Imag[i]);
  criterion = ahat / G->GMax;
  if ((criterion <= *FilterSelCriterion) || (criterion > (float)1.0)) {
    return(0);
  }

 /*
  * Criterion was passed, evaluate the esa
  */
  if (MagDenom < FLOAT_TOO_SMALL) { /* Re & Im were 0 to within float prcsn */
    Amp = (float)0.0;
    U = (float)0.0; 
    A_valid = 0;
    U_valid = 0;
    return(1);
  }
  one_on_MagDenom = (double)1.0 / MagDenom;
  ArgDenom = atan2((double)t->Imag[i],(double)t->Real[i]);

 /*
  * Estimate u
  */
  if (col == t->size-1) {
    j = i;
    k = i - 2;
  }
  else {     
    if (col == 0) {
      k = i;             /* NOTE: j > k */
      j = i+2;
    }
    else {
      k = i - 1;
      j = i + 1; 
    }
  }
 /*
  * Find uCos
  */
  A = (float)0.5 * (t->Real[j] + t->Real[k]);
  B = (float)0.5 * (t->Imag[j] + t->Imag[k]);
  uMagNumer = fNorm(A,B);
  if (uMagNumer < FLOAT_TOO_SMALL) { /* A & B both 0 to within precision */
    U = (float) 0.25;               /* implies freq = pi/2.  book VII p. 2 */
    U_valid = 1;
  }
  else {
    uArgNumer = atan2((double)B, (double)A);
    uCos = (double)uMagNumer * one_on_MagDenom * cos(uArgNumer - ArgDenom);
   if (uCos < (double)-1.0) {
      uCos = (double) -1.0;
    }
    else {
      if (uCos > (double)1.0) {
        uCos = (double)1.0;
      }
    }
    U = (float) 0.5 * acos(uCos) / (double)M_PI;
    U_valid = 1;
  }

 /*
  * Find uSin
  */
/*  A = (float)0.5 * (t->Real[k] - t->Real[j]);*/
/*  B = (float)0.5 * (t->Imag[j] - t->Imag[k]);*/
/*  uMagNumer = fNorm(A,B);*/
/*  if (uMagNumer != (float)0.0) {   */  /* Sine alg numerator not zero */
/*    uArgNumer = atan2((double)A, (double)B);*/
/*    uSin = (double)uMagNumer * one_on_MagDenom * cos(uArgNumer - ArgDenom);*/
/*    if (uSin < (double)-1.0) {*/
/*      uSin = (double) -1.0;*/
/*    }*/
/*    else {*/
/*      if (uSin > (double)1.0) {*/
/*        uSin = (double)1.0;*/
/*      }*/
/*    }*/
/*    if (asin(uSin) < (double)0.0) sgn_uSin = (double) -1.0;*/
/*    else                          sgn_uSin = (double)  1.0;*/
/*    *u *= sgn_uSin;*/
/*  }*/

 /*
  * Estimate a
  */
  if (U_valid) {
    FilterMag = G_EvalMag1D(G,(double)(*u));
    if (((FilterMag / G->GMax) > (float)0.000001)) {   /* was 0.000001 */
      Amp = ahat / FilterMag;
      if (Amp <= (float)1.0) {
        A_valid = 1;
      } else {
        A_valid = Amp = (float)0.0;
      }
    }    /* if FilterMag */
    else {                        /* estimated freq's occur where the filter */
      Amp = U = (float)0.0;       /* is effectively zero. u estimate must */
      A_valid = 0;               /* have been derived from noise. */
    } /* else */
  }   /* if U_valid */
  else {              /* u and/or v not valid, can't  */
    Amp = (float)0.0;  /* estimate a here using this   */
    A_valid = 0;     /* response image.              */
  }  /* else */

 /*
  * Return estimated values if good 
  */
  if (A_valid && U_valid) {
    *a_valid = A_valid;
    *u_valid = U_valid;
    *a = Amp;
    *u = U;
    *FilterSelCriterion = criterion;
  }
  return(1);
} 
#endif
