/*
 * TKfindQ.c
 *
 *  function computes the sequence Q given by Eq. 2.33 on page 36 of my
 *  dissertation.  Edge effects are handled by replication.
 *
 *  Ouput float array Phisk is the TKEO output for the input float sequence
 *  which will be used in the amplitude algorithms given by Eq. 2.32 on page
 *  36 of dissertation.
 *
 *  Output BYTE array Qvalid is "1" where Q is valid and "0" where the
 *  denominator was zero.
 *
 *  The output arrays are allocated on the heap.
 *
 *  jph 10/29/98
 *
 */

#ifndef TKFINDQ_C
#define TKFINDQ_C
#include "teager2D.h"
#include "Esa1D.h"
#include <stdio.h>

float *TKfindQ(sk,size,oPhisk,oQvalid)

  float      *sk;        /* input float signal */
  int         size;      /* size of x array */
  float     **oPhisk;    /* TKEO output for sk */
  BYTE      **oQvalid;   /* 1 if Q is valid */

{
  int         i;         /* loop counter */
  float      *Q;         /* signal Q in Eq. 2.33, pg. 36 of dissertation */
  float      *sknp1;     /* input shifted left by one sample */
  float      *Dskn;      /* Dsk(n) in Eq. 2.33 on pg. 36 of dissertation */
  float      *Dsknp1;    /* Dsk(n+1) in Eq. 2.33 on pg. 36 of dissertation */
  float      *PhiDskn;   /* TKEO output for Dskn */
  float      *PhiDsknp1; /* TKEO output for Dsknp1 */
  float      *Phisk;     /* TKEO output for sk */
  BYTE       *Qvalid;    /* 1 if Q is valid */

 /*
  * Allocate the output and working arrays on the heap
  */
  if ((Q = (float *)malloc(size*sizeof(float))) == NULL) {
    printf("\n\nfTKEO1D: free store exhausted.\n");
    exit(-1);
  }
  if ((*oQvalid = (BYTE *)malloc(size*sizeof(BYTE))) == NULL) {
    printf("\n\nfTKEO1D: free store exhausted.\n");
    exit(-1);
  }
  Qvalid = *oQvalid;

 /*
  * Compute shifted sequence.
  */
  sknp1 = fShiftLeft1D(sk,size);

 /*
  * Compute backward difference sequences.
  */
  Dskn = fBackDiff1D(sk,size);
  Dsknp1 = fBackDiff1D(sknp1,size);

 /*
  * Invoke the TKEO.
  */
  *oPhisk = fTKEO1D(sk,size);
  Phisk = *oPhisk;
  PhiDskn = fTKEO1D(Dskn,size);
  PhiDsknp1 = fTKEO1D(Dsknp1,size);

 /*
  * Compute Q.
  */
  for (i=1; i < size-1; i++) {
    if (Phisk[i] == (float)0.0) {
      Q[i] = (float)0.0;
      Qvalid[i] = 0;
    } else {
      Q[i] = (float)1.0 - (PhiDskn[i] + PhiDsknp1[i]) / ((float)4.0 * Phisk[i]);
      Qvalid[i] = 1;
    }
  }

 /*
  * Handle the end points by replication.
  */
  Q[0] = Q[1];
  Qvalid[0] = Qvalid[1];
  Q[size-1] = Q[size-2];
  Qvalid[size-1] = Qvalid[size-2];

 /*
  * free working arrays
  */
  free(sknp1);
  free(Dskn);
  free(Dsknp1);
  free(PhiDskn);
  free(PhiDsknp1);


  return(Q);

}
#endif
