/*
 * fTKESA1D.c
 *
 *  function implements the 1D discrete Teager-Kaiser ESA shown in
 *  Eq. 2.31 and 2.32 on page 36 of my dissertation.
 *
 *  x:             input float array, holds the signal.
 *  a:             output float array, holds amplitude estimates.
 *  a_valid:       output BYTE array, 1 where a is valid.
 *  phi_dot:       output float array, holds frequency estimates.
 *  phi_dot_valid: output bye array, 1 where phi_dot is valid.
 *
 *  The output arrays are allocated on the heap.
 *
 *  jph 10/30/98
 *
 */

#ifndef FTKESA1D_C
#define FTKESA1D_C
#include "teager2D.h"
#include "Esa1D.h"
#include <stdio.h>

void fTKESA1D(x,size,oa,oa_valid,ophi_dot,ophi_dot_valid)

  float      *x;             /* input float signal */
  int         size;          /* size of x array */
  float     **oa;             /* amplitude estimates */
  BYTE      **oa_valid;       /* 1 where a is valid */
  float     **ophi_dot;       /* frequency estimates */
  BYTE      **ophi_dot_valid; /* 1 where phi_dot is valid */

{
  int         i;         /* loop counter */
  float      *Phisk;     /* TKEO output for input signal x */
  BYTE       *Qvalid;    /* 1 where Q is valid */
  float      *Q;         /* signal Q in Eq. 2.33, pg. 36 of dissertation */
  float      *a;             /* amplitude estimates */
  BYTE       *a_valid;       /* 1 where a is valid */
  float      *phi_dot;       /* frequency estimates */
  BYTE       *phi_dot_valid; /* 1 where phi_dot is valid */

 /*
  * Allocate the output arrays on the heap
  */
  if ((*oa = (float *)malloc(size*sizeof(float))) == NULL) {
    printf("\n\nfTKEO1D: free store exhausted.\n");
    exit(-1);
  }
  a = *oa;
  if ((*ophi_dot = (float *)malloc(size*sizeof(float))) == NULL) {
    printf("\n\nfTKEO1D: free store exhausted.\n");
    exit(-1);
  }
  phi_dot = *ophi_dot;
  if ((*oa_valid = (BYTE *)malloc(size*sizeof(BYTE))) == NULL) {
    printf("\n\nfTKEO1D: free store exhausted.\n");
    exit(-1);
  }
  a_valid = *oa_valid;
  if ((*ophi_dot_valid = (BYTE *)malloc(size*sizeof(BYTE))) == NULL) {
    printf("\n\nfTKEO1D: free store exhausted.\n");
    exit(-1);
  }
  phi_dot_valid = *ophi_dot_valid;

 /*
  * Call TKfindQ to compute the Q array.
  */
  Q = TKfindQ(x,size,&Phisk,&Qvalid);

 /*
  * Loop over samples.  Estimate amplitude and frequency.
  */
  for (i=0; i < size; i++) {
    phi_dot_valid[i] = a_valid[i] = Qvalid[i];

   /*
    * Check for erroneous values of Q.
    */
    if (Q[i] < (float)-1.0) {
      Q[i] = (float)-1.0;
      phi_dot_valid[i] = a_valid[i] = 0;
    } else {
      if (Q[i] > (float)1.0) {
        Q[i] = (float)1.0;
        phi_dot_valid[i] = a_valid[i] = 0;
      }
    }

   /*
    * Check for erroneous values of Phisk.
    */
    if (Phisk[i] < (float)0.0) {
      Phisk[i] = (float)0.0;
      phi_dot_valid[i] = 0;
    }

   /*
    * Estimate frequency.
    */
    phi_dot[i] = (float)acos((double)Q[i]);

   /*
    * Estimate amplitude.
    */
    if (fabs((double)Q[i] == (double)1.0)) {
      a_valid[i] = 0;
      a[i] = (float)0.0;
    } else {
      a[i] = (float)sqrt((double)(Phisk[i]/((float)1.0 - Q[i]*Q[i])));
    }
  }   /* for i */

}
#endif
