/*
 * fft1d.c
 *
 *  Function implements the 1D FFT.  The Numerical Recipes routine "four1"
 *  is called.  The frequency domain coordinates are re-arranged to go from
 *  -0.5 cycles/sample to 0.5 cycles/sample as the index ranges from 0 to N-1.
 *  See book VIII p. 101 for an explanation of the indexing.
 *
 *    The signal length must be a power of two.
 *
 *  FOR THIS FILE, WE NEVER FACTOR THE "n" OUT OF THE SPECTRUM.
 *   - after forward transform, spectral data has the "n" factor in it
 *   - after inverse transform, "n" IS FACTORED OUT of spacial data
 *
 */

#ifndef FFT1D_C
#define FFT1D_C
#include "teager2D.h"
#include "Esa1D.h"
#include <stdio.h>

void fft1d(xReal,xImag,XReal,XImag,size,direction)

  float *xReal;       /* real part of signal */
  float *xImag;       /* imaginary part of signal */
  float *XReal;       /* real part of spectrum */
  float *XImag;       /* imaginary part of spectrum */
  int    size;        /* length of signal */
  int    direction;   /* 1=fwd xform, -1=inverse xform */

{

    float *pass;                  /* array for passing data to/from fourn */
    float one_on_n;               /* 1 / (no. data points in signal) */
    int i;                        /* counter */
    int n;                        /* no. data points in signal */
    int pass_size;                /* number of samples in pass array, twice n */
    int so2;                      /* size over 2 */

    n = size;
    so2 = (size >> 1);
    pass_size = (size << 1);

    if (direction == 1 ) { /* forward transform */
      pass  = (float *) malloc(pass_size*sizeof(float));
     /*
      * Set up the data in the pass array and call fourn
      */
      for (i=0; i<n; i++) {
        pass[2*i] = xReal[i];
        pass[2*i+1] = xImag[i];
      }
      four1(pass-1,size,1);

     /*
      * Set up spectrum in easy to use form:
      *   XReal[0],   XImag[0]   <=> f = -1/2
      *   XReal[N/2], XImag[N/2] <=> f = 0
      *   XReal[N-1], XImag[N-1] <=> f = (N/2-1)/N
      * see book VIII p. 101 for an explanation of the indexing 
      */
      for (i=0; i<so2; i++) {
        XReal[i] = pass[(i<<1) + n];
        XImag[i] = pass[(i<<1) + n + 1];
        XReal[i+so2] = pass[i<<1];
        XImag[i+so2] = pass[(i<<1) + 1];
      }
    }
    else {
      if (direction == -1) { /* reverse transform */ 
        one_on_n = (float)1.0 / (float)n;
        pass  = (float *) malloc(pass_size*sizeof(float));
       /*
        * Set up the data in the pass array and call fourn.
        * see book VIII p. 101 for an explanation of the indexing
        */
        for (i=0; i<so2; i++) {
          pass[(i<<1) + n]     = XReal[i];
          pass[(i<<1) + n + 1] = XImag[i];
          pass[i<<1]           = XReal[i+so2];
          pass[(i<<1) + 1]     = XImag[i+so2];
        }
        four1(pass-1,size,-1);
        for (i=0; i<n; i++) {
          xReal[i] = pass[i<<1]     * one_on_n;
          xImag[i] = pass[(i<<1)+1] * one_on_n;
        }
      }
      else {
        printf("\nERROR: fft2d: unknown value %d specified for direction.\n",
                 direction);
        return;
      }
    }
    free(pass);
    return;

}
#endif 
