/*
 * gauss_noise.c
 * 
 *  Function returns a pointer to a float array that contains samples of an
 *   uncorrelated Gaussian noise process.  
 *
 *  Uses rand() to seed if restart_seed == 0.  Otherwise, use restart_seed
 *   as the seed value.
 *
 *  On return, *restart_seed is equal to the seed that was actually used.
 *  
 *   This function uses the Numerical Recipies function Gassdev.
 *   The array of noise samples is malloc'd in this function.
 *
 */

#ifndef GAUSS_NOISE_C
#define GAUSS_NOISE_C

#include <stdlib.h>
#include "teager2D.h"
#include "Esa1D.h"

float *gauss_noise(mean,sigma,size,restart_seed) 

  float	mean;		/* desired mean of the noise process */
  double sigma;		/* desired std deviation of the noise process */
  int   size;		/* number of samples of the noise process */
  int  *restart_seed;   /* seed value to reproduce a noise sequence */
                        /* pass in 0 to seed from rand() */

{
  float *obs;		/* the observations to be returned */
  unsigned  seed;	/* seed value for rand() */
  int       gseed;	/* seed value for gasdev() */
  int t;		/* index into the obs array */

  if ((obs = (float *)malloc(size*sizeof(float))) == NULL) {
    printf("\n\ngauss_noise: free store exhausted.\n");
    exit(-1);
  }

 /*
  * Compute the seed value.
  */
  if (*restart_seed) {
    if (*restart_seed > 0) {
      printf("\n\ngauss_noise: positive seed value %d specified.\n",
             *restart_seed);
      exit(-1);
    }
    gseed = *restart_seed;
  } else {
    srand(time(NULL));
    gseed = -1 * abs(rand());
    while (gseed == 0) {
      gseed = -1 * abs(rand());
    }
  }
  *restart_seed = gseed;

 /*
  * Make random deviates by calling gasdev()
  */
  for (t=0; t<size; t++){
    obs[t] = (float)((double)gasdev(&gseed) * sigma + (double)mean);
  }
 
  return(obs);

}


#endif

   
 
