/*
 * CreateCornerFilters.c
 *
 *  Function creates the corner filters which cover the high spatial 
 *  frequencies in the corners of the frequency plane.
 *  
 *  jph 10/3/96
 *
 */

#ifndef CREATECORNERFILTERS_C
#define CREATECORNERFILTERS_C
#include "teager2D.h"
#include <stdio.h>

void CreateCornerFilters(G)

  filterbank *G;      /* multiband Gabor filter bank */

{
  filter *f;             /* pointer to a new filter */
  double  twopi;         /* two time pi */

 /*
  * Useful factors
  */
  twopi = (double)2.0 * M_PI;

 /*
  * Allocate FIRST coner filter on the heap
  */
  if ((f = (filter *)malloc(sizeof(filter))) == NULL) {
    printf("\n\nCreateCornerFilters: free store exhausted.\n");
    exit(-1);
  }

 /*
  * Assigned filter parameters:  FIRST corner filter
  */
  f->index       = G->TotNumFilters - 2;
  f->Eta         = G->Eta;
  f->gamma       = G->gamma;
  f->isotropicPF = G->isotropicPF;
  f->kappa       = G->kappa;
  f->kappa_x     = G->kappa_x;
  f->kappa_y     = G->kappa_y;
  f->Ray         = G->NumRays;
  f->IndexOnRay  = 0;
  f->Theta       = (double)-0.25 * M_PI;

 /*
  * Computed filter parameters: FIRST corner filter
  *
  *   The radial frequency of the corner of the frequency plane is 1/sqrt(2).  
  *   The corner filter radial center frequency will be half way between this 
  *   max frequency and the radial center frequency of the last ring of 
  *   filters in the regular filter tesselation.
  */
  f->r0 = G->filt[G->NumFiltersPerRay-1]->r0 +
               ((double)1.0 / pow((double)2.0,(double)0.5) 
                - G->filt[G->NumFiltersPerRay-2]->r0) / (double)3.0;

  f->u0 = f->r0 * cos(f->Theta);
  f->v0 = f->r0 * sin(f->Theta);

 /*
  * The space constant of the corner filter will be the geometric mean of
  * the space constants of the filters in the ultimate ring and the penultimate
  * ring of the regular filter tesselation.
  */
  f->sigma = pow(G->filt[G->NumFiltersPerRay-2]->sigma 
               * G->filt[G->NumFiltersPerRay-1]->sigma,(double)0.5);
  f->OBW = log(
        (f->sigma * twopi * f->r0 + pow(-log(G->Eta),(double)0.5)) /
        (f->sigma * twopi * f->r0 - pow(-log(G->Eta),(double)0.5))
      ) / log((double)2.0);
  f->gMax  = (float) ((double)1.0 / (pow(twopi, (double)0.5) * f->sigma));
  f->GMax  = (float)2.0 / f->gMax;
  f->gArgGaussCoef = (double)-0.25 / pow(f->sigma, (double)2.0);
  f->GArgGaussCoef = M_PI * M_PI / f->gArgGaussCoef;

 /*
  * Insert the new corner filter into the filterbank
  */
  G->filt[f->index] = f;

 /*
  * Allocate SECOND coner filter on the heap
  */
  if ((f = (filter *)malloc(sizeof(filter))) == NULL) {
    printf("\n\nCreateCornerFilters: free store exhausted.\n");
    exit(-1);
  }

 /*
  * Assigned filter parameters:  SECOND corner filter
  */
  f->index       = G->TotNumFilters - 1;
  f->Eta         = G->Eta;
  f->gamma       = G->gamma;
  f->isotropicPF = G->isotropicPF;
  f->kappa       = G->kappa;
  f->kappa_x     = G->kappa_x;
  f->kappa_y     = G->kappa_y;
  f->Ray         = G->NumRays;
  f->IndexOnRay  = 0;
  f->Theta       = (double)0.25 * M_PI;

 /*
  * Computed filter parameters:  SECOND corner filter
  */
  f->r0 = G->filt[G->TotNumFilters - 2]->r0;
  f->u0 = f->r0 * cos(f->Theta);
  f->v0 = f->r0 * sin(f->Theta);
  f->sigma = G->filt[G->NumFiltersPerRay-2]->sigma;
  f->OBW = G->filt[G->TotNumFilters - 2]->OBW;
  f->gMax  = G->filt[G->TotNumFilters - 2]->gMax;
  f->GMax  = G->filt[G->TotNumFilters - 2]->GMax;
  f->gArgGaussCoef = G->filt[G->TotNumFilters - 2]->gArgGaussCoef;
  f->GArgGaussCoef = G->filt[G->TotNumFilters - 2]->GArgGaussCoef;

 /*
  * Insert the new corner filter into the filterbank
  */
  G->filt[f->index] = f;

  return;
}
#endif 
