/*
 * CreateFilterBankAux.c
 *
 *  Function creates a multiband Gabor filter bank in a wavelet tesselation.  
 *  A pointer to the new filter bank is returned through the function name.  
 *  This version INCLUDES the DC filter and the corner filters.  The corner
 *  filters cover the high frequencies in the corners of the frequency plane.
 *
 *  The filter design equations are found on pp. 111 - 112 of Book VI.
 *  REVISION:  8/14/94: pp. 82 - 83 book VIII.
 *
 *  NOTE that the InitialRadialFreq parameter to this function is specified in
 *       units of CYCLES PER PIXEL.
 *
 *  REVISION 9/1/96: added field "DCfilt" to filterbank structure.  It is
 *                   set to NULL in this routine.  A separate routine,
 *                   DesignDCfilt does the design for the DC filter.
 *
 *  jph 9/10/93
 *
 */

#ifndef CREATEFILTERBANKAUX_C
#define CREATEFILTERBANKAUX_C
#include "teager2D.h"
#include <stdio.h>

filterbank *CreateFilterBankAux(r0,R,OBW,Eta,isoPF,kappa,kappa_x,kappa_y)

  double      r0;                  /* radial center freq. of the first filter */                                   /*   on each ray in CYCLES per pixel */
  float       R;                   /* common ratio of radii of adjcnt filtrs */
  float       OBW;                 /* Eta-peak octave bandwidth of filters */
  float       Eta;                 /* OBW is defined in terms of Eta-peak. */
  int         isoPF;               /* 1=use isotropic postfilters */
  float       kappa;               /* postfilter sigma scaling factor */
  float       kappa_x;             /* horiz postfilter sigma scaling factor */
  float       kappa_y;             /* vert  postfilter sigma scaling factor */

{

  filterbank *G;                   /* pointer to the new filter bank */
  int         i;                   /* loop counter */

 /*
  * Allocate the filter bank on the heap
  */
  if ((G = (filterbank *)malloc(sizeof(filterbank))) == NULL) {
    printf("\n\nCreateFilterBankAux: free store exhausted.\n");
    exit(-1);
  }

 /*
  * Assiged filter bank parameters
  */
  G->r0    = r0;
  G->R     = R;
  G->OBW   = OBW;
  G->Eta   = Eta;
  G->isotropicPF = isoPF;
  G->kappa = kappa;
  G->kappa_x = kappa_x;
  G->kappa_y = kappa_y;
  G->DCfilt = NULL;

 /*
  * Computed filter bank parameters
  */
  G->NumFiltersPerRay  = 1 + 
            (int)(floor(-log((double)2.0 * G->r0) / log((double)G->R)));

  G->gamma = pow(pow((double)2.0,(double)G->OBW) - (double)1.0, (double)2.0) /
             pow(pow((double)2.0,(double)G->OBW) + (double)1.0, (double)2.0);

  G->Theta0 = -M_PI / (double)2.0;

  G->Lambda = (double)2.0 * asin(
          pow((double)(4.0 * G->R), (double)(-0.5)) *
          pow((pow((double)G->R, (double)2.0) + (double)1.0) * (G->gamma - 1.0)
                 + (double)(2.0 * G->R) * (G->gamma + 1.0), (double)0.5));

  G->NumRays = (int)floor(M_PI / G->Lambda);

 /*
  * Note: "+3" is for the DC filter and the two corner filters.
  */
  G->TotNumFilters      = G->NumFiltersPerRay * G->NumRays + 3;

 /*
  * Create the filters in the regular tesselation
  */
  if ((G->filt = (FilterPtr *)malloc((G->TotNumFilters+3) * sizeof(FilterPtr))) 
      == NULL) {
    printf("\n\nCreateFilterBankAux: free store exhausted.\n");
    exit(-1);
  }
  for (i=0; i<G->TotNumFilters-3; i++) {
    G->filt[i] = CreateGaborFilter(G,i);
  }

 /*
  * Create the DC filter
  */
  G->filt[G->TotNumFilters-3] = G->DCfilt = DesignDCfilt(G);
  G->DCfilt->index = G->TotNumFilters - 3;

 /*
  * Create the corner filters
  */
  CreateCornerFilters(G);

  return(G);
}
#endif 
