/*
 * CreateGaborFilter.c
 *
 *  Function creates the m'th filter in a multiband Gabor filter bank.  
 *  A pointer to the new filter is returned through the function name.  
 *
 *  The filter design equations are found on pp. 111 - 112 of Book VI.
 *  REVISED 8/14/94.  See book p. 82 - 83 book VIII.
 *
 *  jph 9/10/93
 *
 */

#ifndef CREATEGABORFILTER_C
#define CREATEGABORFILTER_C
#include "teager2D.h"
#include <stdio.h>

FilterPtr CreateGaborFilter(G,m)

  filterbank *G;      /* multiband Gabor filter bank */
  int         m;      /* index in the bank of filter to create */

{
  filter *f;             /* pointer to the new filter */
  double  twopi;         /* two time pi */

 /*
  * Allocate the filter on the heap
  */
  if ((f = (filter *)malloc(sizeof(filter))) == NULL) {
    printf("\n\nCreateGaborFilter: free store exhausted.\n");
    exit(-1);
  }

 /*
  * Useful factors
  */
  twopi = (double)2.0 * M_PI;

 /*
  * Assigned filter parameters
  */
  f->index = m;
  f->Eta   = G->Eta;
  f->OBW   = G->OBW;
  f->gamma = G->gamma;
  f->isotropicPF = G->isotropicPF;
  f->kappa = G->kappa;
  f->kappa_x = G->kappa_x;
  f->kappa_y = G->kappa_y;

 /*
  * Computed filter parameters
  */
  f->Ray        = (int)floor((double)m / (double)G->NumFiltersPerRay); 
  f->IndexOnRay = m % G->NumFiltersPerRay;
  f->Theta      = G->Lambda * ((double)f->Ray + (double)0.5) 
                            - (double)0.5 * M_PI;
  if (f->Theta > twopi) {
    f->Theta -= twopi;
  }

  f->r0 = G->r0 * pow((double)G->R, (double)f->IndexOnRay);
  f->u0 = f->r0 * cos(f->Theta);
  f->v0 = f->r0 * sin(f->Theta);

  f->sigma = pow(-log((double)f->Eta) ,(double)0.5) 
             / (twopi * f->r0 * pow(f->gamma, (double)0.5));

  f->gMax  = (float) ((double)1.0 / (pow(twopi, (double)0.5) * f->sigma));
  f->GMax  = (float)2.0 / f->gMax;

  f->gArgGaussCoef = (double)-0.25 / pow(f->sigma, (double)2.0);
  f->GArgGaussCoef = M_PI * M_PI / f->gArgGaussCoef;

  return(f);
}
#endif 
