/*
 * HardyConvolveGabor.c
 *
 *  Function convolves hardy input image x with a specified gabor filter by
 *  multiplication in the frequency domain.  The image domain and spectrum
 *  representations of Y are computed.
 *
 *  On invocation, it is assumed that the X and Y images have the same size and 
 *  have been fully allocated, and that the X image has been fully 
 *  initialized.  It is also assumed that the size field of the Y image 
 *  has been initialized.
 *
 *  jph 9/13/93
 *
 *  Major Revision 7/11/96: now agrees w/ analytic spectrum as in dissertation.
 */

#ifndef HARDYCONVOLVEGABOR_C
#define HARDYCONVOLVEGABOR_C
#include <stdio.h>
#include "teager2D.h"

void HardyConvolveGabor(Y,X,G)

  pFloatImage   Y;  /* output image */
  pFloatImage   X;  /* output image */
  FilterPtr     G;  /* pointer to the Gabor filter */

{

  double      u;       /* horizontal freq in cycles per pixel */ 
  double      v;       /* horizontal freq in cycles per pixel */ 
  double      fincr;   /* incr in freq for one row or one col in the image */
  int         row;     /* counter for image rows */
  int         col;     /* counter for image cols */
  int         k;       /* array offset index */
  int         looplim; /* terminal/initial value for loop counters */
  int         so2;     /* size over 2 */

 /*
  * Useful values
  */
  fincr = (double)1.0 / (double)X->size;
  so2   = X->size >> 1;
  for (k=0; k < X->size * X->size; k++) {
    Y->FFTReal[k] = Y->FFTImag[k] = (float)0.0;
  }

 /*
  * First row of spectrum
  */
  k=row=col=0;
  v = (double)(row-so2) * fincr;
  u = (double)(col-so2) * fincr;
  fCMult(&Y->FFTReal[k],    &Y->FFTImag[k],
          X->FFTReal[k],     X->FFTImag[k],
          G_EvalReal(G,u,v), G_EvalImag(G,u,v));
  for (col=so2; col < X->size; col++) {
    u = (double)(col-so2) * fincr;
    k = col;
    fCMult(&Y->FFTReal[k],    &Y->FFTImag[k],
            X->FFTReal[k],     X->FFTImag[k],
            G_EvalReal(G,u,v), G_EvalImag(G,u,v));
  }

 /*
  * Rows 1 to so2-1
  */
  for (row=1; row < so2; row++) {
    v = (double)(row-so2) * fincr;
    k = row * X->size;
    for (col=so2+1; col < X->size; col++) {
      u = (double)(col-so2) * fincr;
      fCMult(&Y->FFTReal[k+col],&Y->FFTImag[k+col],
              X->FFTReal[k+col], X->FFTImag[k+col],
              G_EvalReal(G,u,v), G_EvalImag(G,u,v));
    }
  }

 /*
  * Row so2
  */
  k = so2 * X->size;
  row = so2;
  col = 0;
  v = (double)(row-so2) * fincr;
  u = (double)(col-so2) * fincr;
  fCMult(&Y->FFTReal[k],    &Y->FFTImag[k],
          X->FFTReal[k],     X->FFTImag[k],
          G_EvalReal(G,u,v), G_EvalImag(G,u,v));
  for (col=so2; col < X->size; col++) {
    u = (double)(col-so2) * fincr;
    fCMult(&Y->FFTReal[k+col],&Y->FFTImag[k+col],
            X->FFTReal[k+col], X->FFTImag[k+col],
            G_EvalReal(G,u,v), G_EvalImag(G,u,v));
  }

 /*
  * Rows so2+1 to size-1
  */
  for (row=so2+1; row < X->size; row++) {
    k = row * X->size;
    col = 0;
    v = (double)(row-so2) * fincr;
    u = (double)(col-so2) * fincr;
    fCMult(&Y->FFTReal[k],     &Y->FFTImag[k],
            X->FFTReal[k],      X->FFTImag[k],
            G_EvalReal(G,u,v), G_EvalImag(G,u,v));
    for (col=so2; col < X->size; col++) {
      u = (double)(col-so2) * fincr;
      fCMult(&Y->FFTReal[k+col],&Y->FFTImag[k+col],
              X->FFTReal[k+col], X->FFTImag[k+col],
              G_EvalReal(G,u,v), G_EvalImag(G,u,v));
    }
  }
      
 /*
  * Compute the image domain response by inverse fft
  */
  fft2d(Y->Real,Y->Imag,Y->FFTReal,Y->FFTImag,Y->size,-1);

  return;
} 
#endif
