/*
 * MkAnisPF.c
 *
 *   Function returns a pointer to an anisotropic Gaussian postfilter image.  
 *   If zerophase=1, the filter has zero phase, which means that the main 
 *   lobe is split across the four corners of the image instead of being 
 *   centered.  Otherwise, if zerophase=0, the filter is centered in the
 *   returned image.
 *   
 *   See book XII, p. 11.
 *
 *   Isotropic version:
 *
 *   See book VI p. 105, book XI, p. 151.
 *   REVISED 8/14/94: see book VIII p. 83.
 *
 *  jph 4/30/96
 *
 */
#ifndef MKANISPF_C
#define MKANISPF_C
#include <stdio.h>
#include "teager2D.h"

pFloatImage MkAnisPF(sigma,kappa_x,kappa_y,Theta,size,zerophase)

  double      sigma;      /* channel filter space constant */
  float       kappa_x;    /* horiz space constant scaling factor */
  float       kappa_y;    /* vert  space constant scaling factor */
  double      Theta;      /* Orientation of filter in radians */
  int         size;       /* row / col dimension of image */
  int         zerophase;  /* 1=split filter across four corners of image */

{

  pFloatImage g;           /* ptr to filter image being created */
  int         n;           /* size * size, the num of pixels in image */
  float       gMax;        /* max val of filter in space domain */
  double      gFact;       /* phase factor for space domain */
  double      OneOnKxsq;   /* 1 / kappa_x squared */
  double      OneOnKysq;   /* 1 / kappa_y squared */
  double      CosTheta;    /* cosine of rotation angle */
  double      SinTheta;    /* sine of rotation angle */
  int         row;         /* counter for image rows */
  int         col;         /* counter for image cols */
  int         i;           /* counter for image pixels */
  int         j;           /* counter for image pixels */
  int         xx;          /* col argument for zerophase */
  int         yy;          /* row argument for zerophase */
  double      xp;          /* col argument in rotated coordinate system */
  double      yp;          /* row argument in rotated coordinate system */
  int         so2;         /* size / 2 */
  float      *g2;          /* real part of filter image, centered */

 /*
  * Allocate the floating image and its constituents
  */
  if ((g = (pFloatImage)malloc(sizeof(FloatImage))) == NULL) {
    printf("\nMkAnisPF: Free Store Exhausted.\n\n");
    exit(-1);
  }
  n = size * size;
  g->size = g->rows = g->cols = size;
  g->square = 1;
  so2 = size >> 1;

  if ((g->Real = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\nMkAnisPF: Free Store Exhausted.\n\n");
    exit(-1);
  }
  if ((g->Imag = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\nMkAnisPF: Free Store Exhausted.\n\n");
    exit(-1);
  }
  if ((g->FFTReal = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\nMkAnisPF: Free Store Exhausted.\n\n");
    exit(-1);
  }
  if ((g->FFTImag = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\nMkAnisPF: Free Store Exhausted.\n\n");
    exit(-1);
  }

 /*
  * Compute Factors 
  */
  gMax = (float)((double)0.25 
                / ((double)M_PI * (double)(kappa_x*kappa_y) * sigma*sigma));
  gFact = -(double)0.25 / (sigma * sigma);
  OneOnKxsq = (double)1.0 / (double)(kappa_x * kappa_x);
  OneOnKysq = (double)1.0 / (double)(kappa_y * kappa_y);
  CosTheta = cos(Theta - (double)M_PI_2);
  SinTheta = sin(Theta - (double)M_PI_2);

 /*
  * Loop for rows and cols, compute filter in space domain
  */
  for (i=0,row=0; row < size; row++) {
    yy = (row < so2) ? row : row - size;
    for (col=0; col < size; col++,i++) {
      xx = (col < so2) ? col : col - size;
     /*
      * Rotate coordinates to orient principle axes w/ Theta
      */
      xp = (double)xx * CosTheta + (double)yy * SinTheta;
      yp = (double)yy * CosTheta - (double)xx * SinTheta;

      g->Real[i] = gMax*(float)exp(gFact*(xp*xp*OneOnKxsq +  yp*yp*OneOnKysq));
      g->Imag[i] = (float)0.0;
    }
  }

 /*
  * If not zerophase, switch quadrants to center image
  */
  if (!zerophase) {
    if ((g2 = (float *)malloc(n * sizeof(float))) == NULL) {
      printf("\nMkAnisPF: Free Store Exhausted.\n\n");
      exit(-1);
    }

    for (i=0; i < so2; i++) {
      for (j=0; j < so2; j++) {
        g2[i*size + j] = g->Real[(i+so2)*size + j + so2]; /* I   <- IV  */
        g2[i*size + j + so2] = g->Real[(i+so2)*size + j]; /* II  <- III */
        g2[(i+so2)*size + j] = g->Real[i*size + j + so2]; /* III <- II  */
        g2[(i+so2)*size + j + so2] = g->Real[i*size + j]; /* IV  <- I   */
      }
    }

    free(g->Real);
    g->Real = g2;
  }

 /*
  * Compute filter in freq domain by FFT
  */
  fft2d(g->Real,g->Imag,g->FFTReal,g->FFTImag,size,1);

  return(g);
}
#endif
