/*
 * MkPostFilt.c
 *
 *   Function returns a pointer to a Gaussian postfilter image.  The filter 
 *   has zero phase, which means that the main lobe is split across the four 
 *   corners of the image instead of being centered.
 *   See book VI p. 105, book XI, p. 151.
 *   REVISED 8/14/94: see book VIII p. 83.
 *
 *  jph 4/14/94
 *
 */
#ifndef MKPOSTFILT_C
#define MKPOSTFILT_C
#include <stdio.h>
#include "teager2D.h"

pFloatImage MkPostFilt(sigma,kappa,size)

  double      sigma;  /* channel filter space constant */
  float       kappa;  /* postfilter space constant scaling factor */
  int         size;   /* row / col dimension of image */

{

  pFloatImage g;      /* ptr to filter image being created */
  int         n;      /* size * size, the num of pixels in image */
  float       gMax;   /* max val of filter in space domain */
  double      gFact;  /* phase factor for space domain */
  int         row;    /* counter for image rows */
  int         col;    /* counter for image cols */
  int         i;      /* counter for image pixels */
  int         xx;     /* col argument for zerophase */
  int         yy;     /* row argument for zerophase */
  int         so2;    /* size / 2 */

 /*
  * Allocate the floating image and its constituents
  */
  if ((g = (pFloatImage)malloc(sizeof(FloatImage))) == NULL) {
    printf("\nMkPostFilt: Free Store Exhausted.\n\n");
    exit(-1);
  }
  n = size * size;
  g->size = g->rows = g->cols = size;
  g->square = 1;
  so2 = size >> 1;

  if ((g->Real = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\nMkPostFilt: Free Store Exhausted.\n\n");
    exit(-1);
  }
  if ((g->Imag = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\nMkPostFilt: Free Store Exhausted.\n\n");
    exit(-1);
  }
  if ((g->FFTReal = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\nMkPostFilt: Free Store Exhausted.\n\n");
    exit(-1);
  }
  if ((g->FFTImag = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\nMkPostFilt: Free Store Exhausted.\n\n");
    exit(-1);
  }

 /*
  * Compute Factors 
  */
  gMax = (float)((double)0.25 
                / ((double)M_PI * (double)(kappa*kappa) * sigma*sigma));
  gFact = -(double)M_PI * (double)gMax;

 /*
  * Loop for rows and cols, compute filter in space domain
  */
  for (i=0,row=0; row < size; row++) {
    yy = (row < so2) ? row : row - size;
    for (col=0; col < size; col++,i++) {
      xx = (col < so2) ? col : col - size;
      g->Real[i] = gMax * (float)exp(gFact * (double)(xx*xx + yy*yy));
      g->Imag[i] = (float)0.0;
    }
  }

 /*
  * Compute filter in freq domain by FFT
  */
  fft2d(g->Real,g->Imag,g->FFTReal,g->FFTImag,size,1);

  return(g);
}
#endif
