/*
 * PtEsa2D.c
 *
 *  Function evaluates the asilomar esa at one point in a float image.
 *
 *  - This version uses the COMBINED ALGORITHM:  see book VII p. 49
 *
 *  - The magnitude of the response image is divided by the max 
 *    response of the filter. This is compared to the filter selection 
 *    criterion value.  If less, zero is returned and no further action is 
 *    taken (values of all the other parameters are preserved).
 *  - Otherwise, the esa is computed and one is returned.  The amplitude 
 *    estimate is returned in "a" and the frequency estimates are returned 
 *    in "u" and "v".  u_valid and v_valid are one if the corresponding 
 *    frequency estimate is valid, and zero otherwise (e.g. if there is a 
 *    zero denominator in the esa).  The filter selection criterion is updated 
 *    to reflect the new value.
 *  - Edge conditions are handled by reflection.
 *   
 *  jph 9/13/93
 */

#ifndef PTESA2D_C
#define PTESA2D_C
#include <stdio.h>
#include "teager2D.h"

#define FLOAT_TOO_SMALL ((float)0.00001)

int PtEsa2D(t,row,col,G,FilterSelCriterion,u,v,a,u_valid,v_valid,a_valid)

  pFloatImage   t;                  /* response image */
  int           row;                /* image row for esa */
  int           col;                /* image col for esa */
  FilterPtr     G;                  /* pointer to the Gabor filter */
  float        *FilterSelCriterion; /* amp / max filt resp. must exceed this */
  float        *u;                  /* estimated horiz. freq. component */
  float        *v;                  /* estimated vert.  freq. component */
  float        *a;                  /* estimated amplitude */
  BYTE         *u_valid;            /* 1 if u is valid */
  BYTE         *v_valid;            /* 1 if v is valid */
  BYTE         *a_valid;            /* 1 if a is valid */
{

  int           uk,uj,um;           /* cursors for indexing the image */
  int           vk,vj,vm;           /* cursors for indexing the image */
  int           i;                  /* row*size + col */
  float         ahat;               /* magnitude of response image */
  float         criterion;          /* filter selection criterion value */
  float         uMagDenom;          /* magnitude of esa denominator for u */
  float         vMagDenom;          /* magnitude of esa denominator for v */
  double        uArgDenom;          /* argument of esa denominator for u */
  double        vArgDenom;          /* argument of esa denominator for v */
  double        uOne_on_MagDenom;   /* 1 / magnitude of esa denominator for u */
  double        vOne_on_MagDenom;   /* 1 / magnitude of esa denominator for v */
  float         uMagNumer;          /* magnitude of esa Numerator for u calc */
  float         vMagNumer;          /* magnitude of esa Numerator for v calc */
  double        uArgNumer;          /* argument of esa Numerator for u calc */
  double        vArgNumer;          /* argument of esa Numerator for v calc */
  double        uCos;               /* computed cosine of u */
  double        vCos;               /* computed cosine of v */
  double        uSin;               /* computed sine of u */
  double        vSin;               /* computed sine of v */
  double        sgn_uSin;           /* sign of Sine of u */
  double        sgn_vSin;           /* sign of Sine of v */
  float         A,B;                /* numerator terms from u calc */
  float         E,F;                /* numerator terms from v calc */
  float         FilterMag;          /* magnitude of filter at estimated u,v */

 /*
  * Handle Edge Conditions
  */
  if (col == t->size-1) {
    uj = t->size-1 + row * t->size;
    um = t->size-2 + row * t->size;
    uk = t->size-3 + row * t->size;
  } else {
    if (col == 0) {
      uj = 2 + row * t->size;
      um = 1 + row * t->size;            /* NOTE: uj > uk */
      uk = row * t->size;
    } else {
      uj = col + 1 + row * t->size;
      um = col + row * t->size;
      uk = col - 1 + row * t->size;
    }
  }
  if (row == t->size-1) {
    vj = (t->size-1) * t->size + col;
    vm = (t->size-2) * t->size + col;
    vk = (t->size-3) * t->size + col;
  } else {
    if (row == 0) {
      vj = 2 * t->size + col;
      vm = t->size + col;            /* NOTE: vj > vk */
      vk = col;
    } else {
      vj = (row + 1) * t->size + col;
      vm = row * t->size + col;
      vk = (row - 1) * t->size + col;
    }
  }

 /*
  * Evaluate the filter selection criterion
  */
  i = row * t->size + col;
  ahat = fNorm(t->Real[i], t->Imag[i]);
  criterion = ahat / G->GMax;
  if (criterion <= *FilterSelCriterion) {
    return(0);
  }
  *FilterSelCriterion = criterion;
  uMagDenom = fNorm(t->Real[um],t->Imag[um]);
  vMagDenom = fNorm(t->Real[vm],t->Imag[vm]);
  if ((ahat < FLOAT_TOO_SMALL) || (uMagDenom < FLOAT_TOO_SMALL)
                               || (vMagDenom < FLOAT_TOO_SMALL)) {
    *a = (float)0.0;
    *u = (float)0.0;
    *v = (float)0.0;   /* Re & Im were 0 to within float prcsn */
    *a_valid = 0;
    *u_valid = 0;
    *v_valid = 0;
    return;
  }
  uOne_on_MagDenom = (double)1.0 / uMagDenom;
  uArgDenom = atan2((double)t->Imag[um],(double)t->Real[um]);
  vOne_on_MagDenom = (double)1.0 / vMagDenom;
  vArgDenom = atan2((double)t->Imag[vm],(double)t->Real[vm]);

 /*
  * Find uCos
  */
  A = (float)0.5 * (t->Real[uj] + t->Real[uk]);
  B = (float)0.5 * (t->Imag[uj] + t->Imag[uk]);
  uMagNumer = fNorm(A,B);
  if (uMagNumer < FLOAT_TOO_SMALL) { /* A & B both 0 to within precision */
    *u = (float) 0.0;
    *u_valid = 0;
  } else {
    uArgNumer = atan2((double)B, (double)A);
    uCos = (double)uMagNumer * uOne_on_MagDenom * cos(uArgNumer - uArgDenom);
    if (uCos < (double)-1.0) {
      uCos = (double) -1.0;
    } else {
      if (uCos > (double)1.0) {
        uCos = (double)1.0;
      }
    }
    *u = (float) 0.5 * acos(uCos) / (double)M_PI;
    *u_valid = 1;
  }

 /*
  * Find uSin; correct SIGN of u using Sine Alg.
  */
  A = (float)0.5 * (t->Real[uk] - t->Real[uj]);
  B = (float)0.5 * (t->Imag[uj] - t->Imag[uk]);
  uMagNumer = fNorm(A,B);
  if (uMagNumer < FLOAT_TOO_SMALL) {     /* Sine alg numerator zero */
    if (!u_valid) {
      *u = (float)0.0;
      *u_valid = 0;
    }
  } else {
    uArgNumer = atan2((double)A, (double)B);
    uSin = (double)uMagNumer * uOne_on_MagDenom * cos(uArgNumer - uArgDenom);
    if (uSin < (double)-1.0) {
      uSin = (double) -1.0;
    } else {
      if (uSin > (double)1.0) {
        uSin = (double)1.0;
      }
    }
    if (!u_valid) {
      *u = (float) 0.5 * asin(uSin) / (double)M_PI;
      *u_valid = 1;
    } else {
      if (asin(uSin) < (double)0.0) sgn_uSin = (double) -1.0;
      else                          sgn_uSin = (double)  1.0;
      *u *= sgn_uSin;
    }
  }

 /*
  * Find vCos
  */
  E = (float)0.5 * (t->Real[vj] + t->Real[vk]);
  F = (float)0.5 * (t->Imag[vj] + t->Imag[vk]);
  vMagNumer = fNorm(F,E);
  if (vMagNumer < FLOAT_TOO_SMALL) { /* E & F both 0 to within precision */
    *v = (float) 0.0;
    *v_valid = 0;
  } else {
    vArgNumer = atan2((double)F, (double)E);
    vCos = (double)vMagNumer * vOne_on_MagDenom * cos(vArgNumer - vArgDenom);
    if (vCos < (double)-1.0) {
      vCos = (double) -1.0;
    } else {
      if (vCos > (double)1.0) {
        vCos = (double)1.0;
      }
    }
    *v = (float) 0.5 * acos(vCos) / (double)M_PI;
    *v_valid = 1;
  }
 
 /*
  * Find vSin
  */
  E = (float)0.5 * (t->Real[vk] - t->Real[vj]);
  F = (float)0.5 * (t->Imag[vj] - t->Imag[vk]);
  vMagNumer = fNorm(E,F);
  if (vMagNumer < FLOAT_TOO_SMALL) {     /* Sine alg numerator zero */
    if (!v_valid) {
      *v = (float)0.0;
      *v_valid = 0;
    }
  } else {
    vArgNumer = atan2((double)E, (double)F);
     vSin = (double)vMagNumer * vOne_on_MagDenom * cos(vArgNumer - vArgDenom);
     if (vSin < (double)-1.0) {
       vSin = (double) -1.0;
     } else {
       if (vSin > (double)1.0) {
         vSin = (double)1.0;
       }
     }
     if (asin(vSin) < (double)0.0) sgn_vSin = (double) -1.0;
     else                          sgn_vSin = (double)  1.0;
     *v *= sgn_vSin;
  }

 /*
  * Estimate a
  */
  if (*u_valid && *v_valid) {
    FilterMag = G_EvalMag(G,fabs((double)(*u)),(double)(*v));
    if ((FilterMag / G->GMax) > FLOAT_TOO_SMALL) {
      *a = ahat / FilterMag;
      *a_valid = 1;
    } else {
      *a = *u = *v = (float)0.0;  /* is effectively zero. u,v estimates must */
      *a_valid = 0;               /* have been derived from noise. */
    }
  }   else { 
    *a = (float)0.0; 
    *a_valid = 0;   
  }  /* else */

  return(1);
} 
#endif
