/*
 * PtEsa2D.c  THIS VERSION HAS POSTFILTERING... 3/11/94
 *
 *  Function evaluates the asilomar esa at one point in a float image.
 *
 *  - This version uses the COMBINED ALGORITHM:  see book VII p. 49
 *
 *  - The magnitude of the response image is divided by the max response of the filter.
 *    This is compared to the filter selection criterion value.  If less, zero is 
 *    returned and no further action is taken (values of all the other parameters are
 *    preserved).
 *  - Otherwise, the esa is computed and one is returned.  The amplitude estimate is 
 *    returned in "a" and the frequency estimates are returned in "u" and "v".  
 *    u_valid and v_valid are one if the corresponding frequency estimate is valid, and 
 *    zero otherwise (e.g. if there is a zero denominator in the esa).  The
 *    filter selection criterion is updated to reflect the new value.
 *  - Edge conditions are handled by reflection.
 *   
 *  jph 9/13/93
 */

#ifndef PTESA2DPF_C
#define PTESA2DPF_C
#include <stdio.h>
#include "teager2D.h"

#define FLOAT_TOO_SMALL ((float)0.00001)

int PtEsa2DPF(t,amag,row,col,G,FilterSelCriterion,u,v,a,u_valid,v_valid,a_valid)

  pFloatImage   t;                  /* response image */
  float        *amag;               /* postfiltered magnitude */
  int           row;                /* image row for esa */
  int           col;                /* image col for esa */
  FilterPtr     G;                  /* pointer to the Gabor filter */
  float        *FilterSelCriterion; /* amp / max filt resp. must exceed this */
  float        *u;                  /* estimated horiz. freq. component */
  float        *v;                  /* estimated vert.  freq. component */
  float        *a;                  /* estimated amplitude */
  BYTE         *u_valid;            /* 1 if u is valid */
  BYTE         *v_valid;            /* 1 if v is valid */
  BYTE         *a_valid;            /* 1 if a is valid */
{

  int           i,j,k;              /* cursors for indexing the image */
  float         ahat;               /* magnitude of response image */
  float         criterion;          /* filter selection criterion value */
  float         MagDenom;           /* magnitude of esa denominator */
  double        one_on_MagDenom;    /* one over magnitude of esa denominator */
  double        ArgDenom;           /* argument of esa denominator */
  float         uMagNumer;          /* magnitude of esa Numerator for u calc */
  double        uArgNumer;          /* argument of esa Numerator for u calc */
  double        uCos;               /* computed cosine of u */
  double        uSin;               /* computed sine of u */
  double        sgn_uSin;           /* sign of Sine of u */
  float         vMagNumer;          /* magnitude of esa Numerator for v calc */
  double        vArgNumer;          /* argument of esa Numerator for v calc */
  double        vCos;               /* computed cosine of v */
  double        vSin;               /* computed sine of v */
  double        sgn_vSin;           /* sign of Sine of v */
  float         A,B;                /* numerator terms from u calc */
  float         E,F;                /* numerator terms from v calc */
  float         FilterMag;          /* magnitude of filter at estimated u,v */

 /*
  * Evaluate the filter selection criterion
  */
  i = row * t->size + col;
  ahat = MagDenom = amag[i];           /*fNorm(t->Real[i], t->Imag[i]);*/
  criterion = ahat / G->GMax;
  if (criterion <= *FilterSelCriterion) {
    return(0);
  }

 /*
  * Criterion was passed, evaluate the esa
  */
  *FilterSelCriterion = criterion;
  if (MagDenom < FLOAT_TOO_SMALL) { /* Re & Im were zero to within float precision */
    *a = (float)0.0;
    *u = (float)0.0; 
    *v = (float)0.0;
    *a_valid = 0;
    *u_valid = 0;
    *v_valid = 0;
    return(1);
  }
  one_on_MagDenom = (double)1.0 / MagDenom;
  ArgDenom = atan2((double)t->Imag[i],(double)t->Real[i]);

 /*
  * Estimate u
  */
  if (col == t->size-1) {
    j = i;
    k = i - 2;
  }
  else {     
    if (col == 0) {
      k = i;             /* NOTE: j > k */
      j = i+2;
    }
    else {
      k = i - 1;
      j = i + 1; 
    }
  }
 /*
  * Find uCos
  */
  A = (float)0.5 * (t->Real[j] + t->Real[k]);
  B = (float)0.5 * (t->Imag[j] + t->Imag[k]);
  uMagNumer = fNorm(A,B);
  if (uMagNumer < FLOAT_TOO_SMALL) { /* A & B both zero to within float precision */
    *u = (float) 0.25;               /* implies freq = pi/2.  See book VII p. 2 */
    *u_valid = 1;
  }
  else {
    uArgNumer = atan2((double)B, (double)A);
    uCos = (double)uMagNumer * one_on_MagDenom * cos(uArgNumer - ArgDenom);
    if (uCos < (double)-1.0) {
      uCos = (double) -1.0;
    }
    else {
      if (uCos > (double)1.0) {
        uCos = (double)1.0;
      }
    }
    *u = (float) 0.5 * acos(uCos) / (double)M_PI;
    *u_valid = 1;
  }

/*---------------  FOR QUAD I-IV MODEL U IS POSITIVE  ------------------*/
 /*
  * Find uSin
  */
/*-----------------------------------------------------------------------
  A = (float)0.5 * (t->Real[k] - t->Real[j]);
  B = (float)0.5 * (t->Imag[j] - t->Imag[k]);
  uMagNumer = fNorm(A,B);
  if (uMagNumer != (float)0.0) {     *//* Sine alg numerator not zero *//*
    uArgNumer = atan2((double)A, (double)B);
    uSin = (double)uMagNumer * one_on_MagDenom * cos(uArgNumer - ArgDenom);
    if (uSin < (double)-1.0) {
      uSin = (double) -1.0;
    }
    else {
      if (uSin > (double)1.0) {
        uSin = (double)1.0;
      }
    }
    if (asin(uSin) < (double)0.0) sgn_uSin = (double) -1.0;
    else                          sgn_uSin = (double)  1.0;
    *u *= sgn_uSin;
  }
-----------------------------------------------------------------------*/

 /*
  * Estimate v
  */
  if (row == t->size-1) {
    j = i;
    k = i - (t->size << 1);
  }
  else {
    if (row == 0) {
      k = col;
      j = col + (t->size << 1);
    }
    else  {
      k = i - t->size;
      j = i + t->size;
    }
  }

 /*
  * Find vCos
  */
  E = (float)0.5 * (t->Real[j] + t->Real[k]);
  F = (float)0.5 * (t->Imag[j] + t->Imag[k]);
  vMagNumer = fNorm(F,E);
  if (vMagNumer < FLOAT_TOO_SMALL) { /* E & F both zero to within float precision */
    *v = (float) 0.25;               /* implies freq = pi/2.  See book VII p. 2 */
    *v_valid = 1;
  }
  else {
    vArgNumer = atan2((double)F, (double)E);
    vCos = (double)vMagNumer * one_on_MagDenom * cos(vArgNumer - ArgDenom);
    if (vCos < (double)-1.0) {
      vCos = (double) -1.0;
    }
    else {
      if (vCos > (double)1.0) {
        vCos = (double)1.0;
      }
    }
    *v = (float) 0.5 * acos(vCos) / (double)M_PI;
    *v_valid = 1;
  }

 /*
  * Find vSin
  */
  E = (float)0.5 * (t->Real[k] - t->Real[j]);
  F = (float)0.5 * (t->Imag[j] - t->Imag[k]);
  vMagNumer = fNorm(E,F);
  if (vMagNumer != (float)0.0) {       /* Sine alg numerator not zero */
    vArgNumer = atan2((double)E, (double)F);
     vSin = (double)vMagNumer * one_on_MagDenom * cos(vArgNumer - ArgDenom);
     if (vSin < (double)-1.0) {
       vSin = (double) -1.0;
     }
     else {
       if (vSin > (double)1.0) {
         vSin = (double)1.0;
       }
     }
     if (asin(vSin) < (double)0.0) sgn_vSin = (double) -1.0;
     else                          sgn_vSin = (double)  1.0;
     *v *= sgn_vSin;
  }

 /*
  * Estimate a
  */
  if (*u_valid && *v_valid) {
    FilterMag = G_EvalMag(G,(double)(*u),(double)(*v));
    if ((FilterMag / G->GMax) > (float)0.000001) {   /* was 0.000001 */
      *a = ahat / FilterMag;
      if (*a > (float)1.0) *a = (float)1.0;
      *a_valid = 1;
    }    /* if FilterMag */
    else {                        /* estimated freq's occur where the filter */
      *a = *u = *v = (float)0.0;  /* is effectively zero. u,v estimates must */
      *a_valid = 1;               /* have been derived from noise. */
    } /* else */
  }   /* if u_valid && v_valid */
  else {              /* u and/or v not valid, can't  */
    *a = (float)0.0;  /* estimate a here using this   */
    *a_valid = 0;     /* response image.              */
  }  /* else */

  return(1);
} 
#endif
