/*
 * SpbMkBnkSpecAux.c
 *
 * Function creates a binary image suitable for display of the specified 
 * filter Gabor filter bank.  The quantity displayed is c log[1 + |F(u,v)|]
 * The output image is allocated on the heap.  This version INCLUDES the
 * DC filter and the corner filters.  The corner filters cover the high
 * frequencies in the corners of the frequency plane.  
 *
 * This function applies special nonlinear scaling so that all filters will
 * take values between 128 and 255 to facilitate qualitative comparison
 * of their regions of support.
 * 
 * the filterbank parameters are written out to the file "SPbankAux.log"
 *
 *  jph 10/03/96
 *
 */

#ifndef SPBMKBANKSPECTRUMAUX_C
#define SPBMKBANKSPECTRUMAUX_C
#define BYTE unsigned char
#include "teager2D.h"
#include <stdio.h>

unsigned char *SpbMkBankSpectrumAux(size,r0,R,OBW,Eta,kappa)


  int         size;                /* row (col) dimension of the image */
  double      r0;                  /* radial center freq. of the first filter */
                                   /*   on each ray in CYCLES per pixel */
  float       R;                   /* common ratio of radii of adjcnt filtrs */
  float       OBW;                 /* Eta-peak octave bandwidth of filters */
  float       Eta;                 /* OBW is defined in terms of Eta-peak. */
  float       kappa;               /* postfilter sigma scaling factor */

{

  float      *M;         /* Composite Magnitude Image */
  float      *X;         /* magnitude spectrum of one filter in the bank */
  BYTE       *Y;         /* output binary image */
  filterbank *G;         /* Gabor filter bank */
  int         i;         /* loop counter for filters in the bank */
  int         k;         /* loop counter for image coordinates */
  int         N;         /* num pixels in image == size**2  */
  int         row;       /* loop counter for rows */
  int         col;       /* loop counter for cols */
  FILE *logfile;         /* ascii output file */

  N = size * size;
  printf("\n\nCreating Log File SPbankAux.log...\n\n");
  logfile = fopen("SPbankAux.log","w");
  G = CreateFilterBankAux(r0,R,OBW,Eta,1,kappa,kappa,kappa);

 /*
  * Report BANK parameters to log file
  */
  fprintf(logfile,"\nFilterbank:\n");
  fprintf(logfile,"NumRays = %d\n",G->NumRays);
  fprintf(logfile,"NumFiltersPerRay = %d\n",G->NumFiltersPerRay);
  fprintf(logfile,"TotNumFilters = %d\n",G->TotNumFilters);
  fprintf(logfile,"  (Including DC filter and two corner filters).\n");
  fprintf(logfile,"r0 = %lf\n",G->r0);
  fprintf(logfile,"Common Ratio = %lf\n",G->R);
  fprintf(logfile,"Theta0 = %lf\n",G->Theta0);
  fprintf(logfile,"Radial Spacing = %lf\n",G->Lambda);
  fprintf(logfile,"Eta = %f\n",G->Eta);
  fprintf(logfile,"OBW = %f\n",G->OBW);
  fprintf(logfile,"gamma = %f\n",G->gamma);

 /*
  * Report DC filter parameters to log file
  */
  fprintf(logfile,"\n\nDC Filter\n");
  fprintf(logfile,"r0 = %lf\n",G->DCfilt->r0);
  fprintf(logfile,"u0 = %lf\n",G->DCfilt->u0);
  fprintf(logfile,"v0 = %lf\n",G->DCfilt->v0);
  fprintf(logfile,"sigma = %lf\n",G->DCfilt->sigma);
  fprintf(logfile,"GMax = %f\n",G->DCfilt->GMax);

 /*
  * Report filter 0 parameters to log file
  */
  fprintf(logfile,"\n\nFilter %d\n",0);
  fprintf(logfile,"Ray = %d\n",G->filt[0]->Ray);
  fprintf(logfile,"IndexOnRay = %d\n",G->filt[0]->IndexOnRay);
  fprintf(logfile,"Theta = %lf\n",G->filt[0]->Theta);
  fprintf(logfile,"r0 = %lf\n",G->filt[0]->r0);
  fprintf(logfile,"u0 = %lf\n",G->filt[0]->u0);
  fprintf(logfile,"v0 = %lf\n",G->filt[0]->v0);
  fprintf(logfile,"sigma = %lf\n",G->filt[0]->sigma);
  fprintf(logfile,"GMax = %f\n",G->filt[0]->GMax);

 /*
  * Make the image for filter 0
  */
  printf("\nDoing Filter 0...");
  M = GMakeMag(G->filt[0],size);
  for (i=row=0; row < size; row++) {
    for (col=0; col < size; col++,i++) {
      M[i] = (float)(exp((double)M[i]) - (double)1.0);
      if ((M[i] / G->filt[0]->GMax) >= (float)0.5) {
        M[i] = 
         /* in [0.220, 1.0] */ (float)0.220 +
         /* in [0.0, 0.780] */   (float)0.780*(
         /* in [0.0, 1.0]   */             (float)(2.0 *
         /* in [0.0, 0.5]   */                      (
         /* in [0.5, 1.0]   */                        (M[i] / G->filt[0]->GMax) 
                                                       - (float)0.5
                                                    )
                                                  ) 
                                         );
      } else {
        M[i] = (float)0.0;
      } 
    }
  }

 /*
  * Loop for remaining filters in regular filter tesselation
  *  Report parameters to log file and make filter image
  */
  for (i=1; i < G->TotNumFilters-3; i++) {
    fprintf(logfile,"\n\nFilter %d\n",i);
    fprintf(logfile,"Ray = %d\n",G->filt[i]->Ray);
    fprintf(logfile,"IndexOnRay = %d\n",G->filt[i]->IndexOnRay);
    fprintf(logfile,"Theta = %lf\n",G->filt[i]->Theta);
    fprintf(logfile,"r0 = %lf\n",G->filt[i]->r0);
    fprintf(logfile,"u0 = %lf\n",G->filt[i]->u0);
    fprintf(logfile,"v0 = %lf\n",G->filt[i]->v0);
    fprintf(logfile,"sigma = %lf\n",G->filt[i]->sigma);
    fprintf(logfile,"GMax = %f\n",G->filt[i]->GMax);
    printf("\nDoing Filter %d...",i);
    X = GMakeMag(G->filt[i],size);
    for (k=0; k < N; k++) {
      X[k] = (float)(exp((double)X[k]) - (double)1.0);
      if ((X[k] / G->filt[i]->GMax) >= (float)0.5) {
        X[k] = 
         /* in [0.220, 1.0] */ (float)0.220 +
         /* in [0.0, 0.780] */   (float)0.780*(
         /* in [0.0, 1.0]   */             (float)(2.0 *
         /* in [0.0, 0.5]   */                      (
         /* in [0.5, 1.0]   */                        (X[k] / G->filt[i]->GMax)
                                                       - (float)0.5
                                                    )
                                                  ) 
                                         );
        if (X[k] >= M[k]) {
          M[k] = X[k];
        }
      } else {
        X[k] = (float)0.0;
      }
    }
    free(X);
  }

 /*
  * Make DC filter image
  */
  X = GMakeMag(G->DCfilt,size);
/*  The following lines will zero out the filter for negative horiz freq's */
/*  for (row=0; row < size; row++) { */
/*    for (col=0; col < size>>1; col++) { */
/*      X[row*size+col] = (float)0.0; */
/*    } */
/*  } */
  printf("\nDoing DC Filter...");
  for (k=0; k < N; k++) {
    X[k] = (float)(exp((double)X[k]) - (double)1.0);
    if ((X[k] / G->DCfilt->GMax) >= (float)0.5) {
      X[k] = 
       /* in [0.220, 1.0] */ (float)0.220 +
       /* in [0.0, 0.780] */   (float)0.780*(
       /* in [0.0, 1.0]   */             (float)(2.0 *
       /* in [0.0, 0.5]   */                      (
       /* in [0.5, 1.0]   */                        (X[k] / G->DCfilt->GMax)
                                                     - (float)0.5
                                                  )
                                                ) 
                                       );
      if (X[k] >= M[k]) {
        M[k] = X[k];
      }
    } else {
      X[k] = (float)0.0;
    }
  }
  free(X);


 /*
  * Corner filters
  *  Report parameters to log file and make filter image
  */
  for (i=G->TotNumFilters-2; i < G->TotNumFilters; i++) {
    fprintf(logfile,"\n\nFilter %d (Corner Filter)\n",i);
    fprintf(logfile,"Theta = %lf\n",G->filt[i]->Theta);
    fprintf(logfile,"r0 = %lf\n",G->filt[i]->r0);
    fprintf(logfile,"u0 = %lf\n",G->filt[i]->u0);
    fprintf(logfile,"v0 = %lf\n",G->filt[i]->v0);
    fprintf(logfile,"sigma = %lf\n",G->filt[i]->sigma);
    fprintf(logfile,"GMax = %f\n",G->filt[i]->GMax);
    fprintf(logfile,"OBW = %f\n",G->filt[i]->OBW);
    printf("\nDoing Corner Filter %d...",i);
    X = GMakeMag(G->filt[i],size);
    for (k=0; k < N; k++) {
      X[k] = (float)(exp((double)X[k]) - (double)1.0);
/*      if ((X[k] / G->filt[i]->GMax) >= (float)0.5) {*/
        X[k] = 
         /* in [0.220, 1.0] */ (float)0.220 +
         /* in [0.0, 0.780] */   (float)0.780*(
         /* in [0.0, 1.0]   */             (float)(2.0 *
         /* in [0.0, 0.5]   */                      (
         /* in [0.5, 1.0]   */                        (X[k] / G->filt[i]->GMax)
                                                       - (float)0.5
                                                    )
                                                  ) 
                                         );
        if (X[k] >= M[k]) {
          M[k] = X[k];
        }
/*      } else {*/
/*        X[k] = (float)0.0;*/
/*      }*/
    }
    free(X);
  }

  Y = f_dbx(M,size,size);
  free(M);
  return(Y);

}
#endif
