/*
 * d_fft2d.c
 *
 *  Function implements the 2D FFT.  The Numerical Recipes routine "fourn"
 *  is called.  The frequency domain coordinates are as described on 
 *  pp. 114 - 117 of book VI.  
 *
 *    The image must be square and the dimension must be a power of two.
 *
 *  FOR THIS FILE, WE NEVER FACTOR THE "n" OUT OF THE SPECTRUM.
 *   - after forward transform, spectral data has the "n" factor in it
 *   - after inverse transform, "n" IS FACTORED OUT of spacial data
 *
 */

#ifndef D_FFT2D_C
#define D_FFT2D_C
#include "teager2D.h"
#include <stdio.h>

void d_fft2d(xReal,xImag,XReal,XImag,size,direction)

  double *xReal;       /* real part of signal */
  double *xImag;       /* imaginary part of signal */
  double *XReal;       /* real part of spectrum */
  double *XImag;       /* imaginary part of spectrum */
  int    size;        /* row/col dim of image */
  int    direction;   /* 1=fwd xform, -1=inverse xform */

{
    double *pass;                  /* array for passing data to/from fourn */
    double one_on_n;               /* 1 / (no. data points in signal) */
    int i;                        /* counter */
    int n;                        /* no. data points in signal */
    int pass_size;                /* number of samples in pass array, twice n */
    int so2;                      /* size over 2 */
    int st2;                      /* size times 2 */
    int row,col;                  /* loop counters */
    int XQuadI_IIRow;             /* current row in XReal and XImag Quads I & II   */
    int XQuadIII_IVRow;           /* current row in XReal and XImag Quads III & IV */
    int XQuadI_IVCol;             /* current col in XReal and XImag Quads I & II   */
    int XQuadII_IIICol;           /* current col in XReal and XImag Quads II & III */
    int PassQuadI_IIRow;          /* current row in pass Quads I & II   */
    int PassQuadIII_IVRow;        /* current row in pass Quads III & IV */
    int PassQuadI_IVCol;          /* current col in pass Quads I & II   */
    int PassQuadII_IIICol;        /* current col in pass Quads II & III */
    int XIndx;                    /* Index into XReal and XImag arrays */
    int PassIndx;                 /* Index into pass array */
    int pass_dimensions[2];       /* array of dimensions of pass array */

    n = size * size;
    so2 = (size >> 1);
    st2 = (size << 1);
    pass_dimensions[0] = pass_dimensions[1] = size;
    pass_size = n << 1;

    if (direction == 1 ) { /* forward transform */
      pass  = (double *) malloc(pass_size*sizeof(double));
     /*
      * Set up the data in the pass array and call fourn
      */
      for (i=0; i<n; i++) {
        pass[2*i] = xReal[i];
        pass[2*i+1] = xImag[i];
      }
      d_fourn(pass-1,pass_dimensions-1,2,1);

     /*
      * Get the spectrum set up in understandable image coordinates
      * see book VI pp. 144 - 148 for an explanation of the indexing 
      */
      for (row=0; row<so2; row++) {
        XQuadI_IIRow      = row + so2;
        XQuadIII_IVRow    = row;
        PassQuadI_IIRow   = (size - row) % size;
        PassQuadIII_IVRow = so2 - row;

        for (col=0; col<so2; col++) {
          XQuadI_IVCol      = col + so2;
          XQuadII_IIICol    = col;
          PassQuadI_IVCol   = ((size - col) % size) << 1;
          PassQuadII_IIICol = (so2 - col) << 1;

          /* Quad I */
          XIndx    = XQuadI_IIRow    * size + XQuadI_IVCol;
          PassIndx = PassQuadI_IIRow * st2  + PassQuadI_IVCol;
          XReal[XIndx] = pass[PassIndx];
          XImag[XIndx] = pass[PassIndx + 1];
  
          /* Quad II */
          XIndx    = XQuadI_IIRow    * size + XQuadII_IIICol;
          PassIndx = PassQuadI_IIRow * st2  + PassQuadII_IIICol;
          XReal[XIndx] = pass[PassIndx];
          XImag[XIndx] = pass[PassIndx + 1];
  
          /* Quad III */
          XIndx    = XQuadIII_IVRow    * size + XQuadII_IIICol;
          PassIndx = PassQuadIII_IVRow * st2  + PassQuadII_IIICol;
          XReal[XIndx] = pass[PassIndx];
          XImag[XIndx] = pass[PassIndx + 1];
  
          /* Quad IV */
          XIndx    = XQuadIII_IVRow    * size + XQuadI_IVCol;
          PassIndx = PassQuadIII_IVRow * st2  + PassQuadI_IVCol;
          XReal[XIndx] = pass[PassIndx];
          XImag[XIndx] = pass[PassIndx + 1];
        }
      }
    }
    else {
      if (direction == -1) { /* reverse transform */ 
        one_on_n = (double)1.0 / (double)n;
        pass  = (double *) malloc(pass_size*sizeof(double));
       /*
        * Set up the data in the pass array and call fourn.
        * see book VI pp. 144 - 148 for an explanation of the indexing
        */
        for (row=0; row<so2; row++) {
          XQuadI_IIRow      = row + so2;
          XQuadIII_IVRow    = row;
          PassQuadI_IIRow   = (size - row) % size;
          PassQuadIII_IVRow = so2 - row;
  
          for (col=0; col<so2; col++) {
            XQuadI_IVCol      = col + so2;
            XQuadII_IIICol    = col;
            PassQuadI_IVCol   = ((size - col) % size) << 1;
            PassQuadII_IIICol = (so2 - col) << 1;
  
            /* Quad I */
            XIndx    = XQuadI_IIRow    * size + XQuadI_IVCol;
            PassIndx = PassQuadI_IIRow * st2  + PassQuadI_IVCol;
            pass[PassIndx]     = XReal[XIndx];
            pass[PassIndx + 1] = XImag[XIndx];
    
            /* Quad II */
            XIndx    = XQuadI_IIRow    * size + XQuadII_IIICol;
            PassIndx = PassQuadI_IIRow * st2  + PassQuadII_IIICol;
            pass[PassIndx]     = XReal[XIndx];
            pass[PassIndx + 1] = XImag[XIndx];
    
            /* Quad III */
            XIndx    = XQuadIII_IVRow    * size + XQuadII_IIICol;
            PassIndx = PassQuadIII_IVRow * st2  + PassQuadII_IIICol;
            pass[PassIndx]     = XReal[XIndx];
            pass[PassIndx + 1] = XImag[XIndx];
    
            /* Quad IV */
            XIndx    = XQuadIII_IVRow    * size + XQuadI_IVCol;
            PassIndx = PassQuadIII_IVRow * st2  + PassQuadI_IVCol;
            pass[PassIndx]     = XReal[XIndx];
            pass[PassIndx + 1] = XImag[XIndx];
          }
        }
        d_fourn(pass-1,pass_dimensions-1,2,-1);
        for (i=0; i<n; i++) {
          xReal[i] = pass[2*i]   * one_on_n;
          xImag[i] = pass[2*i+1] * one_on_n;
        }
      }
      else {
        printf("\nERROR: d_fft2d: unknown value %d specified for direction.\n",direction);
        exit(-1);
      }
    }
    free(pass);
    return;

}
#endif 
