/*
 * f_interp.c
 *
 *  this function is invoked to interpolate and estimate the value of
 *  one of the ESAs in eq. (16), (17), and (19) on p. 94 of book V
 *  when there is a zero denominator.  Under
 *  this condition, we will refer to the value of the ESA as "indeterminate."
 *
 *  When this function is invoked, it has been determined that ESA
 *  float x[row_dim][col_dim] is indeterminate at (row,col).  
 *  
 *  BYTE image denom[row_dim][col_dim] is zero everywhere that
 *  the denominator of the ESA is zero.  The value of the ESA at
 *  these points is considered to unreliable for use in the interpolation.
 *
 *  The interpolation makes use of estimates U,D,L,R of x[row][col] obtained
 *  by looking up, down, left, and right, respectively.  Associated with
 *  each estimate is a distance; the distances are denoted dU,dD,DL, and
 *  dR, respectively.
 *
 *  Initially, U=D=L=R=0.  Then, for each direction, the routine searches 
 *  in that direction until it finds a reliable estimate or hits the edge of 
 *  the x ESA image.  The distance in that direction is the number of steps
 *  taken.  If the edge was hit, the estimate in that direction remains zero;
 *  otherwise the estimate in that direction is set equal to the first
 *  reliable value of x[i][j] found in that direction.
 *
 *  A weight inversely porportional to each distance is then determined.  For
 *  example the "up" weight wU is defined by:
 *
 *    wU = max{dU, dD, dL, dR} - dU + 1
 *
 *  The interpolation is then calculated according to:
 *
 *    x[row][col] = (wU*U + wD*D + wL*L + wR*R) / (wU + wD + wL + wR)
 *
 *  jph 10 June 1992
 */

#ifndef F_INTERP_C
#define F_INTERP_C
#include <stdio.h>
#include <math.h>
#include "teager2D.h"

void f_interp(x,row,col,row_dim,col_dim,denom)

  float* x;      /* input image */
  int row;        /* row of indeterminate value in x */
  int col;        /* col of indeterminate value in x */
  int row_dim;    /* row dimension of x */
  int col_dim;    /* col dimension of x */
  BYTE* denom;    /* zero where x is indeterminate */
{

  float U;       /* "up"    estimate of x[row][col] */
  float D;       /* "down"  estimate of x[row][col] */
  float L;       /* "left"  estimate of x[row][col] */
  float R;       /* "right" estimate of x[row][col] */
  float dU;      /* distance to U */
  float dD;      /* distance to D */
  float dL;      /* distance to L */
  float dR;      /* distance to R */
  float wU;      /* weight of U */
  float wD;      /* weight of D */
  float wL;      /* weight of L */
  float wR;      /* weight of R */
  float sumw;    /* sum of the weights */
  float dmax;    /* max distance */
  int i;          /* loop counter */
  int found;      /* flag */


 /*
  * detect zero dimension input
  */
  if ((row_dim==0) || (col_dim==0)) return;

 /*
  * find U, dU
  */
  found = 0;
  U = dU = (float) 0.0;
  if (row != 0) {
    for (i=row-1,dU++; i>=0; i--) {
      if ((*(denom+i*col_dim+col)) != 0) {
        U = (*(x+i*col_dim+col));
        i = -1;
        found = 1;
      }
      if (i >= 0) dU++;
    }
    if (!found) dU = 0.0;
  }
 /*
  * find D, dD
  */
  found = 0;
  D = dD = (float) 0.0;
  if (row != row_dim - 1) {
    for (i=row+1,dD++; i<row_dim; i++) {
      if ((*(denom+i*col_dim+col)) != 0) {
        D = *(x+i*col_dim+col);
        i = row_dim;
        found = 1;
      }
      if (i < row_dim) dD++;
    }
    if (!found) dD = 0.0;
  }
 /*
  * find L, dL
  */
  found = 0;
  L = dL = (float) 0.0;
  if (col != 0) {
    for (i=col-1,dL++; i>=0; i--) {
      if ((*(denom+row*col_dim+i)) != 0) {
        L = *(x+row*col_dim+i);
        i = -1;
        found = 1;
      }
      if (i >= 0) dL++;
    }
    if (!found) dL = 0.0;
  }
 /*
  * find R, dR
  */
  found = 0;
  R = dR = (float) 0.0;
  if (col != col_dim - 1) {
    for (i=col+1,dR++; i<col_dim; i++) {
      if ((*(denom+row*col_dim+i)) != 0) {
        R = *(x+row*col_dim+i);
        i = col_dim;
        found = 1;
      }
      if (i < col_dim) dR++;
    }
    if (!found) dR = 0.0;
  }

 /*
  * find dmax
  */
  dmax = dU;
  if (dD > dmax) dmax = dD;
  if (dL > dmax) dmax = dL;
  if (dR > dmax) dmax = dR;

 /*
  * find wU, wD, wL, wR
  */
  wU = (dU==0.0) ? (float) 0.0 : dmax - dU + 1; 
  wD = (dD==0.0) ? (float) 0.0 : dmax - dD + 1; 
  wL = (dL==0.0) ? (float) 0.0 : dmax - dL + 1; 
  wR = (dR==0.0) ? (float) 0.0 : dmax - dR + 1; 
  
 /*
  * interpolate to estimate x[row][col]
  */
  if ((sumw = wU + wD + wL + wR) == 0.0) sumw = 1.0;
  *(x+row*col_dim+col) = (wU*U + wD*D + wL*L + wR*R) / sumw;

} 
#endif
