/*
 * fmedNxN.c 
 *
 *   Do a NxN median filter on a float image.
 *
 *   The output image y is allocated on the heap.
 *
 *   Filtering is performed in a rectangular region with upper left corner
 *   (row1,col1) and lower right corner (row2,col2).
 *
 * jph 6/19/98
 *
 */

#ifndef FMEDNXN_C
#define FMEDNXN_C

#include "teager2D.h"

float *fmedNxN(x,size,WinSize,row1,col1,row2,col2)

  float  *x;                     /* input */
  int     size;                  /* row , col dim */
  int     WinSize;               /* Num rows/cols in window */
  int     row1,col1,row2,col2;   /* limits of where to run filter */
{

  float  *y;             /* output image */
  float  *W;             /* window */
  int     Wsize;         /* num pixels in window */
  int     WHalfWidth;    /* WinSize over 2, the window half-width */
  float   Wtemp;         /* temp floating variable */
  int     k;             /* loop counter */
  int     i;             /* loop counter */
  int     j;             /* loop counter */
  int     row,col;       /* loop counters */
  int     WinRow;        /* current row in window */
  int     WinCol;        /* current col in window */
  int     Wink;          /* current linear offset in window */
  int     MedIdx;        /* No. pixels that must be sorted to find median */

 /*
  * Test for WinSize odd
  */
  if (!(WinSize%2)) {
    printf("\n\nfmedNxN: called with even window size.\n");
    exit(-1);
  }

 /*
  * Allocate y on the heap
  */
  if ((y = (float *)malloc(size*size*sizeof(float))) == NULL) {
    printf("\n\nfmedNxN: free store exhausted.\n");
    exit(-1);
  }

 /*
  * Allocate neighborhood
  */
  Wsize = WinSize * WinSize;
  MedIdx = Wsize>>1;  /* it's the MedIdx+1 largest, so index (in C) is MedIdx*/
  if ((W = (float *)malloc(Wsize*sizeof(float))) == NULL) {
    printf("\n\nfmedNxN: free store exhausted.\n");
    exit(-1);
  }

 /*
  * copy x to y
  */
  for (k=0; k < size*size; k++) {
    y[k] = x[k];
  }

 /*
  * invoke median filter on y in the indicated region
  */
  WHalfWidth = WinSize>>1;
  for (row=row1; row <=row2; row++) {
    for (col=col1; col <= col2; col++) {
      k = row*size+col;

     /*
      * Build Window
      */
      for (Wink=0,WinRow=row-WHalfWidth; WinRow < row+WHalfWidth; WinRow++) {
        for (WinCol=col-WHalfWidth; WinCol < col+WHalfWidth; WinCol++,Wink++) {
          if (WinRow < 0) {
            if (WinCol < 0) {
/*            code for WinRow < 0, WinCol < 0                                */
              W[Wink] = x[0];
            } else {
              if (WinCol >= size) {
/*              code for WinRow < 0, WinCol >= size                          */
                W[Wink] = x[size-1];
              } else {
/*                code for WinRow < 0, WinCol in range                       */
                  W[Wink] = x[WinCol];
              }
            }
          } else {               /* WinRow is > 0 */
            if (WinRow >= size) {
              if (WinCol < 0) {
/*              code for WinRow >= size, WinCol < 0                          */
                W[Wink] = x[(size-1)*size];
              } else {
                if (WinCol >= size) {
/*                code for WinRow >= size, WinCol >= size                    */
                  W[Wink] = x[size*size - 1];
                } else {
/*                  code for WinRow >= size, WinCol in range                 */
                    W[Wink] = x[(size-1)*size + WinCol];
                }
              }
            } else {             /* WinRow is in range */
              if (WinCol < 0) {
/*              code for WinRow in range, WinCol < 0                         */
                W[Wink] = x[WinRow*size];
              } else {
                if (WinCol >= size) {
/*                code for WinRow in range, WinCol >= size                   */
                  W[Wink] = x[WinRow*size + size-1];
                } else {
/*                  code for WinRow in range, WinCol in range                */
                    W[Wink] = x[WinRow*size + WinCol];
                }
              }
            }
          }
        }         /* for WinCol */
      }           /* for WinRow */

     /*
      * Sort the lower half of the window
      */
      for (i=0; i <= MedIdx; i++) {
        for (j=i+1; j < Wsize; j++) {
          if (W[j] > W[i]) {
            Wtemp = W[j];
            W[j]  = W[i];
            W[i]  = Wtemp;
          }
        }
      }
      y[k] = W[MedIdx];
    }
  }

  return(y);
}
#endif
