/*
 * gabor.h
 *
 *  This file contains typedefs and function prototypes for implementation of
 *  a 2D multiband Gabor filter bank.
 *
 *  jph 9/9/93
 */
#ifndef GABOR_H
#define GABOR_H

/*
 * filter: type for an individual Gabor filter
 */
struct filter {

  int      index;             /* index of filter in the bank */
  int      Ray;               /* index of ray filter is on */
  int      IndexOnRay;        /* index of filter on its ray */
  double   Theta;             /* Orientation of filter in radians */
  double   r0;                /* radial center frequency in cycles per pixel */
  double   u0;                /* horizontal center freq in cycles per pixel */
  double   v0;                /* vertical   center freq in cycles per pixel */
  float    Eta;               /* Bandwidth is defined in terms of Eta-peak */
  float    OBW;               /* Eta-peak octave bandwidth of filter */
  double   gamma;             /* root tangent half-orientation BW */
  int      isotropicPF;       /* 1=use isotropic postfilter */
  float    kappa;             /* isotropic postfilter sigma scaling factor */
  float    kappa_x;           /* horiz postfilter sigma scaling factor */
  float    kappa_y;           /* vert  postfilter sigma scaling factor */
  double   sigma;             /* filter space constant */
  float    gMax;              /* peak value of filter mag in space domain */
  float    GMax;              /* peak value of filter mag in frequency domain */
  double   gArgGaussCoef;     /* coef in exp. of Gaussian for space domain */
  double   GArgGaussCoef;     /* coef in exp. of Gaussian for freq  domain */
};
typedef struct filter filter;

/*
 * FilterPtr: type for a pointer to a filter
 */
typedef filter *FilterPtr;

/*
 * filterbank: type for a bank of Gabor filters in a wavelet tesselation
 */
struct filterbank {

  int        NumRays;             /* num of Rays in the filter bank */
  int        NumFiltersPerRay;    /* num of filters on each Ray */
  int        TotNumFilters;       /* tot. num. filters in the filter bank */
  double     r0;                  /* Rad center freq of 1st filt on each Ray */
                                  /*  in cycles per pixel */
  float      R;                   /* common ratio of radii of adjcnt filtrs */
  double     Theta0;              /* orientation of first ray in radians */
  double     Lambda;              /* radial spacing of rays */
  float      Eta;                 /* BW is def'd in terms of Eta-peak, adj */
                                  /* filts on Ray share Eta-peak frequency. */
  float      OBW;                 /* Eta-peak octave bandwidth of filters */
  double     gamma;               /* root tangent half-orientation BW */
  int        isotropicPF;         /* 1=use isotropic postfilter */
  float      kappa;               /* isotropic postfilter sigma scaling factor*/
  float      kappa_x;             /* horiz postfilter sigma scaling factor */
  float      kappa_y;             /* vert  postfilter sigma scaling factor */
  FilterPtr *filt;                /* array of ptr's to filters */
  FilterPtr  DCfilt;              /* pointer to the DC filter, if included */
};
typedef struct filterbank filterbank;

/*
 * Function Prototypes
 */
#ifndef BMAKEBANKSPECTRUM_C
unsigned char *bMakeBankSpectrum();
#endif
#ifndef BSHOWBANKSPECTRUM_C
unsigned char *bShowBankSpectrum();
#endif
#ifndef BMAKEFILTERIMAGE_C
void bMakeFilterImage();
#endif
#ifndef BMAKEFILTERSPECTRUM_C
unsigned char *bMakeFilterSpectrum();
#endif
#ifndef CONVOLVEGABOR_C
void ConvolveGabor();
#endif
#ifndef CREATECORNERFILTERS_C
void CreateCornerFilters();
#endif
#ifndef CREATEFILTERBANK_C
filterbank *CreateFilterBank();
#endif
#ifndef CREATEFILTERBANKAUX_C
filterbank *CreateFilterBankAux();
#endif
#ifndef CREATEGABORFILTER_C
FilterPtr CreateGaborFilter();
#endif
#ifndef DESIGNDCFILT_C
FilterPtr DesignDCfilt();
#endif
#ifndef CREATEDCFILT_C
FilterPtr CreateDCFilter();
#endif
#ifndef DOPOSTFILTER_C
void DoPostFilter();
#endif
#ifndef FINDRADIALDISTANCE_C
double FindRadialDistance();
#endif
#ifndef FINDRADIALQUAL_C
double FindRadialQual();
#endif
#ifndef FINDRADIALSIGMAS_C
double FindRadialSigmas();
#endif
#ifndef FMAKEFILTERIMAGE_C
void fMakeFilterImage();
#endif
#ifndef FMAKEFILTERSPECTRUM_C
float *fMakeFilterSpectrum();
#endif
#ifndef g_EVALIMAG_C
float g_EvalImag();
#endif
#ifndef g_EVALMAG_C
float g_EvalMag();
#endif
#ifndef g_EVALREAL_C
float g_EvalReal();
#endif
#ifndef gMAKEIMAG_C
float *gMakeImag();
#endif
#ifndef gMAKEMAG_C
float *gMakeMag();
#endif
#ifndef gMAKEREAL_C
float *gMakeReal();
#endif
#ifndef G_EVALIMAG_C
float G_EvalImag();
#endif
#ifndef G_EVALMAG_C
float G_EvalMag();
#endif
#ifndef G_EVALREAL_C
float G_EvalReal();
#endif
#ifndef GMAKEIMAG_C
float *GMakeImag();
#endif
#ifndef GMAKEMAG_C
float *GMakeMag();
#endif
#ifndef GMAKEREAL_C
float *GMakeReal();
#endif
#ifndef HARDYCONVOLVEGABOR_C
void HardyConvolveGabor();
#endif
#ifndef ISADJACENT_C
int IsAdjacent();
#endif
#ifndef MKANISPF_C
pFloatImage MkAnisPF();
#endif
#ifndef MKCOSFILT_C
pFloatImage MkCosFilt();
#endif
#ifndef MKPOSTFILT_C
pFloatImage MkPostFilt();
#endif
#ifndef PTESA2D_C
int PtEsa2D();
#endif
#ifndef PTESA2DPF_C
int PtEsa2DPF();
#endif
#ifndef SPBMKBANKSPECTRUM_C
unsigned char *SpbMkBankSpectrum();
#endif
#ifndef SPBMKBANKSPECTRUMDC_C
unsigned char *SpbMkBankSpectrumDC();
#endif
#ifndef SPBMKBANKSPECTRUMAUX_C
unsigned char *SpbMkBankSpectrumAux();
#endif
#endif
