/*
 * int_2D_1component_ESA.c
 *
 *  function computes the 2D energy separation algorithms found in
 *  eq. 16 - 20 on page 94 of book V.  The input image f is delcared
 *  as int f[row_dim][col_dim].  The following output images are
 *  to be declared in the calling routine:
 *
 *  double u[row_dim][col_dim]         d/di phi(i,j)
 *  double v[row_dim][col_dim]         d/dj phi(i,j)
 *  double w[row_dim][col_dim]         |grad  phi(i,j)|
 *  double a[row_dim][col_dim]         amplitude modulation
 *  double theta[row_dim][col_dim]     arg grad phi(i,j)
 *
 *
 *  jph 5 June 1992
 */

#ifndef INT_2D_1COMPONENT_ESA_C
#define INT_2D_1COMPONENT_ESA_C
#include <stdio.h>
#include <math.h>
#include "teager2D.h"

void int_2D_1component_ESA(f,row_dim,col_dim,u,v,w,a,theta)

  int* f;         /* input image */
  int row_dim;    /* row dimension of x */
  int col_dim;    /* col dimension of x */
  double *u;         /* d/di phi(i,j) */
  double *v;         /* d/dj phi(i,j) */
  double *w;         /* |grad  phi(i,j)| */
  double *a;         /* amplitude modulation */
  double *theta;     /* arg grad phi(i,j) */
{

  int *fx;         /* d/di f */
  int *fy;         /* d/dj f */
  int *Phif;       /* 2D teager on f */
  int *Phifx;      /* 2D teager on fx */
  int *Phify;      /* 2D teager on fy */
  int i,j;         /* loop counters */
  double *denom;   /* denominator of eq. 20, p. 94, book V */
  int *i_denom;    /* i_denom = 0 where denom = 0, and 1 elsewhere */

 /*
  * detect zero dimension input
  */
  if ((row_dim==0) || (col_dim==0)) return;


 /*
  * find the partials of f with respect to rows and columns
  */
  fx = d_dRow(f,row_dim,col_dim);
  fy = d_dCol(f,row_dim,col_dim);

 /*
  * invoke the 2D teager operator on f, fx, and fy
  */
  Phif  = teager2D(f ,row_dim,col_dim);
  Phifx = teager2D(fx,row_dim,col_dim);
  Phify = teager2D(fy,row_dim,col_dim);
 /*
  * compute a, u, and v wherever possible
  */
  denom = (double *) malloc(row_dim * col_dim * sizeof(double));
  i_denom = (int *) malloc(row_dim * col_dim * sizeof(int));
  if ((i_denom == NULL) || (denom == NULL)) {
    printf("\nint_2D_1component_ESA.c : out of memory error !");
    return;
  }
  for (i=0; i<row_dim*col_dim; i++) {
    if ( (*(Phif + i)) > 0) {
      *(u + i) = sqrt((double)(*(Phifx + i)) / (double)(*(Phif + i)));
      *(v + i) = sqrt((double)(*(Phify + i)) / (double)(*(Phif + i)));
      if ((*(denom+i)=sqrt((double)(*(Phifx+i))+(double)(*(Phify+i))))
                                                           !=(double) 0.0)
        *(a + i) = (double)(*(Phif + i)) / (*(denom + i));
    }
    else {
      *(denom + i) = 1.0;
      *(a + i) = 0.0;
    }
    if ( (*(denom+i)) == (double) 0.0) (*(i_denom + i)) = 0;
    else (*(i_denom + i)) == 1;
  }

 /*
  * interpolate a, u, and v for cases where the denominator was zero
  */
  for (i=0; i<row_dim; i++) {
    for (j=0; j<col_dim; j++) {
      if ((*(Phif+i*col_dim+j)) == 0) {
        if ((*(Phifx+i*col_dim+j)) == 0) (*(u+i*col_dim+j)) = (double) 0.0;
        else d_interp(u,i,j,row_dim,col_dim,Phif);
        if ((*(Phify+i*col_dim+j)) == 0) (*(v+i*col_dim+j)) = (double) 0.0;
        else d_interp(v,i,j,row_dim,col_dim,Phif);
      }
      if (denom[i] == (double) 0.0) d_interp(a,i,j,row_dim,col_dim,i_denom);
    }
  }

 /*
  * compute w and theta
  */
  for (i=0; i<row_dim*col_dim; i++) {
    if ((*(w+i)=sqrt(pow((*(u+i)),(double)2)+pow((*(v+i)),(double)2)))
                                                                ==(double)0.0)
      *(theta + i) = (double) 0.0;
    else
      *(theta + i) = atan2( (*(u+i)), (*(v+i)) );

  }
 /*
  * free temp variables work space
  */
  free(fx);
  free(fy);
  free(Phif);
  free(Phifx);
  free(Phify);
  free(denom);
  free(i_denom);
} 
#endif
