/*
 *
 * ChanReconAux.c
 *
 *  This is a special version of ChanRecon.c that includes corner filter
 *  channels in addition to the regular filterbank tesselation.
 *
 *  10/3/96 jph
 *
 *===========================================================================
 *
 *  COMMENT BLOCK from ChanRecon.c:
 *
 *  For each channel,
 *      
 *     - run the esa to estimate AM and FM
 *     - reconstruct from the estimates for the channel
 *
 *  Apply the SAME scaling to all reconstructed components and make them
 *  binary images with maximum dynamic range.
 *
 *  Write these binary images out to files.
 *
 *  This program was made by modifying DomComp.c
 *
 *  usage: ChanRecon ParameterFileName
 *
 *  jph 5/3/96
 *
 *  REVISION 9/3/96:  add DC filter.
 *
 */

#ifndef CHANRECONAUX_C
#define CHANRECONAUX_C

#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include <string.h>

#include "teager2D.h"

/*
 * Function prototypes
 */
void Initialize();
void ReadParametersFile();
void dom_esa();
void AllocArrays();
void DomRecon();

main(argc,argv)

  int   argc;
  char *argv[];

{

  pFloatImage        X;              /* input image */
  pFloatImage        Y;              /* filtered image */
  filterbank        *G;              /* the filter bank */
  pFloatImage        g;              /* postfilter image */
  int                size;           /* row/col dim of the square input image */
  int                so2;            /* size / 2 */
  int                n;              /* size squared */
  int                i;              /* counter */
  int                j;              /* counter */
  int                row;            /* image row counter */
  int                col;            /* image col counter */
  int                looplim;        /* terminal value for a loop counter */
  int                terminate;      /* terminal value for a loop counter */
  int                ZeroPad;        /* 1 if zero padding is to be used */
  int                Reflect;        /* 1=fill 0 padded areas w/ reflection */
  int                ReconInterval;  /* boundary condition interval for recon */
  int               *maxfilt;        /* Which channel max'd criterion at i,j */
  BYTE              *btmp;           /* temp byte pointer */
  float             *ftmp;           /* temp float pointer */
  double            *dtmp;           /* temp double pointer */
  float             *criterion;      /* filter selection criterion */
  float             *A;              /* A estimates accross all 40 filters */
  float             *Aresize;        /* resized A estimates */
  float             *U;              /* U estimates accross all 40 filters */
  float             *Uresize;        /* resized U estimates */
  float             *V;              /* V estimates accross all 40 filters */
  float             *Vresize;        /* resized V estimates */
  BYTE              *u_valid;        /* 1 where the u estimate is valid */
  BYTE              *v_valid;        /* 1 where the v estimate is valid */
  BYTE              *a_valid;        /* 1 where the a estimate is valid */
  BYTE              *a_validresize;  /* resized a_valid array */
  BYTE              *u_validresize;  /* resized u_valid array */
  BYTE              *v_validresize;  /* resized v_valid array */
  double            *Theta;          /* phase initial conditions for recon */
  double            *Thetaresize;    /* resized Theta array */
  int                PostFilter;     /* 1 if postfiltering is to be done */
  int                Nfmed;          /* No. times to apply 3x3 median filter */
  int                NfSigCmprs;     /* No. times to apply fSigCmprs */
  float              fSigCmprsSigmas;/* Sigma value to use for fSigCmprs */
  float             *RCI;            /* Dummy imag part of recon components */
  float            **RCR;            /* Real parts of reconstructed comps */
  float              minG;           /* global min of reconstructed comps */
  float              maxG;           /* global max of reconstructed comps */
  char               RootFn[FNSIZE]; /* Root filename for Comp Real data */
  char               Fn[FNSIZE];     /* filename for Component Real data */

 /*
  * Check for right number of arguments
  */
  if (argc != 2) {
    printf("\n\n%s: Reconstruct AM/FM components for each channel.",argv[0]);
    printf("\n     Use DC filter and Corner Filters.");
    printf("\nusage: %s ParameterFileName\n",argv[0]);
    exit(-1);
  }
#ifdef DEBUG
 /*
  * Link to PA so we can use it in the debugger 
  */
  if (!argc) {
    PA((void *)X, 0, -1, 'b');
    PRINTA((void *)X,0,-1,'b',"delete.me");
    WRITEA((void *)X,0,-1,'b',"delete.me");
    image2disk((void *)X,1,1,'*',"delete.me");
    SpecMag2disk(X,"delete.me");
    PC(X,0,0,-1,'b',0);
    PRINTC(X,0,0,-1,'b',0,"delete.me");
    WRITEC(X,0,0,-1,'b',0,"delete.me");
  }
#endif
  X = (pFloatImage)malloc(sizeof(FloatImage));
  Y = (pFloatImage)malloc(sizeof(FloatImage));

  Initialize(argv[1],&size,&ZeroPad,&Reflect,&X->Real,&X->Imag,&X->FFTReal,
             &X->FFTImag,&G,&ReconInterval,&PostFilter,&Nfmed,&NfSigCmprs,
             &fSigCmprsSigmas,RootFn);
  X->size = X->cols = X->rows = size;
  X->square = 1; 
  AllocFloatImage(Y,size);
  n = size * size;
  if (ZeroPad) {
    looplim = size >> 1;
  } else {
    looplim = size;
  }

 /*
  * Allocate working arrays
  */
  AllocArrays(argv,G,looplim,n,PostFilter,&a_valid,&u_valid,&v_valid,&maxfilt,
              &criterion,&A,&U,&V,&Theta,&RCI,&RCR);

  if (PostFilter) {
    printf("\n\nPost Filtering...\n");
  }

 /*
  * Loop for filters
  */
  for (i=0; i < G->TotNumFilters; i++) {
    printf("\nDoing Filter %d...",i);
    HardyConvolveGabor(Y,X,G->filt[i]);

   /*
    * Reset criterion, a, u, v, maxfilt, valid, and Theta arrays
    */
    for (j=0; j < n; j++) {
      criterion[j] = (float)-1.0;
      A[j] = U[j] = V[j] = (float)0.0;
      maxfilt[j] = -1;
      a_valid[j] = u_valid[j] = v_valid[j] = 0;
    }

   /*
    * Dominant ESA
    */
    dom_esa(Y,looplim,G->filt[i],criterion,A,U,V,G,maxfilt,a_valid,u_valid,
            v_valid,ReconInterval,Theta);

   /*
    * Post Filter U, V and A
    */
    if (PostFilter) {
      if (G->filt[i]->isotropicPF) {
        g = MkPostFilt(G->filt[i]->sigma,G->filt[i]->kappa,size);
      } else {
        g = MkAnisPF(G->filt[i]->sigma,G->filt[i]->kappa_x,
                     G->filt[i]->kappa_y,G->filt[i]->Theta,size,1);
      }
      DoPostFilter(g,size,U,U,G,i);
      DoPostFilter(g,size,V,V,G,i);
      ReComputeAsm(Y,U,V,A,G->filt[i]);
      DoPostFilter(g,size,A,A,G,i);
      g = KillFloat(g);
    }      /* if (PostFilter) */

   /*
    * Resize results and valid arrays if ZeroPad 
    */
    if (ZeroPad) {
      so2 = size >> 1;
  
      ftmp = (float *)malloc(so2*so2*sizeof(float));
      btmp = (BYTE  *)malloc(so2*so2*sizeof(BYTE));
      for (row=0; row < so2; row++) {
        memcpy(ftmp+row*so2,A+row*size,so2*sizeof(float));
        memcpy(btmp+row*so2,a_valid+row*size,so2*sizeof(BYTE));
      }
      Aresize = ftmp;
      a_validresize = btmp;
  
      ftmp = (float *)malloc(so2*so2*sizeof(float));
      btmp = (BYTE  *)malloc(so2*so2*sizeof(BYTE));
      for (row=0; row < so2; row++) {
        memcpy(ftmp+row*so2,U+row*size,so2*sizeof(float));
        memcpy(btmp+row*so2,u_valid+row*size,so2*sizeof(BYTE));
      } 
      Uresize = ftmp;
      u_validresize = btmp;
  
      ftmp = (float *)malloc(so2*so2*sizeof(float));
      btmp = (BYTE  *)malloc(so2*so2*sizeof(BYTE));
      for (row=0; row < so2; row++) {
        memcpy(ftmp+row*so2,V+row*size,so2*sizeof(float));
        memcpy(btmp+row*so2,v_valid+row*size,so2*sizeof(BYTE));
      } 
      Vresize = ftmp;
      v_validresize = btmp;
  
      dtmp = (double *)malloc(so2*so2*sizeof(double));
      for (row=0; so2 < size; row++) {
        memcpy(dtmp+row*so2,Theta+row*size,so2*sizeof(double));
      } 
      Thetaresize = dtmp;

     /*
      * Interpolate for yet indeterminate values
      */
      for (j=0,row=0; row<so2; row++) {
        for (col=0; col<so2; col++,j++) {
          if (!u_valid[j]) f_interp(U,row,col,so2,so2,u_valid);
          if (!v_valid[j]) f_interp(V,row,col,so2,so2,v_valid);
          if (!a_valid[j]) f_interp(A,row,col,so2,so2,a_valid);
        }    /* for col */
      }      /* for row */

     /*
      * Reconstruct 
      */
      DomRecon(Aresize,Uresize,Vresize,so2,RCR[i],RCI,ReconInterval,
               Thetaresize);
  
      free(Aresize);
      free(Uresize);
      free(Vresize);
      free(Thetaresize);
      free(a_validresize);
      free(u_validresize);
      free(v_validresize);

    } else {     /* not Zero Padding */

     /*
      * Interpolate for yet indeterminate values
      */
      for (j=0,row=0; row<size; row++) {
        for (col=0; col<size; col++,j++) {
          if (!u_valid[j]) f_interp(U,row,col,size,size,u_valid);
          if (!v_valid[j]) f_interp(V,row,col,size,size,v_valid);
          if (!a_valid[j]) f_interp(A,row,col,size,size,a_valid);
        }    /* for col */
      }      /* for row */
    
     /*
      * Reconstruct 
      */
      DomRecon(A,U,V,size,RCR[i],RCI,ReconInterval,Theta);

    }            /* else */
  }              /* for i */

 /*
  * Invoke fmed3x3 as required
  */
  terminate = G->TotNumFilters;
  if (Nfmed) {
    printf("\nInvoking fmed3x3...");
    for (j=0; j < terminate; j++) {
      printf("\n     Channel %02d...",j);
      for (i=1; i < Nfmed; i++) {
        ftmp = fmed3x3(RCR[j],looplim,0,0,looplim-1,looplim-1);
        free(RCR[j]);
        RCR[j] = ftmp;
      }
    }
  }

 /*
  * Invoke fSigCmprs as required
  */
  if (NfSigCmprs) {
    terminate = G->TotNumFilters;
    printf("\nInvoking fSigCmprs...");
    for (j=0; j < terminate; j++) {
      printf("\n     Channel %02d...",j);
      for (i=1; i < NfSigCmprs; i++) {
        ftmp = fSigCmprs(RCR[j],looplim,fSigCmprsSigmas);
        free(RCR[j]);
        RCR[j] = ftmp;
      }
    }
  }

 /*
  * Find the global min and max of the reconstructed components
  */
/*
  minG = maxG = RCR[0][0];
  for (i=0; i < G->TotNumFilters; i++) {
    for (j=0; j < looplim*looplim; j++) {
      if (RCR[i][j] < minG) {
        minG = RCR[i][j];
      } else {
        if (RCR[i][j] > maxG) {
          maxG = RCR[i][j];
        }
      }
    }
  }
*/
      

 /*
  * Set 1st 2 pixs of each recon comp to minG and maxG for uniform scaling
  */
/*
  for (i=0; i < G->TotNumFilters; i++) {
    RCR[i][0] = minG;
    RCR[i][1] = maxG;
  }
*/
  
 /*
  * Write the reconstructed components out to disk files as BYTE data
  */
  for (i=0; i < G->TotNumFilters; i++) {
    sprintf(Fn,"%s%02d.float",RootFn,i);
    f2disk(RCR[i],looplim,looplim,Fn);
    /*sprintf(Fn,"%s%02d.bin",RootFn,i);*/
    /*image2disk(RCR[i],looplim,looplim,'f',Fn);*/
  }

  printf("\n\n");
  return;

} /*-------------------------------------------- main --------------*/

/*
 * Initialize
 *
 *  Process the paramters file
 *  Read the input image, if real make hardy to get the imaginary part
 *  Compute the input spectrum
 *  Construct the filter bank
 */

void Initialize(ParmFn,size,ZP,Reflect,xReal,xImag,XReal,XImag,G,ReconInterval,
                PostFilter,Nfmed,NfSigCmprs,fSigCmprsSigmas,RootFn)

  char        *ParmFn;           /* filename of parameters file */
  int         *size;             /* row/col dim of the square input image */
  int         *ZP;               /* 1 if zero padding is to be used */
  int         *Reflect;          /* 1=fill 0 padded areas w/ reflection */
  float      **xReal;            /* real part of input image */
  float      **xImag;            /* imag part of input image */
  float      **XReal;            /* real part of input spectrum */
  float      **XImag;            /* imag part of input spectrum */
  filterbank **G;                /* the filter bank */
  int         *ReconInterval;    /* boundary condition interval for recon */
  int         *PostFilter;       /* 1=PostFiltering to be performed */
  int         *Nfmed;            /* No. times to apply 3x3 median filter */
  int         *NfSigCmprs;       /* No. times to apply fSigCmprs */
  float       *fSigCmprsSigmas;  /* Sigma value to use for fSigCmprs */
  char        *RootFn;           /* root filename for recon components */
{

  int         i;                 /* counter */
  int         row;               /* counter for image rows */
  int         col;               /* counter for image cols */
  int         so2;               /* *size / 2 */
  char        datatype;          /* type of in dat: b=byte f=float d=double */
  int         Conjugate;         /* 1=conjugate input.  See Book X, p. 64 */
  char        Rfn[FNSIZE];       /* file for Real part of input image */
  char        Ifn[FNSIZE];       /* file for Imag part of input image */
  double      r0;                /* rad. center freq. of 1st filt on each ray */
  float       R;                 /* common ratio of radii of adjcnt filtrs */
  float       obw;               /* Eta-peak octave bandwidth of filters */
  float       Eta;               /* OBW is defined in terms of Eta-peak. */
  int         isoPF;             /* 1=isotropic post filters */
  float       kappa;             /* Post Filter space constant scale factor */
  float       kappa_x;           /* postfilter horiz space cnst scale factor */
  float       kappa_y;           /* postfilter vert  space cnst scale factor */
  float      *tfloat;            /* temporary float array */
  float      *to;                /* area in memory to copy floats to */
  float      *from;              /* area in memory to copy floats from */
  float      *pfloat;            /* ptr to location to fill w/ zero */

 /*
  * Read the parameter file
  */
  ReadParametersFile(ParmFn,&datatype,&Conjugate,Rfn,Ifn,size,ZP,Reflect,&r0,
                     &R,&Eta,&obw,ReconInterval,PostFilter,&isoPF,
                     &kappa,&kappa_x,&kappa_y,Nfmed,NfSigCmprs,fSigCmprsSigmas,
                     RootFn);
  r0 = r0 / (double)(*size);
 
 /*
  * If zero padding, double size 
  */
  if (*ZP) {
    *size *= 2;
  }
  so2 = *size >> 1;

 /*
  * Make the filter bank
  */ 
  *G = CreateFilterBankAux(r0,R,obw,Eta,isoPF,kappa,kappa_x,kappa_y);

 /*
  * Allocate Spectra
  */
  if ((*XReal = (float *)malloc(*size**size*sizeof(float))) == NULL) {
    printf("\nInitialize: free store exhausted\n");
    exit(-1);
  }
  if ((*XImag = (float *)malloc(*size**size*sizeof(float))) == NULL) {
    printf("\nInitialize: free store exhausted\n");
    exit(-1);
  }

 /*
  * Get the input image and its spectrum
  */
  if (*ZP) {
    tfloat = readasf(so2,so2,Rfn,datatype);
    if ((*xReal = (float *)malloc(*size**size*sizeof(float))) == NULL) {
      printf("\nInitialize: free store exhausted\n");
      exit(-1);
    }
    for (to=*xReal,from=tfloat,row=0; row < so2; row++,to+=*size,from+=so2) {
      memcpy(to, from, so2*sizeof(float));
      for (pfloat=to+so2,col=so2; col < *size; col++) {
        if (!*Reflect) {
          *(pfloat++) = (float)0.0;
        } else {
          *(pfloat++) = *(to + *size-col-1);
        }
      }
    }
    free(tfloat);
    if (!*Reflect) {
      for (pfloat=to,i=so2**size; i < *size**size; i++) {
        *(pfloat++) = (float)0.0;
      }
    } else {
      for (pfloat=to,row=1; row < *size; row+=2) {
        for (col=0; col < *size; col++,pfloat++) {
          *pfloat = *(pfloat - row**size);
        }
      }
    }
  } else {
    *xReal = readasf(*size,*size,Rfn,datatype);
  }

  if (strcmp(Ifn,"none")) {                     /* filename is not "none" */
    if (*ZP) {
      tfloat = readasf(so2,so2,Ifn,datatype);
      if ((*xImag = (float *)malloc(*size**size*sizeof(float))) == NULL) {
        printf("\nInitialize: free store exhausted\n");
        exit(-1);
      }
      for (to=*xImag,from=tfloat,row=0; row < so2; row++,to+=*size,from+=so2) {
        memcpy(to, from, so2*sizeof(float));
        for (pfloat=to+so2,col=so2; col < *size; col++) {
          if (!*Reflect) {
            *(pfloat++) = (float)0.0;
          } else {
            *(pfloat++) = *(to + *size-col-1);
          }
        }
      }
      free(tfloat);
      if (!*Reflect) {
        for (pfloat=to,i=so2**size; i < *size**size; i++) {
          *(pfloat++) = (float)0.0;
        }
      } else {
        for (pfloat=to,row=1; row < *size; row+=2) {
          for (col=0; col < *size; col++,pfloat++) {
            *pfloat = *(pfloat - row**size);
          }
        }
      }
    } else {
      *xImag = readasf(*size,*size,Ifn,datatype);
    }
    if (Conjugate) {
      for (i=0; i < *size**size; i++) {
        (*xImag)[i] *= (float)-1.0;
      }
    }
/*    f_u_norm(*xReal, *xImag, *size, *size);*/
    fft2d(*xReal,*xImag,*XReal,*XImag,*size,1);
  } else {                         /* input is real, must create imag part */
/*    f_u_norm(*xReal, NULL, *size, *size);*/
    if ((*xImag = (float *)malloc(*size**size*sizeof(float))) == NULL) {
      printf("\nInitialize: free store exhausted\n");
      exit(-1);
    }
    MakeHardy(*xReal,*xImag,*XReal,*XImag,*size);
    if (Conjugate) {
      printf("\n\n >>>  CONJUGATING input made by MakeHardy...\n");
      for (i=0; i < *size**size; i++) {
        (*xImag)[i] *= (float)-1.0;
      }
      fft2d(*xReal,*xImag,*XReal,*XImag,*size,1);
    }
  }

  return;
} /*---------------------------------------------- Initialize ----------*/

/*
 * ReadParametersFile
 *
 *  read the data out of the parameters file
 *
 */

void ReadParametersFile(ParmFn,pdatatype,pConjugate,Rfn,Ifn,psize,pZP,pReflect,
                        pr0,pR,pEta,pobw,pReconInterval,pPostFilter,
                        pisoPF,pkappa,pkappa_x,pkappa_y,pNfmed,pNfSigCmprs,
                        pfSigCmprsSigmas,RootFn)
  char   *ParmFn;
  char   *pdatatype;
  int    *pConjugate;
  char   *Rfn;
  char   *Ifn;
  int    *psize;
  int    *pZP;
  int    *pReflect;
  double *pr0;
  float  *pR;
  float  *pEta;
  float  *pobw;
  int    *pReconInterval;
  int    *pPostFilter;
  int    *pisoPF;
  float  *pkappa;
  float  *pkappa_x;
  float  *pkappa_y;
  int    *pNfmed;
  int    *pNfSigCmprs;
  float  *pfSigCmprsSigmas;
  char   *RootFn;
{

  FILE      *parmfp;            /* ptr to parameter file */
  char       instr[256];        /* buffer to read into  */

  if ((parmfp = fopen(ParmFn,"r")) == NULL) {
    printf("\n\nReadParametersFile: unable to open %s\n",ParmFn);
    exit(-1);
  }

  fscanf(parmfp,"Input Data Type= %s\n",instr);
  *pdatatype = instr[0];
  fscanf(parmfp,"Conjugate Input= %s\n",instr);
  if ((instr[0] == 'N') || (instr[0] == 'n')) {
    *pConjugate = 0;
  }
  else {
    *pConjugate = 1;
  }
  fscanf(parmfp,"Real Image File= %s\n",Rfn);
  fscanf(parmfp,"Imag Image File= %s\n",Ifn);
  fscanf(parmfp,"Size= %d\n",psize);
  fscanf(parmfp,"Zero Pad= %s\n",instr);
  if ((instr[0] == 'Y') || (instr[0] == 'y')) {
    *pZP = 1;
  }
  else {
    *pZP = 0;
  }
  fscanf(parmfp,"Reflect= %s\n",instr);
  if ((instr[0] == 'Y') || (instr[0] == 'y')) {
    *pReflect = 1;
  }
  else {
    *pReflect = 0;
  }
  fscanf(parmfp,"r0= %lf\n",pr0);
  fscanf(parmfp,"R= %f\n",pR);
  fscanf(parmfp,"Eta= %f\n",pEta);
  fscanf(parmfp,"OBW= %f\n",pobw);
  fscanf(parmfp,"Reconstruction Interval= %d\n",pReconInterval);
  fscanf(parmfp,"Postfilter= %s\n",instr);
  if ((instr[0] == 'N') || (instr[0] == 'n')) {
    *pPostFilter = 0;
  }
  else {
    *pPostFilter = 1;
  }
  fscanf(parmfp,"Isotropic PF= %s\n",instr);
  if ((instr[0] == 'N') || (instr[0] == 'n')) {
    *pisoPF = 0;
  }
  else {
    *pisoPF = 1;
  }
  fscanf(parmfp,"kappa= %f\n",pkappa);
  fscanf(parmfp,"kappa_x= %f\n",pkappa_x);
  fscanf(parmfp,"kappa_y= %f\n",pkappa_y);
  fscanf(parmfp,"fmed3x3 Applications= %d\n",pNfmed);
  fscanf(parmfp,"fSigCmprs Applications= %d\n",pNfSigCmprs);
  fscanf(parmfp,"Sigma for fSigCmprs= %f\n",pfSigCmprsSigmas);
  fscanf(parmfp,"Root Out Fn= %s\n",RootFn);
  fclose(parmfp);
  return;
} /*--------------------------- ReadParametersFile -------------------------*/

/*
 * dom_esa
 *
 *  function evaluates the esa at every point in an image and checks to see if
 *  if the current response image has the dominant response of those yet seen.
 *
 */

void dom_esa(Y,looplim,G,criterion,a,u,v,Bank,maxfilt,a_valid,u_valid,v_valid,
             ReconInterval,Theta)

  pFloatImage     Y;              /* response image */
  int             looplim;        /* terminal value for a loop counter */
  FilterPtr       G;              /* Y was output from this filter */
  float          *criterion;      /* filter selection criterion */
  float          *a;              /* a estimates */
  float          *u;              /* u estimates */
  float          *v;              /* v estimates */
  filterbank     *Bank;           /* the filterbank */
  int            *maxfilt;        /* which chan max'd criterion @ (i,j) */
  BYTE           *a_valid;        /* 1 where the a estimate is valid */
  BYTE           *u_valid;        /* 1 where the u estimate is valid */
  BYTE           *v_valid;        /* 1 where the v estimate is valid */
  int             ReconInterval;  /* boundary condition interval for recon */
  double         *Theta;          /* phase initial conditions for recon */
{

  int             row,col;        /* loop counters for image coordinates */
  int             k;              /* loop counter for indexing arrays */
 
 /*
  * Apply the ESA at each point in the image
  */
  for (row=0; row < looplim; row++) {
    for (col=0; col < looplim; col++) {
      k = row * Y->size + col;
      if (PtEsa2D(Y,row,col,G,&criterion[k],&u[k],&v[k],&a[k],
                  &u_valid[k],&v_valid[k],&a_valid[k])) {
        maxfilt[k] = G->index;
        if ((!(row % ReconInterval)) && (!(col % ReconInterval))) {
          if ((Y->Real[k] == (float)0.0) && (Y->Imag[k] == (float)0.0)) {
            Theta[k] = (double)0.0;
          } else {
            Theta[k] = atan2((double)Y->Imag[k],(double)Y->Real[k]);
          }
        }
      }
    }
  }
  return;
} /*----------------------------------------- dom_esa -----------------------*/

/*
 * AllocArrays
 *
 *  Allocate Working Arrays.
 */

void AllocArrays(argv,G,looplim,n,PostFilter,a_valid,u_valid,v_valid,maxfilt,
                 criterion,A,U,V,Theta,RCI,RCR)

  char              *argv[];         /* argument list from command line */
  filterbank        *G;              /* the filter bank */
  int                looplim;        /* row/col looplim for estimation */
  int                n;              /* size * size again */
  int                PostFilter;     /* 1=perform postfiltering */
  BYTE             **a_valid;        /* 1 where the a estimate is valid */
  BYTE             **u_valid;        /* 1 where the u estimate is valid */
  BYTE             **v_valid;        /* 1 where the v estimate is valid */
  int              **maxfilt;        /* Which channel max'd criterion at i,j */
  float            **criterion;      /* filter selection criterion */
  float            **A;              /* A estimates accross all filters */
  float            **U;              /* U estimates accross all filters */
  float            **V;              /* V estimates accross all filters */
  double           **Theta;          /* boundary condition interval for recon */
  float            **RCI;            /* dummy imag part of reconstructions */
  float           ***RCR;            /* real parts of reconstructed comps */
{
 
  int                i;              /* loop counter */
 
 /*
  * Allocate working arrays
  */
  if ((*a_valid = (BYTE *) malloc(n * sizeof(BYTE))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*u_valid = (BYTE *) malloc(n * sizeof(BYTE))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*v_valid = (BYTE *) malloc(n * sizeof(BYTE))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*maxfilt = (int *)malloc(n * sizeof(int))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  for (i=0; i < n; i++) (*maxfilt)[i] = -1;
  if ((*criterion = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  for (i=0; i < n; i++) (*criterion)[i] = (float)-1.0;
  if ((*A = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*U = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*V = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*Theta = (double *)malloc(n * sizeof(double))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*RCI = (float *)malloc(looplim*looplim * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }

  if ((*RCR = (float **)malloc(G->TotNumFilters * sizeof(float *))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  for (i=0; i < G->TotNumFilters; i++) {
    if (((*RCR)[i]=(float *)malloc(looplim*looplim*sizeof(float))) == NULL) {
      printf("\n%s: free store exhausted.\n",argv[0]);
      exit(-1);
    }
  }
 
  return;
} /*---------------------  AllocArrays  --------------------------------------*/
 
/*
 * DomRecon
 *
 *  Reconstruct the dominant component.
 *
 */
 
void DomRecon(a,u,v,size,DomCompR,DomCompI,ReconInterval,Theta)
 
  int             size;           /* length of signal for reconstruction */
  float          *a;              /* a estimates */
  float          *u;              /* u estimates */
  float          *v;              /* v estimates */
  float          *DomCompR;      /* Real part of the dominant component */
  float          *DomCompI;      /* Imag part of the dominant component */
  int             ReconInterval;  /* boundary condition interval for recon */
  double         *Theta;          /* phase initial conditions for recon */
{
 
  int             row;            /* loop counter for image rows */
  int             col;            /* loop counter for image cols */
  int             ROW;            /* row counter for recon blocks */
  int             COL;            /* col counter for recon blocks */
  int             R,C;            /* counters for rows and cols within block */
  int             k;              /* counter */
  double         *phase;          /* instantaneous phase */

 /*
  * Allocate phase
  */
  if ((phase = (double *)malloc(size*size*sizeof(double))) == NULL) {
    printf("\nDomRecon:  free store exhausted.\n");
    exit(-1);
  }
 
 /*
  * Outer loops set top left corner of each reconsructed block
  */
  for (ROW=0; ROW < size; ROW+=ReconInterval) {
    for (COL=0; COL < size; COL+=ReconInterval) {
     /*
      * Reconstruct phase on first row of block
      */
      k = ROW*size + COL;
      phase[k] = Theta[k];
      for (k++,C=1,col=COL+1; (col < size) && (C < ReconInterval); 
           col++,C++,k++) {
        phase[k] = phase[k-1] + (double)2.0 * (double)M_PI * (double)u[k-1];
      }
     /*
      * Reconstruct phase on first col of block
      */
      for (k=(ROW+1)*size+COL,R=1,row=ROW+1; 
           (row < size) && (R < ReconInterval);
           row++,R++,k+=size) {
        phase[k] = phase[k-size] + (double)2.0 * (double)M_PI 
                                               * (double)v[k-size];
      }
     /*
      * Reconstruct phase at interior points in the block
      */
      for (R=1,row=ROW+1; (row < size) && (R < ReconInterval); R++,row++) {
        for (C=1,col=COL+1; (col < size) && (C < ReconInterval); C++,col++) {
          k = row*size + col;
          phase[k] = (double)0.5 * (phase[k-1] + phase[k-size] 
                     + (double)2.0 * (double)M_PI 
                       * ((double)u[k-1] + (double)v[k-size]));
        }
      }
    }    /* COL */
  }      /* ROW */

 /*
  * Reconstruct image values
  */
  for (k=row=0; row < size; row++) {
    for (col=0; col < size; col++,k++) {
      DomCompR[k] = a[k] * (float)cos(phase[k]);
      DomCompI[k] = a[k] * (float)sin(phase[k]);
    }
  }
  free(phase);
  return;
} /*----------------------------------------- DomRecon ----------------------*/
#endif
