/*
 * ChkInBand.c
 *
 *   Test a track for a condition where there IS NOT a block of processed
 *   pixels where the filtered frequency is IN BAND for the channel filter
 *   where the observations where taken.
 *
 *   If this condition is TRUE, kill the track.
 *
 *   This implies that BETTER observations where available in another channel.
 *   It also means that a BETTER track should have been started on the other
 *   channel (or else this track would have moved to that channel).
 *
 *   For documentation of the 1D version, see ChkInBand1D.c and Book XI, p. 41.
 *
 *   1. Find the height (width) of the block.
 *   2. If track is new and not enough rows have been tracked to try the test,
 *      then return.
 *   3. See if there is a big enough block in the previously tracked rows
 *      (starting in any column) where a big enough fraction of the filtered
 *      frequencies are in band.  If so, return ok.
 *      A.  check blocks beginning in each column
 *      B.  count number of pixels where filtered frequency is in band
 *      C.  find ratio of pixels in block where frequency is in band
 *      D.  if big enough, return ok
 *      E.  continue until all blocks have been checked.  If the end is
 *          reached w/out finding a good block, kill track.
 *
 *  3/15/96 jph
 */

#ifndef CHKINBAND_C
#define CHKINBAND_C

#define MIN_BLOCK_SIZE 8
#define MIN_NUMBER_BLOCKS 6
#define MIN_PERCENT_IN_BAND ((float)0.9)

#include "track.h"
 
void ChkInBand(T,idx,row,col,size,minCol2Track,MaxCol2Track,G,RptFp)

  trackP       T;               /* Track to be tested */
  int          idx;             /* index of track being tested */
  int          row;             /* Image Row of current observations */
  int          col;             /* Image Col of current observations */
  int          size;            /* size of (possibly zero padded) input image */
  int          minCol2Track;    /* first col to perform track proc. on */
  int          MaxCol2Track;    /* last  col to perform track proc. on */
  filterbank  *G;               /* the filter bank */
  FILE        *RptFp;           /* ptr to stream for track reporting */
{
 
  int     BlockSize;            /* row/col dim of test block */
  int     MaxBlockSize;         /* Maximum size for test blocks */
  int     StartRow;             /* first row in blocks to check */
  int     StopRow;              /* last row in blocks to check */
  int     StartCol;             /* first col in a test block */
  int     StopCol;              /* last  col in a test block */
  int     TestRow;              /* current row in test */
  int     TestCol;              /* current col in test */
  int     k;                    /* cursor into image */
  float   NumPix;               /* number of pixels in test block */
  float   NumInBand;            /* num pix in block w/freq in band */

 /*
  * Do not test inactive tracks
  */
  if ((!T->active) || (T->dormant)) {
    return;
  }

 /*
  * Find BlockSize
  */
  k = row*size + col;
  BlockSize = (int)ceil((double)2.0 * G->filt[(int)T->ObsChan[k]]->sigma);
  MaxBlockSize = MaxCol2Track - minCol2Track - MIN_NUMBER_BLOCKS + 2;
  if (BlockSize > MaxBlockSize) {
    BlockSize = MaxBlockSize;
  } else {
    if (BlockSize < MIN_BLOCK_SIZE) {
      BlockSize = MIN_BLOCK_SIZE;
    }
  }

 /*
  * If track too new to test, return
  */
  if ((row - BlockSize) <= T->Hist.Start.row) {
    return;
  }

 /*
  * Check blocks.  Return if a block is found that passes the test.
  */
  StartRow = row - BlockSize;
  StopRow  = row - 1;
  StartCol = minCol2Track; 
  StopCol  = minCol2Track+BlockSize-1; 
  NumPix   = (float)(BlockSize * BlockSize);

 /*
  * Test first block
  */
  NumInBand = 0;
  for (TestRow=StartRow; TestRow <= StopRow; TestRow++) {
    for (TestCol=StartCol; TestCol <= StopCol; TestCol++) {
      k = TestRow*size + TestCol;
      if (T->ObsChan[k] != (float)-1.0) {
        if ((G_EvalMag(G->filt[(int)T->ObsChan[k]],fabs((double)T->U_filt[k]),
                                                        (double)T->V_filt[k]) 
                  / G->filt[(int)T->ObsChan[k]]->GMax) >= G->Eta / (float)1.5) {
          NumInBand++;
        }
      }
    }
  }
  if ((NumInBand / NumPix) >= MIN_PERCENT_IN_BAND) {
    return;
  }

 /*
  * Loop on test blocks.  Test one block each iteration.  Return if pass test.
  */
  for ( ; StopCol < MaxCol2Track; ) {
   /*
    * Adjust NumInBand for first col of last block tested
    */
    for (TestCol=StartCol,TestRow=StartRow; TestRow <= StopRow; TestRow++) {
      k = TestRow*size + TestCol;
      if (T->ObsChan[k] != (float)-1.0) {
        if ((G_EvalMag(G->filt[(int)T->ObsChan[k]],fabs((double)T->U_filt[k]),
                                                        (double)T->V_filt[k]) 
                 / G->filt[(int)T->ObsChan[k]]->GMax) >= G->Eta / (float)1.5) {
          NumInBand--;
        }
      }
    }
   /*
    * Adjust NumInBand for last col of block now being tested
    */
    StartCol++;
    StopCol++;
    for (TestCol=StopCol,TestRow=StartRow; TestRow <= StopRow; TestRow++) {
      k = TestRow*size + TestCol;
      if (T->ObsChan[k] != (float)-1.0) {
        if ((G_EvalMag(G->filt[(int)T->ObsChan[k]],fabs((double)T->U_filt[k]),
                                                        (double)T->V_filt[k])
                 / G->filt[(int)T->ObsChan[k]]->GMax) >= G->Eta / (float)1.5) {
          NumInBand++;
        }
      }
    }
    if ((NumInBand / NumPix) >= MIN_PERCENT_IN_BAND) {
      return;
    }
  }     /* for blocks */

 /*
  * If here, then NO block passed.  Kill track.
  */
  fprintf(RptFp,"\nChkInBand: Killing Track %d at row=%d col=%d: %s",
           idx,row,col,"too long out of channel BW.");
  KillTrack(T,size);
 
  return;
}/*--------------------------------------------  ChkInBand  ------------------*/
#endif
