/*
 *
 * DomComp.c
 *
 *  Run the ICIP ESA for dominant component analysis.  At each point in the
 *  domain, estimation is performed using the data from the channel that
 *  maximizes the filter selection criterion |t| / Gmax.
 *
 *  This program was made by modifying the program 
 *     /anchovy/export/home/joebob/track/src/MkObs.c
 *
 *  Input parameters are read from a params file specified as the command line
 *  option.  See p. XXX of Book VI for a description of the paramter file.
 *
 *  REVISED 8/14/94: see book VIII p. 82.
 *
 *  REVISED 11/4/94: f_u_norm calls commented out.
 *
 *  MAJOR REVISION: 2/28/96
 *
 *  New Postfiltering scheme added 4/12/96.  Book XI, p. 119, p. 151.
 *
 *  usage: DomComp ParameterFileName
 *
 *  jph 11/24/94
 */

#ifndef DOMCOMP_C
#define DOMCOMP_C

#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include <string.h>

#include "teager2D.h"

/*
 * Function prototypes
 */
void Initialize();
void ReadParametersFile();
void dom_esa();
void AllocArrays();
void DomRecon();
void InsertAsm();

main(argc,argv)

  int   argc;
  char *argv[];

{

  filterbank        *G;              /* the filter bank */
  pFloatImage        g;              /* postfilter image */
  pFloatImage        X;              /* input image */
  pFloatImage        Y;              /* filtered image */
  int                size;           /* row/col dim of the square input image */
  int                n;              /* size squared */
  int                i;              /* counter */
  int                j;              /* counter */
  int                k;              /* loop counter */
  int                row;            /* image row counter */
  int                col;            /* image col counter */
  int                looplim;        /* terminal value for a loop counter */
  int                ZeroPad;        /* 1 if zero padding is to be used */
  int                Reflect;        /* 1=fill 0 padded areas w/ reflection */
  int                ReconInterval;  /* boundary condition interval for recon */
  int               *maxfilt;        /* Which channel max'd criterion at i,j */
  int               *itmp;           /* temp int poineter */
  BYTE              *btmp;           /* temp byte pointer */
  float             *ftmp;           /* temp float pointer */
  double            *dtmp;           /* temp double pointer */
  float             *criterion;      /* filter selection criterion */
  float             *A;              /* A estimates accross all 40 filters */
  float             *Asm;            /* smoothed A estimates */
  float             *U;              /* U estimates accross all 40 filters */
  float             *Usm;            /* smoothed U estimates */
  float             *V;              /* V estimates accross all 40 filters */
  float             *Vsm;            /* smoothed V estimates */
  BYTE              *u_valid;        /* 1 where the u estimate is valid */
  BYTE              *v_valid;        /* 1 where the v estimate is valid */
  BYTE              *a_valid;        /* 1 where the a estimate is valid */
  float             *DomCompR;       /* Real part of the dominant component */
  float             *DomCompI;       /* Imag part of the dominant component */
  double            *Theta;          /* phase initial conditions for recon */
  int               *maxfiltFake;    /* fake maxfilt data */
  float             *criterionFake;  /* fake filter selection criterion */
  float             *AFake;          /* A est from a single channel */
  float             *UFake;          /* U est from a single channel */
  float             *VFake;          /* V est from a single channel */
  BYTE              *a_validFake;    /* 1 where fake a estimate is valid */
  BYTE              *u_validFake;    /* 1 where fake u estimate is valid */
  BYTE              *v_validFake;    /* 1 where fake u estimate is valid */
  double            *ThetaFake;      /* Fake phase initial conds for recon */
  char               Afn[FNSIZE];    /* filename for amplitude estimates */
  char               Ufn[FNSIZE];    /* filename for horiz freq estimates */
  char               Vfn[FNSIZE];    /* filename for vert  freq estimates */
  int                PrintResponse;  /* 1 if we are to print response images */
  char               MaxFiltfn[FNSIZE];  /* filename for maxfilt data */
  char               DomCompRfn[FNSIZE]; /* filename for Dom Comp Real data */
  char               DomCompIfn[FNSIZE]; /* filename for Dom Comp Imag data */
  char               RespRootFn[FNSIZE]; /* root filename for response images */
  int                PrintChannels;  /* 1=print A,U,V est. from each channel */
  char               ChanARootFn[FNSIZE];/* root fn for Channel A estimates */
  char               ChanURootFn[FNSIZE];/* root fn for Channel U estimates */
  char               ChanVRootFn[FNSIZE];/* root fn for Channel V estimates */
  char               TempString[FNSIZE]; /* string for building up filenames */
  int                PostFilter;     /* 1 if postfiltering is to be done */
  char               ChanAsmRootFn[FNSIZE];/* Filename for smoothed amplitude */
  char               ChanUsmRootFn[FNSIZE];/* Filename for smoothed horiz freq*/
  char               ChanVsmRootFn[FNSIZE];/* Filename for smoothed vert  freq*/

 /*
  * Check for right number of arguments
  */
  if (argc != 2) {
    printf("\n\n%s: dominant component AM/FM analysis.",argv[0]);
    printf("\nusage: %s ParameterFileName\n",argv[0]);
    exit(-1);
  }
#ifdef DEBUG
 /*
  * Link to PA so we can use it in the debugger 
  */
  if (!argc) {
    PA((void *)X, 0, -1, 'b');
    PRINTA((void *)X,0,-1,'b',"delete.me");
    WRITEA((void *)X,0,-1,'b',"delete.me");
    image2disk((void *)X,1,1,'*',"delete.me");
    SpecMag2disk(X,"delete.me");
    PC(X,0,0,-1,'b',0);
    PRINTC(X,0,0,-1,'b',0,"delete.me");
    WRITEC(X,0,0,-1,'b',0,"delete.me");
  }
#endif
  X = (pFloatImage)malloc(sizeof(FloatImage));
  Y = (pFloatImage)malloc(sizeof(FloatImage));

  Initialize(argv[1],&size,&ZeroPad,&Reflect,&X->Real,&X->Imag,&X->FFTReal,
             &X->FFTImag,&G,Afn,Ufn,Vfn,MaxFiltfn,DomCompRfn,DomCompIfn,
             &PrintResponse,RespRootFn,&PrintChannels,ChanARootFn,ChanURootFn,
             ChanVRootFn,&ReconInterval,&PostFilter,ChanAsmRootFn,ChanUsmRootFn,
             ChanVsmRootFn);
  X->size = X->cols = X->rows = size;
  X->square = 1; 
  AllocFloatImage(Y,size);
  n = size * size;

 /*
  * Allocate working arrays
  */
  AllocArrays(argv,n,PrintChannels,PostFilter,&a_valid,&u_valid,&v_valid,
              &maxfilt,&criterion,&A,&U,&V,&Asm,&Usm,&Vsm,&a_validFake,
              &u_validFake,&v_validFake,&maxfiltFake,&criterionFake,&AFake,
              &UFake,&VFake,&Theta,&ThetaFake);
  if (PrintResponse) {
    printf("\n\nGenerating Files for Channel Responses...\n");
  }
  if (PrintChannels) {
    printf("\n\nGenerating files for A,U estimates from each channel...\n");
  }
  if (PostFilter) {
    printf("\n\nPost Filtering...\n");
  }

 /*
  * Process with the zeroth filter
  */
  if (ZeroPad) {
    looplim = size >> 1;
  } else {
    looplim = size;
  }
  printf("\nDoing Filter 0...");
  HardyConvolveGabor(Y,X,G->filt[0]);
  if (PrintResponse) {
    if (strcmp(RespRootFn,"none")) {
      sprintf(TempString,"%s%02dR.float",RespRootFn,0);
      f2disk(Y->Real,Y->size,Y->size,TempString);
      sprintf(TempString,"%s%02dI.float",RespRootFn,0);
      f2disk(Y->Imag,Y->size,Y->size,TempString);
    }
  }

 /*
  * Dominant ESA
  */
  dom_esa(Y,looplim,G->filt[0],criterion,A,U,V,G,maxfilt,a_valid,u_valid,
          v_valid,ReconInterval,Theta);

 /*
  * Re-run ESA with fake variables if PrintChannels OR PostFiltering
  */
  if (PrintChannels || PostFilter) {
    dom_esa(Y,looplim,G->filt[0],criterionFake,AFake,UFake,VFake,G,maxfiltFake,
            a_validFake,u_validFake,v_validFake,ReconInterval,ThetaFake);

   /*
    * Post Filter U, V and A
    */
    if (PostFilter) {
      if (G->filt[0]->isotropicPF) {
        g = MkPostFilt(G->filt[0]->sigma,G->filt[0]->kappa,size);
      } else {
        g = MkAnisPF(G->filt[0]->sigma,G->filt[0]->kappa_x,G->filt[0]->kappa_y,
                     G->filt[0]->Theta,size,1);
      }
      DoPostFilter(g,size,UFake,Usm,G,0);
      DoPostFilter(g,size,VFake,Vsm,G,0);
      ReComputeAsm(Y,Usm,Vsm,Asm,G->filt[0]);
      DoPostFilter(g,size,Asm,Asm,G,0);
      InsertAsm(Asm,A,maxfilt,size,0);
      InsertAsm(Usm,U,maxfilt,size,0);
      InsertAsm(Vsm,V,maxfilt,size,0);
      g = KillFloat(g);

     /*
      * Print smoothed A,U,V estimates for this channel
      */
      if (PrintChannels) {
        if (strcmp(ChanAsmRootFn,"none")) {
          sprintf(TempString,"%s%02d.float",ChanAsmRootFn,0);
          f2disk(Asm,Y->size,Y->size,TempString);
        }
        if (strcmp(ChanUsmRootFn,"none")) {
          sprintf(TempString,"%s%02d.float",ChanUsmRootFn,0);
          f2disk(Usm,Y->size,Y->size,TempString);
        }
        if (strcmp(ChanVsmRootFn,"none")) {
          sprintf(TempString,"%s%02d.float",ChanVsmRootFn,0);
          f2disk(Vsm,Y->size,Y->size,TempString);
        }
      }
    }

    if (PrintChannels) {
      if (strcmp(ChanARootFn,"none")) {
        sprintf(TempString,"%s%02d.float",ChanARootFn,0);
        f2disk(AFake,Y->size,Y->size,TempString);
      }
      if (strcmp(ChanURootFn,"none")) {
        sprintf(TempString,"%s%02d.float",ChanURootFn,0);
        f2disk(UFake,Y->size,Y->size,TempString);
      }
      if (strcmp(ChanVRootFn,"none")) {
        sprintf(TempString,"%s%02d.float",ChanVRootFn,0);
        f2disk(VFake,Y->size,Y->size,TempString);
      }
    }       /* if (PrintChannels) */

    for (j=0; j < n; j++) {
      criterionFake[j] = (float)-1.0;
      AFake[j] = (float)0.0;
      UFake[j] = (float)0.0;
      VFake[j] = (float)0.0;
      maxfiltFake[j] = -1;
      a_validFake[j] = 0;
      u_validFake[j] = 0;
      v_validFake[j] = 0;
    }
  }         /* if (PrintChannels || PostFilter) */

 /*
  * Loop for remaining filters
  */
  for (i=1; i < G->TotNumFilters; i++) {
    printf("\nDoing Filter %d...",i);
    HardyConvolveGabor(Y,X,G->filt[i]);
    if (PrintResponse) {
      if (strcmp(RespRootFn,"none")) {
        sprintf(TempString,"%s%02dR.float",RespRootFn,i);
        f2disk(Y->Real,Y->size,Y->size,TempString);
        sprintf(TempString,"%s%02dI.float",RespRootFn,i);
        f2disk(Y->Imag,Y->size,Y->size,TempString);
      }
    }

   /*
    * Dominant ESA
    */
    dom_esa(Y,looplim,G->filt[i],criterion,A,U,V,G,maxfilt,a_valid,u_valid,
            v_valid,ReconInterval,Theta);

   /*
    * Re-run ESA with fake variables if PrintChannels OR PostFiltering
    */
    if (PrintChannels || PostFilter) {
      dom_esa(Y,looplim,G->filt[i],criterionFake,AFake,UFake,VFake,G,
              maxfiltFake,a_validFake,u_validFake,v_validFake,ReconInterval,
              ThetaFake);

     /*
      * Post Filter U, V and A
      */
      if (PostFilter) {
        if (G->filt[i]->isotropicPF) {
          g = MkPostFilt(G->filt[i]->sigma,G->filt[i]->kappa,size);
        } else {
          g = MkAnisPF(G->filt[i]->sigma,G->filt[i]->kappa_x,
                       G->filt[i]->kappa_y,G->filt[i]->Theta,size,1);
        }
        DoPostFilter(g,size,UFake,Usm,G,i);
        DoPostFilter(g,size,VFake,Vsm,G,i);
        ReComputeAsm(Y,Usm,Vsm,Asm,G->filt[i]);
        DoPostFilter(g,size,Asm,Asm,G,i);
        InsertAsm(Asm,A,maxfilt,size,i);
        InsertAsm(Usm,U,maxfilt,size,i);
        InsertAsm(Vsm,V,maxfilt,size,i);
        g = KillFloat(g);

       /*
        * Print smoothed A,U,V estimates for this channel
        */
        if (PrintChannels) {
          if (strcmp(ChanAsmRootFn,"none")) {
            sprintf(TempString,"%s%02d.float",ChanAsmRootFn,i);
            f2disk(Asm,Y->size,Y->size,TempString);
          }
          if (strcmp(ChanUsmRootFn,"none")) {
            sprintf(TempString,"%s%02d.float",ChanUsmRootFn,i);
            f2disk(Usm,Y->size,Y->size,TempString);
          }
          if (strcmp(ChanVsmRootFn,"none")) {
            sprintf(TempString,"%s%02d.float",ChanVsmRootFn,i);
            f2disk(Vsm,Y->size,Y->size,TempString);
          }
        }
      }      /* if (PostFilter) */

      if (PrintChannels) {
        if (strcmp(ChanARootFn,"none")) {
          sprintf(TempString,"%s%02d.float",ChanARootFn,i);
          f2disk(AFake,Y->size,Y->size,TempString);
        }
        if (strcmp(ChanURootFn,"none")) {
          sprintf(TempString,"%s%02d.float",ChanURootFn,i);
          f2disk(UFake,Y->size,Y->size,TempString);
        }
        if (strcmp(ChanVRootFn,"none")) {
          sprintf(TempString,"%s%02d.float",ChanVRootFn,i);
          f2disk(VFake,Y->size,Y->size,TempString);
        }
      }       /* if (PrintChannels) */

      for (j=0; j < n; j++) {
        criterionFake[j] = (float)-1.0;
        AFake[j] = (float)0.0;
        UFake[j] = (float)0.0;
        VFake[j] = (float)0.0;
        maxfiltFake[j] = -1;
        a_validFake[j] = 0;
        u_validFake[j] = 0;
        v_validFake[j] = 0;
      }
    }            /* if (PrintChannels || PostFilter) */
  }              /* for i */

 /*
  * Resize results and valid arrays 
  */
  if (ZeroPad) {
    size >>= 1;
    itmp = (int   *)malloc(size*size*sizeof(int));
    for (row=0; row < size; row++) {
      memcpy(itmp+row*size,maxfilt+row*(size>>1),size*sizeof(int));
    }
    free(maxfilt);
    maxfilt = itmp;

    ftmp = (float *)malloc(size*size*sizeof(float));
    btmp = (BYTE  *)malloc(size*size*sizeof(BYTE));
    for (row=0; row < size; row++) {
      memcpy(ftmp+row*size,A+row*(size<<1),size*sizeof(float));
      memcpy(btmp+row*size,a_valid+row*(size<<1),size*sizeof(BYTE));
    }
    free(A);
    free(a_valid);
    A = ftmp;
    a_valid = btmp;

    ftmp = (float *)malloc(size*size*sizeof(float));
    btmp = (BYTE  *)malloc(size*size*sizeof(BYTE));
    for (row=0; row < size; row++) {
      memcpy(ftmp+row*size,U+row*(size<<1),size*sizeof(float));
      memcpy(btmp+row*size,u_valid+row*(size<<1),size*sizeof(BYTE));
    } 
    free(U);
    free(u_valid);
    U = ftmp;
    u_valid = btmp;

    ftmp = (float *)malloc(size*size*sizeof(float));
    btmp = (BYTE  *)malloc(size*size*sizeof(BYTE));
    for (row=0; row < size; row++) {
      memcpy(ftmp+row*size,V+row*(size<<1),size*sizeof(float));
      memcpy(btmp+row*size,v_valid+row*(size<<1),size*sizeof(BYTE));
    } 
    free(V);
    free(v_valid);
    V = ftmp;
    v_valid = btmp;

    dtmp = (double *)malloc(size*size*sizeof(double));
    for (row=0; row < size; row++) {
      memcpy(dtmp+row*size,Theta+row*(size<<1),size*sizeof(double));
    } 
    free(Theta);
    Theta = dtmp;
  }

 /*
  * Interpolate for yet indeterminate values
  */
  for (i=0,row=0; row<size; row++) {
    for (col=0; col<size; col++,i++) {
      /* here here here */
      if (row==30 && col == 90) {
        col=90;
      }
      if (!u_valid[i]) f_interp(U,row,col,size,size,u_valid);
      if (!v_valid[i]) f_interp(V,row,col,size,size,v_valid);
      if (!a_valid[i]) f_interp(A,row,col,size,size,a_valid);
    }    /* for col */
  }      /* for row */

 /*
  * Reconstruct Dominant Component
  */
  DomRecon(A,U,V,size,&DomCompR,&DomCompI,ReconInterval,Theta);

 /*
  * Write result files
  */
  if (strcmp(Afn,"none")) {
    f2disk(A,looplim,looplim,Afn);
  }
  if (strcmp(Ufn,"none")) {
    f2disk(U,looplim,looplim,Ufn);
  }
  if (strcmp(Vfn,"none")) {
    f2disk(V,looplim,looplim,Vfn);
  }
  if (strcmp(DomCompRfn,"none")) {
    f2disk(DomCompR,looplim,looplim,DomCompRfn);
  }
  if (strcmp(DomCompIfn,"none")) {
    f2disk(DomCompI,looplim,looplim,DomCompIfn);
  }
  if (strcmp(MaxFiltfn,"none")) {
    byte2disk(maxfilt,looplim*sizeof(int),looplim,MaxFiltfn);
  }
 
  printf("\n\n");
  return;

} /*-------------------------------------------- main --------------*/

/*
 * Initialize
 *
 *  Process the paramters file
 *  Read the input image, if real make hardy to get the imaginary part
 *  Compute the input spectrum
 *  Construct the filter bank
 */

void Initialize(ParmFn,size,ZP,Reflect,xReal,xImag,XReal,XImag,G,Afn,Ufn,Vfn,
                MaxFiltfn,DomCompRfn,DomCompIfn,PrintResponse,RespRootFn,
                PrintChannels,ChanARootFn,ChanURootFn,ChanVRootFn,ReconInterval,
                PostFilter,ChanAsmRootFn,ChanUsmRootFn,ChanVsmRootFn)

  char        *ParmFn;           /* filename of parameters file */
  int         *size;             /* row/col dim of the square input image */
  int         *ZP;               /* 1 if zero padding is to be used */
  int         *Reflect;          /* 1=fill 0 padded areas w/ reflection */
  float      **xReal;            /* real part of input image */
  float      **xImag;            /* imag part of input image */
  float      **XReal;            /* real part of input spectrum */
  float      **XImag;            /* imag part of input spectrum */
  filterbank **G;                /* the filter bank */
  char        *Afn;              /* filename for Amplitude Estimates */
  char        *Ufn;              /* filename for horiz frequency estimates */
  char        *Vfn;              /* filename for vert  frequency estimates */
  char        *MaxFiltfn;        /* filename for maxfilt data */
  char        *DomCompRfn;       /* filename for Real dom comp data */
  char        *DomCompIfn;       /* filename for Imag dom comp data */
  int         *PrintResponse;    /* 1 if we are to print the response images */
  char        *RespRootFn;       /* Root filename for response images */
  int         *PrintChannels;    /* 1=print A,U est from each channel */
  char        *ChanARootFn;      /* root fn for A est from each channel */
  char        *ChanURootFn;      /* root fn for U est from each channel */
  char        *ChanVRootFn;      /* root fn for V est from each channel */
  int         *ReconInterval;    /* boundary condition interval for recon */
  int         *PostFilter;       /* 1=PostFiltering to be performed */
  char        *ChanAsmRootFn;    /* Filename for smoothed amplitude */
  char        *ChanUsmRootFn;    /* Filename for smoothed horiz frequency */
  char        *ChanVsmRootFn;    /* Filename for smoothed vert  frequency */
{

  int         i;                 /* counter */
  int         row;               /* counter for image rows */
  int         col;               /* counter for image cols */
  int         so2;               /* *size / 2 */
  char        datatype;          /* type of in dat: b=byte f=float d=double */
  int         Conjugate;         /* 1=conjugate input.  See Book X, p. 64 */
  char        Rfn[FNSIZE];       /* file for Real part of input image */
  char        Ifn[FNSIZE];       /* file for Imag part of input image */
  double      r0;                /* rad. center freq. of 1st filt on each ray */
  float       R;                 /* common ratio of radii of adjcnt filtrs */
  float       obw;               /* Eta-peak octave bandwidth of filters */
  float       Eta;               /* OBW is defined in terms of Eta-peak. */
  int         isoPF;             /* 1=isotropic post filters */
  float       kappa;             /* Post Filter space constant scale factor */
  float       kappa_x;           /* postfilter horiz space cnst scale factor */
  float       kappa_y;           /* postfilter vert  space cnst scale factor */
  float      *tfloat;            /* temporary float array */
  float      *to;                /* area in memory to copy floats to */
  float      *from;              /* area in memory to copy floats from */
  float      *pfloat;            /* ptr to location to fill w/ zero */

 /*
  * Read the parameter file
  */
  ReadParametersFile(ParmFn,&datatype,&Conjugate,Rfn,Ifn,size,ZP,Reflect,&r0,
                     &R,&Eta,&obw,Afn,Ufn,Vfn,MaxFiltfn,DomCompRfn,DomCompIfn,
                     PrintResponse,RespRootFn,PrintChannels,ChanARootFn,
                     ChanURootFn,ChanVRootFn,ReconInterval,PostFilter,&isoPF,
                     &kappa,&kappa_x,&kappa_y,ChanAsmRootFn,ChanUsmRootFn,
                     ChanVsmRootFn);
  r0 = r0 / (double)(*size);
 
 /*
  * If zero padding, double size 
  */
  if (*ZP) {
    *size *= 2;
  }
  so2 = *size >> 1;

 /*
  * Make the filter bank
  */ 
  *G = CreateFilterBank(r0,R,obw,Eta,isoPF,kappa,kappa_x,kappa_y);

 /*
  * Allocate Spectra
  */
  if ((*XReal = (float *)malloc(*size**size*sizeof(float))) == NULL) {
    printf("\nInitialize: free store exhausted\n");
    exit(-1);
  }
  if ((*XImag = (float *)malloc(*size**size*sizeof(float))) == NULL) {
    printf("\nInitialize: free store exhausted\n");
    exit(-1);
  }

 /*
  * Get the input image and its spectrum
  */
  if (*ZP) {
    tfloat = readasf(so2,so2,Rfn,datatype);
    if ((*xReal = (float *)malloc(*size**size*sizeof(float))) == NULL) {
      printf("\nInitialize: free store exhausted\n");
      exit(-1);
    }
    for (to=*xReal,from=tfloat,row=0; row < so2; row++,to+=*size,from+=so2) {
      memcpy(to, from, so2*sizeof(float));
      for (pfloat=to+so2,col=so2; col < *size; col++) {
        if (!*Reflect) {
          *(pfloat++) = (float)0.0;
        } else {
          *(pfloat++) = *(to + *size-col-1);
        }
      }
    }
    free(tfloat);
    if (!*Reflect) {
      for (pfloat=to,i=so2**size; i < *size**size; i++) {
        *(pfloat++) = (float)0.0;
      }
    } else {
      for (pfloat=to,row=1; row < *size; row+=2) {
        for (col=0; col < *size; col++,pfloat++) {
          *pfloat = *(pfloat - row**size);
        }
      }
    }
  } else {
    *xReal = readasf(*size,*size,Rfn,datatype);
  }

  if (strcmp(Ifn,"none")) {                     /* filename is not "none" */
    if (*ZP) {
      tfloat = readasf(so2,so2,Ifn,datatype);
      if ((*xImag = (float *)malloc(*size**size*sizeof(float))) == NULL) {
        printf("\nInitialize: free store exhausted\n");
        exit(-1);
      }
      for (to=*xImag,from=tfloat,row=0; row < so2; row++,to+=*size,from+=so2) {
        memcpy(to, from, so2*sizeof(float));
        for (pfloat=to+so2,col=so2; col < *size; col++) {
          if (!*Reflect) {
            *(pfloat++) = (float)0.0;
          } else {
            *(pfloat++) = *(to + *size-col-1);
          }
        }
      }
      free(tfloat);
      if (!*Reflect) {
        for (pfloat=to,i=so2**size; i < *size**size; i++) {
          *(pfloat++) = (float)0.0;
        }
      } else {
        for (pfloat=to,row=1; row < *size; row+=2) {
          for (col=0; col < *size; col++,pfloat++) {
            *pfloat = *(pfloat - row**size);
          }
        }
      }
    } else {
      *xImag = readasf(*size,*size,Ifn,datatype);
    }
    if (Conjugate) {
      for (i=0; i < *size**size; i++) {
        (*xImag)[i] *= (float)-1.0;
      }
    }
/*    f_u_norm(*xReal, *xImag, *size, *size);*/
    fft2d(*xReal,*xImag,*XReal,*XImag,*size,1);
  } else {                         /* input is real, must create imag part */
/*    f_u_norm(*xReal, NULL, *size, *size);*/
    if ((*xImag = (float *)malloc(*size**size*sizeof(float))) == NULL) {
      printf("\nInitialize: free store exhausted\n");
      exit(-1);
    }
    MakeHardy(*xReal,*xImag,*XReal,*XImag,*size);
    if (Conjugate) {
      printf("\n\n >>>  CONJUGATING input made by MakeHardy...\n");
      for (i=0; i < *size**size; i++) {
        (*xImag)[i] *= (float)-1.0;
      }
      fft2d(*xReal,*xImag,*XReal,*XImag,*size,1);
    }
  }

  return;
} /*---------------------------------------------- Initialize ----------*/

/*
 * ReadParametersFile
 *
 *  read the data out of the parameters file
 *
 */

void ReadParametersFile(ParmFn,pdatatype,pConjugate,Rfn,Ifn,psize,pZP,pReflect,
                        pr0,pR,pEta,pobw,Afn,Ufn,Vfn,MaxFiltfn,DomCompRfn,
                        DomCompIfn,pPrintResponse,RespRootFn,pPrintChannels,
                        ChanARootFn,ChanURootFn,ChanVRootFn,pReconInterval,
                        pPostFilter,pisoPF,pkappa,pkappa_x,pkappa_y,
                        ChanAsmRootFn,ChanUsmRootFn,ChanVsmRootFn)
  char   *ParmFn;
  char   *pdatatype;
  int    *pConjugate;
  char   *Rfn;
  char   *Ifn;
  int    *psize;
  int    *pZP;
  int    *pReflect;
  double *pr0;
  float  *pR;
  float  *pEta;
  float  *pobw;
  char   *Afn;
  char   *Ufn;
  char   *Vfn;
  char   *MaxFiltfn;
  char   *DomCompRfn;
  char   *DomCompIfn;
  int    *pPrintResponse;
  char   *RespRootFn;
  int    *pPrintChannels;
  char   *ChanARootFn;
  char   *ChanURootFn;
  char   *ChanVRootFn;
  int    *pReconInterval;
  int    *pPostFilter;
  int    *pisoPF;
  float  *pkappa;
  float  *pkappa_x;
  float  *pkappa_y;
  char   *ChanAsmRootFn;
  char   *ChanUsmRootFn;
  char   *ChanVsmRootFn;
{

  FILE      *parmfp;            /* ptr to parameter file */
  char       instr[256];        /* buffer to read into  */

  if ((parmfp = fopen(ParmFn,"r")) == NULL) {
    printf("\n\nReadParametersFile: unable to open %s\n",ParmFn);
    exit(-1);
  }

  fscanf(parmfp,"Input Data Type= %s\n",instr);
  *pdatatype = instr[0];
  fscanf(parmfp,"Conjugate Input= %s\n",instr);
  if ((instr[0] == 'N') || (instr[0] == 'n')) {
    *pConjugate = 0;
  }
  else {
    *pConjugate = 1;
  }
  fscanf(parmfp,"Real Image File= %s\n",Rfn);
  fscanf(parmfp,"Imag Image File= %s\n",Ifn);
  fscanf(parmfp,"Size= %d\n",psize);
  fscanf(parmfp,"Zero Pad= %s\n",instr);
  if ((instr[0] == 'Y') || (instr[0] == 'y')) {
    *pZP = 1;
  }
  else {
    *pZP = 0;
  }
  fscanf(parmfp,"Reflect= %s\n",instr);
  if ((instr[0] == 'Y') || (instr[0] == 'y')) {
    *pReflect = 1;
  }
  else {
    *pReflect = 0;
  }
  fscanf(parmfp,"r0= %lf\n",pr0);
  fscanf(parmfp,"R= %f\n",pR);
  fscanf(parmfp,"Eta= %f\n",pEta);
  fscanf(parmfp,"OBW= %f\n",pobw);
  fscanf(parmfp,"A File= %s\n",Afn);
  fscanf(parmfp,"U File= %s\n",Ufn);
  fscanf(parmfp,"V File= %s\n",Vfn);
  fscanf(parmfp,"MaxFilt File= %s\n",MaxFiltfn);
  fscanf(parmfp,"Dominant Component Real File= %s\n",DomCompRfn);
  fscanf(parmfp,"Dominant Component Imag File= %s\n",DomCompIfn);
  fscanf(parmfp,"Print Response Images= %s\n",instr);
  if ((instr[0] == 'N') || (instr[0] == 'n')) {
    *pPrintResponse = 0;
  }
  else {
    *pPrintResponse = 1;
  }
  fscanf(parmfp,"Response Root File Name= %s\n",RespRootFn);
  fscanf(parmfp,"Print Channel Estimates= %s\n",instr);
  if ((instr[0] == 'N') || (instr[0] == 'n')) {
    *pPrintChannels = 0;
  }
  else {
    *pPrintChannels = 1;
  }
  fscanf(parmfp,"Chan Est A Root File Name= %s\n",ChanARootFn);
  fscanf(parmfp,"Chan Est U Root File Name= %s\n",ChanURootFn);
  fscanf(parmfp,"Chan Est V Root File Name= %s\n",ChanVRootFn);
  fscanf(parmfp,"Reconstruction Interval= %d\n",pReconInterval);
  fscanf(parmfp,"Postfilter= %s\n",instr);
  if ((instr[0] == 'N') || (instr[0] == 'n')) {
    *pPostFilter = 0;
  }
  else {
    *pPostFilter = 1;
  }
  fscanf(parmfp,"Isotropic PF= %s\n",instr);
  if ((instr[0] == 'N') || (instr[0] == 'n')) {
    *pisoPF = 0;
  }
  else {
    *pisoPF = 1;
  }
  fscanf(parmfp,"kappa= %f\n",pkappa);
  fscanf(parmfp,"kappa_x= %f\n",pkappa_x);
  fscanf(parmfp,"kappa_y= %f\n",pkappa_y);
  fscanf(parmfp,"Chan Asm Root Fn= %s\n",ChanAsmRootFn);
  fscanf(parmfp,"Chan Usm Root Fn= %s\n",ChanUsmRootFn);
  fscanf(parmfp,"Chan Vsm Root Fn= %s\n",ChanVsmRootFn);
  fclose(parmfp);
  return;
} /*--------------------------- ReadParametersFile -------------------------*/

/*
 * dom_esa
 *
 *  function evaluates the esa at every point in an image and checks to see if
 *  if the current response image has the dominant response of those yet seen.
 *
 */

void dom_esa(Y,looplim,G,criterion,a,u,v,Bank,maxfilt,a_valid,u_valid,v_valid,
             ReconInterval,Theta)

  pFloatImage     Y;              /* response image */
  int             looplim;        /* terminal value for a loop counter */
  FilterPtr       G;              /* Y was output from this filter */
  float          *criterion;      /* filter selection criterion */
  float          *a;              /* a estimates */
  float          *u;              /* u estimates */
  float          *v;              /* v estimates */
  filterbank     *Bank;           /* the filterbank */
  int            *maxfilt;        /* which chan max'd criterion @ (i,j) */
  BYTE           *a_valid;        /* 1 where the a estimate is valid */
  BYTE           *u_valid;        /* 1 where the u estimate is valid */
  BYTE           *v_valid;        /* 1 where the v estimate is valid */
  int             ReconInterval;  /* boundary condition interval for recon */
  double         *Theta;          /* phase initial conditions for recon */
{

  int             row,col;        /* loop counters for image coordinates */
  int             k;              /* loop counter for indexing arrays */
 
 /*
  * Apply the ESA at each point in the image
  */
  for (row=0; row < looplim; row++) {
    for (col=0; col < looplim; col++) {
      /* here here here */
      if (row==30 && col == 90) {
        col=90;
      }
      k = row * Y->size + col;
      if (PtEsa2D(Y,row,col,G,&criterion[k],&u[k],&v[k],&a[k],
                  &u_valid[k],&v_valid[k],&a_valid[k])) {
        maxfilt[k] = G->index;
        if ((!(row % ReconInterval)) && (!(col % ReconInterval))) {
          if ((Y->Real[k] == (float)0.0) && (Y->Imag[k] == (float)0.0)) {
            Theta[k] = (double)0.0;
          } else {
            Theta[k] = atan2((double)Y->Imag[k],(double)Y->Real[k]);
          }
        }
      }
    }
  }
  return;
} /*----------------------------------------- dom_esa -----------------------*/

/*
 * AllocArrays
 *
 *  Allocate Working Arrays.
 */

void AllocArrays(argv,n,PrintChannels,PostFilter,a_valid,u_valid,v_valid,
                 maxfilt,criterion,A,U,V,Asm,Usm,Vsm,a_validFake,u_validFake,
                 v_validFake,maxfiltFake,criterionFake,AFake,UFake,VFake,Theta,
                 ThetaFake)

  char              *argv[];         /* argument list from command line */
  int                n;              /* size * size again */
  int                PrintChannels;  /* 1=print A,U est. from each channel */
  int                PostFilter;     /* 1=perform postfiltering */
  BYTE             **a_valid;        /* 1 where the a estimate is valid */
  BYTE             **u_valid;        /* 1 where the u estimate is valid */
  BYTE             **v_valid;        /* 1 where the v estimate is valid */
  int              **maxfilt;        /* Which channel max'd criterion at i,j */
  float            **criterion;      /* filter selection criterion */
  float            **A;              /* A estimates accross all filters */
  float            **U;              /* U estimates accross all filters */
  float            **V;              /* V estimates accross all filters */
  float            **Asm;            /* smoothed A estimates */
  float            **Usm;            /* smoothed U estimates */
  float            **Vsm;            /* smoothed V estimates */
  BYTE             **a_validFake;    /* 1 where fake a estimate is valid */
  BYTE             **u_validFake;    /* 1 where fake u estimate is valid */
  BYTE             **v_validFake;    /* 1 where fake v estimate is valid */
  int              **maxfiltFake;    /* fake maxfilt data */
  float            **criterionFake;  /* fake filter selection criterion */
  float            **AFake;          /* A est from a single channel */
  float            **UFake;          /* U est from a single channel */
  float            **VFake;          /* V est from a single channel */
  double           **Theta;          /* boundary condition interval for recon */
  double           **ThetaFake;      /* Fake boundary cond interval for recon */
{
 
  int                i;              /* loop counter */
 
 /*
  * Allocate working arrays
  */
  if ((*a_valid = (BYTE *) malloc(n * sizeof(BYTE))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*u_valid = (BYTE *) malloc(n * sizeof(BYTE))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*v_valid = (BYTE *) malloc(n * sizeof(BYTE))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*maxfilt = (int *)malloc(n * sizeof(int))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  for (i=0; i < n; i++) (*maxfilt)[i] = -1;
  if ((*criterion = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  for (i=0; i < n; i++) (*criterion)[i] = (float)-1.0;
  if ((*A = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*U = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*V = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if (PostFilter) {
    if ((*Asm = (float *)malloc(n * sizeof(float))) == NULL) {
      printf("\n%s: free store exhausted.\n",argv[0]);
      exit(-1);
    }
    if ((*Usm = (float *)malloc(n * sizeof(float))) == NULL) {
      printf("\n%s: free store exhausted.\n",argv[0]);
      exit(-1);
    }
    if ((*Vsm = (float *)malloc(n * sizeof(float))) == NULL) {
      printf("\n%s: free store exhausted.\n",argv[0]);
      exit(-1);
    }
  }
  if ((*Theta = (double *)malloc(n * sizeof(double))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*ThetaFake = (double *)malloc(n * sizeof(double))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
 
  if (PrintChannels || PostFilter) {
    if ((*a_validFake = (BYTE *) malloc(n * sizeof(BYTE))) == NULL) {
      printf("\n%s: free store exhausted.\n",argv[0]);
      exit(-1);
    }
    if ((*u_validFake = (BYTE *) malloc(n * sizeof(BYTE))) == NULL) {
      printf("\n%s: free store exhausted.\n",argv[0]);
      exit(-1);
    }
    if ((*v_validFake = (BYTE *) malloc(n * sizeof(BYTE))) == NULL) {
      printf("\n%s: free store exhausted.\n",argv[0]);
      exit(-1);
    }
    if ((*maxfiltFake = (int *)malloc(n * sizeof(int))) == NULL) {
      printf("\n%s: free store exhausted.\n",argv[0]);
      exit(-1);
    }
    for (i=0; i < n; i++) (*maxfiltFake)[i] = -1;
    if ((*criterionFake = (float *)malloc(n * sizeof(float))) == NULL) {
      printf("\n%s: free store exhausted.\n",argv[0]);
      exit(-1);
    }
    for (i=0; i < n; i++) (*criterionFake)[i] = (float)-1.0;
    if ((*AFake = (float *)malloc(n * sizeof(float))) == NULL) {
      printf("\n%s: free store exhausted.\n",argv[0]);
      exit(-1);
    }
    if ((*UFake = (float *)malloc(n * sizeof(float))) == NULL) {
      printf("\n%s: free store exhausted.\n",argv[0]);
      exit(-1);
    }
    if ((*VFake = (float *)malloc(n * sizeof(float))) == NULL) {
      printf("\n%s: free store exhausted.\n",argv[0]);
      exit(-1);
    }
  }
 
  return;
} /*---------------------  AllocArrays  --------------------------------------*/
 
/*
 * DomRecon
 *
 *  Reconstruct the dominant component.
 *
 */
 
void DomRecon(a,u,v,size,pDomCompR,pDomCompI,ReconInterval,Theta)
 
  int             size;           /* length of signal for reconstruction */
  float          *a;              /* a estimates */
  float          *u;              /* u estimates */
  float          *v;              /* v estimates */
  float         **pDomCompR;      /* Real part of the dominant component */
  float         **pDomCompI;      /* Imag part of the dominant component */
  int             ReconInterval;  /* boundary condition interval for recon */
  double         *Theta;          /* phase initial conditions for recon */
{
 
  int             row;            /* loop counter for image rows */
  int             col;            /* loop counter for image cols */
  int             ROW;            /* row counter for recon blocks */
  int             COL;            /* col counter for recon blocks */
  int             R,C;            /* counters for rows and cols within block */
  int             k;              /* counter */
  double         *phase;          /* instantaneous phase */

 /*
  * Allocate arrays
  */
  if ((*pDomCompR = (float *)malloc(size*size*sizeof(float))) == NULL) {
    printf("\nDomRecon:  free store exhausted.\n");
    exit(-1);
  }
  if ((*pDomCompI = (float *)malloc(size*size*sizeof(float))) == NULL) {
    printf("\nDomRecon:  free store exhausted.\n");
    exit(-1);
  }
  if ((phase = (double *)malloc(size*size*sizeof(double))) == NULL) {
    printf("\nDomRecon:  free store exhausted.\n");
    exit(-1);
  }
 
 /*
  * Outer loops set top left corner of each reconsructed block
  */
  for (ROW=0; ROW < size; ROW+=ReconInterval) {
    for (COL=0; COL < size; COL+=ReconInterval) {
     /*
      * Reconstruct phase on first row of block
      */
      k = ROW*size + COL;
      phase[k] = Theta[k];
      for (k++,C=1,col=COL+1; (col < size) && (C < ReconInterval); 
           col++,C++,k++) {
        phase[k] = phase[k-1] + (double)2.0 * (double)M_PI * (double)u[k-1];
      }
     /*
      * Reconstruct phase on first col of block
      */
      for (k=(ROW+1)*size+COL,R=1,row=ROW+1; 
           (row < size) && (R < ReconInterval);
           row++,R++,k+=size) {
        phase[k] = phase[k-size] + (double)2.0 * (double)M_PI 
                                               * (double)v[k-size];
      }
     /*
      * Reconstruct phase at interior points in the block
      */
      for (R=1,row=ROW+1; (row < size) && (R < ReconInterval); R++,row++) {
        for (C=1,col=COL+1; (col < size) && (C < ReconInterval); C++,col++) {
          /* here here here */
          if (row==30 && col == 90) {
            col=90;
          }
          k = row*size + col;
          phase[k] = (double)0.5 * (phase[k-1] + phase[k-size] 
                     + (double)2.0 * (double)M_PI 
                       * ((double)u[k-1] + (double)v[k-size]));
        }
      }
    }    /* COL */
  }      /* ROW */

 /*
  * Reconstruct image values
  */
  for (k=row=0; row < size; row++) {
    for (col=0; col < size; col++,k++) {
      (*pDomCompR)[k] = a[k] * (float)cos(phase[k]);
      (*pDomCompI)[k] = a[k] * (float)sin(phase[k]);
    }
  }
  free(phase);
  return;
} /*----------------------------------------- DomRecon ----------------------*/

/*
 * InsertAsm
 *
 *   Insert smoothed amplitude estimates from channel into the dominant
 *   component amplitude estimate array.
 *
 *
 */
 
void InsertAsm(Asm,A,maxfilt,size,idx)
 
  float          *Asm;            /* smoothed amplitude estimtes */
  float          *A;              /* dominant component amplitude estimates */
  int            *maxfilt;        /* Which channel max'd criterion at i,j */
  int             size;           /* length of signal */
  int             idx;            /* index of channel being processed */
{
 
  int             i;              /* loop counters for image coordinates */
  int             n;              /* size * size */
 
  n = size * size;
  for (i=0; i < n; i++) {
    if (maxfilt[i] == idx) {
      A[i] = Asm[i];
    }
  }
 
  return;
} /*----------------------------------------- InsertAsm ---------------------*/
#endif
