/*
 * ExtractLF.c
 *
 *   Remove low frequency component from a real-valued image.  
 *
 *   An cutoff frequency is given in the parameters file.  The LF image 
 *   contains all Fourier components of the input image with frequency
 *   modulus below the cutoff.
 *
 *   Similarly, the HF image is constructed by taking those Fourier components
 *   of the input image that have magnitude frequency modulus above the cutoff
 *   frequency.
 *
 *  jph 3/17/96
 *
 */

#ifndef EXTRACTLF_C
#define EXTRACTLF_C

#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include <string.h>

#include "teager2D.h"

/*
 * Function prototypes
 */
void ReadParametersFile();
void Initialize();

main(argc,argv)

  int   argc;
  char *argv[];

{

  pFloatImage        X;              /* input image */
  pFloatImage        LF;             /* lo pass image */
  pFloatImage        HF;             /* hi pass image */
  int                size;           /* row/col dim of the square input image */
  int                so2;            /* size / 2 */
  double             one_on_size;    /* 1 / size */
  int                n;              /* size squared */
  double             r0;             /* cutoff frequency */
  char               LFfn[FNSIZE];   /* filename for low pass image */
  char               HFfn[FNSIZE];   /* filename for hi  pass image */
  int                row;            /* image row counter */
  int                col;            /* image col counter */
  int                k;              /* loop counter */
  int                i;              /* counter */
  double             u;              /* horiz frequency */
  double             v;              /* vert  frequency */
  double             f;              /* modulus frequency */

 /*
  * Check for right number of arguments
  */
  if (argc != 2) {
    printf("\n\n%s: Extract Low Freq Component From Image.",argv[0]);
    printf("\nusage: %s ParameterFileName\n",argv[0]);
    exit(-1);
  }
#ifdef DEBUG
 /*
  * Link to PA so we can use it in the debugger 
  */
  if (!argc) {
    PA((void *)X, 0, -1, 'b');
    PRINTA((void *)X,0,-1,'b',"delete.me");
    WRITEA((void *)X,0,-1,'b',"delete.me");
    image2disk((void *)X,1,1,'*',"delete.me");
    SpecMag2disk(X,"delete.me");
    PC(X,0,0,-1,'b',0);
    PRINTC(X,0,0,-1,'b',0,"delete.me");
    WRITEC(X,0,0,-1,'b',0,"delete.me");
  }
#endif
  X = (pFloatImage)malloc(sizeof(FloatImage));
  Initialize(argv[1],&size,&r0,&X->Real,&X->Imag,&X->FFTReal,&X->FFTImag,
             LFfn,HFfn);
  so2 = size >> 1;
  one_on_size = (double)1.0 / (double)size;
  n = size * size;
  X->size = X->cols = X->rows = size;
  X->square = 1; 

 /*
  * Allocate low pass and hi pass images
  */
  LF = (pFloatImage)malloc(sizeof(FloatImage));
  HF = (pFloatImage)malloc(sizeof(FloatImage));
  AllocFloatImage(LF,size);
  AllocFloatImage(HF,size);
  LF->square = HF->square = 1;

 /*
  * Initialize lo pass and hi pass spectra to zero
  */
  for (i=0; i < n; i++) {
    LF->FFTReal[i] = LF->FFTImag[i] = HF->FFTReal[i] 
                   = HF->FFTImag[i] = (float)0.0;
  }

 /*
  * Loop over frequencies.  Copy those below r0 to LF, those above to HF.
  */
  for (k=row=0; row < size; row++) {
    v = (double)(row - so2) * one_on_size;
    for (col=0; col < size; col++,k++) {
      u = (double)(col - so2) * one_on_size;
      f = dNorm(u,v);
      if (f < r0) {
        LF->FFTReal[k] = X->FFTReal[k];
        LF->FFTImag[k] = X->FFTImag[k];
      } else {
        HF->FFTReal[k] = X->FFTReal[k];
        HF->FFTImag[k] = X->FFTImag[k];
      }
    }
  }

 /*
  * Compute LF and HF images by inverse FFT
  */
  fft2d(LF->Real,LF->Imag,LF->FFTReal,LF->FFTImag,size,-1);
  fft2d(HF->Real,HF->Imag,HF->FFTReal,HF->FFTImag,size,-1);
      
 /*
  * Output the LF and HF Images
  */
  f2disk(LF->Real,size,size,LFfn);
  f2disk(HF->Real,size,size,HFfn);

  printf("\n\n");
  return;

} /*-------------------------------------------- main --------------*/

/*
 * Initialize
 *
 *  Process the paramters file
 *  Read the input image, set imaginary part to zero
 *  Compute the input spectrum
 */

void Initialize(ParmFn,size,r0,xReal,xImag,XReal,XImag,LFfn,HFfn)

  char        *ParmFn;           /* filename of parameters file */
  int         *size;             /* row/col dim of the square input image */
  double      *r0;               /* cutoff frequency */
  float      **xReal;            /* real part of input image */
  float      **xImag;            /* imag part of input image */
  float      **XReal;            /* real part of input spectrum */
  float      **XImag;            /* imag part of input spectrum */
  char        *LFfn;             /* filename for low pass image */
  char        *HFfn;             /* filename for hi  pass image */
{

  char        datatype;          /* type of in dat: b=byte f=float d=double */
  char        Rfn[FNSIZE];       /* file for Real part of input image */
  int         i;                 /* counter */


 /*
  * Read the parameter file
  */
  ReadParametersFile(ParmFn,&datatype,Rfn,size,r0,LFfn,HFfn);
                     
  *r0 = *r0 / (double)(*size);
 
 /*
  * Allocate Spectra
  */
  if ((*XReal = (float *)malloc(*size**size*sizeof(float))) == NULL) {
    printf("\nInitialize: free store exhausted\n");
    exit(-1);
  }
  if ((*XImag = (float *)malloc(*size**size*sizeof(float))) == NULL) {
    printf("\nInitialize: free store exhausted\n");
    exit(-1);
  }

 /*
  * Get the input image and its spectrum
  */
  *xReal = readasf(*size,*size,Rfn,datatype);

 /*
  * Set imaginary part of image to zero
  */
  if ((*xImag = (float *)malloc(*size**size*sizeof(float))) == NULL) {
    printf("\nInitialize: free store exhausted\n");
    exit(-1);
  }
  for (i=0; i < *size**size; i++) {
    (*xImag)[i] = (float)0.0;
  }
  fft2d(*xReal,*xImag,*XReal,*XImag,*size,1);

  return;

} /*---------------------------------------------- Initialize ----------*/

/*
 * ReadParametersFile
 *
 *  read the data out of the parameters file
 *
 */

void ReadParametersFile(ParmFn,pdatatype,Rfn,psize,pr0,LFfn,HFfn)

  char   *ParmFn;
  char   *pdatatype;
  char   *Rfn;
  int    *psize;
  double *pr0;
  char   *LFfn;
  char   *HFfn;
{

  FILE      *parmfp;            /* ptr to parameter file */
  char       instr[256];        /* buffer to read into  */

  if ((parmfp = fopen(ParmFn,"r")) == NULL) {
    printf("\n\nReadParametersFile: unable to open %s\n",ParmFn);
    exit(-1);
  }

  fscanf(parmfp,"Input Data Type= %s\n",instr);
  *pdatatype = instr[0];
  fscanf(parmfp,"Image File= %s\n",Rfn);
  fscanf(parmfp,"Size= %d\n",psize);
  fscanf(parmfp,"Zero Pad= %s\n",instr);
  fscanf(parmfp,"r0= %lf\n",pr0);
  fscanf(parmfp,"LF File Name= %s\n",LFfn);
  fscanf(parmfp,"HF File Name= %s\n",HFfn);
  fclose(parmfp);
  return;
} /*--------------------------- ReadParametersFile -------------------------*/
#endif
