/* 
 * GainParms2D.c
 *
 *   Compute the parameters needed for input to program FindGains.c, which 
 *   computes the Kalman gain sequences of the component track filter.
 *
 *   The parameters are the matrix Pi0, the measurement noise variances for
 *   the amplitude and frequency modulations, and the variances of the
 *   second derivatives of the modulating functions.
 *
 *   This program reads a parameter file which contains a filterbank 
 *   specification, the number of input files that will be used in 
 *   the computations, and a description of each input file.
 *
 *   Parameter files for NoFilt.c are constructed, and this program is
 *   used to compute the estimated modulating functions using the unfiltered
 *   algorithm.  Communication between this program and NoFilt is via
 *   temporary files.
 *
 *   Parameter files are also constructed for DomComp.c, which is used
 *   to compute estimated modulating functions using the filtered algorithm.
 *   Communication between this program and DomComp is also via temporary
 *   files.
 *
 *  usage: GainParms ParameterFileName
 *
 *  jph 2/29/96
 */

#ifndef GAINPARMS_C
#define GAINPARMS_C

#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include <string.h>

#include "teager2D.h"

/*
 * Function prototypes
 */
void ReadParametersFile();
void MkNoFiltParmFile();
void MkDomCompParmFile();
void RunNoFilt();
void RunDomComp();
void ComputeStats();
void AllocArrays();
void Report();

main(argc,argv)

  int   argc;
  char *argv[];

{
  double             r0;             /* radial cntr freq of 1st filt, Hz/samp */
  float              R;              /* common ratio of adjcnt filter radii */
  float              Eta;            /* adjcnt filters intersect at Eta-peak */
  float              obw;            /* Eta-peak octave BW of filteres */
  int                Nfiles;         /* no. input files to be processed */
  char              *InDataType;     /* array of type specifiers for inputs */
  int               *Conjugate;      /* flag array.  1=conjugate input */
  char             **Rfn;            /* array of Real input file names */
  char             **Ifn;            /* array of Imag input file names */
  int               *InSize;         /* sizes of the input signals */
  int               *ZP;             /* 1 if zero padding is to be used */
  int               *Reflect;        /* 1=reflect to fill zero padded areas */
  char               RptFnA[FNSIZE]; /* Filename for A parameters report */
  char               RptFnU[FNSIZE]; /* Filename for U parameters report */
  char               RptFnV[FNSIZE]; /* Filename for V parameters report */
  int                FileIdx;        /* index of input file being processed */
  float             *VarNA;          /* A measurement noise variance array */
  float             *VarNU;          /* U measurement noise variance array */
  float             *VarNV;          /* V measurement noise variance array */
  float             *VarApp;         /* variance of 2nd derivative of A array */
  float             *VarUpp;         /* variance of 2nd derivative of U array */
  float             *VarVpp;         /* variance of 2nd derivative of V array */
  float             *Pi0_11A;        /* Pi0 entry 1,1 for A array */
  float             *Pi0_12A;        /* Pi0 entry 1,2 for A array */
  float             *Pi0_22A;        /* Pi0 entry 2,2 for A array */
  float             *Pi0_11U;        /* Pi0 entry 1,1 for U array */
  float             *Pi0_12U;        /* Pi0 entry 1,2 for U array */
  float             *Pi0_22U;        /* Pi0 entry 2,2 for U array */
  float             *Pi0_11V;        /* Pi0 entry 1,1 for V array */
  float             *Pi0_12V;        /* Pi0 entry 1,2 for V array */
  float             *Pi0_22V;        /* Pi0 entry 2,2 for V array */
  int                TotNsamples;    /* total num samples in all inputs */
  int                i;              /* loop counter */
  float              Fact;           /* InSize[i] / TotNsamples */
  float              TotVarNA;       /* VarNA average for all runs */
  float              TotVarNU;       /* VarNU average for all runs */
  float              TotVarNV;       /* VarNV average for all runs */
  float              TotVarApp;      /* VarApp average for all runs */
  float              TotVarUpp;      /* VarUpp average for all runs */
  float              TotVarVpp;      /* VarVpp average for all runs */
  float              TotPi0_11A;     /* Pi0_11A average for all runs */
  float              TotPi0_12A;     /* Pi0_12A average for all runs */
  float              TotPi0_22A;     /* Pi0_22A average for all runs */
  float              TotPi0_11U;     /* Pi0_11U average for all runs */
  float              TotPi0_12U;     /* Pi0_12U average for all runs */
  float              TotPi0_22U;     /* Pi0_22U average for all runs */
  float              TotPi0_11V;     /* Pi0_11V average for all runs */
  float              TotPi0_12V;     /* Pi0_12V average for all runs */
  float              TotPi0_22V;     /* Pi0_22V average for all runs */

 /*
  * Check for right number of arguments
  */
  if (argc != 2) {
    printf("\n\n%s: Generate parameters for FindGains in 2D.",argv[0]);
    printf("\nusage: %s ParameterFileName\n",argv[0]);
    exit(-1);
  }
#ifdef DEBUG
 /*
  * Link to PA so we can use it in the debugger 
  */
  if (!argc) {
    PA((void *)ZP, 0, -1, 'b');
    PRINTA((void *)ZP,0,-1,'b',"delete.me");
    WRITEA((void *)ZP,0,-1,'b',"delete.me");
    image2disk((void *)ZP,1,1,'*',"delete.me");
    SpecMag2disk(ZP,"delete.me");
    PC(ZP,0,0,-1,'b',0);
    PRINTC(ZP,0,0,-1,'b',0,"delete.me");
    WRITEC(ZP,0,0,-1,'b',0,"delete.me");
#ifdef DEBUG_HEAP
    Sheapcheck();
#endif
  }
#endif

 /*
  * Read parameters file
  */
  ReadParametersFile(argv[0],argv[1],&r0,&R,&Eta,&obw,&Nfiles,&InDataType,
                     &Conjugate,&Rfn,&Ifn,&InSize,&ZP,&Reflect,RptFnA,RptFnU,
                     RptFnV);
 /*
  * Allocate working arrays
  */
  AllocArrays(argv,Nfiles,&VarNA,&VarNU,&VarNV,&VarApp,&VarUpp,&VarVpp,
              &Pi0_11A,&Pi0_12A,&Pi0_22A,&Pi0_11U,&Pi0_12U,&Pi0_22U,&Pi0_11V,
              &Pi0_12V,&Pi0_22V);

 /*
  * Compute statistics for each input file
  */
  for (FileIdx=0; FileIdx < Nfiles; FileIdx++) {
    ComputeStats(argv[0],FileIdx,InDataType,Conjugate,Rfn,Ifn,InSize,ZP,Reflect,
                 r0,R,Eta,obw,VarNA,VarNU,VarNV,VarApp,VarUpp,VarVpp,Pi0_11A,
                 Pi0_12A,Pi0_22A,Pi0_11U,Pi0_12U,Pi0_22U,Pi0_11V,Pi0_12V,
                 Pi0_22V);
  }

 /*
  * Average statistics across the runs
  */
  for (TotNsamples=i=0; i < Nfiles; i++) {
    TotNsamples += InSize[i] * InSize[i];
  }
  TotVarNA = TotVarNU   = TotVarNV   = 
             TotVarApp  = TotVarUpp  = TotVarVpp  = 
             TotPi0_11A = TotPi0_12A = TotPi0_22A = 
             TotPi0_11U = TotPi0_12U = TotPi0_22U =
             TotPi0_11V = TotPi0_12V = TotPi0_22V = (float)0.0;
  for (i=0; i < Nfiles; i++) {
    Fact = (float)InSize[i] * (float)InSize[i] / (float)TotNsamples;
    TotVarNA += VarNA[i] * Fact;
    TotVarNU += VarNU[i] * Fact;
    TotVarNV += VarNV[i] * Fact;

    TotVarApp += VarApp[i] * Fact;
    TotVarUpp += VarUpp[i] * Fact;
    TotVarVpp += VarVpp[i] * Fact;

    TotPi0_11A += Pi0_11A[i] * Fact;
    TotPi0_12A += Pi0_12A[i] * Fact;
    TotPi0_22A += Pi0_22A[i] * Fact;

    TotPi0_11U += Pi0_11U[i] * Fact;
    TotPi0_12U += Pi0_12U[i] * Fact;
    TotPi0_22U += Pi0_22U[i] * Fact;

    TotPi0_11V += Pi0_11U[i] * Fact;
    TotPi0_12V += Pi0_12U[i] * Fact;
    TotPi0_22V += Pi0_22U[i] * Fact;
  }

 /*
  * Report
  */
  Report(argv[0],RptFnA,TotVarNA,TotVarApp,TotPi0_11A,TotPi0_12A,TotPi0_22A);
  Report(argv[0],RptFnU,TotVarNU,TotVarUpp,TotPi0_11U,TotPi0_12U,TotPi0_22U);
  Report(argv[0],RptFnV,TotVarNV,TotVarVpp,TotPi0_11V,TotPi0_12V,TotPi0_22V);

  printf("\n%s Done.\n\n",argv[0]);
  return;

} /*-------------------------------------------- main --------------*/

/*
 * ReadParametersFile
 *
 *  read the data out of the parameters file
 *
 */
void ReadParametersFile(CallName,ParmFn,pr0,pR,pEta,pobw,pNfiles,pInDataType,
                        pConjugate,pRfn,pIfn,pInSize,pZP,pReflect,RptFnA,
                        RptFnU,RptFnV)

  char      *CallName;          /* argv[0], invoking cmd name from Unix */
  char      *ParmFn;            /* parameter filename */
  double    *pr0;               /* initial radial frequency in cpi */
  float     *pR;                /* common ratio of adjcnt filter radii */
  float     *pEta;              /* adj. filters intersect at Eta-peak */
  float     *pobw;              /* filter Eta-peak octave bandwidth */
  int       *pNfiles;           /* no. of input files to be processed */
  char     **pInDataType;       /* ptr to input data types array */
  int      **pConjugate;        /* ptr to input conjugate flags array */
  char    ***pRfn;              /* ptr to array of input Real fn's */
  char    ***pIfn;              /* ptr to array of input Imag fn's */
  int      **pInSize;           /* ptr to input size array */
  int      **pZP;               /* ptr to input Zero Pad array */
  int      **pReflect;          /* ptr to input Zero Pad array */
  char      *RptFnA;            /* Filename for A report file */
  char      *RptFnU;            /* Filename for U report file */
  char      *RptFnV;            /* Filename for V report file */
{

  FILE      *parmfp;            /* ptr to parameter file */
  char       instr[256];        /* buffer to read into  */
  int        r0Size;            /* signal size r0 referenced to */
  int        i;                 /* loop counter */
  int        ScanFileIdx;       /* Index for input file number being scanned */

 /*
  * Open parameters files
  */
  if ((parmfp = fopen(ParmFn,"r")) == NULL) {
    printf("\n\n%s:ReadParametersFile: unable to open %s\n",CallName,ParmFn);
    exit(-1);
  }

 /*
  * Scan filterbank parameters
  */
  fscanf(parmfp,"******** Filterbank Parameters %s\n",instr);
  fscanf(parmfp,"r0 Size= %d\n",&r0Size);
  fscanf(parmfp,"r0= %lf\n",pr0);
  *pr0 /= (double)r0Size;
  fscanf(parmfp,"R= %f\n",pR);

  fscanf(parmfp,"Eta= %f\n",pEta);
  fscanf(parmfp,"OBW= %f\n",pobw);

 /*
  * Scan number of input files and allocate arrays
  */
  fscanf(parmfp,"******** Number of Input Files %s\n",instr);
  fscanf(parmfp,"Num Files= %d\n",pNfiles);

  /* Allocate data type array */
  if ((*pInDataType = (char *)malloc(*pNfiles*sizeof(char))) == NULL) {
    printf("\n\n%s:ReadParametersFile:  Free Store Exhausted.\n",CallName);
    exit(-1);
  }

  /* Allocate conjugate array */
  if ((*pConjugate = (int *)malloc(*pNfiles*sizeof(int))) == NULL) {
    printf("\n\n%s:ReadParametersFile:  Free Store Exhausted.\n",CallName);
    exit(-1);
  }

  /* Allocate output Real filenames array */
  if ((*pRfn = (char **)malloc(*pNfiles * sizeof(char*))) == NULL) {
    printf("\n\n%s:ReadParametersFile:  Free Store Exhausted.\n",CallName);
    exit(-1);
  }
  for (i=0; i < *pNfiles; i++) {
    if (((*pRfn)[i] = (char *)malloc(FNSIZE*sizeof(char))) == NULL) {
      printf("\n\n%s:ReadParametersFile:  Free Store Exhausted.\n",CallName);
      exit(-1);
    }
  }

  /* Allocate output Imag filenames array */
  if ((*pIfn = (char **)malloc(*pNfiles * sizeof(char*))) == NULL) {
    printf("\n\n%s:ReadParametersFile:  Free Store Exhausted.\n",CallName);
    exit(-1);
  }
  for (i=0; i < *pNfiles; i++) {
    if (((*pIfn)[i] = (char *)malloc(FNSIZE*sizeof(char))) == NULL) {
      printf("\n\n%s:ReadParametersFile:  Free Store Exhausted.\n",CallName);
      exit(-1);
    }
  }

  /* Allocate InSize array */
  if ((*pInSize = (int *)malloc(*pNfiles*sizeof(int))) == NULL) {
    printf("\n\n%s:ReadParametersFile:  Free Store Exhausted.\n",CallName);
    exit(-1);
  }

  /* Allocate ZeroPad array */
  if ((*pZP = (int *)malloc(*pNfiles*sizeof(int))) == NULL) {
    printf("\n\n%s:ReadParametersFile:  Free Store Exhausted.\n",CallName);
    exit(-1);
  }

  /* Allocate Reflect array */
  if ((*pReflect = (int *)malloc(*pNfiles*sizeof(int))) == NULL) {
    printf("\n\n%s:ReadParametersFile:  Free Store Exhausted.\n",CallName);
    exit(-1);
  }

 /*
  * Scan report file name
  */
  fscanf(parmfp,"******** Report File %s\n",instr);
  fscanf(parmfp,"A Report File= %s\n",RptFnA);
  fscanf(parmfp,"U Report File= %s\n",RptFnU);
  fscanf(parmfp,"V Report File= %s\n",RptFnV);

 /*
  * Loop to scan run parameters for each input file
  */
  for (i=0; i < *pNfiles; i++) { 
   /*
    * Scan index of input file
    */
    fscanf(parmfp,"******** Input File %d ********\n",&ScanFileIdx);
    if (ScanFileIdx != (i+1)) {
      printf("\n\n%s:ReadParametersFile: %s %d %s %d %s\n",CallName,
             "Input File",ScanFileIdx,"Encountered,",(i+1),"Expected.");
      exit(-1);
    }

   /*
    * Scan input data type 
    */
    fscanf(parmfp,"Input Data Type= %s\n",instr);
    (*pInDataType)[i] = instr[0];

   /*
    * Scan conjugate flag
    */
    fscanf(parmfp,"Conjugate Input= %s\n",instr);
    if (instr[0] == 'N') {
      (*pConjugate)[i] = 0;
    }
    else {
      (*pConjugate)[i] = 1;
    }

   /*
    * Scan Real and Imag input filenames
    */
    fscanf(parmfp,"Real Image File= %s\n",(*pRfn)[i]);
    fscanf(parmfp,"Imag Image File= %s\n",(*pIfn)[i]);

   /*
    * Scan input size
    */
    fscanf(parmfp,"Size= %d\n",&(*pInSize)[i]);

   /*
    * Scan Zero Pad flag
    */
    fscanf(parmfp,"Zero Pad= %s\n",instr);
    if ((instr[0] == 'Y') || (instr[0] == 'y')) {
      (*pZP)[i] = 1;
    }
    else {
      (*pZP)[i] = 0;
    }

   /*
    * Scan reflection flag
    */
    fscanf(parmfp,"Reflect= %s\n",instr);
    if ((instr[0] == 'Y') || (instr[0] == 'y')) {
      (*pReflect)[i] = 1;
    }
    else {
      (*pReflect)[i] = 0;
    }
  }    /* for i, loop on input files */
  fclose(parmfp);
  return;
} /*--------------------------- ReadParametersFile -------------------------*/

/*
 * MkNoFiltParmFile
 *
 *  Make a parameter file for NoFilt.c
 *
 *
 */
void MkNoFiltParmFile(CallName,idx,InDataType,Conjugate,Rfn,Ifn,InSize,ZP)

  char      *CallName;          /* argv[0], invoking cmd name from Unix */
  int        idx;               /* index of input file to be processed */
  char      *InDataType;        /* char array of input data types */
  int       *Conjugate;         /* array of input conjugate flags */
  char     **Rfn;               /* array of input Real fn's */
  char     **Ifn;               /* array of input Imag fn's */
  int       *InSize;            /* array of input signal sizes */
  int       *ZP;                /* array of input Zero Pad flags */
{

  FILE      *parmfp;            /* ptr to parameter file */

 /*
  * Open parameter file for writing
  */
  system("touch /export/home2/joebob/tmp/NFtmp.parms");
  system("rm /export/home2/joebob/tmp/NFtmp.parms");
  if ((parmfp = fopen("/export/home2/joebob/tmp/NFtmp.parms","w")) == NULL) {
    printf("\n\n%s:MkNoFiltParmFile: unable to open %s for writing\n",
           CallName,"/export/home2/joebob/tmp/NFtmp.parms");
    exit(-1);
  }

 /*
  * Write parameter file records
  */
  fprintf(parmfp,"Input Data Type= %c\n",InDataType[idx]);
  if (Conjugate[idx]) {
    fprintf(parmfp,"Conjugate Input= Yes\n");
  } else {
    fprintf(parmfp,"Conjugate Input= No\n");
  }
  fprintf(parmfp,"Real Image File= %s\n",Rfn[idx]);
  fprintf(parmfp,"Imag Image File= %s\n",Ifn[idx]);
  fprintf(parmfp,"Size= %d\n",InSize[idx]);
  fprintf(parmfp,"A File= /export/home2/joebob/tmp/NFtmpA.float\n");
  fprintf(parmfp,"U File= /export/home2/joebob/tmp/NFtmpU.float\n");
  fprintf(parmfp,"V File= /export/home2/joebob/tmp/NFtmpV.float\n");
  fclose(parmfp);
  return;
} /*--------------------------- MkNoFiltParmFile -------------------------*/

/*
 * MkDomCompParmFile
 *
 *  Make a parameter file for DomComp.c
 *
 */
void MkDomCompParmFile(CallName,idx,InDataType,Conjugate,Rfn,Ifn,InSize,ZP,
                         Reflect,r0,R,Eta,obw)

  char      *CallName;          /* argv[0], invoking cmd name from Unix */
  int        idx;               /* index of input file to be processed */
  char      *InDataType;        /* char array of input data types */
  int       *Conjugate;         /* array of input conjugate flags */
  char     **Rfn;               /* array of input Real fn's */
  char     **Ifn;               /* array of input Imag fn's */
  int       *InSize;            /* array of input signal sizes */
  int       *ZP;                /* array of input Zero Pad flags */
  int       *Reflect;           /* array of input Reflect flags */
  double     r0;                /* radial ctr freq of 1st filter */
  float      R;                 /* common ratio of adcnt filter radii */
  float      Eta;               /* adjcnt filts intersect at Eta-peak */
  float      obw;               /* filter Eta-peak octave bandwidth */
{

  FILE      *parmfp;            /* ptr to parameter file */

 /*
  * Open parameter file for writing
  */
  system("touch /export/home2/joebob/tmp/Domtmp.parms");
  system("rm /export/home2/joebob/tmp/Domtmp.parms");
  if ((parmfp = fopen("/export/home2/joebob/tmp/Domtmp.parms","w")) == NULL) {
    printf("\n\n%s:MkNoFiltParmFile: unable to open %s for writing\n",
           CallName,"/export/home2/joebob/tmp/Domtmp.parms");
    exit(-1);
  }

 /*
  * Write parameter file records
  */
  fprintf(parmfp,"Input Data Type= %c\n",InDataType[idx]);
  if (Conjugate[idx]) {
    fprintf(parmfp,"Conjugate Input= Yes\n");
  } else {
    fprintf(parmfp,"Conjugate Input= No\n");
  }
  fprintf(parmfp,"Real Image File= %s\n",Rfn[idx]);
  fprintf(parmfp,"Imag Image File= %s\n",Ifn[idx]);
  fprintf(parmfp,"Size= %d\n",InSize[idx]);
  if (ZP[idx]) {
    fprintf(parmfp,"Zero Pad= Yes\n");
  } else {
    fprintf(parmfp,"Zero Pad= No\n");
  }
  if (Reflect[idx]) {
    fprintf(parmfp,"Reflect= Yes\n");
  } else {
    fprintf(parmfp,"Reflect= No\n");
  }
  fprintf(parmfp,"r0= %lf\n",r0*(double)InSize[idx]);
  fprintf(parmfp,"R= %f\n",R);
  fprintf(parmfp,"Eta= %f\n",Eta);
  fprintf(parmfp,"OBW= %f\n",obw);
  fprintf(parmfp,"A File= /export/home2/joebob/tmp/DomtmpA.float\n");
  fprintf(parmfp,"U File= /export/home2/joebob/tmp/DomtmpU.float\n");
  fprintf(parmfp,"V File= /export/home2/joebob/tmp/DomtmpV.float\n");
  fprintf(parmfp,"MaxFilt File= none\n");
  fprintf(parmfp,"Dominant Component Real File= none\n");
  fprintf(parmfp,"Dominant Component Imag File= none\n");
  fprintf(parmfp,"Print Response Images= No\n");
  fprintf(parmfp,"Response Root File Name= none\n");
  fprintf(parmfp,"Print Channel Estimates= No\n");
  fprintf(parmfp,"Chan Est A Root File Name= none\n");
  fprintf(parmfp,"Chan Est U Root File Name= none\n");
  fprintf(parmfp,"Chan Est V Root File Name= none\n");
  fprintf(parmfp,"Reconstruction Interval= %d\n",InSize[idx]);
  fprintf(parmfp,"Postfilter= No\n");
  fprintf(parmfp,"kappa= 1.0\n");
  fprintf(parmfp,"Chan Asm Root Fn= none\n");
  fprintf(parmfp,"Chan Usm Root Fn= none\n");
  fprintf(parmfp,"Chan Vsm Root Fn= none\n");
  fclose(parmfp);
  return;
} /*--------------------------- MkDomCompParmFile ----------------------*/

/*
 * RunNoFilt
 *
 *  Run the program NoFilt
 *
 */
void RunNoFilt(CallName,idx,InDataType,Conjugate,Rfn,Ifn,InSize,ZP,Au,Uu,Vu)

  char      *CallName;          /* argv[0], invoking cmd name from Unix */
  int        idx;               /* index of input file to be processed */
  char      *InDataType;        /* char array of input data types */
  int       *Conjugate;         /* array of input conjugate flags */
  char     **Rfn;               /* array of input Real fn's */
  char     **Ifn;               /* array of input Imag fn's */
  int       *InSize;            /* array of input signal sizes */ 
  int       *ZP;                /* array of input Zero Pad flags */
  float    **Au;                /* unfiltered amp estimates */
  float    **Uu;                /* unfiltered freq estimates */
  float    **Vu;                /* unfiltered freq estimates */
{

 /*
  * Create parameter file
  */
  MkNoFiltParmFile(CallName,idx,InDataType,Conjugate,Rfn,Ifn,InSize,ZP);

 /*
  * Run the program and remove parameter file
  */
  printf("\n%s: Running NoFilt on input file %d...",CallName,idx);
  system("NoFilt /export/home2/joebob/tmp/NFtmp.parms");
  system("rm /export/home2/joebob/tmp/NFtmp.parms");

 /*
  * Allocate result arrays 
  */
  if ((*Au = (float *)malloc(InSize[idx]*InSize[idx]*sizeof(float))) == NULL) {
    printf("\n\n%s:RunNoFilt: Free Store Exhausted.\n",CallName);
    exit(-1);
  }
  if ((*Uu = (float *)malloc(InSize[idx]*InSize[idx]*sizeof(float))) == NULL) {
    printf("\n\n%s:RunNoFilt: Free Store Exhausted.\n",CallName);
    exit(-1);
  }
  if ((*Vu = (float *)malloc(InSize[idx]*InSize[idx]*sizeof(float))) == NULL) {
    printf("\n\n%s:RunNoFilt: Free Store Exhausted.\n",CallName);
    exit(-1);
  }

 /*
  * Read results and remove result files
  */
  disk2f(*Au,InSize[idx],InSize[idx],"/export/home2/joebob/tmp/NFtmpA.float");
  disk2f(*Uu,InSize[idx],InSize[idx],"/export/home2/joebob/tmp/NFtmpU.float");
  disk2f(*Vu,InSize[idx],InSize[idx],"/export/home2/joebob/tmp/NFtmpV.float");
  system("rm /export/home2/joebob/tmp/NFtmpA.float");
  system("rm /export/home2/joebob/tmp/NFtmpU.float");
  system("rm /export/home2/joebob/tmp/NFtmpV.float");

  return;
} /*--------------------------- RunNoFilt --------------------------------*/

/*
 * RunDomComp
 *
 *  Run the program DomComp
 *
 */
void RunDomComp(CallName,idx,InDataType,Conjugate,Rfn,Ifn,InSize,ZP,Reflect,
                  r0,R,Eta,obw,Af,Uf,Vf)

  char      *CallName;          /* argv[0], invoking cmd name from Unix */
  int        idx;               /* index of input file to be processed */
  char      *InDataType;        /* char array of input data types */
  int       *Conjugate;         /* array of input conjugate flags */
  char     **Rfn;               /* array of input Real fn's */
  char     **Ifn;               /* array of input Imag fn's */
  int       *InSize;            /* array of input signal sizes */ 
  int       *ZP;                /* array of input Zero Pad flags */
  int       *Reflect;           /* array of input Reflect flags */
  double     r0;                /* radial ctr freq of first filter */
  float      R;                 /* common ratio of adjcnt filter radii */
  float      Eta;               /* adjcnt fltrs intersect at Eta-peak */
  float      obw;               /* filter Eta-peak octave bandwidth */
  float    **Af;                /* filtered amp estimates */
  float    **Uf;                /* filtered freq estimates */
  float    **Vf;                /* filtered freq estimates */
{

 /*
  * Create parameter file
  */
  MkDomCompParmFile(CallName,idx,InDataType,Conjugate,Rfn,Ifn,InSize,ZP,
                    Reflect,r0,R,Eta,obw);

 /*
  * Run the program and remove parameter file
  */
  printf("\n%s: Running DomComp on input file %d...",CallName,idx);
  system("DomComp /export/home2/joebob/tmp/Domtmp.parms");
  system("rm /export/home2/joebob/tmp/Domtmp.parms");

 /*
  * Allocate result arrays 
  */
  if ((*Af = (float *)malloc(InSize[idx]*InSize[idx]*sizeof(float))) == NULL) {
    printf("\n\n%s:RunDomComp: Free Store Exhausted.\n",CallName);
    exit(-1);
  }
  if ((*Uf = (float *)malloc(InSize[idx]*InSize[idx]*sizeof(float))) == NULL) {
    printf("\n\n%s:RunDomComp: Free Store Exhausted.\n",CallName);
    exit(-1);
  }
  if ((*Vf = (float *)malloc(InSize[idx]*InSize[idx]*sizeof(float))) == NULL) {
    printf("\n\n%s:RunDomComp: Free Store Exhausted.\n",CallName);
    exit(-1);
  }

 /*
  * Read results and remove result files
  */
  disk2f(*Af,InSize[idx],InSize[idx],"/export/home2/joebob/tmp/DomtmpA.float");
  disk2f(*Uf,InSize[idx],InSize[idx],"/export/home2/joebob/tmp/DomtmpU.float");
  disk2f(*Vf,InSize[idx],InSize[idx],"/export/home2/joebob/tmp/DomtmpV.float");
  system("rm /export/home2/joebob/tmp/DomtmpA.float");
  system("rm /export/home2/joebob/tmp/DomtmpU.float");
  system("rm /export/home2/joebob/tmp/DomtmpV.float");

  return;
} /*--------------------------- RunDomComp -------------------------------*/

/*
 * ComputeStats
 *
 *  For one input file, compute:
 *    - amplitude measurement noise variance, sigma_na^2
 *    - frequency measurement noise variance, sigma_nu^2
 *    - frequency measurement noise variance, sigma_nv^2
 *    - amplitude modulation acceleration variance, sigma_adotdot^2
 *    - frequency modulation acceleration variance, sigma_udotdot^2
 *    - frequency modulation acceleration variance, sigma_vdotdot^2
 *    - Pi0_11A initial state vector covariance matrix entry 1,1 for A
 *    - Pi0_12A initial state vector covariance matrix entry 1,2 for A
 *    - Pi0_22A initial state vector covariance matrix entry 2,2 for A
 *    - Pi0_11U initial state vector covariance matrix entry 1,1 for U
 *    - Pi0_12U initial state vector covariance matrix entry 1,2 for U
 *    - Pi0_22U intiial state vecotr covariance matrix entry 2,2 for U
 *    - Pi0_11V initial state vector covariance matrix entry 1,1 for V
 *    - Pi0_12V initial state vector covariance matrix entry 1,2 for V
 *    - Pi0_22V intiial state vecotr covariance matrix entry 2,2 for V
 */
void ComputeStats(CallName,idx,InDataType,Conjugate,Rfn,Ifn,InSize,ZP,Reflect,
                  r0,R,Eta,obw,VarNA,VarNU,VarNV,VarApp,VarUpp,VarVpp,Pi0_11A,
                  Pi0_12A,Pi0_22A,Pi0_11U,Pi0_12U,Pi0_22U,Pi0_11V,Pi0_12V,
                  Pi0_22V)

  char      *CallName;          /* argv[0], invoking cmd name from Unix */
  int        idx;               /* index of input file to be processed */
  char      *InDataType;        /* char array of input data types */
  int       *Conjugate;         /* array of input conjugate flags */
  char     **Rfn;               /* array of input Real fn's */
  char     **Ifn;               /* array of input Imag fn's */
  int       *InSize;            /* array of input signal sizes */ 
  int       *ZP;                /* array of input Zero Pad flags */
  int       *Reflect;           /* array of input Reflect flags */
  double     r0;                /* radial ctr freq of first filter */
  float      R;                 /* common ratio of adcnt filter radii */
  float      Eta;               /* adjcnt fltrs intersect at Eta-peak */
  float      obw;               /* filter Eta-peak octave bandwidth */
  float     *VarNA;             /* A measurement noise variance array */
  float     *VarNU;             /* U measurement noise variance array */
  float     *VarNV;             /* V measurement noise variance array */
  float     *VarApp;            /* variance of 2nd derivative of A array */
  float     *VarUpp;            /* variance of 2nd derivative of U array */
  float     *VarVpp;            /* variance of 2nd derivative of V array */
  float     *Pi0_11A;           /* Pi0 entry 1,1 for A array */
  float     *Pi0_12A;           /* Pi0 entry 1,2 for A array */
  float     *Pi0_22A;           /* Pi0 entry 2,2 for A array */
  float     *Pi0_11U;           /* Pi0 entry 1,1 for U array */
  float     *Pi0_12U;           /* Pi0 entry 1,2 for U array */
  float     *Pi0_22U;           /* Pi0 entry 2,2 for U array */
  float     *Pi0_11V;           /* Pi0 entry 1,1 for V array */
  float     *Pi0_12V;           /* Pi0 entry 1,2 for V array */
  float     *Pi0_22V;           /* Pi0 entry 2,2 for V array */
{

  float     *Au;                /* unfiltered amp estimates */
  float     *Uu;                /* unfiltered freq estimates */
  float     *Vu;                /* unfiltered freq estimates */
  float     *Af;                /* filtered amp estimates */
  float     *Uf;                /* filtered freq estimates */
  float     *Vf;                /* filtered freq estimates */
  float     *n_A;               /* measurement noise sequence for A */
  float     *n_U;               /* measurement noise sequence for U */
  float     *n_V;               /* measurement noise sequence for V */
  int        i;                 /* counter */
  float      mu;                /* sample mean of a float sequence */
  float     *D1A;               /* 1st finite difference of A */
  float     *D2A;               /* 2nd finite difference of A */
  float     *D1U;               /* 1st finite difference of U */
  float     *D2U;               /* 2nd finite difference of U */
  float     *D1V;               /* 1st finite difference of V */
  float     *D2V;               /* 2nd finite difference of V */

 /*
  * Run unfiltered estimation algorithm
  */
  RunNoFilt(CallName,idx,InDataType,Conjugate,Rfn,Ifn,InSize,ZP,&Au,&Uu,&Vu);

 /*
  * Run dominant component estimation algorithm
  */
  RunDomComp(CallName,idx,InDataType,Conjugate,Rfn,Ifn,InSize,ZP,Reflect,r0,R,
             Eta,obw,&Af,&Uf,&Vf);
 /*
  * Allocate measurement noise sequences
  */
  if ((n_A = (float *)malloc(InSize[idx]*InSize[idx]*sizeof(float))) == NULL) {
    printf("\n\n%s:ComputStats: Free Store Exhausted.\n",CallName);
    exit(-1);
  }
  if ((n_U = (float *)malloc(InSize[idx]*InSize[idx]*sizeof(float))) == NULL) {
    printf("\n\n%s:ComputStats: Free Store Exhausted.\n",CallName);
    exit(-1);
  }
  if ((n_V = (float *)malloc(InSize[idx]*InSize[idx]*sizeof(float))) == NULL) {
    printf("\n\n%s:ComputStats: Free Store Exhausted.\n",CallName);
    exit(-1);
  }

 /*
  * Compute measurement noise sequences
  */
  for (i=0; i < InSize[idx]*InSize[idx]; i++) {
    n_A[i] = Af[i] - Au[i];
    n_U[i] = Uf[i] - Uu[i];
    n_V[i] = Vf[i] - Vu[i];
  }
  free(Af);
  free(Uf);
  free(Vf);

 /*
  * Compute measurement noise variances
  */
  mu = fmean(n_A,InSize[idx],InSize[idx]);
  VarNA[idx] = fvariance(n_A,InSize[idx],InSize[idx],mu);
  mu = fmean(n_U,InSize[idx],InSize[idx]);
  VarNU[idx] = fvariance(n_U,InSize[idx],InSize[idx],mu);
  mu = fmean(n_V,InSize[idx],InSize[idx]);
  VarNV[idx] = fvariance(n_V,InSize[idx],InSize[idx],mu);
  free(n_A);
  free(n_U);
  free(n_V);

 /*
  * Allocate arrays for 1st and 2nd finite differences of A, U and V
  */
  if ((D1A = (float *)malloc(InSize[idx]*InSize[idx]*sizeof(float))) == NULL) {
    printf("\n\n%s:ComputStats: Free Store Exhausted.\n",CallName);
    exit(-1);
  }
  if ((D2A = (float *)malloc(InSize[idx]*InSize[idx]*sizeof(float))) == NULL) {
    printf("\n\n%s:ComputStats: Free Store Exhausted.\n",CallName);
    exit(-1);
  }
  if ((D1U = (float *)malloc(InSize[idx]*InSize[idx]*sizeof(float))) == NULL) {
    printf("\n\n%s:ComputStats: Free Store Exhausted.\n",CallName);
    exit(-1);
  }
  if ((D2U = (float *)malloc(InSize[idx]*InSize[idx]*sizeof(float))) == NULL) {
    printf("\n\n%s:ComputStats: Free Store Exhausted.\n",CallName);
    exit(-1);
  }
  if ((D1V = (float *)malloc(InSize[idx]*InSize[idx]*sizeof(float))) == NULL) {
    printf("\n\n%s:ComputStats: Free Store Exhausted.\n",CallName);
    exit(-1);
  }
  if ((D2V = (float *)malloc(InSize[idx]*InSize[idx]*sizeof(float))) == NULL) {
    printf("\n\n%s:ComputStats: Free Store Exhausted.\n",CallName);
    exit(-1);
  }

 /*
  * Compute 1st and 2nd differences of A, U and V
  */
  fFiniteDifP(Au,InSize[idx],InSize[idx],D1A);
  fFiniteDifP(D1A,InSize[idx],InSize[idx],D2A);
  fFiniteDifP(Uu,InSize[idx],InSize[idx],D1U);
  fFiniteDifP(D1U,InSize[idx],InSize[idx],D2U);
  fFiniteDifP(Vu,InSize[idx],InSize[idx],D1V);
  fFiniteDifP(D1V,InSize[idx],InSize[idx],D2V);

 /*
  * Compute modulation acceleration variances
  */
  mu = fmean(D2A,InSize[idx],InSize[idx]);
  VarApp[idx] = fvariance(D2A,InSize[idx],InSize[idx],mu);
  mu = fmean(D2U,InSize[idx],InSize[idx]);
  VarUpp[idx] = fvariance(D2U,InSize[idx],InSize[idx],mu);
  mu = fmean(D2V,InSize[idx],InSize[idx]);
  VarVpp[idx] = fvariance(D2V,InSize[idx],InSize[idx],mu);

 /*
  * Compute Pi0 values for A, U and V
  */
  mu = fmean(Au,InSize[idx],InSize[idx]);
  Pi0_11A[idx] = fvariance(Au,InSize[idx],InSize[idx],mu);
  Pi0_12A[idx] = (float)0.0;  /* WSS m.s. dif. field is uncor w/ derivative */
  mu = fmean(D1A,InSize[idx],InSize[idx]);
  Pi0_22A[idx] = fvariance(D1A,InSize[idx],InSize[idx],mu);

  mu = fmean(Uu,InSize[idx],InSize[idx]);
  Pi0_11U[idx] = fvariance(Uu,InSize[idx],InSize[idx],mu);
  Pi0_12U[idx] = (float)0.0;  /* WSS m.s. dif. field is uncor w/ derivative */
  mu = fmean(D1U,InSize[idx],InSize[idx]);
  Pi0_22U[idx] = fvariance(D1U,InSize[idx],InSize[idx],mu);

  mu = fmean(Vu,InSize[idx],InSize[idx]);
  Pi0_11V[idx] = fvariance(Vu,InSize[idx],InSize[idx],mu);
  Pi0_12V[idx] = (float)0.0;  /* WSS m.s. dif. field is uncor w/ derivative */
  mu = fmean(D1V,InSize[idx],InSize[idx]);
  Pi0_22V[idx] = fvariance(D1V,InSize[idx],InSize[idx],mu);

 /*
  * free arrays
  */
  free(D1A);
  free(D2A);
  free(D1U);
  free(D2U);
  free(D1V);
  free(D2V);
  free(Au);
  free(Uu);
  free(Vu);

  return;
} /*--------------------------- ComputeStats -------------------------------*/

/*
 * AllocArrays
 *
 *  Allocate Working Arrays.
 */
 
void AllocArrays(argv,Nfiles,VarNA,VarNU,VarNV,VarApp,VarUpp,VarVpp,Pi0_11A,
                 Pi0_12A,Pi0_22A,Pi0_11U,Pi0_12U,Pi0_22U,Pi0_11V,Pi0_12V,
                 Pi0_22V)

  char              *argv[];         /* argument list from command line */
  int                Nfiles;         /* no. input files to be processed */
  float            **VarNA;          /* A measurement noise variance array */
  float            **VarNU;          /* U measurement noise variance array */
  float            **VarNV;          /* V measurement noise variance array */
  float            **VarApp;         /* variance of 2nd derivative of A array */
  float            **VarUpp;         /* variance of 2nd derivative of U array */
  float            **VarVpp;         /* variance of 2nd derivative of V array */
  float            **Pi0_11A;        /* Pi0 entry 1,1 for A array */
  float            **Pi0_12A;        /* Pi0 entry 1,2 for A array */
  float            **Pi0_22A;        /* Pi0 entry 2,2 for A array */
  float            **Pi0_11U;        /* Pi0 entry 1,1 for U array */
  float            **Pi0_12U;        /* Pi0 entry 1,2 for U array */
  float            **Pi0_22U;        /* Pi0 entry 2,2 for U array */
  float            **Pi0_11V;        /* Pi0 entry 1,1 for V array */
  float            **Pi0_12V;        /* Pi0 entry 1,2 for V array */
  float            **Pi0_22V;        /* Pi0 entry 2,2 for V array */
{

 /*
  * Allocate working arrays
  */
  if ((*VarNA = (float *)malloc(Nfiles * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*VarNU = (float *)malloc(Nfiles * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*VarNV = (float *)malloc(Nfiles * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*VarApp = (float *)malloc(Nfiles * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*VarUpp = (float *)malloc(Nfiles * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*VarVpp = (float *)malloc(Nfiles * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*Pi0_11A = (float *)malloc(Nfiles * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*Pi0_12A = (float *)malloc(Nfiles * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*Pi0_22A = (float *)malloc(Nfiles * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*Pi0_11U = (float *)malloc(Nfiles * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*Pi0_12U = (float *)malloc(Nfiles * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*Pi0_22U = (float *)malloc(Nfiles * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*Pi0_11V = (float *)malloc(Nfiles * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*Pi0_12V = (float *)malloc(Nfiles * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*Pi0_22V = (float *)malloc(Nfiles * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
 
  return;
} /*---------------------  AllocArrays  --------------------------------------*/

/*
 * Report
 *
 *  Generate a report file giving the parameters needed to compute the 
 *  gains for one Kalman filter.  The format is consistent with the
 *  parameter file format of FindGains.c (see book XI, p. 30).  Some
 *  additional parameters will still have to be added to the file manually.
 */
void Report(CallName,RptFn,VarN,VarMA,Pi0_11,Pi0_12,Pi0_22)

  char      *CallName;          /* argv[0], invoking cmd name from Unix */
  char      *RptFn;             /* report file name */
  float      VarN;              /* measurement noise variance */
  float      VarMA;             /* modulation acceleration variance */
  float      Pi0_11;            /* entry 1,1 of state vector covariance mtrx */
  float      Pi0_12;            /* entry 1,2 of state vector covariance mtrx */
  float      Pi0_22;            /* entry 2,2 of state vector covariance mtrx */
{

  FILE      *RptFp;             /* ptr to Report file */

 /*
  * Open Report file
  */
  if ((RptFp = fopen(RptFn,"w")) == NULL) {
    printf("\n\n%s:Report: unable to open %s for writing\n",CallName,RptFn);
    exit(-1);
  }

 /*
  * Write report
  */
  fprintf(RptFp,"Pi0[1][1]= %f\n",Pi0_11);
  fprintf(RptFp,"Pi0[1][2]= %f\n",Pi0_12);
  fprintf(RptFp,"Pi0[2][1]= %f\n",Pi0_12);
  fprintf(RptFp,"Pi0[2][2]= %f\n",Pi0_22);
  fprintf(RptFp,"Variance n= %f\n",VarN);
  fprintf(RptFp,"Variance udotdot= %f\n",VarMA);
  fprintf(RptFp,"MaxNumGains= *\n");
  fprintf(RptFp,"Abs Threshold= *\n");
  fprintf(RptFp,"Rel Threshold= *\n");
  fprintf(RptFp,"Alpha Fn= *\n");
  fprintf(RptFp,"Beta  Fn= *\n");
  fclose(RptFp);
  return;
} /*--------------------------- Report -------------------------------------*/
#endif
