/*
 *
 * MkObs.c
 *
 *  This program makes the ESA observations to input to the frequency tracker.
 *
 *  It was made by modifying the program 
 *     /anchovy/export/home/joebob/asilomar/asilomar.c
 *  which is an upgraded implementation of the asilomar ESA.
 *
 *  In this program, the ESA is evaluated for all frequency channels at every 
 *  point in the images.  
 *
 *  Input parameters are read from a params file specified as the command line
 *  option.  See p. XXX of Book VI for a description of the paramter file.
 *
 *  This program is quick and dirty.
 *
 *  FOR POSTFILTERING:     leave POSTFILTER #defined
 *  FOR NO POSTFILTERING:  #undef POSTFILTER
 * 
 *  REVISED 8/14/94: see book VIII p. 82.
 *
 *  REVISED 11/4/94: f_u_norm calls commented out.
 *
 *  MAJOR MODS: 2/29/96
 *
 *  4/15/96: Postfiltering added.  See book XI, pp. 119 - 151.
 *
 *  usage: MkObs ParameterFileName
 *
 *  jph 2/13/94
 *
 */

#ifndef MKOBS_C
#define MKOBS_C

#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include <string.h>

#include "teager2D.h"

#define NO_OBSERVATION ((float)-10.0)

/*
 * Function prototypes
 */
void Initialize();
void ReadParametersFile();
void dom_esa();
void AllocArrays();

main(argc,argv)

  int   argc;
  char *argv[];

{

  filterbank        *G;              /* the filter bank */
  pFloatImage        g;              /* postfilter image */
  pFloatImage        X;              /* input image */
  pFloatImage        Y;              /* filtered image */
  int                size;           /* row/col dim of the square input image */
  int                n;              /* size squared */
  int                i;              /* counter */
  int                k;              /* loop counter */
  int                row;            /* image row counter */
  int                col;            /* image col counter */
  int                looplim;        /* terminal value for a loop counter */
  int                FILTER;         /* loop counter */
  int                writesize;      /* num bytes to output in a single write */
  int                ZeroPad;        /* 1 if zero padding is to be used */
  int                Reflect;        /* 1=fill 0 padded areas w/ reflection */
  int                fd;             /* file descriptor */
  float             *A;              /* A estimates accross all 40 filters */
  float             *U;              /* U estimates accross all 40 filters */
  float             *V;              /* V estimates accross all 40 filters */
  BYTE              *a_valid;        /* 1 where the a estimate is valid */
  BYTE              *u_valid;        /* 1 where the u estimate is valid */
  BYTE              *v_valid;        /* 1 where the v estimate is valid */
  float             *OutDat[3];      /* ptrs to array of data to output */
  float             *OutVector;      /* vector of obs ready to output */
  int                PrintResponse;  /* 1 if we are to print response images */
  char              *OutFn[3];       /* names for observation files */
  char               RespRootFn[FNSIZE]; /* root filename for response images */
  int                PrintChannels;  /* 1=print A,U,V est. from each channel */
  char               ChanARootFn[FNSIZE];/* root fn for Channel A estimates */
  char               ChanURootFn[FNSIZE];/* root fn for Channel U estimates */
  char               ChanVRootFn[FNSIZE];/* root fn for Channel V estimates */
  char               AobsFn[FNSIZE]; /* filename for A observations */
  char               UobsFn[FNSIZE]; /* filename for U observations */
  char               VobsFn[FNSIZE]; /* filename for V observations */
  char               TempString[FNSIZE]; /* string for building up filenames */
  int                PostFilter;     /* 1 if postfiltering is to be done */
  char               ChanAsmRootFn[FNSIZE];/* Filename for smoothed amplitude */
  char               ChanUsmRootFn[FNSIZE];/* Filename for smoothed horiz freq*/
  char               ChanVsmRootFn[FNSIZE];/* Filename for smoothed vert  freq*/

 /*
  * Check for right number of arguments
  */
  if (argc != 2) {
    printf("\n\n%s: Make observations for 2D freq tracking algorithm.",argv[0]);
    printf("\nusage: %s ParameterFileName\n",argv[0]);
    exit(-1);
  }
#ifdef DEBUG
 /*
  * Link to PA so we can use it in the debugger 
  */
  if (!argc) {
    PA((void *)X, 0, -1, 'b');
    PRINTA((void *)X,0,-1,'b',"delete.me");
    WRITEA((void *)X,0,-1,'b',"delete.me");
    image2disk((void *)X,1,1,'*',"delete.me");
    SpecMag2disk(X,"delete.me");
    PC(X,0,0,-1,'b',0);
    PRINTC(X,0,0,-1,'b',0,"delete.me");
    WRITEC(X,0,0,-1,'b',0,"delete.me");
  }
#endif
  X = (pFloatImage)malloc(sizeof(FloatImage));
  Y = (pFloatImage)malloc(sizeof(FloatImage));

  Initialize(argv[1],&size,&ZeroPad,&Reflect,&X->Real,&X->Imag,&X->FFTReal,
             &X->FFTImag,&G,&PrintResponse,RespRootFn,&PrintChannels,
             ChanARootFn,ChanURootFn,ChanVRootFn,AobsFn,UobsFn,VobsFn,
             &PostFilter,ChanAsmRootFn,ChanUsmRootFn,ChanVsmRootFn);
  X->size = X->cols = X->rows = size;
  X->square = 1; 
  AllocFloatImage(Y,size);
  n = size * size;

 /*
  * Check for erroneous parameters  (incompatibility should be fixed some day)
  */
  if (PostFilter && ZeroPad) {
    printf("\n\n  >>>>  Post filtering is NOT compatible with Zero Padding!\n");
    exit(-1);
  } else {
    if (PostFilter && Reflect) {
      printf("\n\n  >>>>  Post filtering is NOT compatible with Reflection!\n");
      exit(-1);
    }
  }

 /*
  * Allocate working arrays
  */
  AllocArrays(argv,n,G,&a_valid,&u_valid,&v_valid,&A,&U,&V);

 /*
  * Report if auxilliary processing will be performed
  */
  if (PrintResponse) {
    printf("\n\nGenerating Files for Channel Responses...\n");
  }
  if (PrintChannels) {
    printf("\n\nGenerating files for A,U,V estimates from each channel...\n");
  }
  if (PostFilter) {
    printf("\n\nPost Filtering...\n");
  }

 /*
  * Process with the zeroth filter
  */
  if (ZeroPad) {
    looplim = size >> 1;
  } else {
    looplim = size;
  }
  printf("\nDoing Filter 0...");
  HardyConvolveGabor(Y,X,G->filt[0]);
  if (PrintResponse) {
    if (strcmp(RespRootFn,"none")) {
      sprintf(TempString,"%s%02dR.float",RespRootFn,0);
      f2disk(Y->Real,Y->size,Y->size,TempString);
      sprintf(TempString,"%s%02dI.float",RespRootFn,0);
      f2disk(Y->Imag,Y->size,Y->size,TempString);
    }
  }
 /*
  * Dominant ESA
  */
  dom_esa(Y,looplim,G->filt[0],&A[0],&U[0],&V[0],G,a_valid,u_valid,v_valid,
          PostFilter);
  if (PrintChannels) {
    if (strcmp(ChanARootFn,"none")) {
      sprintf(TempString,"%s%02d.float",ChanARootFn,0);
      f2disk(&A[0],Y->size,Y->size,TempString);
    }
    if (strcmp(ChanURootFn,"none")) {
      sprintf(TempString,"%s%02d.float",ChanURootFn,0);
      f2disk(&U[0],Y->size,Y->size,TempString);
    }
    if (strcmp(ChanVRootFn,"none")) {
      sprintf(TempString,"%s%02d.float",ChanVRootFn,0);
      f2disk(&V[0],Y->size,Y->size,TempString);
    }
  }
 /*
  * Post Filter if required
  */
  if (PostFilter) {
    if (G->filt[0]->isotropicPF) {
      g = MkPostFilt(G->filt[0]->sigma,G->filt[0]->kappa,size);
    } else {
      g = MkAnisPF(G->filt[0]->sigma,G->filt[0]->kappa_x,G->filt[0]->kappa_y,
                   G->filt[0]->Theta,size,1);
    }
    DoPostFilter(g,size,&U[0],&U[0],G,0);
    DoPostFilter(g,size,&V[0],&V[0],G,0);
    ReComputeAsm(Y,&U[0],&V[0],&A[0],G->filt[0]);
    DoPostFilter(g,size,&A[0],&A[0],G,0);
    g = KillFloat(g);
   /*
    * Print smoothed A,U,V estimates for this channel
    */
    if (PrintChannels) {
      if (strcmp(ChanAsmRootFn,"none")) {
        sprintf(TempString,"%s%02d.float",ChanAsmRootFn,0);
        f2disk(&A[0],Y->size,Y->size,TempString);
      }
      if (strcmp(ChanUsmRootFn,"none")) {
        sprintf(TempString,"%s%02d.float",ChanUsmRootFn,0);
        f2disk(&U[0],Y->size,Y->size,TempString);
      }
      if (strcmp(ChanVsmRootFn,"none")) {
        sprintf(TempString,"%s%02d.float",ChanVsmRootFn,0);
        f2disk(&V[0],Y->size,Y->size,TempString);
      }
    }
  }                   /* if (PostFilter) */

 /*
  * Loop for remaining filters
  */
  for (i=1; i < G->TotNumFilters; i++) {
    printf("\nDoing Filter %d...",i);
    HardyConvolveGabor(Y,X,G->filt[i]);
    if (PrintResponse) {
      if (strcmp(RespRootFn,"none")) {
        sprintf(TempString,"%s%02dR.float",RespRootFn,i);
        f2disk(Y->Real,Y->size,Y->size,TempString);
        sprintf(TempString,"%s%02dI.float",RespRootFn,i);
        f2disk(Y->Imag,Y->size,Y->size,TempString);
      }
    }
   /*
    * Dominant ESA
    */
    dom_esa(Y,looplim,G->filt[i],&A[i*n],&U[i*n],&V[i*n],G,
            a_valid,u_valid,v_valid,PostFilter);
    if (PrintChannels) {
      if (strcmp(ChanARootFn,"none")) {
        sprintf(TempString,"%s%02d.float",ChanARootFn,i);
        f2disk(&A[i],Y->size,Y->size,TempString);
      }
      if (strcmp(ChanURootFn,"none")) {
        sprintf(TempString,"%s%02d.float",ChanURootFn,i);
        f2disk(&U[i],Y->size,Y->size,TempString);
      }
      if (strcmp(ChanVRootFn,"none")) {
        sprintf(TempString,"%s%02d.float",ChanVRootFn,i);
        f2disk(&V[i],Y->size,Y->size,TempString);
      }
    }
   /*
    * Post Filter if required
    */
    if (PostFilter) {
      if (G->filt[i]->isotropicPF) {
        g = MkPostFilt(G->filt[i]->sigma,G->filt[i]->kappa,size);
      } else {
        g = MkAnisPF(G->filt[i]->sigma,G->filt[i]->kappa_x,G->filt[i]->kappa_y,
                     G->filt[i]->Theta,size,1);
      }
      DoPostFilter(g,size,&U[i*n],&U[i*n],G,i);
      DoPostFilter(g,size,&V[i*n],&V[i*n],G,i);
      ReComputeAsm(Y,&U[i*n],&V[i*n],&A[i*n],G->filt[i]);
      DoPostFilter(g,size,&A[i*n],&A[i*n],G,i);
      g = KillFloat(g);
     /*
      * Print smoothed A,U,V estimates for this channel
      */
      if (PrintChannels) {
        if (strcmp(ChanAsmRootFn,"none")) {
          sprintf(TempString,"%s%02d.float",ChanAsmRootFn,i);
          f2disk(&A[i*n],Y->size,Y->size,TempString);
        }
        if (strcmp(ChanUsmRootFn,"none")) {
          sprintf(TempString,"%s%02d.float",ChanUsmRootFn,i);
          f2disk(&U[i*n],Y->size,Y->size,TempString);
        }
        if (strcmp(ChanVsmRootFn,"none")) {
          sprintf(TempString,"%s%02d.float",ChanVsmRootFn,i);
          f2disk(&V[i*n],Y->size,Y->size,TempString);
        }
      }
    }
  } /* for i */

 /*
  * Free valid arrays
  */
  free(a_valid);
  free(u_valid);
  free(v_valid);

 /*
  * Write result files
  */
  writesize = G->TotNumFilters * looplim * sizeof(float);
  if ((OutVector=(float *)malloc(writesize))==NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  OutFn[0] = AobsFn;
  OutFn[1] = UobsFn;
  OutFn[2] = VobsFn;
  OutDat[0] = A;
  OutDat[1] = U;
  OutDat[2] = V;

 /* 
  * i loop is on the quantity being tracked : A, U, or V
  */
  for (i=0; i < 3; i++) {
   /*
    * open file 
    */
    if ((fd = open(OutFn[i], O_WRONLY | O_CREAT | O_TRUNC, 0644))==-1) {
      printf("%s: could not open output file %s!\n",argv[0],OutFn[i]);
      exit(-1);
    }
    for (row=0; row < looplim; row++) { /* output one row on each iteration */
      for (k=col=0; col < looplim; col++) { /* output one col each time */
         for (FILTER=0; FILTER < G->TotNumFilters; FILTER++,k++) {
           OutVector[k] = OutDat[i][FILTER*n + row*size + col];
         }
      }
     /*
      * Write out one rows worth of observations (all filters)
      */
      if (write(fd,OutVector,writesize) != writesize) {
        printf("\n%s: complete disk write did not succeed.\n",argv[0]);
      }
    }    /* for row */
   /*
    * Close file
    */
    if (close(fd) == -1) {
      printf("\n%s: error closing file %s.\n",argv[0],OutFn[i]);
    }
  }      /* for i */

  printf("\n\n");
  return;

} /*-------------------------------------------- main --------------*/

/*
 * Initialize
 *
 *  Process the paramters file
 *  Read the input image, if real make hardy to get the imaginary part
 *  Compute the input spectrum
 *  Construct the filter bank
 */

void Initialize(ParmFn,size,ZP,Reflect,xReal,xImag,XReal,XImag,G,PrintResponse,
                RespRootFn,PrintChannels,ChanARootFn,ChanURootFn,ChanVRootFn,
                AobsFn,UobsFn,VobsFn,PostFilter,ChanAsmRootFn,ChanUsmRootFn,
                ChanVsmRootFn)

  char        *ParmFn;           /* filename of parameters file */
  int         *size;             /* row/col dim of the square input image */
  int         *ZP;               /* 1 if zero padding is to be used */
  int         *Reflect;          /* 1=fill 0 padded areas w/ reflection */
  float      **xReal;            /* real part of input image */
  float      **xImag;            /* imag part of input image */
  float      **XReal;            /* real part of input spectrum */
  float      **XImag;            /* imag part of input spectrum */
  filterbank **G;                /* the filter bank */
  int         *PrintResponse;    /* 1 if we are to print the response images */
  char        *RespRootFn;       /* Root filename for response images */
  int         *PrintChannels;    /* 1=print A,U est from each channel */
  char        *ChanARootFn;      /* root fn for A est from each channel */
  char        *ChanURootFn;      /* root fn for U est from each channel */
  char        *ChanVRootFn;      /* root fn for V est from each channel */
  char        *AobsFn;           /* filename for A observations */
  char        *UobsFn;           /* filename for U observations */
  char        *VobsFn;           /* filename for V observations */
  int         *PostFilter;       /* 1=PostFiltering to be performed */
  char        *ChanAsmRootFn;    /* Filename for smoothed amplitude */
  char        *ChanUsmRootFn;    /* Filename for smoothed horiz frequency */
  char        *ChanVsmRootFn;    /* Filename for smoothed vert  frequency */
{

  int         i;                 /* counter */
  int         row;               /* counter for image rows */
  int         col;               /* counter for image cols */
  int         so2;               /* *size / 2 */
  char        datatype;          /* type of in dat: b=byte f=float d=double */
  int         Conjugate;         /* 1=conjugate input.  See Book X, p. 64 */
  char        Rfn[FNSIZE];       /* file for Real part of input image */
  char        Ifn[FNSIZE];       /* file for Imag part of input image */
  double      r0;                /* rad. center freq. of 1st filt on each ray */
  float       R;                 /* common ratio of radii of adjcnt filtrs */
  float       obw;               /* Eta-peak octave bandwidth of filters */
  float       Eta;               /* OBW is defined in terms of Eta-peak. */
  int         isoPF;             /* 1=isotropic post filters */
  float       kappa;             /* Post Filter space constant scale factor */
  float       kappa_x;           /* postfilter horiz space cnst scale factor */
  float       kappa_y;           /* postfilter vert  space cnst scale factor */
  float      *tfloat;            /* temporary float array */
  float      *to;                /* area in memory to copy floats to */
  float      *from;              /* area in memory to copy floats from */
  float      *pfloat;            /* ptr to location to fill w/ zero */


 /*
  * Read the parameter file
  */
  ReadParametersFile(ParmFn,&datatype,&Conjugate,Rfn,Ifn,size,ZP,Reflect,&r0,
                     &R,&Eta,&obw,PrintResponse,RespRootFn,PrintChannels,
                     ChanARootFn,ChanURootFn,ChanVRootFn,AobsFn,UobsFn,VobsFn,
                     PostFilter,&isoPF,&kappa,&kappa_x,&kappa_y,ChanAsmRootFn,
                     ChanUsmRootFn,ChanVsmRootFn);
  r0 = r0 / (double)(*size);
 
 /*
  * If zero padding, double size 
  */
  if (*ZP) {
    *size *= 2;
  }
  so2 = *size >> 1;

 /*
  * Make the filter bank
  */ 
  *G = CreateFilterBank(r0,R,obw,Eta,isoPF,kappa,kappa_x,kappa_y);

 /*
  * Allocate Spectra
  */
  if ((*XReal = (float *)malloc(*size**size*sizeof(float))) == NULL) {
    printf("\nInitialize: free store exhausted\n");
    exit(-1);
  }
  if ((*XImag = (float *)malloc(*size**size*sizeof(float))) == NULL) {
    printf("\nInitialize: free store exhausted\n");
    exit(-1);
  }

 /*
  * Get the input image and its spectrum
  */
  if (*ZP) {
    tfloat = readasf(so2,so2,Rfn,datatype);
    if ((*xReal = (float *)malloc(*size**size*sizeof(float))) == NULL) {
      printf("\nInitialize: free store exhausted\n");
      exit(-1);
    }
    for (to=*xReal,from=tfloat,row=0; row < so2; row++,to+=*size,from+=so2) {
      memcpy(to, from, so2*sizeof(float));
      for (pfloat=to+so2,col=so2; col < *size; col++) {
        if (!*Reflect) {
          *(pfloat++) = (float)0.0;
        } else {
          *(pfloat++) = *(to + *size-col-1);
        }
      }
    }
    free(tfloat);
    if (!*Reflect) {
      for (pfloat=to,i=so2**size; i < *size**size; i++) {
        *(pfloat++) = (float)0.0;
      } 
    } else {
      for (pfloat=to,row=1; row < *size; row+=2) {
        for (col=0; col < *size; col++,pfloat++) {
          *pfloat = *(pfloat - row**size);
        }
      }
    }
  }
  else {
    *xReal = readasf(*size,*size,Rfn,datatype);
  }

  if (strcmp(Ifn,"none")) {                     /* filename is not "none" */
    if (*ZP) {
      tfloat = readasf(so2,so2,Ifn,datatype);
      if ((*xImag = (float *)malloc(*size**size*sizeof(float))) == NULL) {
        printf("\nInitialize: free store exhausted\n");
        exit(-1);
      }
      for (to=*xImag,from=tfloat,row=0; row < so2; row++,to+=*size,from+=so2) {
        memcpy(to, from, so2*sizeof(float));
        for (pfloat=to+so2,col=so2; col < *size; col++) {
          if (!*Reflect) {
            *(pfloat++) = (float)0.0;
          } else {
            *(pfloat++) = *(to + *size-col-1);
          }
        }
      }
      free(tfloat);
      if (!*Reflect) {
        for (pfloat=to,i=so2**size; i < *size**size; i++) {
          *(pfloat++) = (float)0.0;
        }
      } else {
        for (pfloat=to,row=1; row < *size; row+=2) {
          for (col=0; col < *size; col++,pfloat++) {
            *pfloat = *(pfloat - row**size);
          }
        }
      }
    }
    else {
      *xImag = readasf(*size,*size,Ifn,datatype);
    }
    if (Conjugate) {
      for (i=0; i < *size**size; i++) {
        (*xImag)[i] *= (float)-1.0;
      }
    }
/*    f_u_norm(*xReal, *xImag, *size, *size);*/
    fft2d(*xReal,*xImag,*XReal,*XImag,*size,1);
  }
  else {                         /* input is real, must create imag part */
/*    f_u_norm(*xReal, NULL, *size, *size);*/
    if ((*xImag = (float *)malloc(*size**size*sizeof(float))) == NULL) {
      printf("\nInitialize: free store exhausted\n");
      exit(-1);
    }
    MakeHardy(*xReal,*xImag,*XReal,*XImag,*size);
    if (Conjugate) {
      printf("\n\n >>>  CONJUGATING input made by MakeHardy...\n");
      for (i=0; i < *size**size; i++) {
        (*xImag)[i] *= (float)-1.0;
      }
      fft2d(*xReal,*xImag,*XReal,*XImag,*size,1);
    }
  }

  return;
} /*---------------------------------------------- Initialize ----------*/

/*
 * ReadParametersFile
 *
 *  read the data out of the parameters file
 *
 */

void ReadParametersFile(ParmFn,pdatatype,pConjugate,Rfn,Ifn,psize,pZP,pReflect,
                        pr0,pR,pEta,pobw,pPrintResponse,RespRootFn,
                        pPrintChannels,ChanARootFn,ChanURootFn,ChanVRootFn,
                        AobsFn,UobsFn,VobsFn,pPostFilter,pisoPF,pkappa,pkappa_x,
                        pkappa_y,ChanAsmRootFn,ChanUsmRootFn,ChanVsmRootFn)
  char   *ParmFn;
  char   *pdatatype;
  int    *pConjugate;  
  char   *Rfn;
  char   *Ifn;
  int    *psize;
  int    *pZP;
  int    *pReflect;
  double *pr0;
  float  *pR;
  float  *pEta;
  float  *pobw;
  int    *pPrintResponse;
  char   *RespRootFn;
  int    *pPrintChannels;
  char   *ChanARootFn;
  char   *ChanURootFn;
  char   *ChanVRootFn;
  char   *AobsFn;
  char   *UobsFn;
  char   *VobsFn;
  int    *pPostFilter;
  int    *pisoPF;
  float  *pkappa;
  float  *pkappa_x;
  float  *pkappa_y;
  char   *ChanAsmRootFn;
  char   *ChanUsmRootFn;
  char   *ChanVsmRootFn;
{

  FILE      *parmfp;            /* ptr to parameter file */
  char       instr[256];        /* buffer to read into  */

  if ((parmfp = fopen(ParmFn,"r")) == NULL) {
    printf("\n\nReadParametersFile: unable to open %s\n",ParmFn);
    exit(-1);
  }

  fscanf(parmfp,"Input Data Type= %s\n",instr);
  *pdatatype = instr[0];
  fscanf(parmfp,"Conjugate Input= %s\n",instr);
  if ((instr[0] == 'N') || (instr[0] == 'n')) {
    *pConjugate = 0;
  }
  else {
    *pConjugate = 1;
  }
  fscanf(parmfp,"Real Image File= %s\n",Rfn);
  fscanf(parmfp,"Imag Image File= %s\n",Ifn);
  fscanf(parmfp,"Size= %d\n",psize);
  fscanf(parmfp,"Zero Pad= %s\n",instr);
  if ((instr[0] == 'Y') || (instr[0] == 'y')) {
    *pZP = 1;
  }
  else {
    *pZP = 0;
  }
  fscanf(parmfp,"Reflect= %s\n",instr);
  if ((instr[0] == 'Y') || (instr[0] == 'y')) {
    *pReflect = 1;
  }
  else {
    *pReflect = 0;
  }
  fscanf(parmfp,"r0= %lf\n",pr0);
  fscanf(parmfp,"R= %f\n",pR);
  fscanf(parmfp,"Eta= %f\n",pEta);
  fscanf(parmfp,"OBW= %f\n",pobw);
  fscanf(parmfp,"Print Response Images= %s\n",instr);
  if ((instr[0] == 'N') || (instr[0] == 'n')) {
    *pPrintResponse = 0;
  }
  else {
    *pPrintResponse = 1;
  }
  fscanf(parmfp,"Response Root File Name= %s\n",RespRootFn);
  fscanf(parmfp,"Print Channel Estimates= %s\n",instr);
  if ((instr[0] == 'N') || (instr[0] == 'n')) {
    *pPrintChannels = 0;
  }
  else {
    *pPrintChannels = 1;
  }
  fscanf(parmfp,"Chan Est A Root File Name= %s\n",ChanARootFn);
  fscanf(parmfp,"Chan Est U Root File Name= %s\n",ChanURootFn);
  fscanf(parmfp,"Chan Est V Root File Name= %s\n",ChanVRootFn);

  fscanf(parmfp,"Postfilter= %s\n",instr);
  if ((instr[0] == 'N') || (instr[0] == 'n')) {
    *pPostFilter = 0;
  }
  else {
    *pPostFilter = 1;
  }
  fscanf(parmfp,"Isotropic PF= %s\n",instr);
  if ((instr[0] == 'N') || (instr[0] == 'n')) {
    *pisoPF = 0;
  }
  else {
    *pisoPF = 1;
  }
  fscanf(parmfp,"kappa= %f\n",pkappa);
  fscanf(parmfp,"kappa_x= %f\n",pkappa_x);
  fscanf(parmfp,"kappa_y= %f\n",pkappa_y);
  fscanf(parmfp,"Chan Asm Root Fn= %s\n",ChanAsmRootFn);
  fscanf(parmfp,"Chan Usm Root Fn= %s\n",ChanUsmRootFn);
  fscanf(parmfp,"Chan Vsm Root Fn= %s\n",ChanVsmRootFn);
  fscanf(parmfp,"A Obs File Name= %s\n",AobsFn);
  fscanf(parmfp,"U Obs File Name= %s\n",UobsFn);
  fscanf(parmfp,"V Obs File Name= %s\n",VobsFn);
  fclose(parmfp);
  return;
} /*--------------------------- ReadParametersFile -------------------------*/

/*
 * dom_esa
 *
 *  function evaluates the esa at every point in an image and checks to see if
 *  if the current response image has the dominant response of those yet seen.
 *
 */

void dom_esa(Y,looplim,G,a,u,v,Bank,a_valid,u_valid,v_valid,PostFilter)

  pFloatImage     Y;              /* response image */
  int             looplim;        /* terminal value for loop counters */
  FilterPtr       G;              /* Y was output from this filter */
  float          *a;              /* a estimates */
  float          *u;              /* u estimates */
  float          *v;              /* v estimates */
  filterbank     *Bank;           /* the filterbank */
  BYTE           *a_valid;        /* 1 where the a estimate is valid */
  BYTE           *u_valid;        /* 1 where the u estimate is valid */
  BYTE           *v_valid;        /* 1 where the v estimate is valid */
  int             PostFilter;     /* 1 if postfiltering is to be done */
{

  int             row,col;        /* loop counters for image coordinates */
  int             k;              /* loop counter for indexing arrays */
  float           criterion;      /* filter selection criterion */

 /*
  * Initialize a_valid, u_valid, and v_valid
  */
  for (k=0; k < Y->size*Y->size; k++) {
    a_valid[k] = u_valid[k] = v_valid[k] = 0;
  }

 /*
  * Apply the ESA at each point in the image
  */
  for (row=0; row < looplim; row++) {
    for (col=0; col < looplim; col++) {
      k = row * Y->size + col;
      criterion = (float) -1.0;
      PtEsa2D(Y,row,col,G,&criterion,&u[k],&v[k],&a[k],&u_valid[k],&v_valid[k],
              &a_valid[k]);
    }
  }

 /*
  * Return NO_OBSERVATION for invalid estimates
  */
  if (!PostFilter) {
    for (k=row=0; row < looplim; row++) {
      for (col=0; col < looplim; col++,k++) {
        if (!u_valid[k]) {
          u[k] = NO_OBSERVATION;
        }
        if (!v_valid[k]) {
          v[k] = NO_OBSERVATION;
        }
        if (!a_valid[k]) {
          a[k] = NO_OBSERVATION;
        }
      }    /* for col */
    }
  }

  return;
} /*----------------------------------------- dom_esa -----------------------*/

/*
 * AllocArrays
 *
 *  Allocate Working Arrays.
 */
 
void AllocArrays(argv,n,G,a_valid,u_valid,v_valid,A,U,V)
 
  char              *argv[];         /* argument list from command line */
  int                n;              /* size * size again */
  filterbank        *G;              /* the filter bank */
  BYTE             **a_valid;        /* 1 where the a estimate is valid */
  BYTE             **u_valid;        /* 1 where the u estimate is valid */
  BYTE             **v_valid;        /* 1 where the v estimate is valid */
  float            **A;              /* A estimates accross all filters */
  float            **U;              /* U estimates accross all filters */
  float            **V;              /* V estimates accross all filters */
{
 
 /*
  * Allocate working arrays
  */
  if ((*a_valid = (BYTE *) malloc(n * sizeof(BYTE))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*u_valid = (BYTE *) malloc(n * sizeof(BYTE))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*v_valid = (BYTE *) malloc(n * sizeof(BYTE))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*A = (float *)malloc(G->TotNumFilters * n * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*U = (float *)malloc(G->TotNumFilters * n * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((*V = (float *)malloc(G->TotNumFilters * n * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
 
  return;
} /*---------------------  AllocArrays  --------------------------------------*/
#endif
