/*
 *
 * NFRecon.c
 *
 *  Reconstruct real image from A,U,V estimates obtained by demodulating
 *  image directly, without filtering.
 *
 *  ASSUMES that frequency estimates are in cycles/pixel...  also see 
 *    NFReconTK.c which assumes they are in radians!
 *  
 *
 *  jph 7/13/96
 */

#ifndef NFRECON_C
#define NFRECON_C

#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include <string.h>

#include "teager2D.h"


main(argc,argv)

  int   argc;
  char *argv[];

{

  int                size;           /* row/col dim of the square input image */
  float             *x;              /* input image */
  float             *y;              /* reconstructed image */
  char              *Rfn;            /* Real image input file */
  char              *Afn;            /* File name for A estimates */
  char              *Ufn;            /* File name for U estimates */
  char              *Vfn;            /* File name for V estimates */
  char              *Outfn;          /* File name for reconstruction */
  int                ReconInterval;  /* boundary condition interval for recon */
  float             *A;              /* A estimates */
  float             *U;              /* U estimates */
  float             *V;              /* V estimates */
  int                row;            /* loop counter for image rows */
  int                col;            /* loop counter for image cols */
  int                ROW;            /* row counter for recon blocks */
  int                COL;            /* col counter for recon blocks */
  int                R,C;            /* counters for rows & cols within block */
  int                k;              /* counter */
  double            *phase;          /* instantaneous phase */

 /*
  * Check for right number of arguments
  */
  if (argc != 8) {
    printf("\n\n%s: Reconstruct Real image from unfiltered A,U,V estimates.",
           argv[0]);
    printf("\nusage: %s size RealImagefn Afn Ufn Vfn Interval Outfn\n",argv[0]);
    exit(-1);
  }
#ifdef DEBUG
 /*
  * Link to PA so we can use it in the debugger 
  */
  if (!argc) {
    PA((void *)x, 0, -1, 'b');
    PRINTA((void *)x,0,-1,'b',"delete.me");
    WRITEA((void *)x,0,-1,'b',"delete.me");
    image2disk((void *)x,1,1,'*',"delete.me");
    SpecMag2disk(x,"delete.me");
    PC(x,0,0,-1,'b',0);
    PRINTC(x,0,0,-1,'b',0,"delete.me");
    WRITEC(x,0,0,-1,'b',0,"delete.me");
  }
#endif

  size = atoi(argv[1]);
  Rfn = argv[2];
  Afn = argv[3];
  Ufn = argv[4];
  Vfn = argv[5];
  ReconInterval = atoi(argv[6]);
  Outfn = argv[7];

  if ((x = (float *)malloc(size*size*sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((y = (float *)malloc(size*size*sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((A = (float *)malloc(size*size*sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((U = (float *)malloc(size*size*sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((V = (float *)malloc(size*size*sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((phase = (double *)malloc(size*size*sizeof(double))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }

  disk2f(x,size,size,Rfn);
  disk2f(A,size,size,Afn);
  disk2f(U,size,size,Ufn);
  disk2f(V,size,size,Vfn);

 /*
  * Outer loops set top left corner of each reconsructed block
  */
  for (ROW=0; ROW < size; ROW+=ReconInterval) {
    for (COL=0; COL < size; COL+=ReconInterval) {
     /*
      * Reconstruct phase on first row of block
      */
      k = ROW*size + COL;
      if (A[k] <= (float)0.0000001) {
        phase[k] = (float)0.0;
      } else {
        phase[k] = acos((double)x[k]/(double)A[k]);
     }
      if (COL != 0) {
        if (phase[k-1] < (double)0.0) {
          phase[k] *= (double)-1.0;
        }
      } else {
        if (ROW != 0) {
          if (phase[k-row] < (double)0.0) {
            phase[k] *= (double)-1.0;
          }
        }
      }
      for (k++,C=1,col=COL+1; (col < size) && (C < ReconInterval); 
           col++,C++,k++) {
        phase[k] = phase[k-1] + (double)2.0 * (double)M_PI * (double)U[k-1];
      }
     /*
      * Reconstruct phase on first col of block
      */
      for (k=(ROW+1)*size+COL,R=1,row=ROW+1; 
           (row < size) && (R < ReconInterval);
           row++,R++,k+=size) {
        phase[k] = phase[k-size] + (double)2.0 * (double)M_PI 
                                               * (double)V[k-size];
      }
     /*
      * Reconstruct phase at interior points in the block
      */
      for (R=1,row=ROW+1; (row < size) && (R < ReconInterval); R++,row++) {
        for (C=1,col=COL+1; (col < size) && (C < ReconInterval); C++,col++) {
          k = row*size + col;
          phase[k] = (double)0.5 * (phase[k-1] + phase[k-size] 
                     + (double)2.0 * (double)M_PI 
                       * ((double)U[k-1] + (double)V[k-size]));
        }
      }
    }    /* COL */
  }      /* ROW */

 /*
  * Reconstruct image values
  */
  for (k=row=0; row < size; row++) {
    for (col=0; col < size; col++,k++) {
      y[k] = A[k] * (float)cos(phase[k]);
    }
  }
  free(phase);

  f2disk(y,size,size,Outfn);

  printf("\n\n");
  return;
 
} /*-------------------------------------------- main --------------*/
#endif
