/*
 * ProcTrack.c
 *
 *   Contains function track, which will
 *   Process existing tracks against the observations, then start new
 *   new tracks if there are significant observations left that could
 *   not be associated with an existing track.
 *
 *   MAJOR REVISION   8-17-94.  Book VIII, p. 88, 89.
 *   MAJOR REVISIONS  3-12-96.
 *   4-26-96: added freq scaling to track start criterion... see book XII p. 3.
 *   9-9-96: MAJOR REVISIONS.
 *
 * 4/25/94 jph
 */

#ifndef PROCTRACK_C
#define PROCTRACK_C

#include "track.h"
 
/*-----------------------------------------------------------------------------
   TRACK
-----------------------------------------------------------------------------*/
void Track(jt,row,col,delta,minCol2Track,MaxCol2Track,T,ChannelQ,G,Aob,Uob,Vob,
           ChannelStatus,Qorder,ChanQcopy,size,LastStart,Alast,Ulast,Vlast,KA,
           KU,KV,minRow2Track,MaxRow2Track,LimitRays,fScale,TEScale,RptFp)

  int          jt;
  int          row;             /* Image Row of current observations */
  int          col;             /* Image Col of current observations */
  int          delta;           /* -1 if proc'ing col R to L, 1 otherwise */
  int          minCol2Track;    /* first col to perform track proc. on */
  int          MaxCol2Track;    /* last  col to perform track proc. on */
  trackP       T;               /* tracks */
  float       *ChannelQ;        /* channel quality */
  filterbank  *G;               /* the filter bank */
  float       *Aob;             /* row of observations of A */
  float       *Uob;             /* row of observations of u */
  float       *Vob;             /* row of observations of v */
  int         *ChannelStatus;   /* track assoc status of channel */
  int         *Qorder;          /* no.s of channels sorted by Q */
  float       *ChanQcopy;       /* copy of channel quality for sorting */
  int          size;            /* row/col dimension of images */
  image_point *LastStart;       /* for each chan, pix of last trackstart*/
  float       *Alast;           /* last row of observations of A */
  float       *Ulast;           /* last row of observations of u */
  float       *Vlast;           /* last row of observations of v */
  pABgain      KA;              /* gains for A tracker */
  pABgain      KU;              /* gains for U tracker */
  pABgain      KV;              /* gains for V tracker */
  int          minRow2Track;    /* first row to perform track proc. on */
  int          MaxRow2Track;    /* last row to perform track proc. on */
  int          LimitRays;       /* 1=limit num rays visited by a track */
  int          fScale;          /* 1=scale track start criterion by freq */
  int          TEScale;         /* 1=scale trkstrt crit by Teager Energy */
  FILE        *RptFp;           /* ptr to stream for track reporting */
{
 
  int     i;                    /* loop counter */
  int     j;                    /* counter */
  int     k;                    /* counter */
  int     Upd;                  /* active track counter */
  int     m;                    /* observations index */
  int     FirstPixOnRow;        /* TRUE if this is the first pixel on a row */
  int     LastPix;              /* index of last tracked pixel */
  int     Ntracks;              /* current number of tracks */
  int     Nactive;              /* current number of active tracks */
  int     Ndormant;             /* current number of dormant tracks */
  float   FloatTemp;            /* temporary float value */
  int     IntTemp;              /* temporary int value */
  int     CurSizeIndx;          /* element of Qorder currntly being processed */
  int     CurChan;              /* no. of channel currently being processed */
  int     CurTrk;               /* loop counter for tracks */
  int     TrksProcQ13;          /* Number of Tracks updated in Quads I,III */
  int     TrksProcQ24;          /* Number of Tracks updated in Quads II,IV */
  int     UpdChan;              /* channel to update track from */
  int     CandChan;             /* candidate channel for track updating */
  int     warp;                 /* 1 if chan is upd across 1st & last rays */
  int     warp2;                /* another version of warp */
  int     UpdOrder[MAXNTRACK];  /* update order for tracks */
  int     TrkInserted;          /* 1 if CurTrk inserted into UpdOrder list */
  float   UpdQual[MAXNTRACK];   /* quality metric for existing tracks */
  int     InBW[MAXNTRACK];      /* InBW[i]=1 if track UpdOrder[i] in chan BW */
  int     TrkInBW;              /* 1 if CurTrk is in channel BW */
  double  Dist;                 /* distance from pred u,v to filter center */
  double  MaxQual;              /* Quality of association btwn track and obs */
  double  CanQual;              /* Quality of association btwn track and chan */
  float   TQual;                /* Quality of last update of a track */

 /*
  * Report
  */
  if (col == minCol2Track) {
    for (i=Ntracks=Nactive=Ndormant=0; i < MAXNTRACK; i++) {
      if (T[i].in_use) {
        Ntracks++;
        if (T[i].active) Nactive++;
        if (T[i].dormant) Ndormant++;
      } 
    }
    fprintf(RptFp,"\nTracking row %d: %d tracks, %d active, %d dormant.",
           row,Ntracks,Nactive,Ndormant);
  }

 /*
  * Take Care of Housekeeping if first column tracked in row
  */
  FirstPixOnRow = 0;
  if (((delta ==  1) && (col == minCol2Track)) ||
      ((delta == -1) && (col == MaxCol2Track))) {
    FirstPixOnRow = 1;
    for (k=0; k < MAXNTRACK; k++) {
      T[k].AveQualRow = 0;
      T[k].NgoodUpdRow = 0;
      if (T[k].dormant) T[k].dormantOnRow = 1;
    }
    if (!(row%4)) {
      if (!(col == minCol2Track)) {
        for (i=Ntracks=Nactive=Ndormant=0; i < MAXNTRACK; i++) {
          if (T[i].in_use) {
            Ntracks++;
            if (T[i].active) Nactive++;
            if (T[i].dormant) Ndormant++;
          }
        }
      }
      printf("\n>>>>>  Tracking row %d: %d tracks, %d active.",
              row,Ntracks,Nactive);
    }
  }

 /*
  * Set processed field to show that tracks have not been updated
  */
  TrksProcQ13 = 0;
  TrksProcQ24 = 0;
  for (i=0; i < MAXNTRACK; i++) {
    T[i].updated = 0;
  }
  
 /*
  * Initialize channel quality
  */
  for (j=col*G->TotNumFilters,i=0; i < G->TotNumFilters; i++,j++) {
    if ((Aob[j] == NO_OBSERVATION) || (Uob[j] == NO_OBSERVATION) 
                                   || (Vob[j] == NO_OBSERVATION)) {
      ChannelQ[i] = (float)0.0;
    } else {
      ChannelQ[i] = Aob[j]
                     * G_EvalMag(G->filt[i],fabs((double)Uob[j]),(double)Vob[j])
                     / G->filt[i]->GMax;
    }
    if (fScale) {
      ChannelQ[i] *= fNorm(Uob[j],Vob[j]);
    } else {
      if (TEScale) {
        ChannelQ[i] *= Aob[j] * fNorm(Uob[j],Vob[j]) * fNorm(Uob[j],Vob[j]);
      }
    }
  }
 /*
  * Initialize channel status
  */
  for (i=0; i < G->TotNumFilters; i++) {
    ChannelStatus[i] = -4;
  }
 /*
  * Order the channels by quality: insertion sort
  */
  for (i=0; i < G->TotNumFilters; i++) {
    Qorder[i] = i;
    ChanQcopy[i] = ChannelQ[i];
  }
  for (j=1; j < G->TotNumFilters; j++) {
    FloatTemp = ChanQcopy[j];
    IntTemp   = Qorder[j];
    i = j - 1;
    while (i >= 0 && ChanQcopy[i] < FloatTemp) {
      ChanQcopy[i+1] = ChanQcopy[i];
      Qorder[i+1]    = Qorder[i];
      i--;
    }
    ChanQcopy[i+1] = FloatTemp;
    Qorder[i+1]    = IntTemp;
  }

 /*
  * Order tracks by size of tracked A, those in Eta-peak BW first. Count tracks.
  */
  if (FirstPixOnRow) {
    LastPix = (row-1)*size + col;
  } else {
    if (delta==1) {
      LastPix = row*size + col - 1;
    } else {
      LastPix = row*size + col + 1;
    }
  }
  for (i=0; i < MAXNTRACK; i++) {
    InBW[i] = UpdOrder[i] = -1;
  }

  for (Nactive=i=0; i < MAXNTRACK; i++) {
    if (T[i].active) {
      if (FirstPixOnRow) {
        CopyUpdRec(T[i].HorUpd,T[i].VertUpd[col]);
      }
      TQual = G_EvalMag(G->filt[T[i].HorUpd->LastChan],
                       (double)(T[i].U_filt[LastPix]),
                       (double)(T[i].V_filt[LastPix]))
               / G->filt[T[i].HorUpd->LastChan]->GMax;
      if (TQual >= G->Eta) {    /* if tracked freq in Eta-peak BW */
        TrkInBW = 1;
      } else {
        TrkInBW = 0;
      }
      TQual *= T[i].A_filt[LastPix];
      if (!Nactive) {
        UpdOrder[0] = i;
        InBW[0] = TrkInBW;
        UpdQual[0] = TQual;
      } else {
        for (TrkInserted=CurTrk=0; (CurTrk < Nactive) && (!TrkInserted); 
             CurTrk++) {
          if ((TrkInBW && (!InBW[CurTrk])) 
             || (TrkInBW && InBW[CurTrk] && (TQual > UpdQual[CurTrk])) 
             || ((!TrkInBW) && (!InBW[CurTrk]) && (TQual > UpdQual[CurTrk]))) {
           /*
            * Insert track i in UpdOrder here
            */
            for (j=Nactive; j > CurTrk; j--) {
              UpdOrder[j] = UpdOrder[j-1];
              InBW[j] = InBW[j-1];
              UpdQual[j] = UpdQual[j-1];
            }
            UpdOrder[CurTrk] = i;
            InBW[CurTrk] = TrkInBW;
            UpdQual[CurTrk] = TQual;
            TrkInserted = 1;
          }  /* if (insert track here) */
        }    /* for CurTrk */
       /*
        * If track not inserted in UpdOrder, put it at the end
        */
        if (!TrkInserted) {
          UpdOrder[Nactive] = i;
          InBW[Nactive] = TrkInBW;
          UpdQual[Nactive] = TQual;
        }
      }     /* else of if (!Nactive) */
      Nactive++;
    }       /* if (T[i].active) */
  }         /* for i */

 /*
  * Update tracks in order according to UpdOrder
  */
  for (Upd=0; Upd < Nactive; Upd++) {
    CurTrk = UpdOrder[Upd];
    warp  = 0;
    warp2 = 0;

   /*
    * First candidate channel for update is last channel used
    */
    UpdChan = T[CurTrk].HorUpd->LastChan;
    T[CurTrk].Upd = T[CurTrk].HorUpd;
    m = col*G->TotNumFilters + UpdChan;
    MaxQual = FindRadialQual(G->filt[UpdChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
   /*
    * If UpdChan already used, set MaxQual to a value smaller than it can be
    */
    if (ChannelStatus[UpdChan] != -4) {
      MaxQual = (double)-1.0;
    }

   /*
    * Check if previous channel on ray is a viable candidate for update
    */
    if ((G->filt[T[CurTrk].HorUpd->LastChan]->IndexOnRay)) {
      CandChan = T[CurTrk].HorUpd->LastChan - 1;
      m = col*G->TotNumFilters + CandChan;
      if (ChannelStatus[CandChan] == -4) {
     /*
      * Check if viable candidate for Horizontal update
      */
        if ((double)fNorm(T[CurTrk].HorUpd->pred.U,T[CurTrk].HorUpd->pred.V) 
              < G->filt[T[CurTrk].HorUpd->LastChan]->r0) {
          CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
          if (CanQual > MaxQual) {
            T[CurTrk].Upd = T[CurTrk].HorUpd;
            UpdChan = CandChan;
            MaxQual = CanQual;
          }
        }
       /*
        * Check if viable candidate for Vertical update
        */
        if (T[CurTrk].VertUpd[col]->LastChan != -1) {
          if ((double)fNorm(T[CurTrk].VertUpd[col]->pred.U,
                            T[CurTrk].VertUpd[col]->pred.V)
                < G->filt[T[CurTrk].HorUpd->LastChan]->r0) {
            CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
            if (CanQual > MaxQual) {
              T[CurTrk].Upd = T[CurTrk].VertUpd[col];
              UpdChan = CandChan;
              MaxQual = CanQual;
            }
          } 
        }   /* if LastChan != -1 */
      }     /* if ChannelStatus == -4 */
    }       /* if IndexOnRay */

   /*
    * Check if next channel on ray is a viable candidate for update
    */
    if (G->filt[T[CurTrk].HorUpd->LastChan]->IndexOnRay 
          < G->NumFiltersPerRay - 1) {
      CandChan = T[CurTrk].HorUpd->LastChan + 1;
      m = col*G->TotNumFilters + CandChan;
      if (ChannelStatus[CandChan] == -4) {
       /*
        * Check if viable candidate for Horizontal update
        */
        if ((double)fNorm(T[CurTrk].HorUpd->pred.U,T[CurTrk].HorUpd->pred.V) 
              > G->filt[T[CurTrk].HorUpd->LastChan]->r0) {
          CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
          if (CanQual > MaxQual) {
            T[CurTrk].Upd = T[CurTrk].HorUpd;
            UpdChan = CandChan;
            MaxQual = CanQual;
          }
        }
       /*
        * Check if viable candidate for Vertical update
        */
        if (T[CurTrk].VertUpd[col]->LastChan != -1) {
          if ((double)fNorm(T[CurTrk].VertUpd[col]->pred.U,
                            T[CurTrk].VertUpd[col]->pred.V)
                > G->filt[T[CurTrk].HorUpd->LastChan]->r0) {
            CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
            if (CanQual > MaxQual) {
              T[CurTrk].Upd = T[CurTrk].VertUpd[col];
              UpdChan = CandChan;
              MaxQual = CanQual;
            }
          }
        }   /* if LastChan != -1 */
      }     /* if ChannelStatus == -4 */
    }       /* if IndexONRay < NumFiltersPerRay - 1 */

   /*
    * Check if channels on previous ray are viable candidates for update
    */
    if (G->filt[T[CurTrk].HorUpd->LastChan]->Ray) {
     /*
      * Same channel on previous ray
      */
      CandChan = T[CurTrk].HorUpd->LastChan - G->NumFiltersPerRay;
      m = col*G->TotNumFilters + CandChan;
      if (ChannelStatus[CandChan] == -4) {
       /*
        * Check if viable candidate for Horizontal update
        */
        if (atan2((double)T[CurTrk].HorUpd->pred.V,
                  (double)T[CurTrk].HorUpd->pred.U) 
            < G->filt[T[CurTrk].HorUpd->LastChan]->Theta) {
          CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
          if (CanQual > MaxQual) {
            T[CurTrk].Upd = T[CurTrk].HorUpd;
            UpdChan = CandChan;
            MaxQual = CanQual;
          }
        }   /* if atan2(pred.V,pred.U) < Theta */
       /*
        * Check if viable candidate for Vertical update
        */
        if (T[CurTrk].VertUpd[col]->LastChan != -1) {
          if (atan2((double)T[CurTrk].VertUpd[col]->pred.V,
                    (double)T[CurTrk].VertUpd[col]->pred.U) 
              < G->filt[T[CurTrk].HorUpd->LastChan]->Theta) {
            CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
            if (CanQual > MaxQual) {
              T[CurTrk].Upd = T[CurTrk].VertUpd[col];
              UpdChan = CandChan;
              MaxQual = CanQual;
            }
          } /* if atan2(pred.V,pred.U) < Theta */
        }   /* if LastChan != -1 */
      }     /* if ChannelStatus == -4 */

     /*
      * Previous channel on previous ray
      */
      if ((G->filt[T[CurTrk].HorUpd->LastChan]->IndexOnRay)) {
        CandChan = T[CurTrk].HorUpd->LastChan - G->NumFiltersPerRay - 1;
        m = col*G->TotNumFilters + CandChan;
        if (ChannelStatus[CandChan] == -4) {
         /*
          * Check if viable candidate for Horizontal update
          */
          if ((atan2((double)T[CurTrk].HorUpd->pred.V,
                    (double)T[CurTrk].HorUpd->pred.U) 
                  < G->filt[T[CurTrk].HorUpd->LastChan]->Theta) 
            && ((double)fNorm(T[CurTrk].HorUpd->pred.U,T[CurTrk].HorUpd->pred.V)
                < G->filt[T[CurTrk].HorUpd->LastChan]->r0)) {
            CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
            if (CanQual > MaxQual) {
              T[CurTrk].Upd = T[CurTrk].HorUpd;
              UpdChan = CandChan;
              MaxQual = CanQual;
            }
          }   /* if atan2(pred.V,pred.U) < Theta && fNorm(pred.U,pred.V) < r0 */
         /*
          * Check if viable candidate for Vertical update
          */
          if (T[CurTrk].VertUpd[col]->LastChan != -1) {
            if ((atan2((double)T[CurTrk].VertUpd[col]->pred.V,
                    (double)T[CurTrk].VertUpd[col]->pred.U) 
                  < G->filt[T[CurTrk].HorUpd->LastChan]->Theta) 
              && ((double)fNorm(T[CurTrk].VertUpd[col]->pred.U,
                                T[CurTrk].VertUpd[col]->pred.V)
                < G->filt[T[CurTrk].HorUpd->LastChan]->r0)) {
              CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
              if (CanQual > MaxQual) {
                T[CurTrk].Upd = T[CurTrk].VertUpd[col];
                UpdChan = CandChan;
                MaxQual = CanQual;
              }
            } /* if atan2(pred.V,pred.U) < Theta && fNorm(pred.U,pred.V) < r0 */
          }   /* if LastChan != -1 */
        }     /* if ChannelStatus == -4 */
      }       /* if IndexOnRay */
     /*
      * Next channel on previous ray
      */
      if (G->filt[T[CurTrk].HorUpd->LastChan]->IndexOnRay 
            < G->NumFiltersPerRay - 1) {
        CandChan = T[CurTrk].HorUpd->LastChan - G->NumFiltersPerRay + 1;
        m = col*G->TotNumFilters + CandChan;
        if (ChannelStatus[CandChan] == -4) {
         /*
          * Check if viable candidate for Horizontal update
          */
          if ((atan2((double)T[CurTrk].HorUpd->pred.V,
                  (double)T[CurTrk].HorUpd->pred.U) 
                < G->filt[T[CurTrk].HorUpd->LastChan]->Theta) 
            && ((double)fNorm(T[CurTrk].HorUpd->pred.U,T[CurTrk].HorUpd->pred.V)
              > G->filt[T[CurTrk].HorUpd->LastChan]->r0)) {
            CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
            if (CanQual > MaxQual) {
              T[CurTrk].Upd = T[CurTrk].HorUpd;
              UpdChan = CandChan;
              MaxQual = CanQual;
            }
          }   /* if atan2(pred.V,pred.U) < Theta && fNorm(pred.U,pred.V) > r0 */
         /*
          * Check if viable candidate for Vertical update
          */
          if (T[CurTrk].VertUpd[col]->LastChan != -1) {
            if ((atan2((double)T[CurTrk].VertUpd[col]->pred.V,
                    (double)T[CurTrk].VertUpd[col]->pred.U) 
                  < G->filt[T[CurTrk].HorUpd->LastChan]->Theta) 
              && ((double)fNorm(T[CurTrk].VertUpd[col]->pred.U,
                                T[CurTrk].VertUpd[col]->pred.V)
                > G->filt[T[CurTrk].HorUpd->LastChan]->r0)) {
              CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
              if (CanQual > MaxQual) {
                T[CurTrk].Upd = T[CurTrk].VertUpd[col];
                UpdChan = CandChan;
                MaxQual = CanQual;
              }
            } /* if atan2(pred.V,pred.U) < Theta && fNorm(pred.U,pred.V) > r0 */
          }   /* if LastChan != -1 */
        }     /* if ChannelStatus == -4 */
      }       /* if IndexOnRay < NumFilterPerRay - 1*/
    }         /* if Ray */
    else {    /* previous ray is the last ray */
     /*
      * Same channel on last ray
      */
      CandChan = G->TotNumFilters - G->NumFiltersPerRay 
                 + G->filt[T[CurTrk].HorUpd->LastChan]->IndexOnRay;
      m = col*G->TotNumFilters + CandChan;
      if (ChannelStatus[CandChan] == -4) {
       /*
        * Check if viable candidate for Horizontal update
        */
        if ((double)T[CurTrk].HorUpd->pred.U 
            < G->filt[T[CurTrk].HorUpd->LastChan]->u0) {
          CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
          if (CanQual > MaxQual) {
            T[CurTrk].Upd = T[CurTrk].HorUpd;
            UpdChan = CandChan;
            MaxQual = CanQual;
          }
        }   /* if pred.U < G->u0 */
       /*
        * Check if viable candidate for Vertical update
        */
        if (T[CurTrk].VertUpd[col]->LastChan != -1) {
          if ((double)T[CurTrk].VertUpd[col]->pred.U
              < G->filt[T[CurTrk].HorUpd->LastChan]->u0) {
            CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
            if (CanQual > MaxQual) {
              T[CurTrk].Upd = T[CurTrk].VertUpd[col];
              UpdChan = CandChan;
              MaxQual = CanQual;
            }
          } /* if pred.U < G->u0 */
        }   /* if LastChan != -1 */
      }     /* if ChannelStatus == -4 */

     /*
      * Previous channel on last ray
      */
      if ((G->filt[T[CurTrk].HorUpd->LastChan]->IndexOnRay)) {
        CandChan = G->TotNumFilters - G->NumFiltersPerRay
                 + G->filt[T[CurTrk].HorUpd->LastChan]->IndexOnRay - 1;
        m = col*G->TotNumFilters + CandChan;
        if (ChannelStatus[CandChan] == -4) {
         /*
          * Check if viable candidate for Horizontal update
          */
          if (((double)T[CurTrk].HorUpd->pred.U
              < G->filt[T[CurTrk].HorUpd->LastChan]->u0)
            && ((double)fNorm(T[CurTrk].HorUpd->pred.U,T[CurTrk].HorUpd->pred.V)
              < G->filt[T[CurTrk].HorUpd->LastChan]->r0)) {
            CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
            if (CanQual > MaxQual) {
              T[CurTrk].Upd = T[CurTrk].HorUpd;
              UpdChan = CandChan;
              MaxQual = CanQual;
            }
          }   /* if pred.U < G->u0 && fNorm(pred.U,pred.V) < r0 */
         /*
          * Check if viable candidate for Vertical update
          */
          if (T[CurTrk].VertUpd[col]->LastChan != -1) {
            if (((double)T[CurTrk].VertUpd[col]->pred.U
                < G->filt[T[CurTrk].VertUpd[col]->LastChan]->u0)
              && ((double)fNorm(T[CurTrk].VertUpd[col]->pred.U,
                                T[CurTrk].VertUpd[col]->pred.V)
                < G->filt[T[CurTrk].HorUpd->LastChan]->r0)) {
              CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
              if (CanQual > MaxQual) {
                T[CurTrk].Upd = T[CurTrk].VertUpd[col];
                UpdChan = CandChan;
                MaxQual = CanQual;
              }
            } /* if pred.U < G->u0 && fNorm(pred.U,pred.V) < r0 */
          }   /* if LastChan != -1 */
        }     /* if ChannelStatus == -4 */
      }       /* if IndexOnRay */
     /*
      * Next channel on last ray
      */
      if (G->filt[T[CurTrk].HorUpd->LastChan]->IndexOnRay 
            < G->NumFiltersPerRay - 1) {
        CandChan = G->TotNumFilters - G->NumFiltersPerRay
                 + G->filt[T[CurTrk].HorUpd->LastChan]->IndexOnRay + 1;
        m = col*G->TotNumFilters + CandChan;
        if (ChannelStatus[CandChan] == -4) {
         /*
          * Check if viable candidate for Horizontal update
          */
          if (((double)T[CurTrk].HorUpd->pred.U
              < G->filt[T[CurTrk].HorUpd->LastChan]->u0)
            && ((double)fNorm(T[CurTrk].HorUpd->pred.U,T[CurTrk].HorUpd->pred.V)
              > G->filt[T[CurTrk].HorUpd->LastChan]->r0)) {
            CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
            if (CanQual > MaxQual) {
              T[CurTrk].Upd = T[CurTrk].HorUpd;
              UpdChan = CandChan;
              MaxQual = CanQual;
            }
          }   /* if pred.U < G->u0 && fNorm(pred.U,pred.V) > r0 */
         /*
          * Check if viable candidate for Vertical update
          */
          if (T[CurTrk].VertUpd[col]->LastChan != -1) {
            if (((double)T[CurTrk].VertUpd[col]->pred.U
                < G->filt[T[CurTrk].VertUpd[col]->LastChan]->u0)
              && ((double)fNorm(T[CurTrk].VertUpd[col]->pred.U,
                                T[CurTrk].VertUpd[col]->pred.V)
                > G->filt[T[CurTrk].HorUpd->LastChan]->r0)) {
              CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
              if (CanQual > MaxQual) {
                T[CurTrk].Upd = T[CurTrk].VertUpd[col];
                UpdChan = CandChan;
                MaxQual = CanQual;
              }
            } /* if pred.U < G->u0 && fNorm(pred.U,pred.V) > r0 */
          }   /* if LastChan != -1 */
        }     /* if ChannelStatus == -4 */
      }       /* if IndexOnRay < NumFilterPerRay - 1*/
    }         /* else (previous ray is the last ray) */

   /*
    * Check if channels on next ray are viable candidates for update
    */
    if (G->filt[T[CurTrk].HorUpd->LastChan]->Ray < G->NumRays - 1) {
     /*
      * Same channel on next ray
      */
      CandChan = T[CurTrk].HorUpd->LastChan + G->NumFiltersPerRay;
      m = col*G->TotNumFilters + CandChan;
      if (ChannelStatus[CandChan] == -4) {
       /*
        * Check if viable candidate for Horizontal update
        */
        if (atan2((double)T[CurTrk].HorUpd->pred.V,
                  (double)T[CurTrk].HorUpd->pred.U) 
            > G->filt[T[CurTrk].HorUpd->LastChan]->Theta) {
          CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
          if (CanQual > MaxQual) {
            T[CurTrk].Upd = T[CurTrk].HorUpd;
            UpdChan = CandChan;
            MaxQual = CanQual;
          }
        }   /* if atan2(pred.V,pred.U) > Theta */
       /*
        * Check if viable candidate for Vertical update
        */
        if (T[CurTrk].VertUpd[col]->LastChan != -1) {
          if (atan2((double)T[CurTrk].VertUpd[col]->pred.V,
                    (double)T[CurTrk].VertUpd[col]->pred.U) 
              > G->filt[T[CurTrk].HorUpd->LastChan]->Theta) {
            CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
            if (CanQual > MaxQual) {
              T[CurTrk].Upd = T[CurTrk].VertUpd[col];
              UpdChan = CandChan;
              MaxQual = CanQual;
            }
          } /* if atan2(pred.V,pred.U) > Theta */
        }   /* if LastChan != -1 */
      }     /* if ChannelStatus == -4 */

     /*
      * Previous channel on next ray
      */
      if ((G->filt[T[CurTrk].HorUpd->LastChan]->IndexOnRay)) {
        CandChan = T[CurTrk].HorUpd->LastChan + G->NumFiltersPerRay - 1;
        m = col*G->TotNumFilters + CandChan;
        if (ChannelStatus[CandChan] == -4) {
         /*
          * Check if viable candidate for Horizontal update
          */
          if ((atan2((double)T[CurTrk].HorUpd->pred.V,
                  (double)T[CurTrk].HorUpd->pred.U) 
                > G->filt[T[CurTrk].HorUpd->LastChan]->Theta) 
            && ((double)fNorm(T[CurTrk].HorUpd->pred.U,T[CurTrk].HorUpd->pred.V)
              < G->filt[T[CurTrk].HorUpd->LastChan]->r0)) {
            CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
            if (CanQual > MaxQual) {
              T[CurTrk].Upd = T[CurTrk].HorUpd;
              UpdChan = CandChan;
              MaxQual = CanQual;
            }
          }   /* if atan2(pred.V,pred.U) > Theta && fNorm(pred.U,pred.V) < r0 */
         /*
          * Check if viable candidate for Vertical update
          */
          if (T[CurTrk].VertUpd[col]->LastChan != -1) {
            if ((atan2((double)T[CurTrk].VertUpd[col]->pred.V,
                    (double)T[CurTrk].VertUpd[col]->pred.U) 
                  > G->filt[T[CurTrk].HorUpd->LastChan]->Theta) 
              && ((double)fNorm(T[CurTrk].VertUpd[col]->pred.U,
                                T[CurTrk].VertUpd[col]->pred.V)
                < G->filt[T[CurTrk].HorUpd->LastChan]->r0)) {
              CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
              if (CanQual > MaxQual) {
                T[CurTrk].Upd = T[CurTrk].VertUpd[col];
                UpdChan = CandChan;
                MaxQual = CanQual;
              }
            } /* if atan2(pred.V,pred.U) > Theta && fNorm(pred.U,pred.V) < r0 */
          }   /* if LastChan != -1 */
        }     /* if ChannelStatus == -4 */
      }       /* if IndexOnRay */
     /*
      * Next channel on next ray
      */
      if (G->filt[T[CurTrk].HorUpd->LastChan]->IndexOnRay 
            < G->NumFiltersPerRay - 1) {
        CandChan = T[CurTrk].HorUpd->LastChan - G->NumFiltersPerRay + 1;
        m = col*G->TotNumFilters + CandChan;
        if (ChannelStatus[CandChan] == -4) {
         /*
          * Check if viable candidate for Horizontal update
          */
          if ((atan2((double)T[CurTrk].HorUpd->pred.V,
                  (double)T[CurTrk].HorUpd->pred.U) 
                > G->filt[T[CurTrk].HorUpd->LastChan]->Theta) 
            && ((double)fNorm(T[CurTrk].HorUpd->pred.U,T[CurTrk].HorUpd->pred.V)
              > G->filt[T[CurTrk].HorUpd->LastChan]->r0)) {
            CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
            if (CanQual > MaxQual) {
              T[CurTrk].Upd = T[CurTrk].HorUpd;
              UpdChan = CandChan;
              MaxQual = CanQual;
            }
          }   /* if atan2(pred.V,pred.U) > Theta && fNorm(pred.U,pred.V) > r0 */
         /*
          * Check if viable candidate for Vertical update
          */
          if (T[CurTrk].VertUpd[col]->LastChan != -1) {
            if ((atan2((double)T[CurTrk].VertUpd[col]->pred.V,
                    (double)T[CurTrk].VertUpd[col]->pred.U) 
                  > G->filt[T[CurTrk].HorUpd->LastChan]->Theta) 
              && ((double)fNorm(T[CurTrk].VertUpd[col]->pred.U,
                                T[CurTrk].VertUpd[col]->pred.V)
                > G->filt[T[CurTrk].HorUpd->LastChan]->r0)) {
              CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
              if (CanQual > MaxQual) {
                T[CurTrk].Upd = T[CurTrk].VertUpd[col];
                UpdChan = CandChan;
                MaxQual = CanQual;
              }
            } /* if atan2(pred.V,pred.U) > Theta && fNorm(pred.U,pred.V) > r0 */
          }   /* if LastChan != -1 */
        }     /* if ChannelStatus == -4 */
      }       /* if IndexOnRay < NumFilterPerRay - 1*/
    }         /* if Ray < NumRays - 1*/
    else {    /* next ray is the first ray */
     /*
      * Same channel on first ray
      */
      CandChan = G->filt[T[CurTrk].HorUpd->LastChan]->IndexOnRay;
      m = col*G->TotNumFilters + CandChan;
      if (ChannelStatus[CandChan] == -4) {
       /*
        * Check if viable candidate for Horizontal update
        */
        if ((double)T[CurTrk].HorUpd->pred.U 
            < G->filt[T[CurTrk].HorUpd->LastChan]->u0) {
          CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
          if (CanQual > MaxQual) {
            T[CurTrk].Upd = T[CurTrk].HorUpd;
            UpdChan = CandChan;
            MaxQual = CanQual;
          }
        }   /* if pred.U < G->u0 */
       /*
        * Check if viable candidate for Vertical update
        */
        if (T[CurTrk].VertUpd[col]->LastChan != -1) {
          if ((double)T[CurTrk].VertUpd[col]->pred.U
              < G->filt[T[CurTrk].HorUpd->LastChan]->u0) {
            CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
            if (CanQual > MaxQual) {
              T[CurTrk].Upd = T[CurTrk].VertUpd[col];
              UpdChan = CandChan;
              MaxQual = CanQual;
            }
          } /* if pred.U < G->u0 */
        }   /* if LastChan != -1 */
      }     /* if ChannelStatus == -4 */

     /*
      * Previous channel on first ray
      */
      if ((G->filt[T[CurTrk].HorUpd->LastChan]->IndexOnRay)) {
        CandChan = G->filt[T[CurTrk].HorUpd->LastChan]->IndexOnRay - 1;
        m = col*G->TotNumFilters + CandChan;
        if (ChannelStatus[CandChan] == -4) {
         /*
          * Check if viable candidate for Horizontal update
          */
          if (((double)T[CurTrk].HorUpd->pred.U
              < G->filt[T[CurTrk].HorUpd->LastChan]->u0)
            && ((double)fNorm(T[CurTrk].HorUpd->pred.U,T[CurTrk].HorUpd->pred.V)
              < G->filt[T[CurTrk].HorUpd->LastChan]->r0)) {
            CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
            if (CanQual > MaxQual) {
              T[CurTrk].Upd = T[CurTrk].HorUpd;
              UpdChan = CandChan;
              MaxQual = CanQual;
            }
          }   /* if pred.U < G->u0 && fNorm(pred.U,pred.V) < r0 */
         /*
          * Check if viable candidate for Vertical update
          */
          if (T[CurTrk].VertUpd[col]->LastChan != -1) {
            if (((double)T[CurTrk].VertUpd[col]->pred.U
                < G->filt[T[CurTrk].VertUpd[col]->LastChan]->u0)
              && ((double)fNorm(T[CurTrk].VertUpd[col]->pred.U,
                                T[CurTrk].VertUpd[col]->pred.V)
                < G->filt[T[CurTrk].HorUpd->LastChan]->r0)) {
              CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
              if (CanQual > MaxQual) {
                T[CurTrk].Upd = T[CurTrk].VertUpd[col];
                UpdChan = CandChan;
                MaxQual = CanQual;
              }
            } /* if pred.U < G->u0 && fNorm(pred.U,pred.V) < r0 */
          }   /* if LastChan != -1 */
        }     /* if ChannelStatus == -4 */
      }       /* if IndexOnRay */
     /*
      * Next channel on first ray
      */
      if (G->filt[T[CurTrk].HorUpd->LastChan]->IndexOnRay 
            < G->NumFiltersPerRay - 1) {
        CandChan = G->filt[T[CurTrk].HorUpd->LastChan]->IndexOnRay + 1;
        m = col*G->TotNumFilters + CandChan;
        if (ChannelStatus[CandChan] == -4) {
         /*
          * Check if viable candidate for Horizontal update
          */
          if (((double)T[CurTrk].HorUpd->pred.U
              < G->filt[T[CurTrk].HorUpd->LastChan]->u0)
            && ((double)fNorm(T[CurTrk].HorUpd->pred.U,T[CurTrk].HorUpd->pred.V)
              > G->filt[T[CurTrk].HorUpd->LastChan]->r0)) {
            CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
            if (CanQual > MaxQual) {
              T[CurTrk].Upd = T[CurTrk].HorUpd;
              UpdChan = CandChan;
              MaxQual = CanQual;
            }
          }   /* if pred.U < G->u0 && fNorm(pred.U,pred.V) > r0 */
         /*
          * Check if viable candidate for Vertical update
          */
          if (T[CurTrk].VertUpd[col]->LastChan != -1) {
            if (((double)T[CurTrk].VertUpd[col]->pred.U
                < G->filt[T[CurTrk].VertUpd[col]->LastChan]->u0)
              && ((double)fNorm(T[CurTrk].VertUpd[col]->pred.U,
                                T[CurTrk].VertUpd[col]->pred.V)
                > G->filt[T[CurTrk].HorUpd->LastChan]->r0)) {
              CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
              if (CanQual > MaxQual) {
                T[CurTrk].Upd = T[CurTrk].VertUpd[col];
                UpdChan = CandChan;
                MaxQual = CanQual;
              }
            } /* if pred.U < G->u0 && fNorm(pred.U,pred.V) > r0 */
          }   /* if LastChan != -1 */
        }     /* if ChannelStatus == -4 */
      }       /* if IndexOnRay < NumFilterPerRay - 1*/
    }         /* next (previous ray is the first ray) */

   /*
    * If the preferred Upd Channel was already used by another track, 
    * try to update from a neighboring channel
    */
    if (ChannelStatus[UpdChan] != -4) {
      for (CandChan=0; CandChan < G->TotNumFilters; CandChan++) {
        if (ChannelStatus[CandChan] == -4) {
          if (IsAdjacent(CandChan,T[CurTrk].Upd->LastChan,G,&warp)) {
            m = col*G->TotNumFilters + CandChan;
           /*
            * Check if viable candidate for Horizontal update
            */
            CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
            if (CanQual > MaxQual) {
              T[CurTrk].Upd = T[CurTrk].HorUpd;
              UpdChan = CandChan;
              MaxQual = CanQual;
            }
           /*
            * Check if viable candidate for Vertical update
            */
            if (T[CurTrk].VertUpd[col]->LastChan != -1) {
              CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
              if (CanQual > MaxQual) {
                T[CurTrk].Upd = T[CurTrk].VertUpd[col];
                UpdChan = CandChan;
                MaxQual = CanQual;
              }
           }/* if LastChan != -1 */
          } /* if IsAdjacent */
        }   /* if ChannelStatus == -4 */
      }     /* for CandChan */
    }       /* if ChannelStatus[UpdChan] != -4 */

   /*
    * If it was not possible to update from a neighbor, try any channel
    */
    if (ChannelStatus[UpdChan] != -4) {
      for (CandChan=0; CandChan < G->TotNumFilters; CandChan++) {
        if (ChannelStatus[CandChan] == -4) {
          m = col*G->TotNumFilters + CandChan;
         /*
          * Check if viable candidate for Horizontal update
          */
          CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
          if (CanQual > MaxQual) {
            T[CurTrk].Upd = T[CurTrk].HorUpd;
            UpdChan = CandChan;
            MaxQual = CanQual;
          }
         /*
          * Check if viable candidate for Vertical update
          */
          if (T[CurTrk].VertUpd[col]->LastChan != -1) {
            CanQual = FindRadialQual(G->filt[CandChan],
                                   fabs((double)T[CurTrk].Upd->pred.U),
                                        (double)T[CurTrk].Upd->pred.V);
            if (CanQual > MaxQual) {
              T[CurTrk].Upd = T[CurTrk].VertUpd[col];
              UpdChan = CandChan;
              MaxQual = CanQual;
            }
          } /* if LastChan != -1 */
        }   /* if ChannelStatus == -4 */
      }     /* for CandChan */
    }       /* if ChannelStatus[UpdChan] != -4 */

   /*
    * By this point in the code, UpdChan has been selected
    */
    if (ChannelStatus[UpdChan] != -4) {
      printf("\nTHis Cannot Happen!!!!!\n");
      printf("\n   CurTrk=%d",CurTrk);
      printf("\n   ChanStatus[UpdChan]=%d",ChannelStatus[UpdChan]);
      printf("\n   UpdChan=%d",UpdChan);
      printf("\n   Row=%d  Col=%d",row,col);
      exit(0);
    }
    ChannelStatus[UpdChan] = CurTrk;

   /*
    * Update Track from UpdChan
    */
    m = col*G->TotNumFilters + UpdChan;
    CopyUpdRec(T[CurTrk].Upd,T[CurTrk].HorUpd);
    UpdateTrackFromObs(T+CurTrk,CurTrk,UpdChan,ChannelQ[UpdChan],
            Aob[m],Uob[m],Vob[m],row*size+col,MaxCol2Track,minCol2Track,
            KA,KU,KV,size,warp,RptFp);

    if (T[CurTrk].active) {
      CopyUpdRec(T[CurTrk].Upd,T[CurTrk].NextVertUpd[col]);
    }
  }         /* for Upd */

 /*
  * Loop over observations by size of ChannelQ To Start New Tracks
  *  For any channel that is in the top half or the Qorder, try to start a new
  *  track if the channel has not been associated with an existing track.
  */
  for (CurSizeIndx=0; CurSizeIndx < (G->TotNumFilters >> 1); CurSizeIndx++) {
   /*
    * If the chan did NOT assoc, try to START NEW TRACK
    */
    CurChan = Qorder[CurSizeIndx];
    m = col*G->TotNumFilters + CurChan;
    if ((ChannelStatus[CurChan] < 0) && (Aob[m] > TRACK_TOO_SMALL)) {
      TrackStart(CurChan,row,col,delta,size,ChannelStatus,LastStart,G,T,
                 Aob,Uob,Vob,MaxCol2Track,minCol2Track,Alast,Ulast,Vlast,
                 ChannelQ,minRow2Track,RptFp);
    }
  }             /* for CurSizeIndx */

 /*
  * Merge existing tracks
  */
  Merge(row,col,T,minCol2Track,MaxCol2Track,minRow2Track,size,delta,RptFp);

 /*
  * Deactivate tracks with amplitude too small
  */
/*-------------------------------------------------------------------
  for (CurTrk=0; CurTrk < MAXNTRACK; CurTrk++) {
    if (!T[CurTrk].in_use) continue;
    if (T[CurTrk].active) {
      if (T[CurTrk].HorUpd->pred.A < TRACK_TOO_SMALL) {
        if (T[CurTrk].age < (4*(MaxCol2Track - minCol2Track + 1))) {
          if (!T[CurTrk].protect) {
            fprintf(RptFp,
                    "\nProcTrack: Killing Track %d on A < TOO SMALL.",CurTrk);
            KillTrack(T+CurTrk,size);
          }
        }
      }
    }
  }
-------------------------------------------------------------------*/

 /*
  * Take Care of Housekeeping if last column tracked in row
  */
  if (((delta ==  1) && (col == MaxCol2Track)) ||
      ((delta == -1) && (col == minCol2Track))) {
   /*
    * Copy NextVert to Vert and re-initialize NextVert
    */
    for (k=0; k < MAXNTRACK; k++) {
      if (T[k].active && T[k].in_use) {
        ChkInBand(T+k,k,row,col,size,minCol2Track,MaxCol2Track,G,RptFp);
        if (T[k].active) {
          AgeVerticalUpdRecs(T+k,size);
          T[k].AveChanQual += T[k].AveQualRow /
                              (float)(MaxRow2Track-minRow2Track);
          if (T[k].dormantOnRow) {
            T[k].NRowsDormant++;
            if (T[k].NRowsDormant > MAX_ROWS_DORMANT) {
              if (T[k].candidate && !T[k].protect) {
                fprintf(RptFp,"\n%s %d: candidate w/ too many dormant rows.",
                       "ProcTrack Kill T",k);
                KillTrack(T+k,size);
              } else {
                DeactivateTrack(T,row,col);
              }  /* if candidate */
            }    /* if NRowsDormant > MAX */
          } else {
            T[k].NRowsDormant = 0;
          } /* else of if dormant on Row */
        } /* if active */
      }   /* if active */
    }     /* for k: loop on tracks */
  }       /* if last col tracked on row */

  return;

}/*--------------------------------------------  Track  ---------------------*/
#endif
