/*
 *
 * RunPostFilt.c
 *
 *  Program to run one isotropic Gaussian postfilter on a real image.
 *
 *  jph 5/10/96
 *
 */

#ifndef RUNPOSTFILT_C
#define RUNPOSTFILT_C

#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include <string.h>

#include "teager2D.h"


/*
 * Function prototypes
 */
void Initialize();
void ReadParametersFile();

main(argc,argv)

  int   argc;
  char *argv[];

{

  filterbank        *G;              /* the filter bank */
  pFloatImage        g;              /* postfilter image */
  pFloatImage        X;              /* input image */
  pFloatImage        Y;              /* filtered image */
  int                size;           /* row/col dim of the square input image */
  int                n;              /* size squared */
  int                k;              /* loop counter */
  char               OutFn[FNSIZE];  /* output Filename */

 /*
  * Check for right number of arguments
  */
  if (argc != 2) {
    printf("\n\n%s: Run one Gaussian isotropic postfilter on real image."
           ,argv[0]);
    printf("\nusage: %s ParameterFileName\n",argv[0]);
    exit(-1);
  }
#ifdef DEBUG
 /*
  * Link to PA so we can use it in the debugger 
  */
  if (!argc) {
    PA((void *)X, 0, -1, 'b');
    PRINTA((void *)X,0,-1,'b',"delete.me");
    WRITEA((void *)X,0,-1,'b',"delete.me");
    image2disk((void *)X,1,1,'*',"delete.me");
    SpecMag2disk(X,"delete.me");
    PC(X,0,0,-1,'b',0);
    PRINTC(X,0,0,-1,'b',0,"delete.me");
    WRITEC(X,0,0,-1,'b',0,"delete.me");
  }
#endif
  X = (pFloatImage)malloc(sizeof(FloatImage));
  Y = (pFloatImage)malloc(sizeof(FloatImage));

  Initialize(argv[1],&size,&X->Real,&X->Imag,&X->FFTReal,&X->FFTImag,&G,
             &k,OutFn);
  X->size = X->cols = X->rows = size;
  X->square = 1; 
  AllocFloatImage(Y,size);
  n = size * size;

  g = MkPostFilt(G->filt[k]->sigma,G->filt[k]->kappa,size);
  DoPostFilter(g,size,X->Real,Y->Real,G,0);

  f2disk(Y->Real,Y->size,Y->size,OutFn);

  printf("\n\n");
  return;

} /*-------------------------------------------- main --------------*/

/*
 * Initialize
 *
 *  Process the paramters file
 *  Read the input image.
 *  Compute the input spectrum
 *  Construct the filter bank
 */

void Initialize(ParmFn,size,xReal,xImag,XReal,XImag,G,k,OutFn)

  char        *ParmFn;           /* filename of parameters file */
  int         *size;             /* row/col dim of the square input image */
  float      **xReal;            /* real part of input image */
  float      **xImag;            /* imag part of input image */
  float      **XReal;            /* real part of input spectrum */
  float      **XImag;            /* imag part of input spectrum */
  filterbank **G;                /* the filter bank */
  int         *k;                /* filt to use for computing postilter */
  char        *OutFn;            /* Output Filename */
{

  int         i;                 /* counter */
  char        datatype;          /* type of in dat: b=byte f=float d=double */
  char        Rfn[FNSIZE];       /* file for Real part of input image */
  double      r0;                /* rad. center freq. of 1st filt on each ray */
  float       R;                 /* common ratio of radii of adjcnt filtrs */
  float       obw;               /* Eta-peak octave bandwidth of filters */
  float       Eta;               /* OBW is defined in terms of Eta-peak. */
  float       kappa;             /* Post Filter space constant scale factor */

 /*
  * Read the parameter file
  */
  ReadParametersFile(ParmFn,&datatype,Rfn,size,&r0,&R,&Eta,&obw,&kappa,k,OutFn);
  r0 = r0 / (double)(*size);
 
 /*
  * Make the filter bank
  */ 
  *G = CreateFilterBank(r0,R,obw,Eta,1,kappa,1.0,1.0);

 /*
  * Allocate Spectra
  */
  if ((*XReal = (float *)malloc(*size**size*sizeof(float))) == NULL) {
    printf("\nInitialize: free store exhausted\n");
    exit(-1);
  }
  if ((*XImag = (float *)malloc(*size**size*sizeof(float))) == NULL) {
    printf("\nInitialize: free store exhausted\n");
    exit(-1);
  }

 /*
  * Get the input image and its spectrum
  */
  *xReal = readasf(*size,*size,Rfn,datatype);
  if ((*xImag = (float *)malloc(*size**size*sizeof(float))) == NULL) {
    printf("\nInitialize: free store exhausted\n");
    exit(-1);
  }
  for (i=0; i < *size**size; i++) {
    (*xImag)[i] = (float)0.0;
  }
  fft2d(*xReal,*xImag,*XReal,*XImag,*size,1);

  return;
} /*---------------------------------------------- Initialize ----------*/

/*
 * ReadParametersFile
 *
 *  read the data out of the parameters file
 *
 */

void ReadParametersFile(ParmFn,pdatatype,Rfn,psize,pr0,pR,pEta,pobw,pkappa,
                        pk,OutFn)
  char   *ParmFn;
  char   *pdatatype;
  char   *Rfn;
  int    *psize;
  double *pr0;
  float  *pR;
  float  *pEta;
  float  *pobw;
  float  *pkappa;
  int    *pk;
  char   *OutFn;
{

  FILE      *parmfp;            /* ptr to parameter file */
  char       instr[256];        /* buffer to read into  */

  if ((parmfp = fopen(ParmFn,"r")) == NULL) {
    printf("\n\nReadParametersFile: unable to open %s\n",ParmFn);
    exit(-1);
  }

  fscanf(parmfp,"Input Data Type= %s\n",instr);
  *pdatatype = instr[0];
  fscanf(parmfp,"Real Image File= %s\n",Rfn);
  fscanf(parmfp,"Size= %d\n",psize);
  fscanf(parmfp,"r0= %lf\n",pr0);
  fscanf(parmfp,"R= %f\n",pR);
  fscanf(parmfp,"Eta= %f\n",pEta);
  fscanf(parmfp,"OBW= %f\n",pobw);
  fscanf(parmfp,"kappa= %f\n",pkappa);
  fscanf(parmfp,"Filter= %d\n",pk);
  fscanf(parmfp,"Output File= %s\n",OutFn);
  fclose(parmfp);
  return;
} /*--------------------------- ReadParametersFile -------------------------*/
#endif
