/*
 * Stiffen.c
 *
 *  Function StiffenUpdChan.  Enforces the constraint that a component
 *  must be slowly varying.  This function limits how rapidly a component
 *  can jump accross channels in the filterbank.  The channel is about
 *  to be updated from UpdChan when we apply the stiffness test.  If 
 *  UpdChan = LastChan then no action is taken.  If UpdChan != LastChan,
 *  and this is the first change of channel in the last STIFFNESS pixels,
 *  then no action is taken.  If a change of channel has taken place in the
 *  last STIFFNESS pixels, but UpdChan is equal to the number of an update
 *  channel in the last STIFFNESS pixels, then no action is taken.  
 *  Otherwise, UpdChan is changed to Last Chan.
 *
 *
 * 12/7/94 jph
 *
 *    mod 12/17/94   parameter G is added for compatibility with nStiffen.c
 */

#ifndef STIFFEN_C
#define STIFFEN_C
                            
#include "track.h"
 
/*-----------------------------------------------------------------------------
   STIFFENUPDCHAN
-----------------------------------------------------------------------------*/
void StiffenUpdChan(T,G,UpdChan,row,col,minCol2Track,MaxCol2Track,size,delta,
                    wrap)

  trackP       T;               /* track */
  filterbank  *G;               /* filter bank; not used in this version */
  int         *UpdChan;         /* proposed channel to update from */
  int          row;             /* image row of pixel being updated */
  int          col;             /* image col of pixel being updated */
  int          minCol2Track;    /* min column where tracking is performed */
  int          MaxCol2Track;    /* Max column where tracking is performed */
  int          size;            /* row/col dim of image */
  int          delta;           /* 1=proc row left to right, -1=right to left*/
  int         *wrap;            /* set to zero if UpdChan is changed */

{

  int          change;          /* 1 if there has been a chan change on row */
  int          Row;             /* row counter */
  int          Col;             /* col counter */
  int          k;               /* loop counter */
  int          LocalminCol;     /* local copy of minCol2Track */
  int          LocalMaxCol;     /* local copy of MaxCol2Track */
  float       *LastChan;        /* channel last update came from */
  float       *UpdPt;           /* ptr for traversing T->ObsChan */

  LocalminCol = minCol2Track;
  LocalMaxCol = MaxCol2Track;
  if (row==T->Hist.Start.row) {
    if (delta== 1) LocalminCol = T->Hist.Start.col;
    if (delta==-1) LocalMaxCol = T->Hist.Stop.col;
  }

  if ((delta==1)   && (col==LocalminCol)) return;
  if ((delta==-1)  && (col==LocalMaxCol)) return;
  LastChan = &T->ObsChan[row*size+col-delta];

  if (*UpdChan == *LastChan) return;

 /*
  * If UpdChan != LastChan, but no chan change in last STIFFNESS pts, ok.
  *   or
  * If UpdChan == update channel on one of the last STIFFNESS pts, then ok.
  */
  for (k=2,change=0,Col=col-2*delta,UpdPt=LastChan;
       (Col >= LocalminCol) && (Col <= LocalMaxCol) && (k < STIFFNESS);
       k++,Col-=delta) {
    if (*UpdChan == T->ObsChan[row*size+Col]) return;
    if (T->ObsChan[row*size+Col] != *UpdPt) {
      change++;
      UpdPt = &T->ObsChan[row*size+Col];
    } 
  }
  if (!change) return;

 /*
  * If you got here, UpdChan != LastChan, there have been channel changes
  *   in the last STIFFNESS points, and UpdChan is not one of the channels
  *   that was already used in the last STIFFNESS pts.  This is inconsistent
  *   with the fact that the modulating functions are slowly varying.  We
  *   assume that the track is being pulled away from where it should be
  *   by interference from another component or noise.
  */

  *UpdChan = *LastChan;
  *wrap    = 0;
 
  return;

}/*--------------------------------------------  StiffenUpdChan  -------------*/
#endif
