
/*
 * TrackStart.c
 *
 *   Tries to Start a new track on the given channel observations
 *   If track is started return 1
 *   If new track fails some test, return 0.
 *
 *   MAJOR MODS  3/15/96
 *
 * 4/25/94 jph
 */

#ifndef TRACKSTART_C
#define TRACKSTART_C

#include "track.h"
 
/*-----------------------------------------------------------------------------
   TRACK START
-----------------------------------------------------------------------------*/

int TrackStart(chan,row,col,delta,size,ChannelStatus,LastStart,G,T,Aob,Uob,Vob,
               MaxCol2Track,minCol2Track,Alast,Ulast,Vlast,ChannelQ,
               minRow2Track,RptFp)
 
  int          chan;            /* channel number of observations */
  int          row;             /* Image Row of current observations */
  int          col;             /* Image Col of current observations */
  int          delta;           /* -1 if processing col R to L, 1 otherwise */
  int          size;            /* row/col dimension of images */
  int         *ChannelStatus;   /* track assoc status of channel */
  image_point *LastStart;       /* for each chan, pix of last trackstart*/
  filterbank  *G;               /* the filter bank */
  trackP      T;                /* tracks */
  float      *Aob;              /* row of observations of A */
  float      *Uob;              /* row of observations of u */
  float      *Vob;              /* row of observations of v */
  int         MaxCol2Track;     /* last  col to perform track proc. on */
  int         minCol2Track;     /* first col to perform track proc. on */
  float       *Alast;           /* last row of observations of A */
  float       *Ulast;           /* last row of observations of u */
  float       *Vlast;           /* last row of observations of v */
  float       *ChannelQ;        /* channel quality */
  int         minRow2Track;     /* first row to perform track proc. on */
  FILE       *RptFp;            /* ptr to stream for track reporting */
 
{
 
  int     N;                    /* size * size */
  int     i;                    /* loop counter */
  int     j;                    /* counter */
  int     k;                    /* counter */
  int     m;                    /* counter */
  int     n;                    /* counter */
  float  *Apt;                  /* ptr to track A value */
  float  *Upt;                  /* ptr to track U value */
  float  *Vpt;                  /* ptr to track V value */
 
  N = size * size;

 /*
  * Do not start a new track if it is passed the last row to start a track
  */
  if (row > LAST_ROW_START_TRACK+minRow2Track) {
    ChannelStatus[chan] = -4;
    return(0);
  }
 /*
  * Do not start a new track if any observations are NO_OBSERVSTION
  */
  m = col*G->TotNumFilters + chan;
  if ((Aob[m] == NO_OBSERVATION) || (Uob[m] == NO_OBSERVATION) ||
                                    (Vob[m] == NO_OBSERVATION)) {
    ChannelStatus[chan] = -4;
    return(0);
  }
 
 /*
  * Do not start a new track at row,col using the observations from channel
  *  chan if (row,col) is within a distance of NEW_START_RADIUS of the
  *  last track that was started using observations from channel chan.
  */
  if ((abs(row - LastStart[chan].row) <= NEW_START_RADIUS) &&
      (abs(col - LastStart[chan].col) <= NEW_START_RADIUS)) {
    ChannelStatus[chan] = -4;
    return(0);
  }

 /*
  * Do Not start a track here if the observed amplitude is exactly 1.0...
  *   This indicates that the ESA estimate was greater than 1 and clipped.
  *
  *  11/2/94: now that there is no clipping, I won't start if > 1.2, i.e. an
  *  obvious error of 20% or more.
  */
/*--------------------------------------------------------------------
  m = col*G->TotNumFilters + chan;
  if (Aob[m] >= (float)1.2) {
    ChannelStatus[chan] = -4;
    return(0);
  }
--------------------------------------------------------------------*/
 
 /*
  * Do not start a new track if we think this observation is of a component
  * we are already tracking.  We think that if there is an existing track
  * that is not a candidate such that the filter where that track's last
  * observation was taken from is at greater than or equal to half peak
  * at the frequencies being estimated by the observation we are
  * now considering starting a track with.
  */
  j = row*size + col;
  m = col*G->TotNumFilters + chan;
  for (i=0; i < MAXNTRACK; i++) {
    if ((T[i].in_use) && (T[i].active)
         /* && (!T[i].candidate) */  ) {
      if ((G_EvalMag(G->filt[T[i].HorUpd->LastChan],(double)Uob[m],
                                       (double)Vob[m]) /
        G->filt[T[i].HorUpd->LastChan]->GMax) >= (float) 0.5) {
          ChannelStatus[chan] = -4;
          return(0);
      }
    }
  }    /* for i */
 
 /*
  * Do not start a new track if we think this observation is of a component
  * we are already tracking: a second test.  Bong the new track start if
  * the observed values of U and V are within 30 percent of the
  * filtered values of U and V for an existing track AND THE EXISTING
  * TRACK HAS LARGER AMPLITUDE..
  * THIS TEST SHOULD BE REDUNDANT W/ THE NEW ALGORITHM ON P. 110 OF BOOK VII
  */
  j = row*size + col;
  m = col*G->TotNumFilters + chan;
  for (i=0; i < MAXNTRACK; i++) {
    if ((T[i].in_use) && (T[i].active)) {
      if (T[i].updated) {
        Upt = &T[i].U_filt[j];
        Vpt = &T[i].V_filt[j];
        Apt = &T[i].A_filt[j];
      }
      else {
        Upt = &T[i].HorUpd->pred.U;
        Vpt = &T[i].HorUpd->pred.V;
        Apt = &T[i].HorUpd->pred.A;
      }
      if (fabs(Uob[m] - *Upt)
                      < (double)0.3 * fabs((double)*Upt)) {
        if (fabs(Vob[m] - *Vpt)
                        < (double)0.3 * fabs((double)*Vpt)) {
          if (Aob[m] < *Apt) {
            ChannelStatus[chan] = -4;
            return(0);
          }
        }
      }
    }
  }     /* for i */
 
 /*
  * Find the index of the first free track
  */
  n = 0;
  while ((T[n].in_use) && (n < MAXNTRACK)) {
    n++;
  }
  if (n >= MAXNTRACK) {
   return(0);
  }

 /*
  * Report
  */
  fprintf(RptFp,"\nNew Track Start: track=%d row=%d col=%d channel=%d",
          n,row,col,chan);
 
  LastStart[chan].row = row;
  LastStart[chan].col = col;
  ChannelStatus[chan] = n;
  T[n].in_use         = 1;
  T[n].candidate      = 1;
  T[n].active         = 1;
  T[n].protect        = 0;
  T[n].updated        = 1;
  T[n].coast          = 0;
  T[n].Ncoast         = 0;
  T[n].TotNcoast      = 0;
  T[n].dormant        = 0;
  T[n].Ndormant       = 0;
  T[n].dormantOnRow   = 0;
  T[n].NRowsDormant   = 0;
  for (i=0; i < MAXNTRACK; i++) {
    T[n].MergeCount[i] = 0;
    T[n].MergeRows[i] = 0;
  }
  T[n].age            = 1;
  T[n].NgoodUpd       = 1;
  T[n].NgoodUpdRow    = 1;
  T[n].Hist.Start.row = row;
  T[n].Hist.Start.col = col;
  T[n].Hist.Merged    = 0;
  T[n].Hist.CandidateWhenMerged = 0;
  T[n].Hist.MTrack    = -1;
  T[n].Hist.SmallAmpDeac = 0;
  T[n].Hist.NRaysVis  = 1;
  if ((T[n].Hist.RaysVis   = (int *)malloc(G->NumRays * sizeof(int))) == NULL) {
    printf("\nTrackStart: Free store exhausted.\n");
    return(0);
  }
  for (i=0; i < G->NumRays; i++) {
    T[n].Hist.RaysVis[i] = 0;
  }
  T[n].Hist.RaysVis[G->filt[chan]->Ray] = 1;
  T[n].MaxQual        = ChannelQ[chan];
  T[n].AveQualRow     = ChannelQ[chan] / (float)(MaxCol2Track - minCol2Track);
  T[n].AveChanQual    = (float)0.0;
  T[n].SumInnov       = (float)0.0;
  if ((T[n].A_filt = (float *)malloc(N * sizeof(float))) == NULL) {
    printf("\nTrackStart: Free store exhausted.\n");
  }
  if ((T[n].U_filt = (float *)malloc(N * sizeof(float))) == NULL ) {
    printf("\nTrackStart: Free store exhausted.\n");
    return(0);
  }
  if ((T[n].V_filt = (float *)malloc(N * sizeof(float))) == NULL ) {
    printf("\nTrackStart: Free store exhausted.\n");
  }
  if ((T[n].ObsChan = (float *)malloc(N * sizeof(float))) == NULL ) {
    printf("\nTrackStart: Free store exhausted.\n");
    return(0);
  }
  for (i=0,k=0; i < size; i++) {   /* i counts rows, k counts pixels */
    for (j=0; j < size; k++,j++) { /* j counts cols */
      T[n].ObsChan[k] = -1;
      T[n].A_filt[k] = T[n].U_filt[k] = T[n].V_filt[k] = (float)0.0;
      if ((i < row) && (j >= minCol2Track) 
                    && (j <= MaxCol2Track)) T[n].Ndormant++;
      else {
        if (i == row) {
          if (((delta == -1) && (j <= MaxCol2Track) && (j > col)) ||
              ((delta ==  1) && 
              (j >= minCol2Track) && (j < col))) T[n].Ndormant++;
        } /* if (i == row) */
      }   /* else */
    }
  }
  T[n].HorUpd = (UpdRecP)malloc(sizeof(UpdRec));
  T[n].VertUpd = (UpdRecP *)malloc(size * sizeof(UpdRecP));
  T[n].NextVertUpd = (UpdRecP *)malloc(size * sizeof(UpdRecP));
  for (i=0; i < size; i++) {
    T[n].VertUpd[i] = (UpdRecP)malloc(sizeof(UpdRec));
    T[n].NextVertUpd[i] = (UpdRecP)malloc(sizeof(UpdRec));
    T[n].VertUpd[i]->LastChan = -1;
    T[n].NextVertUpd[i]->LastChan = -1;
  }

  T[n].Upd = T[n].HorUpd;
  T[n].HorUpd->k = -1;
  T[n].HorUpd->LastChan      = chan;
  T[n].ObsChan[row*size+col] = chan;

 /*
  * Initialize filtered estimates at row,col (-1 | -1)
  */
  j = row*size + col;
  m = col*G->TotNumFilters + chan;
  T[n].A_filt[j] = Aob[m];
  T[n].U_filt[j] = Uob[m];
  T[n].V_filt[j] = Vob[m];
 
 /*
  * Initialize predicted velocities at row,col ( 0 | -1)
  */
  if (((col == minCol2Track) && (delta ==-1)) ||
      ((col == MaxCol2Track) && (delta == 1))) {
    T[n].HorUpd->v_pred.A = Aob[m] - Alast[m];  /* next move is down */
    T[n].HorUpd->v_pred.U = Uob[m] - Ulast[m];
    T[n].HorUpd->v_pred.V = Vob[m] - Vlast[m];
  }
  else {
    T[n].HorUpd->v_pred.A = Aob[m] - Aob[m - delta*G->TotNumFilters];
    T[n].HorUpd->v_pred.U = Uob[m] - Uob[m - delta*G->TotNumFilters];
    T[n].HorUpd->v_pred.V = Vob[m] - Vob[m - delta*G->TotNumFilters];
  }
 
 /*
  * Predict ahead one time step to k=0
  */
  T[n].HorUpd->pred.A = T[n].A_filt[j] + T[n].HorUpd->v_pred.A;
  if (T[n].HorUpd->pred.A < (float)0.0) {
      T[n].HorUpd->pred.A = (float)0.0;
  }
  if (T[n].HorUpd->pred.A > (float)1.0) {
    T[n].HorUpd->pred.A = (float)1.0;
  }
  T[n].HorUpd->pred.U = T[n].U_filt[j] + T[n].HorUpd->v_pred.U;
  if (T[n].HorUpd->pred.U < (float)-0.5) {
      T[n].HorUpd->pred.U = (float)-0.5;
  }
  if (T[n].HorUpd->pred.U > (float)0.5) {
    T[n].HorUpd->pred.U = (float)0.5;
  }
  T[n].HorUpd->pred.V = T[n].V_filt[j] + T[n].HorUpd->v_pred.V;
  if (T[n].HorUpd->pred.V < (float)-0.5) {
    T[n].HorUpd->pred.V = (float)-0.5;
  }
  if (T[n].HorUpd->pred.V > (float)0.5) {
    T[n].HorUpd->pred.V = (float)0.5;
  }

  return(1);
 
}/*--------------------------------------------  TrackStart  ----------------*/
#endif
