/*
 * nStiffen.c
 *
 *  This is a specialized version of Stiffen.c for use with some of the
 *  Brodatz textures.  It will only let a track be updated from channels
 *  on MAXRAYVISIT rays of the filterbank.  This will hopefully allow tracks to
 *  remain in the correct channels across regions where the component is
 *  not supported, rather than being drawn into channels occupied by
 *  components that are supported in the region.
 *
 *  Function StiffenUpdChan.  Enforces the constraint that a component
 *  must be slowly varying.  This function limits how rapidly a component
 *  can jump accross channels in the filterbank.  The channel is about
 *  to be updated from UpdChan when we apply the stiffness test.  If 
 *  UpdChan = LastChan then no action is taken.  If UpdChan != LastChan,
 *  and this is the first change of channel in the last STIFFNESS pixels,
 *  then no action is taken.  If a change of channel has taken place in the
 *  last STIFFNESS pixels, but UpdChan is equal to the number of an update
 *  channel in the last STIFFNESS pixels, then no action is taken.  
 *  Otherwise, UpdChan is changed to Last Chan.
 *
 *
 * 12/17/94 jph
 */

#ifndef NSTIFFEN_C
#define NSTIFFEN_C
                            
#include "track.h"
 
/*-----------------------------------------------------------------------------
   STIFFENUPDCHAN
-----------------------------------------------------------------------------*/
void nStiffenUpdChan(T,G,UpdChan,row,col,minCol2Track,MaxCol2Track,size,delta,
                     wrap)

  trackP       T;               /* track */
  filterbank  *G;               /* filter bank */
  int         *UpdChan;         /* proposed channel to update from */
  int          row;             /* image row of pixel being updated */
  int          col;             /* image col of pixel being updated */
  int          minCol2Track;    /* min column where tracking is performed */
  int          MaxCol2Track;    /* Max column where tracking is performed */
  int          size;            /* row/col dim of image */
  int          delta;           /* 1=proc row left to right, -1=right to left*/
  int         *wrap;            /* set to zero if UpdChan is changed */

{

  int          change;          /* 1 if there has been a chan change on row */
  int          Row;             /* row counter */
  int          Col;             /* col counter */
  int          k;               /* loop counter */
  int          LocalminCol;     /* local copy of minCol2Track */
  int          LocalMaxCol;     /* local copy of MaxCol2Track */
  float       *LastChan;        /* channel last update came from */
  float       *UpdPt;           /* ptr for traversing T->ObsChan */
  int          UpdRay;          /* ray of filter UpdChan */


  LocalminCol = minCol2Track;
  LocalMaxCol = MaxCol2Track;
  if (row==T->Hist.Start.row) {
    LocalminCol = T->Hist.Start.col;
  }
  if ((delta==1)   && (col==LocalminCol)) return;
  if ((delta==-1)  && (col==LocalMaxCol)) return;
  LastChan = &T->ObsChan[row*size+col-delta];
  if (*UpdChan == *LastChan) return;

 /*
  * Test to see if we will allow update from this RAY of the filterbank
  */
  UpdRay = G->filt[*UpdChan]->Ray;
  if (!T->Hist.RaysVis[UpdRay]) {   /* if we have not visited the ray before */
    if (T->Hist.NRaysVis >= MAXRAYVISIT) {
      *UpdChan = *LastChan;
      *wrap = 0;
      return;
    }
   /*
    * if this ray not adjacent to a previously visited ray
    */

    if ((UpdRay == 0) && (!T->Hist.RaysVis[1])
                      && (!T->Hist.RaysVis[G->NumRays-1])) {
       *UpdChan = *LastChan;  /* this ray not allowed */
       *wrap = 0;
       return;
    }
    if ((UpdRay+1 == G->NumRays) && (!T->Hist.RaysVis[0]) 
                                 && (!T->Hist.RaysVis[G->NumRays-2])) {
       *UpdChan = *LastChan;  /* this ray not allowed */
       *wrap = 0;
       return;
    }
    if ((UpdRay != 0) && (UpdRay+1 != G->NumRays) &&
                         (!T->Hist.RaysVis[UpdRay-1]) &&
                         (!T->Hist.RaysVis[UpdRay+1])) {
       *UpdChan = *LastChan;  /* this ray not allowed */
       *wrap = 0;
       return;
    }
    T->Hist.RaysVis[UpdRay] = 1;  /* this ray is allowed */
    T->Hist.NRaysVis++;
  }    /* if (!T->Hist.RaysVis[UpdRay]) */

 /*
  * If UpdChan != LastChan, but no chan change in last STIFFNESS pts, ok.
  *   or
  * If UpdChan == update channel on one of the last STIFFNESS pts, then ok.
  */
  for (k=2,change=0,Col=col-2*delta,UpdPt=LastChan;
       (Col >= LocalminCol) && (Col <= LocalMaxCol) && (k < STIFFNESS);
       k++,Col-=delta) {
    if (*UpdChan == T->ObsChan[row*size+Col]) return;
    if (T->ObsChan[row*size+Col] != *UpdPt) {
      change++;
      UpdPt = &T->ObsChan[row*size+Col];
    } 
  }
  if (!change) return;

 /*
  * If you got here, UpdChan != LastChan, there have been channel changes
  *   in the last STIFFNESS points, and UpdChan is not one of the channels
  *   that was already used in the last STIFFNESS pts.  This is inconsistent
  *   with the fact that the modulating functions are slowly varying.  We
  *   assume that the track is being pulled away from where it should be
  *   by interference from another component or noise.
  */

  *UpdChan = *LastChan;
  *wrap = 0;
 
  return;

}/*--------------------------------------------  StiffenUpdChan  -------------*/
#endif
