/*
 * track.h
 *
 *  header file for track.c
 *
 *  contains global definitions, #define 's, and declarations.
 *
 * 4/24/94 jph
 */

#ifndef TRACK_H
#define TRACK_H

#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include <string.h>

#include "teager2D.h"

/*-----------------------------------------------------------------------------
  Define Constants Macros
-----------------------------------------------------------------------------*/
#define MAXNTRACK 16
#define NO_OBSERVATION ((float)-10.0)
#define NEW_START_RADIUS 32
#define TRACK_TOO_SMALL 0.001
#define NOMINAL_NUM_TRACKS  8
#define Q_THRESHOLD 0.30
#define A_THRESHOLD 0.03
#define MAX_COAST_FRAMES 64
#define BETA_GATE 0.2
#define N_COAST_DORMANT 6
#define MAX_ROWS_DORMANT 6
#define CAND_MIN_GOOD_UPD_ROW 16
#define MERGE_NEW_AGE_DIF 16
#define LAST_ROW_START_TRACK 64
#define STIFFNESS 16
#define MAXRAYVISIT 4

/*-----------------------------------------------------------------------------
  Local Typedefs
-----------------------------------------------------------------------------*/
 
/*
 * Type for a point in the image
 */
struct image_point_t {
  int          row;
  int          col;
};
typedef struct image_point_t image_point;

/*
 * Type for an A,U,V triple
 */
struct AUV_triple_t {
  float        A;                     /* Amplitude */ 
  float        U;                     /* Horiz Inst Freq */
  float        V;                     /* Vert  Inst Freq */
};
typedef struct AUV_triple_t AUV_triple;
typedef AUV_triple *AUV_tripleP;

/*
 * Type for gain sequence
 */
struct Kgain_t {
  int          Ngains;                /* number of gains in the sequence */
  float       *gain;                  /* the gain sequence */
};
typedef struct Kgain_t Kgain;
typedef Kgain *pKgain;

/*
 * Type for Alpha-Beta filter gains
 */
struct ABgain_t {
  pKgain       alpha;                 /* the alpha gain sequence */
  pKgain       beta;                  /* the beta  gain sequence */
};
typedef struct ABgain_t ABgain;
typedef ABgain *pABgain;

/*
 * Type for a freq gate
 */
struct freq_gate_t {
  float        radius;                /* radius of gate */
  float        Aradius;               /* radius of gate in A direction */
  float        Uradius;               /* radius of gate in U direction */
  float        Vradius;               /* radius of gate in V direction */
  float        eA_sm;                 /* smoothed innovation in A */
  float        eU_sm;                 /* smoothed innovation in U */
  float        eV_sm;                 /* smoothed innovation in V */
};
typedef struct freq_gate_t gate_t;

/*
 * Type for a track update record
 */
struct UpdRec_t {
  int          k;                     /* index of next Kalman gain to use */
  AUV_triple   pred;                  /* horizontal predicted A,U,V */
  AUV_triple   v_pred;                /* horizontal velocity in A,U,V */
  float        alpha;                 /* weight for A term in epsilon */
  float        beta;                  /* weight for U term in epsilon */
  float        gamma;                 /* weight for V term in epsilon */
  gate_t       gate;                  /* track frequency gate */
  int          loosegate;             /* allows gate to relax at ends of col */
  int          LastChan;              /* Num of channel of last observation */
};
typedef struct UpdRec_t UpdRec;
typedef UpdRec *UpdRecP;

/*
 * Type for a track history record
 */
struct TrkHist_t {
   image_point  Start;                 /* Track Start Location */
   image_point  Stop;                  /* Track Stop  Location */
   int          Merged;                /* 1 if track has been merged */
   int          CandidateWhenMerged;   /* 1 if candidate when merged */
   int          MTrack;                /* track merged with (if merged) */
   int          SmallAmpDeac;          /* 1 if deactivated for sml ampltd */
   int          NRaysVis;              /* Number of rays visited by track */
   int         *RaysVis;               /* 1 if trk has been upd frm ray k */
};
typedef struct TrkHist_t TrkHist;
typedef TrkHist *TrkHistP;

 
/*
 * Type for a track
 */
struct track_t {
 
  int          in_use;                /* 1 if track has been started */
  int          candidate;             /* 1 if still only a candidate track */
  int          active;                /* 1 if the track is active */
  int          updated;               /* 1 if track has been updated this pix */
  int          protect;               /* for debugging... */
  int          coast;                 /* 1 if the track is coasting */
  int          Ncoast;                /* num observations track has coasted */
  int          TotNcoast;             /* total number of coasted observations */
  int          NObOutBW;              /* num obs outside channel BW, bk XI p41*/
  int          dormant;               /* 1 if track is dormant */
  int          Ndormant;              /* No pixels at which track is dormant */
  int          dormantOnRow;          /* 1 while dormant on current row */
  int          NRowsDormant;          /* No complete contiguous rows dormant */
  int          age;                   /* Age of the track */
  int          NgoodUpd;              /* No pixels Trk had a good update */
  int          NgoodUpdRow;           /* NgoodUpd for this row only */
  int          MergeCount[MAXNTRACK]; /* Num updates overlap w/ other tracks */
  int          MergeRows[MAXNTRACK];  /* no rows supporting merge */
  float        MaxQual;               /* max chan quality observed for track */
  float        AveQualRow;            /* ave chan qual along current row */
  float        AveChanQual;           /* ave chan qual for track */
  float        SumInnov;              /* Sum of norm of innovations */
  float       *A_filt;                /* filtered A estimates */
  float       *U_filt;                /* filtered U estimates */
  float       *V_filt;                /* filtered V estimates */
  float       *ObsChan;               /* Channel from which observations came*/
  UpdRecP      HorUpd;                /* Ptr to horizontal UpdRec */
  UpdRecP     *VertUpd;               /* Row of Ptr to prev row Vert UpdRecs */
  UpdRecP     *NextVertUpd;           /* Row of Ptr to curnt row Vert UpdRecs */
  UpdRecP      Upd;                   /* Ptr to a single UpdRec */
  TrkHist      Hist;                  /* Track History Info */
};
typedef struct track_t track_t;
typedef track_t *trackP;

/*-----------------------------------------------------------------------------
  Typedefs for Reconstruction
-----------------------------------------------------------------------------*/

/*
 * Type for a list of domain points
 */
struct DomainPtList {
  int          NDomainPts;            /* Num domain pts in list */
  int         *domain;                /* array of domain coordinates */
  int         *Bidx;                  /* BC index of each domain coordinate */
};
typedef struct DomainPtList DomainPtList;
typedef DomainPtList *pDomainPtList;

/*
 * Type for a boundary condition control record
 */
struct BCctrlRecord {
  int            NtrksNeed;          /* Num tracks needing bc's from channel */
  pDomainPtList  DPL;                /* array of domain pt lists, 1 for ea trk*/
};
typedef struct BCctrlRecord BCctrlRecord;
typedef BCctrlRecord *pBCctrlRecord;

/*
 * Type for boundary conditions to reconstruct a single component
 */
struct BoundCon {
  int         firstrow;        /* image row component started on */
  int         fr_startcol;     /* 1st  col component exists on in startrow */
  int         fr_stopcol;      /* last col component exists on in startrow */
  int         lastrow;         /* image row component stopped on */
  int         lr_startcol;     /* 1st  col component exists on in lastrow */
  int         lr_stopcol;      /* last col component exists on in lastrow */
  int         minCol2Track;    /* min column tracked */
  int         MaxCol2Track;    /* Max column tracked */
  int         BCinterval;      /* spacing interval between boundary conditions*/
  int         Nfirst;          /* Number of BCs for first row */
  int         Nlast;           /* Number of BCs for last  row */
  int         Npts;            /* Tot Number of boundary cond points for comp */
  float      *Re;              /* Real boundary condition samples */
  float      *ReFirst;         /* Pts to Re samples for first  row */
  float      *ReLast;          /* Pts to Re samples for last   row */
  float      *ReRest;          /* Pts to Re samples for other  rows */
  float      *Im;              /* Imag boundary condition samples */
  float      *ImFirst;         /* Pts to Im samples for first  row */
  float      *ImLast;          /* Pts to Im samples for last   row */
  float      *ImRest;          /* Pts to Im samples for other  rows */
};
typedef struct BoundCon BoundCon;
typedef BoundCon *pBoundCon;

/*-----------------------------------------------------------------------------
  Function Prototypes
-----------------------------------------------------------------------------*/
#ifndef AGEVERTICALUPDRECS_C
void AgeVerticalUpdRecs();
#endif
#ifndef ALLOCWAB_C
void AllocWAb();
#endif
#ifndef ASSOC_C
float Assoc();
#endif
#ifndef BINIT_C
void  bInitialize();
#endif
#ifndef BRDPFILE_C
void  bReadParametersFile();
#endif
#ifndef CHKINBAND_C
void ChkInBand();
#endif
#ifndef COASTORKILL_C
int  CoastOrKill();
#endif
#ifndef COPYUPDREC_C
void CopyUpdRec();
#endif
#ifndef DEACTIVATETRACK_C
void DeactivateTrack();
#endif
#ifndef EPIB_C
void Epilogue();
#endif
#ifndef EVALCAN_C
int   EvaluateCandidate();
#endif
#ifndef FINDABC_C
void FindABC();
#endif
#ifndef FREEWAB_C
void FreeWAb();
#endif
#ifndef FSIGN_C
float Fsign();
#endif
#ifndef HIPASSDOMESA_C
void HiPassDomEsa();
#endif
#ifndef HIPASSINIT_C
void HiPassInit();
#endif
#ifndef HIPASSRPF_C
void HiPassRPF();
#endif
#ifndef INGATE_C
int InGate();
#endif
#ifndef INIT_C
void  Initialize();
#endif
#ifndef INITABGAIN_C
pABgain InitABgain();
#endif
#ifndef KIDX_C
int Kidx();
#endif
#ifndef KILLABGAIN_C
pABgain  KillABgain();
#endif
#ifndef KILLTRACK_C
void  KillTrack();
#endif
#ifndef MERGE_C
void  Merge();
#endif
#ifndef NSTIFFEN_C
void  nStiffenUpdChan();
#endif
#ifdef USE_MAXFILT1
#ifndef OBS2TMPFILE_C
void Obs2TmpFile();
#endif
#endif
#ifndef PROCTRACK_C
void  Track();
#endif
#ifndef RDPFILE_C
void  ReadParametersFile();
#endif
#ifndef RDNXTROW_C
void  ReadNextRowObservations();
#endif
#ifndef RESIZEWRITE_C
void ResizeWrite();
#endif
#ifndef STIFFEN_C
void  StiffenUpdChan();
#endif
#ifdef USE_MODEL
#ifndef RPT1_C
FILE *Rpt1();
#endif
#ifndef RPTAQ_C
void RptAQ();
#endif
#ifndef RPTCHANQ_C
void RptChanQ();
#endif
#endif
#ifndef TRACKSTART_C
int   TrackStart();
#endif
#ifndef UPDGATERAD_C
void UpdGateRad();
#endif
#ifndef UPDTRKFRMOBS_C
void  UpdateTrackFromObs();
#endif

void Epilogue();

#endif
