/*
 * RelError1D.c
 *
 *  Computes the relative error in a 1D component reconstruction.
 *  The relative error is defined as the ratio of the absolute reconstruction
 *  error to the instantaneous amplitude.  This quantity is computed at every
 *  point.  Simple statistics of the relative error are reported.  The
 *  common logarithm of the relative error is also computed at each point.
 *  Simple statistics of the common logarithm are also reported (these may
 *  be interpreted as the "order of magnitude" error.
 *
 *  Data Format:
 *   
 *     The instantaneous amplitude and reconstruction error files are assumed
 *     to hold ascii data in a format suitable for xvgr.
 *
 *  Special Cases:
 *
 *     if the absolute error is zero, then the relative error is zero.
 *     if the instantaneous amplitude is zero, the relative error is zero.
 *     if the relative error is zero, the log relative error is zero.
 *
 *  6/11/97 jph
 *  
 */

#ifndef RELERROR1D_C
#define RELERROR1D_C
#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include "teager2D.h"

main(argc,argv)

  int   argc;
  char *argv[];

{

  int    size;          /* Number of data points to be considered */
  char  *AmpFn;         /* Filename for component instantaneous amplitude */
  char  *ErrorFn;       /* Filename for component reconstruction error */
  float *Amp;           /* Instantaneous amplitude of component */
  float *Error;         /* Component reconstruction error */
  float *RelError;      /* Relative reconstruction error */
  float *LogRelError;   /* Common log of relative reconstruction error */
  FILE  *infile;        /* input file pointer */
  int    i;             /* Counter */
  int    j;             /* Counter */
  float  REmin;         /* min of RelError */
  float  REmed;         /* median of RelError */
  float  REave;         /* mean of RelError */
  float  REmax;         /* max of RelError */
  float  REsig;         /* std deviation of RelError */
  float  LREmin;        /* min of LogRelError */
  float  LREmed;        /* median of LogRelError */
  float  LREave;        /* mean of LogRelError */
  float  LREmax;        /* max of LogRelError */
  float  LREsig;        /* std deviation of LogRelError */
 
 /*
  * Parse args
  */
  if (argc !=  4) {
    printf("\n\nusage: %s Size AmpFile  ReconErrorFile\n",argv[0]);
    return;
  }
  size = atoi(argv[1]);
  AmpFn = argv[2];
  ErrorFn = argv[3];

 /*
  * Allocate working arrays
  */
  if ((Amp  = (float *) malloc(size*sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((Error = (float *) malloc(size*sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((RelError = (float *) malloc(size*sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((LogRelError = (float *) malloc(size*sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }

 /*
  * Read input files
  */
  if ((infile = fopen(AmpFn,"r")) == NULL) {
    printf("\nCannot open file %s",AmpFn);
    exit(0);
  }
  for (i=0; i < size; i++) {
    fscanf(infile,"%d\t%f\n",&j,&Amp[i]);
  }
  fclose(infile);
  if ((infile = fopen(ErrorFn,"r")) == NULL) {
    printf("\nCannot open file %s",ErrorFn);
    exit(0);
  }
  for (i=0; i < size; i++) {
    fscanf(infile,"%d\t%f\n",&j,&Error[i]);
  }
  fclose(infile);

 /*
  * Compute relative error
  */
  for (i=0; i < size; i++) {
    if (Error[i] == (float)0.0) {
      RelError[i] = (float)0.0;
    } else {
      if (Amp[i] == (float)0.0) {
        RelError[i] = (float)0.0;
      } else {
        RelError[i] = (float)fabs(((double)Error[i])/((double)Amp[i]));
      }
    }
  }     /* for i */

 /*
  * Compute common log of relative error
  */
  for (i=0; i < size; i++) {
    if (RelError[i] == (float)0.0) {
      LogRelError[i] = (float)0.0;
    } else {
      LogRelError[i] = (float)fabs(log10((double)RelError[i]));
    }
  }

 /*
  * Compute simple statistics of relative error
  */
  fsstat(RelError,1,size,&REmin,&REmed,&REmax,&REave,&REsig);

 /*
  * Compute simple statistics of log relative error
  */
  fsstat(LogRelError,1,size,&LREmin,&LREmed,&LREmax,&LREave,&LREsig);
 /*
  * if there were points with zero error, then LREmin will be zero.  But,
  * we want it to be instead the worst "order of magnitude error", which
  * is the smallest NONZERO value in LogRelError.
  */
  if (LREmin == (float)0.0) {
    LREmin = LREmax;
    for (i=0; i < size; i++) {
      if ((LogRelError[i] != (float)0.0) && (LogRelError[i] < LREmin)) {
        LREmin = LogRelError[i];
      }
    }
  }

 /*
  * Report
  */
  printf("\n");
  printf("\nSimple Statistics of Relative Error:");
  printf("\nMax= %f",REmax);
  printf("\nAve= %f",REave);
  printf("\nMed= %f",REmed);
  printf("\nMin= %f",REmin);
  printf("\nStd= %f",REsig);
  printf("\n");
  printf("\nSimple Statistics of Log Relative Error:");
  printf("\nMax= %f",LREmax);
  printf("\nAve= %f",LREave);
  printf("\nMed= %f",LREmed);
  printf("\nMin= %f",LREmin);
  printf("\nStd= %f",LREsig);
  printf("\n\n");
  
  return;
} 
#endif
