/* 
 * DomComp1D.c
 *
 *  Dominant component analysis in 1D.  At each point in the domain, estimation
 *  is performed for one dominanat component using the data from the channel
 *  that maximizes the filter selection criterion |t| / Gmax.
 *
 *  This program was made by modifying the program 
 *    anchovy:/export/home2/joebob/tracor/track1D/src/MkObs1D.c
 *
 *  8/16/95 jph
 *
 *
 *----------  COMMENT BLOCK FROM MkObs1D.c  ---------------------------------
 *
 *  This program makes the ESA observations to input to the frequency tracker
 *  for the 1D version of the ICIP'94 Algorithm.
 *  
 *  It was made by modifying the program 
 *     /anchovy/export/home/joebob/track/MkObs.c
 *
 *  In this program, the ESA is evaluated for all frequency channels at every 
 *  point in the input signal.  A value of -10 = NO_OBSERVATION for any ESA 
 *  output means that the computation of the ESA did not yield a valid 
 *  observation at that point  (for reasons of, e.g. zero denominator).
 *
 *  Input parameters are read from a params file specified as the command line
 *  option.  
 *
 * See p. 102 of Book VIII for a description of the paramter file.
 *
 * 
 *  usage: MkObs1D ParameterFileName
 *
 *  jph 9/19/94
 */

#ifndef DOMCOMP1D_C
#define DOMCOMP1D_C

#define NO_OBSERVATION -10

#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include <string.h>

#include "teager2D.h"
#include "Esa1D.h"

/*
 * Function prototypes
 */
void Initialize();
void ReadParametersFile();
void dom_esa();

main(argc,argv)

  int   argc;
  char *argv[];

{

  filterbank        *G;              /* the filter bank */
  pFloatImage        X;              /* input signal */
  pFloatImage        Y;              /* filtered signal */
  int                size;           /* size of the input signal */
  int                n;              /* size again */
  int                i;              /* counter */
  int                k;              /* loop counter */
  int                col;            /* loop counter */
  int                looplim;        /* terminal value for a loop counter */
  int                filter;         /* loop counter */
  int                writesize;      /* num bytes to output in a single write */
  int                ZeroPad;        /* 1 if zero padding is to be used */
  int                fd;             /* file descriptor */
  int               *maxfilt;        /* Which channel max'd criterion at i,j */
  float             *criterion;      /* filter selection criterion */
  float             *A;              /* A estimates accross all filters */
  float             *U;              /* U estimates accross all filters */
  BYTE              *u_valid;        /* 1 where the u estimate is valid */
  BYTE              *a_valid;        /* 1 where the a estimate is valid */
  float             *DCReal;         /* Real part of the dominant component */
  float             *DCImag;         /* Imag part of the dominant component */
  float             *OutDat[2];      /* ptrs to array of data to output */
  float             *OutVector;      /* vector of obs ready to output */
  int                PrintResponse;  /* 1 if printing response signals */
  char              *OutFn[2];       /* names for observation files */
  char               RespRootFn[FNSIZE];/* root filename for response signals */
  char               TempString[FNSIZE];/* string for building up filenames */

 /*
  * Check for right number of arguments
  */
  if (argc != 2) {
    printf("\n\n%s: 1D Dominant component AM/FM analysis.",argv[0]);
    printf("\nusage: %s ParameterFileName\n",argv[0]);
    exit(-1);
  }
#ifdef DEBUG
 /*
  * Link to PA so we can use it in the debugger 
  */
  if (!argc) {
    PA((void *)X, 0, -1, 'b');
    PRINTA((void *)X,0,-1,'b',"delete.me");
    WRITEA((void *)X,0,-1,'b',"delete.me");
    image2disk((void *)X,1,1,'*',"delete.me");
    SpecMag2disk(X,"delete.me");
    PC(X,0,0,-1,'b',0);
    PRINTC(X,0,0,-1,'b',0,"delete.me");
    WRITEC(X,0,0,-1,'b',0,"delete.me");
  }
#endif
  X = (pFloatImage)malloc(sizeof(FloatImage));
  Y = (pFloatImage)malloc(sizeof(FloatImage));
  Initialize(argv[1],&size,&ZeroPad,&X->Real,&X->Imag,&X->FFTReal,&X->FFTImag,
             &G,&PrintResponse,RespRootFn);
  X->size = X->cols = size;
  X->rows = 1;
  X->square = 0;
  AllocFloatSequence(Y,size);
  n = size;

 /*
  * Allocate working arrays
  */
  u_valid = (BYTE *) malloc(Y->size*Y->size*sizeof(BYTE));
  a_valid = (BYTE *) malloc(Y->size*Y->size*sizeof(BYTE));
 
  if ((maxfilt = (int *)malloc(n * sizeof(int))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
  }
  for (i=0; i < n; i++) maxfilt[i] = -1;
 
  if ((criterion = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
  }
  for (i=0; i < n; i++) criterion[i] = (float)-1.0;

  if ((A = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
  }
  if ((U = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
  }
  if ((DCReal = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
  }
  if ((DCImag = (float *)malloc(n * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
  }

 /*
  * Process with the zeroth filter
  */
  if (ZeroPad) {
    looplim = size >> 1;
  } else {
    looplim = size;
  }
  printf("\nDoing Filter 0...");
  HardyConvolveGabor1D(Y,X,G->filt[0]);
  dom_esa(Y,G->filt[0],criterion,U,A,G,ZeroPad,maxfilt,u_valid,a_valid,
          DCReal,DCImag);

 /*
  * Loop for remaining filters
  */
  for (i=1; i < G->TotNumFilters; i++) {
    printf("\nDoing Filter %d...",i);
    HardyConvolveGabor1D(Y,X,G->filt[i]);
    dom_esa(Y,G->filt[i],criterion,U,A,G,ZeroPad,maxfilt,u_valid,a_valid,
            DCReal,DCImag);
  } /* for i */

 /*
  * Interpolate for yet indeterminate values
  */
  for (i=0; i<looplim; i++) {
    if (!u_valid[i]) f_interp1D(U,i,looplim,u_valid);
    if (!a_valid[i]) f_interp1D(A,i,looplim,a_valid);
  }    /* for i */

 /*
  * Write result files
  */
  printf("\nFilename for A estimates ? (none for no write) ");
  scanf("%s",TempString);
  if (strcmp(TempString,"none")) {
    f2disk(A,looplim,1,TempString);
  }
  printf("\nFilename for U estimates ? (none for no write) ");
  scanf("%s",TempString);
  if (strcmp(TempString,"none")) {
    f2disk(U,looplim,1,TempString);
  }
  printf("\nFilename for Real Dominant Component ? (none for no write) ");
  scanf("%s",TempString);
  if (strcmp(TempString,"none")) {
    f2disk(DCReal,looplim,1,TempString);
  }
  printf("\nFilename for Imag Dominant Component ? (none for no write) ");
  scanf("%s",TempString);
  if (strcmp(TempString,"none")) {
    f2disk(DCImag,looplim,1,TempString);
  }
  printf("\nFilename for maxfilt data ? (none for no write) ");
  scanf("%s",TempString);
  if (strcmp(TempString,"none")) {
    byte2disk(maxfilt,looplim*sizeof(int),1,TempString);
  }
 
  printf("\n\n");
  return;

} /*-------------------------------------------- main --------------*/

/*
 * Initialize
 *
 *  Process the paramters file
 *  Read the input signal, if real make hardy to get the imaginary part
 *  Normalize the input signal to have values in [-1, +1]
 *  Compute the input spectrum
 *  Construct the filter bank
 */

void Initialize(ParmFn,size,ZP,xReal,xImag,XReal,XImag,G,PrintResponse,
                RespRootFn)

  char        *ParmFn;           /* filename of parameters file */
  int         *size;             /* size of the input signal */
  int         *ZP;               /* 1 if zero padding is to be used */
  float      **xReal;            /* real part of input signal */
  float      **xImag;            /* imag part of input signal */
  float      **XReal;            /* real part of input spectrum */
  float      **XImag;            /* imag part of input spectrum */
  filterbank **G;                /* the filter bank */
  int         *PrintResponse;    /* 1 if we are to print the response images */
  char        *RespRootFn;       /* Root filename for response signals */
{

  int         comps;             /* dummy num freq comps to estimate */
  int         i;                 /* counter */
  char        datatype;          /* type of in dat: b=byte f=float d=double */
  int         Conjugate;         /* 1=conjugate input.  See Book X, p. 64 */
  char        Rfn[FNSIZE];       /* file for Real part of input signal */
  char        Ifn[FNSIZE];       /* file for Imag part of input signal */
  double      r0;                /* center freq. of 1st filt */
  float       obw;               /* Eta-peak octave bandwidth of filters */
  float       Eta;               /* OBW is defined in terms of Eta-peak. */
  float       kappa;             /* postfilter sigma scaling factor */
  float      *tfloat;            /* temporary float array */

 /*
  * Read the parameter file
  */
  ReadParametersFile(ParmFn,&datatype,&Conjugate,Rfn,Ifn,size,ZP,&comps,
                     PrintResponse,&r0,&Eta,&obw,&kappa,RespRootFn);
  r0 = r0 / (double)(*size);

 /*
  * If zero padding, double size and also r0 (which is in cpi)
  */
  if (*ZP) {
    *size *= 2;
  }

 /*
  * Make the filter bank
  */ 
  *G = CreateFilterBank1D(r0,obw,Eta,kappa);

 /*
  * Allocate Spectra
  */
  if ((*XReal = (float *)malloc(*size*sizeof(float))) == NULL) {
    printf("\nInitialize: free store exhausted\n");
    exit(-1);
  }
  if ((*XImag = (float *)malloc(*size*sizeof(float))) == NULL) {
    printf("\nInitialize: free store exhausted\n");
    exit(-1);
  }

 /*
  * Get the input image and its spectrum
  */
  if (*ZP) {
    tfloat = readasf((*size >> 1),1,Rfn,datatype);  
    if ((*xReal = (float *)malloc(*size*sizeof(float))) == NULL) {
      printf("\nInitialize: free store exhausted\n");
      exit(-1);
    }
    memcpy(*xReal,tfloat,(*size >> 1) * sizeof(float));
    free(tfloat);
    for (i=(*size >> 1); i < *size; i++) {
      (*xReal)[i] = (float)0.0;
    }
  } 
  else {
    *xReal = readasf(*size,1,Rfn,datatype);  
  }

  if (strcmp(Ifn,"none")) {                     /* filename is not "none" */
    if (*ZP) {
      tfloat = readasf((*size>>1),1,Ifn,datatype);
      if ((*xImag = (float *)malloc(*size*sizeof(float))) == NULL) {
        printf("\nInitialize: free store exhausted\n");
        exit(-1);
      }
      memcpy(*xImag,tfloat,(*size >> 1) * sizeof(float));
      free(tfloat);
      for (i=(*size >> 1); i < *size; i++) {
        (*xImag)[i] = (float)0.0;
      }
    } 
    else {
      *xImag = readasf(*size,1,Ifn,datatype);
    }
    if (Conjugate) {
      for (i=0; i < *size; i++) {
        (*xImag)[i] *= (float)-1.0;
      }
    }
    f_u_norm(*xReal, *xImag, *size, 1);
    fft1d(*xReal,*xImag,*XReal,*XImag,*size,1);
  }
  else {                         /* input is real, must create imag part */
    f_u_norm(*xReal, NULL, *size, 1);
    if ((*xImag = (float *)malloc(*size*sizeof(float))) == NULL) {
      printf("\nInitialize: free store exhausted\n");
      exit(-1);
    }
    MakeHardy1D(*xReal,*xImag,*XReal,*XImag,*size);
  }

  return;
} /*---------------------------------------------- Initialize ----------*/

/*
 * ReadParametersFile
 *
 *  read the data out of the parameters file
 *
 */
void ReadParametersFile(ParmFn,pdatatype,pConjugate,Rfn,Ifn,psize,pZP,pcomps,
                        pPrintResponse,pr0,pEta,pobw,pkappa,RespRootFn)

  char   *ParmFn;
  char   *pdatatype;
  int    *pConjugate;
  char   *Rfn;
  char   *Ifn;
  int    *psize;
  int    *pZP;
  int    *pcomps;
  int    *pPrintResponse;
  double *pr0;
  float  *pEta;
  float  *pobw;
  float  *pkappa;
  char   *RespRootFn;
{

  FILE      *parmfp;            /* ptr to parameter file */
  char       instr[256];        /* buffer to read into  */

  if ((parmfp = fopen(ParmFn,"r")) == NULL) {
    printf("\n\nReadParametersFile: unable to open %s\n",ParmFn);
    exit(-1);
  }

  fscanf(parmfp,"Input Data Type= %s\n",instr);
  *pdatatype = instr[0];
  fscanf(parmfp,"Conjugate Input= %s\n",instr);
  if (instr[0] == 'N') {
    *pConjugate = 0;
  }
  else {
    *pConjugate = 1;
  }
  fscanf(parmfp,"Real Sequence File= %s\n",Rfn);
  fscanf(parmfp,"Imag Sequence File= %s\n",Ifn);
  fscanf(parmfp,"Size= %d\n",psize);
  fscanf(parmfp,"Zero Pad= %s\n",instr);
  if ((instr[0] == 'Y') || (instr[0] == 'y')) {
    *pZP = 1;
  }
  else {
    *pZP = 0;
  }
  fscanf(parmfp,"Components= %d\n",pcomps);
  fscanf(parmfp,"Print Response Sequences= %s\n",instr);
  if (instr[0] == 'N') {
    *pPrintResponse = 0;
  }
  else {
    *pPrintResponse = 1;
  }
  fscanf(parmfp,"r0= %lf\n",pr0);
  fscanf(parmfp,"Eta= %f\n",pEta);
  fscanf(parmfp,"OBW= %f\n",pobw);
  fscanf(parmfp,"kappa= %f\n",pkappa);
  fscanf(parmfp,"Response Root File Name= %s\n",RespRootFn);
  fclose(parmfp);
  return;
} /*--------------------------- ReadParametersFile -------------------------*/

/*
 * dom_esa
 *
 *  function evaluates the esa at every point in a signal and checks to see if
 *  if the current response signal has the dominant response of those yet seen.
 *
 */

void dom_esa(Y,G,criterion,u,a,Bank,ZP,maxfilt,u_valid,a_valid,DCReal,DCImag)

  pFloatImage     Y;              /* response image */
  FilterPtr       G;              /* Y was output from this filter */
  float          *criterion;      /* filter selection criterion */
  float          *u;              /* u estimates */
  float          *a;              /* a estimates */
  filterbank     *Bank;           /* the filterbank */
  int             ZP;             /* one if we are doing zero-padding */
  int            *maxfilt;        /* which chan max'd criterion @ (i,j) */
  BYTE           *u_valid;        /* 1 where the u estimate is valid */
  BYTE           *a_valid;        /* 1 where the a estimate is valid */
  float          *DCReal;         /* Real part of the dominant component */
  float          *DCImag;         /* Imag part of the dominant component */
{

  int             i;              /* loop counters for image coordinates */
  int             looplim;        /* terminal value for loop counters */
  double          theta;          /* inst phase of component */

 /*
  * Apply the ESA at each point in the image
  */
  if (ZP) {
    looplim = Y->size >> 1;
  } else {
    looplim = Y->size;
  }
  for (i=0; i < looplim; i++) {
    if (PtEsa1D(Y,Y,i,G,&criterion[i],&u[i],&a[i],&u_valid[i],&a_valid[i])) {
      maxfilt[i] = G->index;
      if ((Y->Real[i] == (float)0.0) && (Y->Imag[i] == (float)0.0)) {
        theta = (double)0.0;
      } else {
        theta = atan2(Y->Imag[i],Y->Real[i]);
      }
      DCReal[i] = a[i] * (float)cos(theta);
      DCImag[i] = a[i] * (float)sin(theta);
    }
  }
  return;
} /*----------------------------------------- dom_esa -----------------------*/
#endif
