
/*
 * aAve4+DownSamp.c
 *
 *  Four point causal average and downsample by 4 for ascii data.
 *  - read an ascii data file in two column x-y format.
 *  - discard the x values.
 *  - output array is 1/4 the length of the input array.
 *  - each output sample is the average of four input samples.
 *    - the first output is the average of the first four inputs.
 *    - the second output is the average of 5th - 8th inputs.
 *  - write output in two column x-y ascii format with new zero-based 
 *    ordinates.
 *
 *  If the number of input points is not divisible by 4, then the final
 *  output will be an average of less than four points.  For example, if
 *  there are 10 input points, the last output will be the average of the
 *  last two inputs.
 *
 * 2/23/99 jph
 *
 */

#ifndef AAVE4PDOWNSAMP_C
#define AAVE4PDOWNSAMP_C

#include "teager2D.h"

main(argc,argv)

  int   argc;
  char *argv[];

{


  int           size;            /* row , col dim */
  int           so4;             /* size / 4 */
  int           LastSize;        /* num. pts. to average for last output */
  int           OutSize;         /* size of output array */
  char         *InFn;            /* input filename */
  FILE         *infile;          /* input file pointer */
  char         *OutFn;           /* output filename */
  FILE         *outfile;         /* output file pointer */
  float        *x;               /* input data array */
  float        *y;               /* output data array */
  int           junk;            /* temporary junk variable */
  int           i;               /* loop counter for input array */
  int           j;               /* loop counter for adding 4 inputs */
  int           k;               /* loop counter for output array */
 
 /*
  * parse args
  */
  if (argc !=  4) {
    printf("\n%s: 4 point causal average and downsample by 4 for ascii data",
           argv[0]);
    printf("\nusage: %s size infile outfile \n",argv[0]);
    return(-1);
  }
  size = atoi(argv[1]);
  so4  = size >> 2;
  LastSize = size % 4;
  OutSize = so4;
  if (LastSize) OutSize++;
  InFn  = argv[2];
  OutFn = argv[3];

 /*
  * Allocate working arrays
  */
  if ((x = (float *)malloc(size * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.",argv[0]);
    exit(-1);
  }
  if ((y = (float *)malloc(OutSize * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.",argv[0]);
    exit(-1);
  }
  
 /*
  * Open input file and read.
  */
  if ((infile = fopen(InFn,"r")) == NULL) {
    printf("\n%s: cannot open file %s",InFn);
    exit(-1);
  }
  for (i=0; i < size; i++) {
    fscanf(infile,"%d\t%f\n",&junk,&x[i]);
  }
  fclose(infile);

 /*
  * Compute output
  */
  for (i=k=0; k < so4; k++,i+=4) {
    y[k] = (float)0.0;
    for (j=0; j < 4; j++) {
      y[k] += x[i+j];
    }
    y[k] /= (float)4;
  }
  if (LastSize) {
    for (j=0; j < LastSize; j++) {
      y[k] += x[i+j];
    }
    y[k] /= (float)LastSize;
  }

 /*
  * Write output file
  */
  if ((outfile = fopen(OutFn,"w")) == NULL) {
    printf("\n%s: cannot open file %s",OutFn);
    exit(-1);
  }
  for (k=0; k < OutSize; k++) {
    fprintf(outfile,"%d\t%f\n",k,y[k]);
  }
  fclose(outfile);

  return;

} 
#endif
