/* 
 * aShowSpec1D.c 
 *
 *  This program writes out the spectrum of each 1D gabor filter in the 
 *  filterbank to an ascii file suitable for display with xvgr.
 *  
 *  Input parameters are read from a params file specified as the command line
 *  option.  
 *
 *  usage: MkObs1D ParameterFileName
 *
 *  jph 9/19/94
 */

#ifndef ASHOWSPEC1D_C
#define ASHOWSPEC1D_C

#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include <string.h>

#include "teager2D.h"
#include "Esa1D.h"

/*
 * Function prototypes
 */
void Initialize();
void ReadParametersFile();

main(argc,argv)

  int   argc;
  char *argv[];

{

  filterbank        *G;              /* the filter bank */
  int                size;           /* size of the input signal */
  int                n;              /* size again */
  int                i;              /* counter */
  int                k;              /* loop counter */
  float             *X;              /* filter spectrum */
  double             fincr;          /* increment in frequency per sample */
  double             f;              /* frequency */
  FILE              *fd;             /* file descriptor */
  FILE *logfile;         /* ascii output file */

  char               RootFn[FNSIZE];   /* root filename for response signals */
  char               TempString[FNSIZE]; /* string for building up filenames */

 /*
  * Check for right number of arguments
  */
  if (argc != 2) {
    printf("\n%s: Write filter spectra to files in ascii for xvgr.",argv[0]);
    printf("\nusage: %s ParameterFileName\n",argv[0]);
    exit(-1);
  }
#ifdef DEBUG
 /*
  * Link to PA so we can use it in the debugger 
  */
  if (!argc) {
    PA((void *)X, 0, -1, 'b');
    PRINTA((void *)X,0,-1,'b',"delete.me");
    WRITEA((void *)X,0,-1,'b',"delete.me");
    image2disk((void *)X,1,1,'*',"delete.me");
    SpecMag2disk(X,"delete.me");
    PC(X,0,0,-1,'b',0);
    PRINTC(X,0,0,-1,'b',0,"delete.me");
    WRITEC(X,0,0,-1,'b',0,"delete.me");
  }
#endif
  Initialize(argv[1],&size,&G,RootFn);
  n = size;
  fincr = (double)1.0 / (double) size;

 /*
  * Report bank data to acsii log file.
  */ 
  printf("\nWriting log data to bank.log.");
  if ((logfile = fopen("bank.log","w")) == NULL) {
    printf("\n\n%s: unable to open %s\n",argv[0],"bank.log");
    exit(-1);
  }
  fprintf(logfile,"\nFilterbank:\n");
  fprintf(logfile,"TotNumFilters = %d\n",G->TotNumFilters);
  fprintf(logfile,"r0 = %lf\n",G->r0);
  fprintf(logfile,"Common Ratio = %lf\n",G->R);
  fprintf(logfile,"Eta = %f\n",G->Eta);
  fprintf(logfile,"OBW = %f\n",G->OBW);
  fprintf(logfile,"gamma = %f\n",G->gamma);

 /*
  * Compute each filter and output.
  */
  printf("\nThere will be %d filters.",G->TotNumFilters);
  for (i=0; i < G->TotNumFilters; i++) {
    printf("\nDoing Filter %d...",i);
    X = GMakeReal1D(G->filt[i],size);
    sprintf(TempString,"%s%02d.ascii",RootFn,i);
    if ((fd = fopen(TempString,"w")) == NULL) {
      printf("\n\n%s: unable to open %s\n",argv[0],TempString);
      exit(-1);
    }
    for (k=0, f=-0.5; k < size; k++,f+=fincr) {
      fprintf(fd,"%lf\t%f\n",f,X[k]);
    }
    fclose(fd);
    free(X);
    fprintf(logfile,"\n\nFilter %d\n",i);
    fprintf(logfile,"IndexOnRay = %d\n",G->filt[i]->IndexOnRay);
    fprintf(logfile,"r0 = %lf\n",G->filt[i]->r0);
    fprintf(logfile,"sigma = %lf\n",G->filt[i]->sigma);
    fprintf(logfile,"GMax = %f\n",G->filt[i]->GMax);
  } /* for i */

  fclose(logfile);
  printf("\n\n");
  return;

} /*-------------------------------------------- main --------------*/

/*
 * Initialize
 *
 *  Process the paramters file
 *  Read the input signal, if real make hardy to get the imaginary part
 *  Normalize the input signal to have values in [-1, +1]
 *  Compute the input spectrum
 *  Construct the filter bank
 */

void Initialize(ParmFn,size,G,RootFn)

  char        *ParmFn;           /* filename of parameters file */
  int         *size;             /* size of the input signal */
  filterbank **G;                /* the filter bank */
  char        *RootFn;           /* Root filename for response signals */
{

  double      r0;                /* center freq. of 1st filt */
  float       obw;               /* Eta-peak octave bandwidth of filters */
  float       Eta;               /* OBW is defined in terms of Eta-peak. */

 /*
  * Read the parameter file
  */
  ReadParametersFile(ParmFn,size,&r0,&Eta,&obw,RootFn);
  r0 = r0 / (double)(*size);

 /*
  * Make the filter bank
  */ 
  *G = CreateFilterBank1D(r0,obw,Eta,1.0);

  return;
} /*---------------------------------------------- Initialize ----------*/

/*
 * ReadParametersFile
 *
 *  read the data out of the parameters file
 *
 */
void ReadParametersFile(ParmFn,psize,pr0,pEta,pobw,RootFn)

  char   *ParmFn;
  int    *psize;
  double *pr0;
  float  *pEta;
  float  *pobw;
  char   *RootFn;
{

  FILE      *parmfp;            /* ptr to parameter file */

  if ((parmfp = fopen(ParmFn,"r")) == NULL) {
    printf("\n\nReadParametersFile: unable to open %s\n",ParmFn);
    exit(-1);
  }

  fscanf(parmfp,"Size= %d\n",psize);
  fscanf(parmfp,"r0= %lf\n",pr0);
  fscanf(parmfp,"Eta= %f\n",pEta);
  fscanf(parmfp,"OBW= %f\n",pobw);
  fscanf(parmfp,"Root File Name= %s\n",RootFn);
  fclose(parmfp);
  return;
} /*--------------------------- ReadParametersFile -------------------------*/

#endif
