/*
 * bentchirp.c
 *
 *   Make a bent chirp image.  See Book X, p. 115.  The phase is quadratic
 *   with respect to a rotating coordinate system.  Initially, the angle
 *   between the x,y coordinate system and the p,q coordinate system is
 *   gamma_i degrees = theta_i radians.  In the last row of the image,
 *   this angle has progressed linearly to gamma_f degrees - theta_f radians.
 *
 *   The quadratic phase in the p,q coordinate system begins at a frequency
 *   of cpi_i cycles per image = f_i cycles per sample and ends at frequency
 *   cpi_f cycles per image = f_f cycles per sample.
 * 
 *   12/7/95 jph
 *
 */

#include <stdio.h>
#include "teager2D.h"

main(argc,argv)

  int   argc;
  char *argv[];

{

  double  cpi_i;         /* initial chirp freq in cycles per image */
  double  cpi_f;         /* final   chirp freq in cycles per image */
  double  gamma_i;       /* initial rotation in degrees */
  double  gamma_f;       /* final rotation in degrees */
  double  theta_i;       /* initial rotation in radians */
  double  theta_f;       /* final rotation in radians */
  int     size;          /* image is size x size */
  int     row;           /* image row counter */
  int     col;           /* image col counter */
  int     ystart;        /* row to start bending in */
  char   *outfnR;        /* Real output file name */
  char   *outfnI;        /* Imag output file name */
  double  r;             /* length of coordinate vector */
  double  alpha;         /* orientation of coordinate vector w/respect to x,y */
  double  theta;         /* angle of rotation of p,q w/respect to x,y */
  double  p;             /* rotated x axis */
  double  q;             /* rotated y axis */
  double *phi;           /* instantaneous phase */
  float  *a;             /* instantaneous amplitude */
  float  *xR;            /* Real part of image */
  float  *xI;            /* Imag part of image */


  if (argc != 9) {
    printf("\n\n%s: Make bent chirp image. See book X p. 115.",
              argv[0]);
    printf("\nusage: %s size cpi_i cpi_f rot_i rot_f ystart Rfn Ifn\n\n",
           argv[0]);
    exit(0);
  }

 /*
  * Parse args
  */
  size = atoi(argv[1]);
  cpi_i = atof(argv[2]);
  cpi_f = atof(argv[3]);
  gamma_i = atof(argv[4]);
  gamma_f = atof(argv[5]);
  ystart = atoi(argv[6]);
  outfnR = argv[7];
  outfnI = argv[8];

 /*
  * Init and Allocate
  */
  theta_i = (double)2.0 * (double)M_PI * gamma_i / (double)360.0;
  theta_f = (double)2.0 * (double)M_PI * gamma_f / (double)360.0;
  if ((phi = (double *)malloc(size*size*sizeof(double))) == NULL) {
    printf("\n\n%s: free store exhausted.\n\n",argv[0]);
    exit(-1);
  }
  if ((a = (float *)malloc(size*size*sizeof(float))) == NULL) {
    printf("\n\n%s: free store exhausted.\n\n",argv[0]);
    exit(-1);
  }
  if ((xR = (float *)malloc(size*size*sizeof(float))) == NULL) {
    printf("\n\n%s: free store exhausted.\n\n",argv[0]);
    exit(-1);
  }
  if ((xI = (float *)malloc(size*size*sizeof(float))) == NULL) {
    printf("\n\n%s: free store exhausted.\n\n",argv[0]);
    exit(-1);
  }

 /*
  * Compute instantaneous phase and amplitude
  */
  for (row=0; row < size; row++) {
    theta = theta_i;
    if (row >= ystart) theta += (theta_f - theta_i) / (double)(size - ystart)
                                * (double)(row - ystart);
    for (col=0; col < size; col++) {
      r = dNorm((double)row,(double)col);
      if (row || col) {
        alpha = atan2((double)row,(double)col);
      } else {
        alpha = (double)0.0;
      }

      p = r * cos(alpha - theta);
      q = r * sin(alpha - theta);
      phi[row*size + col] = (double)M_PI / (double)(size*size)
                                        * (cpi_f - cpi_i) * q * q
                          + (double)2.0 * (double)M_PI / (double)size
                                        * cpi_i * q;
      a[row*size + col] = (float)1.0;
    }
  }

 /*
  * Compute image values
  */
  for (row=0; row < size; row++) {
    for (col=0; col < size; col++) {
      xR[row*size+col] = a[row*size+col] * (float) cos(phi[row*size+col]);
      xI[row*size+col] = a[row*size+col] * (float) sin(phi[row*size+col]);
    }
  }

 /*
  * Write image to disk
  */
  f2disk(xR,size,size,outfnR);
  f2disk(xI,size,size,outfnI);

  return;
}
