/*
 *
 * fHVdecomp.c
 *
 *  Do horizontal/vertical decomposition of float image.
 *
 *  The horizontal image is found by applying a triangular filter.
 *  The vertical image is found by subtracting the horizontal image from
 *  the original.
 *
 *  jph 7/31/97
 *
 */

#ifndef FHVDECOMP_C
#define FHVDECOMP_C

#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include <string.h>

#include "teager2D.h"

/*
 * Function prototypes
 */
void Initialize();

main(argc,argv)

  int   argc;
  char *argv[];

{

  pFloatImage        X;              /* input image */
  pFloatImage        H;              /* horizontal image */
  pFloatImage        V;              /* vertical image */
  int                size;           /* row/col dim of the square input image */
  int                so2;            /* size / 2  */
  char              *InFn;           /* input image filename */
  char              *HFn;            /* horizontal image filename */
  char              *VFn;            /* vertical image filename */
  char              *CallName;       /* name of routine called from Unix */
  int                n;              /* size squared */
  int                i;              /* counter */
  int                row;            /* image row counter */
  int                col;            /* image col counter */

 /*
  * Check for right number of arguments and parse args
  */
  if (argc != 5) {
    printf("\n\n%s: Horiz/Vert decomposition of a float image.",argv[0]);
    printf("\nusage: %s size infile HorizFile VertFile\n",argv[0]);
    exit(-1);
  }
  CallName = argv[0];
  size = atoi(argv[1]);
  InFn = argv[2];
  HFn  = argv[3];
  VFn  = argv[4];

#ifdef DEBUG
 /*
  * Link to PA so we can use it in the debugger 
  */
  if (!argc) {
    PA((void *)X, 0, -1, 'b');
    PRINTA((void *)X,0,-1,'b',"delete.me");
    WRITEA((void *)X,0,-1,'b',"delete.me");
    image2disk((void *)X,1,1,'*',"delete.me");
    SpecMag2disk(X,"delete.me");
    PC(X,0,0,-1,'b',0);
    PRINTC(X,0,0,-1,'b',0,"delete.me");
    WRITEC(X,0,0,-1,'b',0,"delete.me");
  }
#endif

 /*
  * Allocate images
  */
  X = (pFloatImage)malloc(sizeof(FloatImage));
  H = (pFloatImage)malloc(sizeof(FloatImage));
  V = (pFloatImage)malloc(sizeof(FloatImage));

  Initialize(CallName,InFn,size,&X->Real,&X->Imag,&X->FFTReal,&X->FFTImag);
  X->size = X->cols = X->rows = size;
  X->square = 1; 
  AllocFloatImage(H,size);
  AllocFloatImage(V,size);
  n = size * size;
  so2 = size >> 1;

 /*
  * Compute horizontal image
  */
  for (i=0; i < n; i++) {
    H->Real[i] = H->Imag[i] = H->FFTReal[i] = H->FFTImag[i] = (float)0.0;
  }
  for (row=0; row < so2; row++) {
    for (col=0; col <= row; col++) {
      i = row*size + col;
      H->FFTReal[i] = X->FFTReal[i];
      H->FFTImag[i] = X->FFTImag[i];
    }
    for (col=size-row; col < size; col++) {
      i = row*size + col;
      H->FFTReal[i] = X->FFTReal[i];
      H->FFTImag[i] = X->FFTImag[i];
    }
  }
  for (row=so2; row < size; row++) {
    for (col=0; col <= size-row; col++) {
      i = row*size + col;
      H->FFTReal[i] = X->FFTReal[i];
      H->FFTImag[i] = X->FFTImag[i];
    }
    for (col=row+1; col < size; col++) {
      i = row*size + col;
      H->FFTReal[i] = X->FFTReal[i];
      H->FFTImag[i] = X->FFTImag[i];
    }
  }
  for (i=0; i < size; i++) {
    H->FFTReal[i*size + i] *= (float)0.5;
    H->FFTImag[i*size + i] *= (float)0.5;
  }
  for (row=size-1,col=1; row > so2; row--,col++) {
    H->FFTReal[row*size + col] *= (float)0.5;
    H->FFTImag[row*size + col] *= (float)0.5;
  }
  for (row=so2-1,col=so2+1; row > 0; row--,col++) {
    H->FFTReal[row*size + col] *= (float)0.5;
    H->FFTImag[row*size + col] *= (float)0.5;
  }

  fft2d(H->Real,H->Imag,H->FFTReal,H->FFTImag,size,-1);

 /*
  * Compute vertical image
  */
  for (i=0; i < n; i++) {
    V->Real[i] = X->Real[i] - H->Real[i];
  }
  
 /*
  * Write result.
  */
  f2disk(H->Real,size,size,HFn);
  f2disk(V->Real,size,size,VFn);

  printf("\n\n");
  return;

} /*-------------------------------------------- main --------------*/

/*
 * Initialize
 *
 *  Read the input image. 
 *  Allocate and compute input spectrum.
 */

void Initialize(CallName,InFn,size,xReal,xImag,XReal,XImag)

  char        *CallName;         /* name of routine called from Unix */
  char        *InFn;             /* input filename */
  int          size;             /* row/col dim of the square input image */
  float      **xReal;            /* real part of input image */
  float      **xImag;            /* imag part of input image */
  float      **XReal;            /* real part of input spectrum */
  float      **XImag;            /* imag part of input spectrum */
{

  int         i;                 /* counter */

 /*
  * Allocate Spectra
  */
  if ((*XReal = (float *)malloc(size*size*sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted\n",CallName);
    exit(-1);
  }
  if ((*XImag = (float *)malloc(size*size*sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted\n",CallName);
    exit(-1);
  }

 /*
  * Get the input image and its spectrum
  */
  *xReal = readasf(size,size,InFn,'f');
  if ((*xImag = (float *)malloc(size*size*sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted\n",CallName);
    exit(-1);
  }
  for (i=0; i < size*size; i++) {
    (*xImag)[i] = (float)0.0;
  }

  fft2d(*xReal,*xImag,*XReal,*XImag,size,1);

  return;
} /*---------------------------------------------- Initialize ----------*/
#endif
