/*
 *
 * fHiPass.c
 *
 *  Hi pass filter an image.
 *
 *  Read image, apply usual low pass postfilter, subtract postfiltered image
 *  image from original to get a high pass version.
 *
 *  jph 5/27/97
 *
 */

#ifndef FHIPASS_C
#define FHIPASS_C

#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include <string.h>

#include "teager2D.h"

/*
 * Function prototypes
 */
void Initialize();
void ReadParametersFile();

main(argc,argv)

  int   argc;
  char *argv[];

{

  filterbank        *G;              /* the filter bank */
  pFloatImage        g;              /* postfilter image */
  pFloatImage        X;              /* input image */
  pFloatImage        LP;             /* low-pass image */
  pFloatImage        Y;              /* filtered image */
  int                size;           /* row/col dim of the square input image */
  int                n;              /* size squared */
  int                PFiltNum;       /* number of chan of postfilter to use */
  int                i;              /* counter */
  char               OutFn[FNSIZE];  /* names for observation files */

 /*
  * Check for right number of arguments
  */
  if (argc != 2) {
    printf("\n\n%s: High pass filter a float image.",argv[0]);
    printf("\nusage: %s ParameterFileName\n",argv[0]);
    exit(-1);
  }
#ifdef DEBUG
 /*
  * Link to PA so we can use it in the debugger 
  */
  if (!argc) {
    PA((void *)X, 0, -1, 'b');
    PRINTA((void *)X,0,-1,'b',"delete.me");
    WRITEA((void *)X,0,-1,'b',"delete.me");
    image2disk((void *)X,1,1,'*',"delete.me");
    SpecMag2disk(X,"delete.me");
    PC(X,0,0,-1,'b',0);
    PRINTC(X,0,0,-1,'b',0,"delete.me");
    WRITEC(X,0,0,-1,'b',0,"delete.me");
  }
#endif
  X  = (pFloatImage)malloc(sizeof(FloatImage));
  LP = (pFloatImage)malloc(sizeof(FloatImage));
  Y  = (pFloatImage)malloc(sizeof(FloatImage));

  Initialize(argv[1],&size,&X->Real,&X->Imag,&X->FFTReal,&X->FFTImag,&G,
             &PFiltNum,OutFn);
  X->size = X->cols = X->rows = size;
  X->square = 1; 
  AllocFloatImage(LP,size);
  AllocFloatImage(Y,size);
  n = size * size;

 /*
  * Compute the low-pass filtered image using the specified postfilter.
  */
  g = MkPostFilt(G->filt[PFiltNum]->sigma,G->filt[PFiltNum]->kappa,size);
  DoPostFilter(g,size,&X->Real[0],&LP->Real[0],G,PFiltNum);
  g = KillFloat(g);
  for (i=0; i < n; i++) {
    LP->Imag[i] = (float)0.0;
  }

 /*
  * Compute the hi-pass filtered image.
  */
  for (i=0; i < n; i++) {
    Y->Real[i] = X->Real[i] - LP->Real[i];
  }

 /*
  * Write result.
  */
  f2disk(Y->Real,size,size,OutFn);

  printf("\n\n");
  return;

} /*-------------------------------------------- main --------------*/

/*
 * Initialize
 *
 *  Process the paramters file
 *  Read the input image. 
 *  Compute the input spectrum
 *  Construct the filter bank
 */

void Initialize(ParmFn,size,xReal,xImag,XReal,XImag,G,PFiltNum,OutFn)

  char        *ParmFn;           /* filename of parameters file */
  int         *size;             /* row/col dim of the square input image */
  float      **xReal;            /* real part of input image */
  float      **xImag;            /* imag part of input image */
  float      **XReal;            /* real part of input spectrum */
  float      **XImag;            /* imag part of input spectrum */
  filterbank **G;                /* the filter bank */
  int         *PFiltNum;         /* number of chan of postfilter to use */
  char        *OutFn;            /* Filename for filtered output image */
{

  int         i;                 /* counter */
  char        Rfn[FNSIZE];       /* file for Real part of input image */
  char        Ifn[FNSIZE];       /* file for Imag part of input image */
  double      r0;                /* rad. center freq. of 1st filt on each ray */
  float       R;                 /* common ratio of radii of adjcnt filtrs */
  float       obw;               /* Eta-peak octave bandwidth of filters */
  float       Eta;               /* OBW is defined in terms of Eta-peak. */
  float       kappa;             /* Post Filter space constant scale factor */

 /*
  * Read the parameter file
  */
  ReadParametersFile(ParmFn,Rfn,size,&r0,&R,&Eta,&obw,&kappa,PFiltNum,OutFn);
  r0 = r0 / (double)(*size);
 
 /*
  * Make the filter bank
  */ 
  *G = CreateFilterBank(r0,R,obw,Eta,1,kappa,kappa,kappa);

 /*
  * Allocate Spectra
  */
  if ((*XReal = (float *)malloc(*size**size*sizeof(float))) == NULL) {
    printf("\nInitialize: free store exhausted\n");
    exit(-1);
  }
  if ((*XImag = (float *)malloc(*size**size*sizeof(float))) == NULL) {
    printf("\nInitialize: free store exhausted\n");
    exit(-1);
  }

 /*
  * Get the input image and its spectrum
  */
  *xReal = readasf(*size,*size,Rfn,'f');
  if ((*xImag = (float *)malloc(*size**size*sizeof(float))) == NULL) {
    printf("\nInitialize: free store exhausted\n");
    exit(-1);
  }
  for (i=0; i < *size**size; i++) {
    (*xImag)[i] = (float)0.0;
  }

  return;
} /*---------------------------------------------- Initialize ----------*/

/*
 * ReadParametersFile
 *
 *  read the data out of the parameters file
 *
 */

void ReadParametersFile(ParmFn,Rfn,psize,pr0,pR,pEta,pobw,pkappa,pPFiltNum,
                        OutFn)
  char   *ParmFn;
  char   *Rfn;
  int    *psize;
  double *pr0;
  float  *pR;
  float  *pEta;
  float  *pobw;
  float  *pkappa;
  int    *pPFiltNum;
  char   *OutFn;
{

  FILE      *parmfp;            /* ptr to parameter file */
  char       instr[256];        /* buffer to read into  */

  if ((parmfp = fopen(ParmFn,"r")) == NULL) {
    printf("\n\nReadParametersFile: unable to open %s\n",ParmFn);
    exit(-1);
  }

  fscanf(parmfp,"Real Image File= %s\n",Rfn);
  fscanf(parmfp,"Size= %d\n",psize);
  fscanf(parmfp,"r0= %lf\n",pr0);
  fscanf(parmfp,"R= %f\n",pR);
  fscanf(parmfp,"Eta= %f\n",pEta);
  fscanf(parmfp,"OBW= %f\n",pobw);
  fscanf(parmfp,"kappa= %f\n",pkappa);
  fscanf(parmfp,"Postfilter Num= %d\n",pPFiltNum);
  fscanf(parmfp,"Output File= %s\n",OutFn);
  fclose(parmfp);
  return;
} /*--------------------------- ReadParametersFile -------------------------*/
#endif
