/*
 * fRotate.c
 *
 *  Rotate a float image in the counterclockwise direction.
 * 
 * 8/30/96 jph
 *
 */

#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include <string.h>

#include "teager2D.h"

main(argc,argv) 
 
  int   argc;
  char *argv[];
 
{

  int       size;              /* row/col dim of image */
  int       so2;               /* size over 2 */
  int       N;                 /* size * size */
  int       i;                 /* loop counter */
  int       row;               /* image row counter */
  int       col;               /* image col counter */
  int       ulcol;             /* col of upper left  neighbor of (x',y') */
  int       ulrow;             /* row of upper left  neighbor of (x',y') */
  int       urcol;             /* col of upper right neighbor of (x',y') */
  int       urrow;             /* row of upper right neighbor of (x',y') */
  int       llcol;             /* col of lower left  neighbor of (x',y') */
  int       llrow;             /* row of lower left  neighbor of (x',y') */
  int       lrcol;             /* col of lower right neighbor of (x',y') */
  int       lrrow;             /* row of lower right neighbor of (x',y') */
  float     theta;             /* rotation angle */
  float     costheta;          /* cosine(theta) */
  float     sintheta;          /* sine(theta) */
  float     ycol_int;          /* integer part of ycol[i] */
  float     ycol_frac;         /* fractional part of ycol[i] */
  float     yrow_int;          /* integer part of yrow[i] */
  float     yrow_frac;         /* fractional part of ycol[i] */
  float     uldist;            /* distance from (x',y') to ul neighbor */
  float     urdist;            /* distance from (x',y') to ur neighbor */
  float     lldist;            /* distance from (x',y') to ll neighbor */
  float     lrdist;            /* distance from (x',y') to lr neighbor */
  float     sumdist;           /* uldist + urdist + lldist + lrdist */
  float     ulfact;            /* weight factor for ul neighbor */
  float     urfact;            /* weight factor for ur neighbor */
  float     llfact;            /* weight factor for ll neighbor */
  float     lrfact;            /* weight factor for lr neighbor */
  float    *x;                 /* input image */
  float    *y;                 /* output image */
  float    *yrow;              /* which row in x maps to a row of y */
  float    *ycol;              /* which col in x maps to a col of y */
  char     *infn;              /* input filename */
  char     *outfn;             /* output filename */

 /*
  * parse args
  */
  if (argc != 5) {
    printf("\n%s: rotate a float image counterclockwise.",argv[0]);
    printf("\nusage: %s size infile theta_deg outfile\n",argv[0]);
    exit(0);
  }
  size = atoi(argv[1]);
  so2 = size >> 1;
  N = size * size;
  infn = argv[2];
  theta = atof(argv[3]);
  theta /= (float)360.0;
  theta *= (float)2.0 * (float)M_PI;
  outfn = argv[4];

 /*
  * Allocate working arrays
  */
  if ((x = (float *)malloc(N * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n");
    exit(-1);
  }
  if ((y = (float *)malloc(N * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n");
    exit(-1);
  }
  if ((yrow = (float *)malloc(N * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n");
    exit(-1);
  }
  if ((ycol = (float *)malloc(N * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n");
    exit(-1);
  }

 /*
  * Read Image
  */
  disk2f(x,size,size,infn);

 /*
  * For each pixel of y, find out the corresponding row,col in x
  */
  costheta = (float)cos((double)theta);
  sintheta = (float)sin((double)theta);
  for (i=row=0; row < size; row++) {
    for (col=0; col < size; col++,i++) {
      ycol[i] = (float)(col-so2)*costheta + (float)so2 
                - (float)(row-so2)*sintheta;
      yrow[i] = (float)(col-so2)*sintheta + (float)so2 
                + (float)(row-so2)*costheta;
    }
  }

 /*
  * Rotate
  */
  for (i=row=0; row < size; row++) {
    for (col=0; col < size; col++,i++) {
      y[i] = (float)0.0;
     /*
      * check for in bounds
      */
      if ((ycol[i] >= (float)0.0) && (yrow[i] >= (float)0.0) 
                                  && (ycol[i] < (float)size)
                                  && (yrow[i] < (float)size)) {
       /*
        * find neighbors
        */
        ycol_int = (float)floor((double)ycol[i]);
        yrow_int = (float)floor((double)yrow[i]);
        ycol_frac = ycol[i] - ycol_int;
        yrow_frac = yrow[i] - yrow_int;
  
        ulrow = urrow = yrow_int;
        llrow = lrrow = yrow_int + 1;
        if (llrow >= size) {
          llrow--;
          lrrow--;
        }
        ulcol = llcol = ycol_int;
        urcol = lrcol = ycol_int + 1;
        if (urcol >= size) {
          urcol--;
          lrcol--;
        }

       /*
        * find distances to neighbors
        */
        uldist = fNorm(ycol[i]-(float)ulcol, yrow[i]-(float)ulrow);
        urdist = fNorm(ycol[i]-(float)urcol, yrow[i]-(float)urrow);
        lldist = fNorm(ycol[i]-(float)llcol, yrow[i]-(float)llrow);
        lrdist = fNorm(ycol[i]-(float)lrcol, yrow[i]-(float)lrrow);
        sumdist = uldist + urdist + lldist + lrdist;
     
       /*
        * Compute weight factors for neighbors
        */
        ulfact = uldist / sumdist;
        urfact = urdist / sumdist;
        llfact = lldist / sumdist;
        lrfact = lrdist / sumdist;
  
       /*
        * Compute output pixel value 
        */      
        y[i] = ulfact * x[ulrow*size + ulcol] 
               + urfact * x[urrow*size + urcol]
               + llfact * x[llrow*size + llcol] 
               + lrfact * x[lrrow*size + lrcol];
      }
    }
  }
      
 /*
  * Write result file
  */
  f2disk(y,size,size,outfn);

  printf("\n\n");
  exit(0);

}
