/*
 * fUniform.c
 *
 * Makes a float file containing U(0,1) noise.  Uses the Numerical Recipes
 * routine Ran3.
 *
 *  8/17/95 jph
 *
 */

#ifndef FUNIFORM_C
#define FUNIFORM_C
#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include "teager2D.h"

/*
 * Local Defines
 */
#define MBIG 1000000000
#define MSEED 161803398
#define MZ 0
#define FAC (1.0/MBIG)

/*
 * Function Prototypes
 */
float ran3();

/*-----------------------------------------------------------------------------
 * MAIN routine
 *---------------------------------------------------------------------------*/
main(argc,argv)

  int   argc;
  char *argv[];

{

  int       size;              /* number of deviates to generate */
  int       seed;              /* seed for random generator */
  int       i;                 /* loop counter */
  char     *outfile;           /* ptr to output filename */
  float    *x;                 /* array of random deviates */

 /*
  * Check for proper number of command line arguments.
  */
  if ((argc !=  3) && (argc != 4)) {
    printf("\n\n%s: Make a float file of U(0,1) random deviates.",argv[0]);
    printf(  "\n      Seed using rand() if int seed argument not specified.");
    printf("\n\nusage: %s size [seed] outfile \n",argv[0]);
    return;
  }

 /*
  * Parse command line.  If no seed specified, generate seed using rand().
  */
  size = atoi(argv[1]);
  if (argc == 4) {       /* a seed WAS specified */
    seed = atoi(argv[2]);
    if (seed >= 0) {
      printf("\n\n%s: seed must be < 0.  %d specified.",argv[0],seed);
    } else {
      printf("\n\n%s: using %d for seed value.",argv[0],seed);
    }
    outfile = argv[3];
  } else {               /* seed NOT specified */
    srand(time(NULL));   /* seed rand() as on p. 473 of ABC */
    seed = rand();
    if (seed >= 0) seed *= -1;
    printf("\n\n%s: using %d for seed value.",argv[0],seed);
    outfile = argv[2];
  }

 /*
  * Allocate working array and fill with N(0,1) random deviates.
  */
  if ((x = (float *)malloc(size * sizeof(float))) == NULL) {
    printf("\n\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  for (i=0; i < size; i++) {
    x[i] = ran3(&seed);
  }
  f2disk(x,size,1,outfile);
  
  printf("\n\n");
  return;
} 

/*
 * Numerical Recipes Routine ran3.
 */

float ran3(idum)
int *idum;
{
        static int inext,inextp;
        static long ma[56];
        static int iff=0;
        long mj,mk;
        int i,ii,k;
 
        if (*idum < 0 || iff == 0) {
                iff=1;
                mj=MSEED-(*idum < 0 ? -*idum : *idum);
                mj %= MBIG;
                ma[55]=mj;
                mk=1;
                for (i=1;i<=54;i++) {
                        ii=(21*i) % 55;
                        ma[ii]=mk;
                        mk=mj-mk;
                        if (mk < MZ) mk += MBIG;
                        mj=ma[ii];
                }
                for (k=1;k<=4;k++)
                        for (i=1;i<=55;i++) {
                                ma[i] -= ma[1+(i+30) % 55];
                                if (ma[i] < MZ) ma[i] += MBIG;
                        }
                inext=0;
                inextp=31;
                *idum=1;
        }
        if (++inext == 56) inext=1;
        if (++inextp == 56) inextp=1;
        mj=ma[inext]-ma[inextp];
        if (mj < MZ) mj += MBIG;
        ma[inext]=mj;
        return mj*FAC;
}
 
#undef MBIG
#undef MSEED
#undef MZ
#undef FAC
#endif
