/*
 * mkchirp1D.c
 *
 *  This program makes a 1D chirp signal with Gaussian amplitude modulation.
 *  It is a re-creation of the "mkchirp" program described on pp. 136 - 138 
 *  of book VI.  I somehow lost the source code for the original.
 *
 * 5/5/95 jph
 *
 */

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "teager2D.h"
#include "Esa1D.h"

void ReadParametersFile();

/*char     fnReal[FNSIZE]; *//* Real output file name */

main(argc,argv) 
  int argc;
  char* argv[];
{

  float   *xReal;      /* Real part of the signal */
  float   *xImag;      /* Imag part of the signal */
  float   *g;          /* gaussian window data */
  float    A;          /* amplitude scaling factor */
  float    ConjFact;   /* factor for computing conjuage signal */
  double   C;          /* Gaussian normalizing factor */
  double   Fi;         /* Lowest freq chirp will realize in cycles per sample */
  double   Ff;         /* Highest freq chirp will realize in cycles per sample*/
  double   a;          /* linear phase term */
  double   b;          /* quadratic phase term */
  double   phi;        /* phase */
  double   twopi;      /* twice pi */
  double   sigma;      /* std dev of the Gaussian window */
  double   sigmasq;    /* variance of the Gaussian window */
  int      Ti;         /* time to start the chirp */
  int      Tf;         /* time to stop the chirp  */
  int      i;          /* loop counter */
  int      size;       /* Num samples in the test signal */
  int      useGauss;   /* 1 if a Gaussian window is to be used */
  int      conjugate;  /* 1 means conjugate signal before writing it out */
  int      t_offset;   /* center of the gaussian curve */
  char     ParmFn[FNSIZE]; /* Parameter file name */
  char     fnReal[FNSIZE]; /* Real output file name */
  char     fnImag[FNSIZE]; /* Imag output file name */
 
 /*
  * Check for right number of command line arguments
  */
  if (argc != 2) {
    printf("\n\nusage: %s paramter_file_name\n\n",argv[0]);
    return;
  }

 /*
  * Read parameter file
  */
  ReadParametersFile(argv[1],&size,&Ti,&Tf,&A,&sigma,
                     &useGauss,&Fi,&Ff,&conjugate,fnReal,fnImag);

 /*
  * Check parameters 
  */
  if (Tf >= size) {
    printf("\n%s: Chirp off time (%d) must be less than size (%d)\n",
           argv[0],Tf,size);
    exit(-1);
  }
  if (Tf < Ti) {
    printf("\n%s: Chirp off time (%d) must be >= Chirp on time (%d)\n",
            argv[0],Tf,Ti);
    exit(-1);
  }

 /*
  * Initialize Parameters
  */
  twopi = (double)2.0 * M_PI;
  sigmasq = pow(sigma,(double)2.0);
  
 /*
  * Allocate data arrays
  */
  if ((xReal = (float *) malloc(size * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((xImag = (float *) malloc(size * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }
  if ((g = (float *) malloc(size * sizeof(float))) == NULL) {
    printf("\n%s: free store exhausted.\n",argv[0]);
    exit(-1);
  }

 /*
  * Initialize the data outside the support of the chirp
  */
  for (i=0; i < Ti; i++) {
    g[i] = xReal[i] = xImag[i] = (float)0.0;
  }
  for (i=Tf+1; i < size; i++) {
    g[i] = xReal[i] = xImag[i] = (float)0.0;
  }

 /*
  * Make the Gaussian window 
  */
  t_offset = Ti + (int)((double)(Tf - Ti) / (double)2.0);
  C = (double) 1.0 / (sigma * pow(twopi,(double)0.5));
  for (i=Ti; i <= Tf; i++) {
    if (useGauss) {
      g[i] = (float) exp(
       ((double)-0.5 * pow((double)(i - t_offset),(double)2.0) / sigmasq));
    } else {
      g[i] = (float)1.0;
    }
  }

 /*
  * Make the 1D chirp data, see book VI pp. 136 - 137.
  */
  b = twopi * (double)(Ff - Fi) / (double)(Tf - Ti);
  a = twopi * (double)Fi - b * (double)Ti;
  if (conjugate) ConjFact = (float)-1.0;
  else           ConjFact = (float) 1.0;
  for (i=Ti; i <= Tf; i++) {
    phi = (double) a * (double)i + 
          (double)0.5 * b * pow((double)i, (double)2.0);
    xReal[i] = g[i] * A * (float)cos(phi);
    xImag[i] = ConjFact * g[i] * A * (float)sin(phi);
  }

  f2disk(xReal,size,1,fnReal);
  f2disk(xImag,size,1,fnImag);

  
  return;
} /*--------------------------------------------------  main  --------------*/

/*
 * ReadParametersFile
 *
 *  read the data out of the parameters file
 *
 */
void ReadParametersFile(ParmFn,psize,pTi,pTf,pA,psigma,puseGauss,
                        pFi,pFf,pconjugate,fnReal,fnImag)
  char    *ParmFn;
  int     *psize;
  int     *pTi;
  int     *pTf;
  float   *pA;
  double  *psigma;
  int     *puseGauss;
  double  *pFi;
  double  *pFf;
  int     *pconjugate;
  char    *fnReal;
  char    *fnImag;
 
{
 
  FILE      *parmfp;            /* ptr to parameter file */
  char       instr[256];        /* buffer to read into  */
  char      *why;               /* battle axe variable */
  char       fn_len;            /* length of file name */
 
  if ((parmfp = fopen(ParmFn,"r")) == NULL) {
    printf("\n\nmkchirp: unable to open %s\n",ParmFn);
    exit(-1);
  }
 
  fscanf(parmfp,"Size= %d\n",psize);
  fscanf(parmfp,"Chirp On Time= %d\n",pTi);
  fscanf(parmfp,"Chirp Off Time= %d\n",pTf);
  fscanf(parmfp,"Chirp Initial Freq= %lf\n",pFi);
  fscanf(parmfp,"Chirp Final Freq= %lf\n",pFf);
  fscanf(parmfp,"Amplitude Scaling Factor= %f\n",pA);
  fscanf(parmfp,"Use Gaussian Window= %s\n",instr);
  if (instr[0] == 'N' || instr[0] == 'n') {
    *puseGauss = 0;
  }
  else {
    *puseGauss = 1;
  }
  if (*puseGauss) {
    fscanf(parmfp,"Gaussian Window Std Dev= %lf\n",psigma);
  }
  fscanf(parmfp,"Conjugate= %s\n",instr);
  if (instr[0] == 'N' || instr[0] == 'n') {
    *pconjugate = 0;
  }
  else {
    *pconjugate = 1;
  }
  fscanf(parmfp,"Real Output Data File Name= %s\n",fnReal);
  fscanf(parmfp,"Imag Output Data File Name= %s\n",fnImag);
  fclose(parmfp);
  return;
} /*------------------------------- ReadParametersFile -------------------------*/
